/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 16.12.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   std includes   */
#include <iostream>

/*   project headers   */
#include <bits/mc-engine-opts.h>


namespace duct {
  
  std::ostream& operator<<(std::ostream& os, const shower_engine_opts& opts)
  {
    os<<"\033[1mShower engine options:\033[0m\n";
    
    os<<"  \033[1mYou know what you are doing, don't you?\033[0m "
    <<(opts.is_insane ? "\033[31mI hope so!\033[0m" : "\033[32mThe program corrects me, what can go wrong?\033[0m")<<"\n\n"
    <<"  Transverse momentum cutoff            : "<<opts.kT_cutoff<<"GeV\n"
    <<"  Virtuality cutoff                     : "<<opts.virtuality_cutoff<<"GeV\n"
    <<"  Max number of the color basis vectors : "<<opts.max_number_of_color_basis_states<<"\n"
    <<"  Type of color evolution               : ";
    
    switch(opts.type_of_color_evolution)
    {
      case evolution_type::typeI: os<<"typeI\n"; break;
      case evolution_type::typeII: os<<"typeII\n"; break;
    }
    

    os<<"\n  Color options :\n";
    
    double tmP = -1.0;
    
    for(auto& [tmD, c] : opts.color)
    {
      os<<"    On the evolution range : [";
      if(tmP < 0.0) os<<"muH";
      else os<<tmP;
      os<<", "<<tmD<<"] GeV\n";
      
      os
      <<"      Max color suppression                 : "<<c.max_color_suppression<<"\n"
      <<"      Order of soft corrections             : "<<c.soft_corrections<<"\n"
      <<"      Exponentiated phase effects           : "<<(c.ipi_effects ? "yes" : "no")<<"\n";
      
      tmP = tmD;
    }
                                                  
    os<<"\n  Threshold options :\n"
    <<"    Threshold effects         : "<<(opts.threshold.enabled ? "Enabled\n" : "Disabled\n")
    <<"    Order of soft corrections : "<<opts.threshold.order_of_soft_corrections<<"\n";
    
    
    return os;
  }
  
  
  shower_engine_opts& shower_engine_opts::sanitize()
  {
    /*   don't run the full sanitizer if it is an insane option    */
    if(is_insane)
    {
      for(auto& [tmD, c] : color) {
        c.max_number_of_basis_states = max_number_of_color_basis_states;
        c.type_of_evolution = type_of_color_evolution;
      }
      
      return *this;
    }
    
    
    /*   if there is no color option specified we add one and that results an LC shower   */
    if(color.empty()) color[0.8] = {{}};
    
    
    /*   try to make the options consistent    */
    unsigned max_supp = UINT_MAX, real = UINT_MAX;
    bool ipi = true;
    
    for(auto& [tmD, c] : color)
    {
      if(c.max_color_suppression < max_supp) max_supp = c.max_color_suppression;
      if(c.soft_corrections < real) real = c.soft_corrections;
      ipi &= c.ipi_effects;
      
      c.max_color_suppression = max_supp;
      c.soft_corrections = real;
      c.ipi_effects = ipi;

      c.max_number_of_basis_states = max_number_of_color_basis_states;
      c.type_of_evolution = type_of_color_evolution;
   }
    
    
    return *this;
  }
  
}   //   namespace duct
