/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 19/02/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_color_state_h__
#define __mc_color_state_h__ 1


/*   std includes   */
#include <algorithm>
#include <functional>
#include <iostream>
#include <climits>
#include <complex>
#include <random>
#include <type_traits>
#include <initializer_list>
#include <vector>
#include <list>


/*   project headers   */
#include <cxx/binary-io.h>
#include <cxx/memory-resource.h>
#include <cxx/misc/cxx-iterator.h>




namespace duct {
  
  struct parton_cstate
  {
    /*   color connections   */
    struct colorlink
    {
      /*   make a dead link, it means the particle is not QCD parton   */
      static colorlink dead() {
        return {INT_MAX, INT_MIN};
      }
      
      /*    make a quark    */
      static colorlink quark(int n) {
        return {quarkend, n};
      }
      
      /*   make an antiquark    */
      static colorlink antiquark(int p) {
        return {p, antiquarkend};
      }
      
      //----- test if it is a quark -----
      bool is_quark() const { return prev == quarkend;}
      bool is_triplet() const { return prev == quarkend;}
      
      //----- test if it is a quark -----
      bool is_antiquark() const { return next == antiquarkend;}
      bool is_antitriplet() const { return next == antiquarkend;}
      
      //----- test if it is a quark -----
      bool is_gluon() const { return next > antiquarkend && prev < quarkend;}
      bool is_octet() const { return next > antiquarkend && prev < quarkend;}
      
      //----- test if it is connection ----
      bool is_qcd() const {
        return next > INT_MIN && prev < INT_MAX && !(prev == quarkend && next == antiquarkend);
      }
      
      /*   test if it is connected to parton i  */
      bool is_connected(int i) const {
        return prev == i || next == i;
      }
      
      //---- labeling the quark and antiquark ends -----
      static constexpr int quarkend =  INT_MAX-1;
      static constexpr int antiquarkend = INT_MIN+1;
      
      /*  The label of the previous and next parton on the color string.
       *  this should follow our labeling convetion (tagged integers).
       */
      int prev, next;
    } colorL, colorR;
    
    /*  Healthiness of a gluon:
     *    => '-' not applicable
     *    => 'f' frail gluon (or color 8 object)
     *    => 'h' healthy gluon (or color 8 object)
     */
    char health;
    
    /*   helps to calculate the color suppression index   */
    mutable bool active;
  };
  
  
  /*    printing a color link   */
  inline std::ostream& operator<<(std::ostream& os, const parton_cstate::colorlink& c)
  {
    if(c.is_quark()) os<<"[Q, "<<c.next<<")";
    else if(c.is_antiquark()) os<<"("<<c.prev<<", A]";
    else if(c.is_gluon()) os<<"("<<c.prev<<","<<c.next<<")";
    else os<<"(-,-)";
    
    return os;
  }
  
  /*   comparing color links   */
  inline bool operator==(const parton_cstate::colorlink& x, const parton_cstate::colorlink& y) {
    return (x.prev == y.prev && x.next == y.next);
  }
  
  inline bool operator!=(const parton_cstate::colorlink& x, const parton_cstate::colorlink& y) {
    return (x.prev != y.prev || x.next != y.next);
  }
  
  
  /*   This class represent one component of a generic color state.
   */
  template<ProcessTag _Proc>
  class color_state
  {
  private:
    /*   private types   */
    using _Size = std::size_t;
    using _Diff = std::ptrdiff_t;
    
    /*   number of QCD incomings   */
    static constexpr _Size _n_in = number_of_qcd_in<_Proc>;
    static constexpr _Size _n_reserve = 50u;
    
  public:
    /*   usual typedefs     */
    using parton = parton_cstate;
    using size_type = _Size;
    using difference_type = _Diff;
    
    using reference = parton&;
    using pointer = parton*;
    using iterator = parton*;
    
    using const_reference = const parton&;
    using const_pointer = const parton*;
    using const_iterator = const parton*;
    
    using allocator_type = pmr::polymorphic_allocator<parton>;
    
  public:
    /*   weight of the color state    */
    std::complex<double> weight;
    
    /*   Color suppression index of the "operator".
     *   This is basically the suppression index that always increses
     *   and helps us to truncate the 1/Nc^2 series.
     */
    unsigned color_suppression_index;
    
    /*   Color suppression index of the color overlap.
     *   This helps us to optimize the probability of the basis states.
     */
    int overlap_color_suppression_index;
    
    
    /*    number of the delta and phase insertions   */
    unsigned order_of_soft_corrections;
    unsigned order_of_softH_corrections;
    unsigned order_of_softV_corrections;
    unsigned order_of_softT_corrections;

    
    /*   It is true if the state is consired as U(Nc) state.   */
    bool is_color_u3;
    
    
  public:
    /*    default constructors   */
    color_state()
    : color_state{allocator_type{}} {}
    
    explicit color_state(allocator_type a)
    : _M_memory{a} { _M_create_storage_for(_n_reserve);}
    
    /*   destructor   */
    ~color_state() { _M_deallocate();}
    
    
    /*    non-standard constructor  (no initialization)  */
    explicit color_state(_Size _n_qcd, allocator_type a = {})
    : _M_memory{a}
    {
      _M_create_storage_for(_n_in + _n_qcd);
      _M_lay_down_for(_n_qcd);
    }
    
    
    /*   copy   */
    color_state& operator=(const color_state&);
    
    color_state(const color_state& x, allocator_type a = {})
    : _M_memory{a}
    {
      if(x.begin())
      {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of_out());
        
        _M_copy_head(x);
        std::copy(x.begin(), x.end(), _M_begin);
      }
      else _M_begin = _M_end = _M_end_of_storage = nullptr;
    }
    
    /*   move   */
    color_state(color_state&&, allocator_type);
    color_state& operator=(color_state&&);
    
    color_state(color_state&& x)
    : _M_memory{x._M_memory}
    , _M_begin{x._M_begin}
    , _M_end{x._M_end}
    , _M_end_of_storage{x._M_end_of_storage}
    {
      _M_copy_head(x);
      x._M_begin = x._M_end = x._M_end_of_storage = nullptr;
    }
    
    
    /*    element access    */
    reference operator[](_Diff k) { return *((_M_begin+_n_in) + k-1);}
    const_reference operator[](_Diff k) const { return *((_M_begin+_n_in) + k-1);}
    
    
    /*   iterators    */
    iterator begin() { return _M_begin;}
    const_iterator begin() const { return _M_begin;}
    const_iterator cbegin() const { return _M_begin;}
    
    iterator end() { return _M_end;}
    const_iterator end() const { return _M_end;}
    const_iterator cend() const { return _M_end;}
    
    
    /*   observables   */
    bool empty() const { return (_M_begin == _M_end);}
    _Size size() const { return _M_end-_M_begin;}
    _Size capacity() const { return _M_end_of_storage-_M_begin;}
    _Diff upper() const { return _M_end-(_M_begin+_n_in);}
    _Size number_of_out() const { return static_cast<_Size>(upper());}
    
    constexpr _Diff lower() const { return 1-static_cast<_Diff>(_n_in);}
    constexpr _Size number_of_in() const { return _n_in;}
    
    
    /*   resreving storage    */
    void reserve(_Size _n_qcd)
    {
      auto count = _n_in + _n_qcd;
      if(!_M_begin) _M_create_storage_for(count);
      else if(capacity() < count) _M_rebase_reserve(_n_qcd);
    }
    
    
    /*    resizing the array    */
    void resize(_Size _n_qcd)
    {
      if(!_M_begin) {
        _M_create_storage_for(_n_in + _n_qcd);
        _M_lay_down_for(_n_qcd);
        return;
      }
      
      if(capacity() < _n_in + _n_qcd) {
        _M_rebase_resize(_n_qcd);
        return;
      }
      
      _M_lay_down_for(_n_qcd);
    }
    
    /*   resize the state by n element   */
    void resize_by(_Diff k) {
      resize(static_cast<_Size>(std::max<_Diff>(0, upper()+k)));
    }
    
    /*   shrink the array   */
    void shrink_to_fit() {
      if(_M_begin && _M_end != _M_end_of_storage)
        _M_rebase_reserve(number_of_out());
    }
    
    
    /*    functions for inserting open and closed color strings   */
    void insert_open_stringL(const int *, const int *);
    void insert_open_stringR(const int *, const int *);
    void insert_closed_stringL(const int *, const int *);
    void insert_closed_stringR(const int *, const int *);
    
    /*    inserting open and closed color strings using initializer list    */
    void insert_open_stringL(std::initializer_list<int> ll) { insert_open_stringL(ll.begin(), ll.end());}
    void insert_open_stringR(std::initializer_list<int> ll) { insert_open_stringR(ll.begin(), ll.end());}
    void insert_closed_stringL(std::initializer_list<int> ll) { insert_closed_stringL(ll.begin(), ll.end());}
    void insert_closed_stringR(std::initializer_list<int> ll) { insert_closed_stringR(ll.begin(), ll.end());}
    
    /*    check the LC+ the color connections   */
    double chi(int k, int l) const
    {
      double cfac = (!is_color_u3 && !is_octet(l) ? 0.5*qcd::Cf : 0.25*qcd::Ca);
      
      if(l == k) return (is_octet(l) ? 4.0 : 2.0)*cfac;
      
      int ret = 0;
      auto& cL = operator[](l).colorL;
      auto& cR = operator[](l).colorR;
      
      if(cL.prev == k) ++ret;
      if(cL.next == k) ++ret;
      
      if(cR.prev == k) ++ret;
      if(cR.next == k) ++ret;
      
      
      return ret*cfac;
    }
    
    /*    characteristic function for the averaged state   */
    double xi(int k, int l) const
    {
      const double cfac = (is_octet(l) ? qcd::Ca : qcd::Cf);
      
      if(k == l) return cfac;
      return cfac/(upper()-lower());
    }
    

    /*   test if it is triplet, antitriplet or octet   */
    bool is_octet(int l) const { return operator[](l).colorL.is_octet();}
    bool is_triplet(int l) const { return operator[](l).colorL.is_triplet();}
    bool is_antitriplet(int l) const { return operator[](l).colorL.is_antitriplet();}
    
    int color(int i) const
    {
      if(is_octet(i)) return 8;
      if(is_triplet(i)) return 3;
      if(is_antitriplet(i)) return -3;
      
      return 0;
    }
    
    /*   make the health report of color 8 objects   */
    char health_report(int) const;
    void make_health_report();
    
  private:
    /*    allocator   */
    allocator_type _M_memory;
    
    /*    it is like an std::vector   */
    parton_cstate *_M_begin, *_M_end, *_M_end_of_storage;
    
    
  private:
    /*   create storage    */
    void _M_create_storage_for(_Size _count)
    {
      auto count = std::max(_n_reserve, _count);
      _M_begin = _M_end = _M_memory.allocate(count);
      _M_end_of_storage = _M_begin + count;
    }
    
    /*   allocate memory    */
    void _M_lay_down_for(_Size _n_qcd) {
      _M_end = _M_begin + _n_in + _n_qcd;
    }
    
    /*    rebase    */
    void _M_rebase_resize(_Size _n_qcd)
    {
      auto count = std::max(2*size(), _n_in + _n_qcd);
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_end = _M_begin + _n_in + _n_qcd;
    }
    
    void _M_rebase_reserve(_Size nqcd)
    {
      auto _n_qcd = number_of_out();
      auto count = std::max(2*size(), _n_in + nqcd);
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_end = _M_begin + _n_in + _n_qcd;
    }
    
    /*   deallocate memory    */
    void _M_deallocate() {
      if(_M_begin) _M_memory.deallocate(_M_begin, capacity());
    }
    
    /*    copy the head infomation     */
    void _M_copy_head(const color_state& x)
    {
      weight = x.weight;

      color_suppression_index = x.color_suppression_index;
      overlap_color_suppression_index = x.overlap_color_suppression_index;

      order_of_soft_corrections = x.order_of_soft_corrections;
      order_of_softH_corrections = x.order_of_softH_corrections;
      order_of_softV_corrections = x.order_of_softV_corrections;
      order_of_softT_corrections = x.order_of_softT_corrections;
 
      is_color_u3 = x.is_color_u3;
    }
  };
  
  
  /*   move constructor    */
  template<ProcessTag _Proc>
  color_state<_Proc>::color_state(color_state&& x, allocator_type u)
  : _M_memory{u}
  {
    _M_begin = _M_end = _M_end_of_storage = nullptr;
    
    if(x.begin())
    {
      _M_copy_head(x);
      
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin, x._M_begin);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of_out());
        std::copy(x.begin(), x.end(), _M_begin);
      }
    }
  }
  
  
  /*  copy assignment doesn't change the memory management,
   *  it only copies the values and adjust the boundaries
   */
  template<ProcessTag _Proc>
  color_state<_Proc>& color_state<_Proc>::operator=(const color_state<_Proc>& x)
  {
    if(this != &x && x.begin())
    {
      if(auto count = x.size(); capacity() < count) {
        _M_deallocate();
        _M_create_storage_for(count);
      }
      
      _M_copy_head(x);
      _M_lay_down_for(x.number_of_out());
      std::copy(x.begin(), x.end(), _M_begin);
    }
    
    return *this;
  }
  
  
  /*   move assignment    */
  template<ProcessTag _Proc>
  color_state<_Proc>& color_state<_Proc>::operator=(color_state<_Proc>&& x)
  {
    if(this != &x && x.begin())
    {
      _M_copy_head(x);
      
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin, x._M_begin);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else
      {
        if(auto count = x.size(); capacity() < count) {
          _M_deallocate();
          _M_create_storage_for(count);
        }
        
        _M_lay_down_for(x.number_of_out());
        std::copy(x.begin(), x.end(), _M_begin);
      }
    }
    
    return *this;
  }
  
  
  /*****************************************************************************************************************/
  
  
  /*    inserting color strings   */
  template<ProcessTag _Proc>
  void color_state<_Proc>::insert_closed_stringL(const int *begin, const int *end)
  {
    color_state<_Proc>& p = *this;
    const int *i;
    
    p[*begin].colorL.prev = *(end-1);
    p[*begin].colorL.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorL.prev = *(i-1);
      p[*i].colorL.next = *(i+1);
    }
    
    p[*i].colorL.prev = *(i-1);
    p[*i].colorL.next = *begin;
  }
  
  template<ProcessTag _Proc>
  void color_state<_Proc>::insert_closed_stringR(const int *begin, const int *end)
  {
    color_state<_Proc>& p = *this;
    const int *i;
    
    p[*begin].colorR.prev = *(end-1);
    p[*begin].colorR.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorR.prev = *(i-1);
      p[*i].colorR.next = *(i+1);
    }
    
    p[*i].colorR.prev = *(i-1);
    p[*i].colorR.next = *begin;
  }
  
  template<ProcessTag _Proc>
  void color_state<_Proc>::insert_open_stringL(const int *begin, const int *end)
  {
    color_state<_Proc>& p = *this;
    const int *i;
    
    p[*begin].colorL.prev = parton_cstate::colorlink::quarkend;
    p[*begin].colorL.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorL.prev = *(i-1);
      p[*i].colorL.next = *(i+1);
    }
    
    p[*i].colorL.prev = *(i-1);
    p[*i].colorL.next = parton_cstate::colorlink::antiquarkend;
  }
  
  template<ProcessTag _Proc>
  void color_state<_Proc>::insert_open_stringR(const int *begin, const int *end)
  {
    color_state<_Proc>& p = *this;
    const int *i;
    
    p[*begin].colorR.prev = parton_cstate::colorlink::quarkend;
    p[*begin].colorR.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorR.prev = *(i-1);
      p[*i].colorR.next = *(i+1);
    }
    
    p[*i].colorR.prev = *(i-1);
    p[*i].colorR.next = parton_cstate::colorlink::antiquarkend;
  }
  
  
  template<ProcessTag _Proc>
  inline char color_state<_Proc>::health_report(int l) const
  {
    //---- the health is defined only for color 8 objects -----
    color_state<_Proc>& p = *this;
    
    if(!(p[l].colorL.is_octet())) return '-';
    
    //---- it is a color 8, find the heath status -----
    int rprev = l, lnext;
    do {
      if((lnext = p[rprev].colorL.next) == l) return 'f';
      if((rprev = p[lnext].colorR.prev) == l) return 'h';
    } while(true);
  }
  
  
  template<ProcessTag _Proc>
  inline void color_state<_Proc>::make_health_report()
  {
    color_state<_Proc>& p = *this;
    
    for(int l = p.lower(); l <= p.upper(); ++l)
    {
      //---- the health is defined only for color 8 objects -----
      if(!(p[l].colorL.is_octet())) p[l].health = '-';
      
      //---- it is a color 8, find the heath status -----
      else {
        int rprev = l, lnext;
        do {
          if((lnext = p[rprev].colorL.next) == l) { p[l].health = 'f'; break;}
          if((rprev = p[lnext].colorR.prev) == l) { p[l].health = 'h'; break;}
        } while(true);
      }
    }
  }
  
  
  /*        Elementary color operators       */
  template<ProcessTag _Proc>
  void color_a_dagger(int sigma, int emitter, color_state<_Proc>& p)
  {
    static constexpr int QUARK = parton_cstate::colorlink::quarkend;
    static constexpr int ANTI_QUARK = parton_cstate::colorlink::antiquarkend;
    
    int spectator, m = p.upper();
    
    switch(sigma) {
      case 1:
        spectator = p[emitter].colorL.next;
        p[emitter].colorL.next  = m;
        p[m].colorL.prev = emitter;
        p[m].colorL.next = spectator;
        p[spectator].colorL.prev = m;
        break;
      case -1:
        spectator = p[emitter].colorL.prev;
        p[spectator].colorL.next = m;
        p[m].colorL.prev = spectator;
        p[m].colorL.next = emitter;
        p[emitter].colorL.prev = m;
        break;
      case 2:
        spectator = p[emitter].colorL.prev;
        p[emitter].colorL.prev  = QUARK;
        p[spectator].colorL.next  = m;
        p[m].colorL.prev = spectator;
        p[m].colorL.next = ANTI_QUARK;
        break;
      case -2:
        p[p[emitter].colorL.prev].colorL.next = p[emitter].colorL.next;
        p[p[emitter].colorL.next].colorL.prev = p[emitter].colorL.prev;
        p[emitter].colorL.prev = QUARK;
        p[emitter].colorL.next  = m;
        p[m].colorL.prev = emitter;
        p[m].colorL.next = ANTI_QUARK;
        break;
      case 3:
        spectator = p[emitter].colorL.next;
        p[emitter].colorL.next = ANTI_QUARK;
        p[spectator].colorL.prev = m;
        p[m].colorL.next = spectator;
        p[m].colorL.prev = QUARK;
        break;
      case -3:
        p[p[emitter].colorL.prev].colorL.next = p[emitter].colorL.next;
        p[p[emitter].colorL.next].colorL.prev = p[emitter].colorL.prev;
        p[emitter].colorL.next = ANTI_QUARK;
        p[emitter].colorL.prev  = m;
        p[m].colorL.next = emitter;
        p[m].colorL.prev = QUARK;
        break;
      case 4:
        p[emitter].colorL.prev  = m;
        p[m].colorL.next = emitter;
        p[m].colorL.prev = QUARK;
        break;
      case -4:
        p[emitter].colorL.next  = m;
        p[m].colorL.prev = emitter;
        p[m].colorL.next = ANTI_QUARK;
        break;
      default:
        //   do nothing
        break;
    }
  }
  
  
  template<ProcessTag _Proc>
  void color_a(int sigma, int emitter, color_state<_Proc>& p)
  {
    static constexpr int QUARK = parton_cstate::colorlink::quarkend;
    static constexpr int ANTI_QUARK = parton_cstate::colorlink::antiquarkend;
    
    int spectator, m = p.upper();
    
    switch(sigma) {
      case 1:
        spectator = p[emitter].colorR.next;
        p[emitter].colorR.next  = m;
        p[m].colorR.prev = emitter;
        p[m].colorR.next = spectator;
        p[spectator].colorR.prev = m;
        break;
      case -1:
        spectator = p[emitter].colorR.prev;
        p[spectator].colorR.next = m;
        p[m].colorR.prev = spectator;
        p[m].colorR.next = emitter;
        p[emitter].colorR.prev = m;
        break;
      case 2:
        spectator = p[emitter].colorR.prev;
        p[emitter].colorR.prev  = QUARK;
        p[spectator].colorR.next  = m;
        p[m].colorR.prev = spectator;
        p[m].colorR.next = ANTI_QUARK;
        break;
      case -2:
        p[p[emitter].colorR.prev].colorR.next = p[emitter].colorR.next;
        p[p[emitter].colorR.next].colorR.prev = p[emitter].colorR.prev;
        p[emitter].colorR.prev = QUARK;
        p[emitter].colorR.next  = m;
        p[m].colorR.prev = emitter;
        p[m].colorR.next = ANTI_QUARK;
        break;
      case 3:
        spectator = p[emitter].colorR.next;
        p[emitter].colorR.next = ANTI_QUARK;
        p[spectator].colorR.prev = m;
        p[m].colorR.next = spectator;
        p[m].colorR.prev = QUARK;
        break;
      case -3:
        p[p[emitter].colorR.prev].colorR.next = p[emitter].colorR.next;
        p[p[emitter].colorR.next].colorR.prev = p[emitter].colorR.prev;
        p[emitter].colorR.next = ANTI_QUARK;
        p[emitter].colorR.prev  = m;
        p[m].colorR.next = emitter;
        p[m].colorR.prev = QUARK;
        break;
      case 4:
        p[emitter].colorR.prev  = m;
        p[m].colorR.next = emitter;
        p[m].colorR.prev = QUARK;
        break;
      case -4:
        p[emitter].colorR.next  = m;
        p[m].colorR.prev = emitter;
        p[m].colorR.next = ANTI_QUARK;
        break;
      default:
        //  do nothing
        break;
    }
  }
  
  
  /*   binary I/O   */
  template<ProcessTag _Proc>
  std::ostream& write(std::ostream& os, const color_state<_Proc>& c)
  {
    duct::write(os, c.weight, c.color_suppression_index, c.overlap_color_suppression_index,
                c.order_of_soft_corrections, c.order_of_softH_corrections, c.order_of_softV_corrections, c.order_of_softT_corrections,
                c.is_color_u3, c.number_of_out());
    
    return os.write(static_cast<const char *>(c.begin()), c.size()*sizeof(parton_cstate));
  }

  template<ProcessTag _Proc>
  std::istream& read(std::istream& is, color_state<_Proc>& c)
  {
    std::size_t nout;
  
    duct::read(is, c.weight, c.color_suppression_index, c.overlap_color_suppression_index,
               c.order_of_soft_corrections, c.order_of_softH_corrections, c.order_of_softV_corrections, c.order_of_softT_corrections,
               c.is_color_u3, nout);
    
    c.resize(nout);
    
    return is.read(static_cast<char *>(c.begin()), c.size()*sizeof(parton_cstate));
  }
  

  
  /*    print operator for the color state   */
  template<ProcessTag _Proc>
  std::ostream& operator<<(std::ostream& os, const color_state<_Proc>& c)
  {
    os
    <<"Weight                          : "<<c.weight<<'\n'
    <<"Is color U(3)?                  : "<<c.is_color_u3<<'\n'
    <<"Color supression index          : "<<c.color_suppression_index<<'\n'
    <<"Overlap color suppression index : "<<c.overlap_color_suppression_index<<'\n'
    <<"Number of Delta insertions      : "<<c.order_of_soft_corrections<<'\n'
    <<" - DeltaH insertions            : "<<c.order_of_softH_corrections<<'\n'
    <<" - DeltaV insertions            : "<<c.order_of_softV_corrections<<'\n'
    <<" - DeltaT insertions            : "<<c.order_of_softT_corrections<<'\n';
    
    os<<"-------------------------------------\n";
    for(auto k = c.lower(); k <= c.upper(); ++k)
      os<<k<<": |"<<c[k].colorL<<"><"<<c[k].colorR<<"|  "<<c[k].health<<'\n';
    
    return os;
  }
  
  
  /*    calculate the of an gluon insertion    */
  template<ProcessTag _Proc>
  char health_with_gluon_loop(int i, int j, color_state<_Proc>& p)
  {
    using link = parton_cstate::colorlink;
    
    /*   insert a gluon after i and j   */
    int m = p.upper();
    
    /*   a^\dagger_+(i) operator   */
    int s = p[i].colorL.next;
    link ci = p[i].colorL, cs = p[s].colorL;
    
    p[i].colorL.next = m;
    p[m].colorL.prev = i;
    p[m].colorL.next = s;
    p[s].colorL.prev = m;
    
    
    /*   a_+(emitter) operator   */
    int r = p[j].colorR.next;
    link cj = p[j].colorR, cr = p[r].colorR;
    
    p[j].colorR.next = m;
    p[m].colorR.prev = j;
    p[m].colorR.next = r;
    p[r].colorR.prev = m;
    
    
    /*   find out the health of the inserted gluon    */
    char ret = '-';
    int rprev = m, lnext;
    
    do {
      if((lnext = p[rprev].colorL.next) == m) { ret = 'f'; break;}
      if((rprev = p[lnext].colorR.prev) == m) { ret = 'h'; break;}
    } while(true);
    
    
    /*   restore the color state    */
    p[i].colorL = ci; p[s].colorL = cs;
    p[j].colorR = cj; p[r].colorR = cr;
    
    
    return ret;
  }
  
  
  template<ProcessTag _Proc>
  bool check_for_tadpoles(const color_state<_Proc>& c)
  {
    for(auto k = c.lower(); k <= c.upper(); ++k)
    {
      if(c[k].colorL.next == k && c[k].colorL.prev == k) return true;
      if(c[k].colorR.next == k && c[k].colorR.prev == k) return true;
    }
    
    return false;
  }
    
  
  template<ProcessTag _Proc>
  int color_suppression_index_u3(const color_state<_Proc>& c)
  {
    /*    fresh start   */
    int nloops = 0;
    
    for(auto& p : c)
      if(!p.colorL.is_antiquark())
      {
        p.active = true;
        --nloops;
      } else p.active = false;
      
    
    /*    counting loops    */
    for(int l = c.lower(); l <= c.upper(); ++l)
      if(c[l].active)
      {
        int i = l, n;
        
        do
        {
          c[i].active = false;
          n = c[i].colorL.next;
          i = c[n].colorR.prev;
          c[i].active = false;
        } while(i != l);
        
        ++nloops;
      }
    
    return -nloops;
  }
  
  /**************************************************************************************************************************/
  
  
  /*  Splitting variables
   *  This class is a collection of variables for reconstracting
   *  the event after a splitting in a real splitting.
   */
  struct splitting_cvars_real
  {
    int sigmaL, sigmaR;
    int emitL, emitR;
    
    double weight;
    unsigned state_idx;
    int overlap_idx;
  };
  
  inline std::ostream& operator<<(std::ostream& os, const splitting_cvars_real& v)
  {
    os<<"["<<v.sigmaL<<", "<<v.sigmaR<<", "<<v.emitL<<", "<<v.emitR<<", "<<v.weight<<", "<<v.overlap_idx<<", "<<v.state_idx<<"]";
    return os;
  }

 
  template<ProcessTag _Proc>
  void shower_mapping(color_state<_Proc>& p, const splitting_cvars_real& vars)
  {
    /*   resize and copy   */
    p.resize_by(1);
    
    /*    assigning the new color state    */
    color_a_dagger(vars.sigmaL, vars.emitL, p);
    color_a(vars.sigmaR, vars.emitR, p);
    
    /*    issuing the health report for the new state    */
    p.make_health_report();
    
    /*   control parameters   */
    p.color_suppression_index += vars.state_idx;
    p.overlap_color_suppression_index += vars.overlap_idx;

    /*   weight parameters   */
    p.weight *= vars.weight;
  }
  
  
  template<ProcessTag _Proc> inline
  void shower_mapping(color_state<_Proc>& p, const color_state<_Proc>& q, const splitting_cvars_real& vars)
  {
    p = q;
    shower_mapping(p, vars);
  }
  
  
  template<ProcessTag _Proc, class _Fn, class... _Args>
  requires std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
  void shower_mapping(color_state<_Proc>& p, const splitting_cvars_real& vars, _Fn&& func, _Args&& ...args)
  {
    /*   resize and copy   */
    p.resize_by(1);
    
    /*    assigning the new color state    */
    color_a_dagger(vars.sigmaL, vars.emitL, p);
    color_a(vars.sigmaR, vars.emitR, p);
    
    /*    issuing the health report for the new state    */
    p.make_health_report();
    
    /*   weight and other parameters   */
    p.weight *= vars.weight;
    p.color_suppression_index += vars.state_idx;
    p.overlap_color_suppression_index += vars.overlap_idx;

    /*   callback on the transformed state for further modifiacations    */
    std::invoke(std::forward<_Fn>(func), std::forward<_Args>(args)..., p);
  }

  
  template<ProcessTag _Proc, class _Fn, class... _Args>
  requires std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
  inline void shower_mapping(color_state<_Proc>& p, const color_state<_Proc>& q, const splitting_cvars_real& vars, _Fn&& func, _Args&& ...args)
  {
    p = q;
    shower_mapping(p, vars, std::forward<_Fn>(func), std::forward<_Args>(args)...);
  }
  

  
  

  /*  Splitting variables
   *  This class is a collection of variables for reconstracting
   *  the event after a splitting in a virtual splitting.
   */
  struct splitting_cvars_virt
  {
    char oper, side;
    int emit, absorb;
    
    double weight;
    unsigned state_idx;
    int overlap_idx;
  };
  

  template<ProcessTag _Proc>
  void shower_mapping(color_state<_Proc>& p, const splitting_cvars_virt& vars)
  {
    p.weight *= vars.weight;
    p.color_suppression_index += vars.state_idx;
    p.overlap_color_suppression_index += vars.overlap_idx;


    
    /*   trivial case when we basically insert a self energy on the fermion line   */
    if(vars.oper == 'G')
    {
      if(vars.emit == vars.absorb)
      {
        p.weight *= qcd::Nc;
      }
      
      /*   otherwise we make an extra closed loop   */
      else
      {
        /*    assigning the new color state    */
        const int &i = vars.emit, &j = vars.absorb;
        
        if(vars.side == 'L')
        {
          p[p[i].colorL.next].colorL.prev = j;
          p[p[j].colorL.next].colorL.prev = i;
          
          std::swap(p[i].colorL.next, p[j].colorL.next);
        }
        
        else if(vars.side == 'R')
        {
          p[p[i].colorR.next].colorR.prev = j;
          p[p[j].colorR.next].colorR.prev = i;
          
          std::swap(p[i].colorR.next, p[j].colorR.next);
        }
        
        /*    issuing the health report for the new state    */
        p.make_health_report();
      }
    }
  }
  
  
  template<ProcessTag _Proc> inline
  void shower_mapping(color_state<_Proc>& p, const color_state<_Proc>& q, const splitting_cvars_virt& vars)
  {
    /*   here we don't have to resize, just copy   */
    p = q;
    shower_mapping(p, vars);
  }
  
  
  template<ProcessTag _Proc, class _Fn, class... _Args>
  requires (std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
            || std::is_invocable_v<_Fn, _Args..., const splitting_cvars_virt&, color_state<_Proc>&>)
  void shower_mapping(color_state<_Proc>& p, const splitting_cvars_virt& vars, _Fn&& func, _Args&& ...args)
  {
    p.weight *= vars.weight;
    p.color_suppression_index += vars.state_idx;
    p.overlap_color_suppression_index += vars.overlap_idx;

    
    /*    inserting a gluon after i and j    */
    if(vars.oper == 'G')
    {
      /*    assigning the new color state    */
      const int &i = vars.emit, &j = vars.absorb;
      
      /*   trivial case when we basically insert a self energy on the fermion line   */
      if(i == j)
      {
        p.weight *= qcd::Nc;
      }
      
      /*   otherwise we make an extra closed loop   */
      else
      {
        if(vars.side == 'L')
        {
          p[p[i].colorL.next].colorL.prev = j;
          p[p[j].colorL.next].colorL.prev = i;
          
          std::swap(p[i].colorL.next, p[j].colorL.next);
        }
        
        else if(vars.side == 'R')
        {
          p[p[i].colorR.next].colorR.prev = j;
          p[p[j].colorR.next].colorR.prev = i;
          
          std::swap(p[i].colorR.next, p[j].colorR.next);
        }
        
        /*    issuing the health report for the new state    */
        p.make_health_report();
      }
    }
    
    
    /*   callback on the transformed state for further modifiacations    */
    if constexpr(std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>)
      std::invoke(std::forward<_Fn>(func), std::forward<_Args>(args)..., p);
    else std::invoke(std::forward<_Fn>(func), std::forward<_Args>(args)..., vars, p);
  }
  
  
  template<ProcessTag _Proc, class _Fn, class... _Args>
  requires (std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
            || std::is_invocable_v<_Fn, _Args..., const splitting_cvars_virt&, color_state<_Proc>&>)
  inline void shower_mapping(color_state<_Proc>& p, const color_state<_Proc>& q, const splitting_cvars_virt& vars, _Fn&& func, _Args&& ...args)
  {
    p = q;
    shower_mapping(p, vars, std::forward<_Fn>(func), std::forward<_Args>(args)...);
  }
  
  
}   //    namespace duct




/*  We need a container to represent a generic color vector.
 */
namespace duct {
  

  /*   the shower color state is nothing else just a vector of color_state<_Proc>   */
  template<ProcessTag _Proc>
  class shower_color_state
  {
    using _Cstate = color_state<_Proc>;
    using _List = pmr::list<_Cstate>;
    
  public:
    /*    public types   */
    using cstate = _Cstate;
    
    using pointer = typename _List::pointer;
    using reference = typename _List::reference;
    using iterator = typename _List::iterator;
    
    using const_pointer = typename _List::const_pointer;
    using const_reference = typename _List::const_reference;
    using const_iterator = typename _List::const_iterator;
    
    
    /*  this has to be an allocator awaring class   */
    using size_type = typename _List::size_type;
    using difference_type = typename _List::difference_type;
    using allocator_type = typename _List::allocator_type;
    
    
    /*   default constructor   */
    shower_color_state()
    : shower_color_state{allocator_type{}} {}
    
    explicit shower_color_state(allocator_type a)
    : _M_psi{a}, _M_junk{a}, _TMP_wP{a}, _TMP_wM{a}
    {
      _TMP_wP.reserve(50u);
      _TMP_wM.reserve(50u);
    }
    
    /*   destructor    */
    ~shower_color_state() = default;
    
    
    /*    copy   */
    shower_color_state(const shower_color_state& c, allocator_type a = {})
    : _M_psi{c._M_psi, a}, _M_junk{a}, _TMP_wP{a}, _TMP_wM{a}
    {
      _TMP_wP.reserve(50u);
      _TMP_wM.reserve(50u);
    }
    
    shower_color_state& operator=(const shower_color_state& c)
    {
      if(&c != this)
      {
        /*   move the nodes to junk   */
        if(!_M_psi.empty())
          _M_junk.splice(_M_junk.cbegin(), _M_psi);
        
        
        /*   copy from c   */
        for(auto& ci : c) push(ci);
      }
      
      return *this;
    }
    
    
    /*   move   */
    shower_color_state(shower_color_state&&) = default;
    
    shower_color_state(shower_color_state&& c, allocator_type a)
    : _M_psi{std::move(c._M_psi), a}, _M_junk{a}, _TMP_wP{std::move(c._TMP_wP), a}, _TMP_wM{std::move(c._TMP_wM), a}
    {
      if(_M_junk.get_allocator() == c._M_junk.get_allocator())
        _M_junk = std::move(c._M_junk);
    }
    
    shower_color_state& operator=(shower_color_state&& c)
    {
      if(&c != this)
      {
        /*   move the nodes to junk   */
        if(!_M_psi.empty())
          _M_junk.splice(_M_junk.cbegin(), _M_psi);
        
        /*    steal the states and even the junk    */
        _M_psi = std::move(c._M_psi);
        
        if(_M_junk.get_allocator() == c._M_junk.get_allocator())
          _M_junk.splice(_M_junk.cbegin(), c._M_junk);
        
        
        /*   moving the temporaries     */
        _TMP_wP = std::move(c._TMP_wP);
        _TMP_wM = std::move(c._TMP_wM);
      }
      
      return *this;
    }
    
    
    /*   iterators   */
    iterator begin() { return _M_psi.begin();}
    iterator end() { return _M_psi.end();}
    
    const_iterator begin() const { return _M_psi.cbegin();}
    const_iterator end() const { return _M_psi.cend();}
    
    const_iterator cbegin() const { return _M_psi.cbegin();}
    const_iterator cend() const { return _M_psi.cend();}
    
    
    /*   clear the list
     *    moves every nodes into the junk
     */
    void clear()
    {
      if(!_M_psi.empty())
        _M_junk.splice(_M_junk.begin(), _M_psi);
    }
    
    /*   push(_back) an unitialized color state in   */
    iterator push()
    {
      if(_M_junk.empty())
        return _M_psi.emplace(_M_psi.cend());
      
      iterator n = _M_junk.begin();
      
      _M_psi.splice(_M_psi.cend(), _M_junk, n);
      
      return n;
    }
    
    /*   push(_back) a color state in (with move)   */
    iterator push(cstate&& c)
    {
      if(_M_junk.empty())
        return _M_psi.emplace(_M_psi.cend(), std::move(c));
      
      iterator n = _M_junk.begin();
      
      _M_psi.splice(_M_psi.cend(), _M_junk, n);
      *n = std::move(c);
      
      return n;
    }
    
    /*   push(_back) a color state in (with copy)   */
    iterator push(const cstate& c)
    {
      if(_M_junk.empty())
        return _M_psi.emplace(_M_psi.cend(), c);
      
      iterator n = _M_junk.begin();
      
      _M_psi.splice(_M_psi.cend(), _M_junk, n);
      *n = c;
      
      return n;
    }
    
    
    /*    Real operators
     *    it replaces the basis state phi with one or more states
     *      => |phi) -> O|phi)
     *      => |phi) -> (O1 + O2 + ... )|phi)
     *      => invalidates chi matrix
     */
    void transform(const_iterator phi, const splitting_cvars_real& vars) {
      shower_mapping(const_cast<_Cstate&>(*phi), vars);
    }
    
    
    template<InputIterator _Iter>
    requires std::is_same_v<std::decay_t<typename std::iterator_traits<_Iter>::value_type>, splitting_cvars_real>
    void transform(const_iterator phi, _Iter vfirst, _Iter vlast)
    {
      /*   nothing to do   */
      if(vfirst == vlast) return;
      
      /*   transforms the state   */
      const auto& v = *(vfirst++);
      
      for(; vfirst != vlast; ++vfirst)
      {
        auto n = push(*phi);
        shower_mapping(*n, *vfirst);
      }
      
      shower_mapping(const_cast<_Cstate&>(*phi), v);
    }
    
    
    template<class _Fn, class... _Args>
    requires std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
    void transform(const_iterator phi, const splitting_cvars_real& vars, _Fn&& func, _Args&& ...args)
    {
      shower_mapping(const_cast<_Cstate&>(*phi), vars, std::forward<_Fn>(func), std::forward<_Args>(args)...);
    }
    
    
    template<InputIterator _Iter, class _Fn, class... _Args>
    requires (std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
              && std::is_same_v<std::decay_t<typename std::iterator_traits<_Iter>::value_type>, splitting_cvars_real>)
    void transform(const_iterator phi, _Iter vfirst, _Iter vlast, _Fn&& func, _Args&& ...args)
    {
      /*   nothing to do   */
      if(vfirst == vlast) return;
      
      
      /*   transforms the state   */
      const auto& v = *(vfirst++);
      
      for(; vfirst != vlast; ++vfirst)
      {
        auto n = push(*phi);
        shower_mapping(*n, *vfirst, std::forward<_Fn>(func), std::forward<_Args>(args)...);
      }
      
      shower_mapping(const_cast<_Cstate&>(*phi), v, std::forward<_Fn>(func), std::forward<_Args>(args)...);
    }
    
    
    /*    Real operators
     *    it replaces the basis state phi with one or more states
     *      => |phi) -> |phi) + O|phi)
     *      => |phi) -> |phi) + (O1 + O2 + ... )|phi)
     *      => invalidates chi matrix
     */
    void transform_keep(const_iterator phi, const splitting_cvars_real& vars)
    {
      auto n = push(*phi);
      shower_mapping(*n, vars);
    }
    
    
    template<InputIterator _Iter>
    requires std::is_same_v<std::decay_t<typename std::iterator_traits<_Iter>::value_type>, splitting_cvars_real>
    void transform_keep(const_iterator phi, _Iter vfirst, _Iter vlast)
    {
      /*   nothing to do   */
      if(vfirst == vlast) return;
      
      /*   transforms the state   */
      for(; vfirst != vlast; ++vfirst)
      {
        auto n = push(*phi);
        shower_mapping(*n, *vfirst);
      }
    }
    
    
    template<class _Fn, class... _Args>
    requires std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
    void transform_keep(const_iterator phi, const splitting_cvars_real& vars, _Fn&& func, _Args&& ...args)
    {
      auto n = push(*phi);
      shower_mapping(*n, vars, std::forward<_Fn>(func), std::forward<_Args>(args)...);
    }
    
    
    template<InputIterator _Iter, class _Fn, class... _Args>
    requires (std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
              && std::is_same_v<std::decay_t<typename std::iterator_traits<_Iter>::value_type>, splitting_cvars_real>)
    void transform_keep(const_iterator phi, _Iter vfirst, _Iter vlast, _Fn&& func, _Args&& ...args)
    {
      /*   nothing to do   */
      if(vfirst == vlast) return;
      
      /*   transforms the state   */
      for(; vfirst != vlast; ++vfirst)
      {
        auto n = push(*phi);
        shower_mapping(*n, *vfirst, std::forward<_Fn>(func), std::forward<_Args>(args)...);
      }
    }
    
    
    /*    Virtual operators
     *    it replaces the basis state phi with one or more states
     *      => |phi) -> (1+O)|phi)
     *      => |phi) -> (1 + O1 + O2 + ... )|phi)
     *      => invalidates chi matrix
     */
    void transform(const_iterator phi, const splitting_cvars_virt& vars)
    {
      auto n = push(*phi);
      shower_mapping(*n, vars);
    }
    
    
    template<InputIterator _Iter>
    requires std::is_same_v<std::decay_t<typename std::iterator_traits<_Iter>::value_type>, splitting_cvars_virt>
    void transform(const_iterator phi, _Iter vfirst, _Iter vlast)
    {
      /*   nothing to do   */
      if(vfirst == vlast) return;
      
      /*   transforms the state   */
      for(; vfirst != vlast; ++vfirst)
      {
        auto n = push(*phi);
        shower_mapping(*n, *vfirst);
      }
    }
    
    
    template<class _Fn, class... _Args>
    requires (std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
              || std::is_invocable_v<_Fn, _Args..., const splitting_cvars_virt&, color_state<_Proc>&>)
    void transform(const_iterator phi, const splitting_cvars_virt& vars, _Fn&& func, _Args&& ...args)
    {
      auto n = push(*phi);
      shower_mapping(*n, vars, std::forward<_Fn>(func), std::forward<_Args>(args)...);
    }
    

    template<InputIterator _Iter, class _Fn, class... _Args>
    requires ((std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
               || std::is_invocable_v<_Fn, _Args..., const splitting_cvars_virt&, color_state<_Proc>&>)
              && std::is_same_v<std::decay_t<typename std::iterator_traits<_Iter>::value_type>, splitting_cvars_virt>)
    void transform(const_iterator phi, _Iter vfirst, _Iter vlast, _Fn&& func, _Args&& ...args)
    {
      /*   nothing to do   */
      if(vfirst == vlast) return;
      
      
      /*   transforms the state   */
      for(; vfirst != vlast; ++vfirst)
      {
        auto n = push(*phi);
        shower_mapping(*n, *vfirst, std::forward<_Fn>(func), std::forward<_Args>(args)...);
      }
    }
    
    
    
    /*   transform each component by a callback function
     *   it doesn't add extra components
     */
    template<class _Fn, class... _Args>
    requires std::is_invocable_v<_Fn, _Args..., color_state<_Proc>&>
    void transform(_Fn&& func, _Args&& ...args)
    {
      /*   nothing to do   */
      if(empty()) return;
      
      /*   apply the transformation for the current state   */
      for(auto& ci : _M_psi)
        std::invoke(std::forward<_Fn>(func), std::forward<_Args>(args)..., ci);
    }
    
    
    /*    reduce the color state by performing the sum by Monte Carlo    */
    template<class _RandomEngine>
    void mc_reduce(double, _RandomEngine&, double, std::size_t);
    
    template<class _RandomEngine>
    void mc_reduce(_RandomEngine&);

    /*    removes the negligable components   */
    void reduce(double axe)
    {
      /*   nothing to do   */
      if(empty()) return;
      
      /*   reduce algorithm   */
      const double tiny = std::max(1e-15, axe)/(_M_psi.size()+1);
      
      auto ic = _M_psi.begin();
      auto last = _M_psi.end();
      
      /*   try to remove components    */
      while(ic != last)
        if(std::abs(ic->weight) < tiny)
          _M_junk.splice(_M_junk.cbegin(), _M_psi, ic++);
        else ++ic;
    }
    
    
    /*   some usual container observers    */
    bool empty() const { return _M_psi.empty();}
    size_type size() const { return _M_psi.size();}
    size_type capacity() const { return _M_psi.size() + _M_junk.size();}
    
    
  private:
    /*    linked list to store the color states    */
    _List _M_psi, _M_junk;
    
    
  private:
    /*   this helps to prune the small contributions   */
    template<class _RandomEngine, class _Predicate>
    unsigned _M_mc_reduce(_RandomEngine&, double, std::size_t, _Predicate);
    
    
    /*   some workspace to do MC reduction   */
    struct state_ref
    {
      iterator current;
      double wprob;
      unsigned selected;
    };
    
    pmr::vector<state_ref> _TMP_wP, _TMP_wM;
  };
  
  /*    binary I/O    */
  template<ProcessTag _Proc>
  inline std::ostream& write(std::ostream& os, const shower_color_state<_Proc>& c)
  {
    duct::write(os, static_cast<std::size_t>(c.size()));
    
    for(const auto& ci : c)
      duct::write(os, ci);
    
    return os;
  }
  
  template<ProcessTag _Proc>
  inline std::istream& read(std::istream& is, shower_color_state<_Proc>& c)
  {
    std::size_t nc;
    
    c.clear();
    duct::read(is, nc);
    
    for(std::size_t k = 0; k < nc; k++)
      duct::read(is, *c.push());
    
    return is;
  }
  
  
  
  /*    print operator for the color state   */
  template<ProcessTag _Proc>
  inline std::ostream& operator<<(std::ostream& os, const shower_color_state<_Proc>& c)
  {
    for(const auto& ci : c)
      os<<ci<<'\n';
    return os;
  }
  

  template<ProcessTag _Proc>
  template<class _RandomEngine>
  void shower_color_state<_Proc>::mc_reduce(double axe, _RandomEngine& rne, double rate, std::size_t nmax)
  {
    /*   trivial case   */
    if(size() < 5u) {
      reduce(axe);
      return;
    }
    
    
    /*   LC+ part    */
    auto maxdelta= _M_mc_reduce(rne, rate, nmax, [](const color_state<_Proc>& c) -> bool {
      return c.order_of_soft_corrections == 0u;
    });
    
    /*   first order delta corrections    */
    if(maxdelta > 0u)
      _M_mc_reduce(rne, rate, nmax, [](const color_state<_Proc>& c) -> bool {
        return c.order_of_soft_corrections == 1u;
      });
    
    /*   second order delta corrections    */
    if(maxdelta > 1u)
      _M_mc_reduce(rne, rate, nmax, [](const color_state<_Proc>& c) -> bool {
        return c.order_of_soft_corrections == 2u;
      });
    
    /*   higher order delta corrections    */
    if(maxdelta > 2u)
      _M_mc_reduce(rne, rate, nmax, [](const color_state<_Proc>& c) -> bool {
        return c.order_of_soft_corrections == 3u;
      });
    
    /*   higher order delta corrections    */
    if(maxdelta > 3u)
      _M_mc_reduce(rne, rate, nmax, [](const color_state<_Proc>& c) -> bool {
        return c.order_of_soft_corrections > 3u;
      });

    
    /*    clean up the color state    */
    reduce(axe);
  }

  
  template<ProcessTag _Proc>
  template<class _RandomEngine, class _Predicate>
  unsigned shower_color_state<_Proc>::_M_mc_reduce(_RandomEngine& rne, double rate, std::size_t nmax, _Predicate pred)
  {
    using complex_t = std::complex<double>;

    /*   finding the thrust axis of the complex weights    */
    complex_t n{0.0, 0.0};
    
    auto zdot = [](const complex_t& a, const complex_t& b) -> double {
      return std::real(a)*std::real(b) + std::imag(a)*std::imag(b);
    };
    
    
    for(const auto& ci : _M_psi)
      if(pred(ci))
      {
        if(zdot(n, ci.weight) < 0.0) n -= ci.weight;
        else n += ci.weight;
      }
    
    
    /*    corting the basis vectors accoring to ther weights into two hemisphere    */
    double normP = 0.0, normM = 0.0;
    unsigned maxdelta = 0u;
    
    _TMP_wP.clear();
    _TMP_wM.clear();
    
    
    /*    selecting and sorting the relevant basis states    */
    for(iterator ic = _M_psi.begin(); ic != _M_psi.end(); ++ic)
    {
      if(ic->order_of_soft_corrections > maxdelta)
        maxdelta = ic->order_of_soft_corrections;
      
      if(pred(*ic))
      {
        double wprob = std::abs(ic->weight)/std::pow(qcd::Nc, ic->overlap_color_suppression_index);
        
        if(wprob > 0.0)
        {
          if(zdot(n, ic->weight) > 0.0)
          {
            normP += wprob;
            _TMP_wP.push_back({ic, wprob, 0u});
          }
          else
          {
            normM += wprob;
            _TMP_wM.push_back({ic, wprob, 0u});
          }
        }
      }
    }
    
    
    /*   nothing to do    */
    if(_TMP_wP.empty() && _TMP_wM.empty()) return maxdelta;
    
    
    /*   finding out the number of important basis states    */
    std::size_t nP = 0u, nM = 0u;
    
    auto compare_refs = [](const state_ref& x, const state_ref& y) -> bool {
      return x.wprob > y.wprob;
    };

    if(!_TMP_wP.empty())
    {
      double sum = 0.0, rr = rate*normP;

      std::sort(_TMP_wP.begin(), _TMP_wP.end(), compare_refs);

      while(sum < rr) sum += _TMP_wP[nP++].wprob;
      if(nP > nmax) nP = nmax;
    }
    
    if(!_TMP_wM.empty())
    {
      double sum = 0.0, rr = rate*normM;
      
      std::sort(_TMP_wM.begin(), _TMP_wM.end(), compare_refs);
      
      while(sum < rr) sum += _TMP_wM[nM++].wprob;
      if(nM > nmax) nM = nmax;
    }
    

    /*     selecting basis states randomly    */
    std::uniform_real_distribution<double> rng(0.0, 0.999999);
    
    for(std::size_t k = 0; k < std::max(nP, nM); ++k)
    {
      double r = rng(rne);
      
      /*   select a component with positive weight   */
      if(k < nP)
      {
        auto io = _TMP_wP.begin();
        double rn = normP*r, oprob = io->wprob;
        
        while(oprob < rn) oprob += (++io)->wprob;
        io->selected++;
      }
      
      /*   select a component with negative weight   */
      if(k < nM)
      {
        auto io = _TMP_wM.begin();
        double rn = normM*r, oprob = io->wprob;
        
        while(oprob < rn) oprob += (++io)->wprob;
        io->selected++;
      }
    }

    
    /*     assign the correct weights    */
    for(const auto& ci : _TMP_wP)
      if(ci.selected)
      {
        ci.current -> weight /= std::abs(ci.current -> weight);
        ci.current -> weight *= (ci.selected*std::pow(qcd::Nc, ci.current->overlap_color_suppression_index))*normP/nP;
      }
      else ci.current -> weight = 0.0;
    
    for(const auto& ci : _TMP_wM)
      if(ci.selected)
      {
        ci.current -> weight /= std::abs(ci.current -> weight);
        ci.current -> weight *= (ci.selected*std::pow(qcd::Nc, ci.current->overlap_color_suppression_index))*normM/nM;
      }
      else ci.current -> weight = 0.0;
    
    
    return maxdelta;
  }

  
  template<ProcessTag _Proc>
  template<class _RandomEngine>
  void shower_color_state<_Proc>::mc_reduce(_RandomEngine& rne)
  {
    /*   trivial case   */
    if(size() == 1u) return;
    
    /*   we have more than one basis state   */
    double norm = 0.0;
    
    for(const auto& ci : _M_psi)
      norm += std::abs(ci.weight)/std::pow(qcd::Nc, ci.overlap_color_suppression_index);
    
    auto selected = select_randomly(rne, _M_psi.begin(), _M_psi.end(), [=](const color_state<_Proc>& ci) -> double {
      return std::abs(ci.weight)/std::pow(qcd::Nc, ci.overlap_color_suppression_index)/norm;
    });
    
    
    for(auto it = _M_psi.begin(); it != _M_psi.end(); ++it)
      if(it == selected)
      {
        it -> weight /= std::abs(it -> weight);
        it -> weight *= std::pow(qcd::Nc, it->overlap_color_suppression_index)*norm;
        
      } else it->weight = 0.0;
    
    /*    clean up the color state    */
    reduce(1.0e-5*std::abs(selected->weight));
  }
  
}   //  namespace duct




namespace duct {
  
  
  
  /*   This class represent quantum color states.
   */
  template<ProcessTag _Proc>
  class quantum_color_state
  {
  private:
    /*   private types   */
    using _Size = std::size_t;
    using _Diff = std::ptrdiff_t;
    
    /*   number of QCD incomings   */
    static constexpr _Size _n_in = number_of_qcd_in<_Proc>;
    static constexpr _Size _n_reserve = 50u;
    
  public:
    /*   usual typedefs     */
    using parton = parton_cstate::colorlink;
    using size_type = _Size;
    using difference_type = _Diff;
    
    using reference = parton&;
    using pointer = parton*;
    using iterator = parton*;
    
    using const_reference = const parton&;
    using const_pointer = const parton*;
    using const_iterator = const parton*;
    
    using allocator_type = pmr::polymorphic_allocator<parton>;
    
  public:
    /*   weight of the color state    */
    std::complex<double> weight;
    
    
  public:
    /*    default constructors   */
    quantum_color_state()
    : quantum_color_state{allocator_type{}} {}
    
    explicit quantum_color_state(allocator_type a)
    : _M_memory{a} { _M_create_storage_for(_n_reserve);}
    
    /*   destructor   */
    ~quantum_color_state() { _M_deallocate();}
    
    
    /*    non-standard constructor  (no initialization)  */
    explicit quantum_color_state(_Size _n_qcd, allocator_type a = {})
    : _M_memory{a}
    {
      _M_create_storage_for(_n_in + _n_qcd);
      _M_lay_down_for(_n_qcd);
    }
    
    
    /*   copy   */
    quantum_color_state& operator=(const quantum_color_state&);
    
    quantum_color_state(const quantum_color_state& x, allocator_type a = {})
    : _M_memory{a}
    {
      if(x.begin())
      {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of_out());
        
        _M_copy_head(x);
        std::copy(x.begin(), x.end(), _M_begin);
      }
      else _M_begin = _M_end = _M_end_of_storage = nullptr;
    }
    
    /*   move   */
    quantum_color_state(quantum_color_state&&, allocator_type);
    quantum_color_state& operator=(quantum_color_state&&);
    
    quantum_color_state(quantum_color_state&& x)
    : _M_memory{x._M_memory}
    , _M_begin{x._M_begin}
    , _M_end{x._M_end}
    , _M_end_of_storage{x._M_end_of_storage}
    {
      _M_copy_head(x);
      x._M_begin = x._M_end = x._M_end_of_storage = nullptr;
    }
    
    
    /*    element access    */
    reference operator[](_Diff k) { return *((_M_begin+_n_in) + k-1);}
    const_reference operator[](_Diff k) const { return *((_M_begin+_n_in) + k-1);}
    
    
    /*   iterators    */
    iterator begin() { return _M_begin;}
    const_iterator begin() const { return _M_begin;}
    const_iterator cbegin() const { return _M_begin;}
    
    iterator end() { return _M_end;}
    const_iterator end() const { return _M_end;}
    const_iterator cend() const { return _M_end;}
    
    
    /*   observables   */
    bool empty() const { return (_M_begin == _M_end);}
    _Size size() const { return _M_end-_M_begin;}
    _Size capacity() const { return _M_end_of_storage-_M_begin;}
    _Diff upper() const { return _M_end-(_M_begin+_n_in);}
    _Size number_of_out() const { return static_cast<_Size>(upper());}
    
    constexpr _Diff lower() const { return 1-static_cast<_Diff>(_n_in);}
    constexpr _Size number_of_in() const { return _n_in;}
    
    
    /*   resreving storage    */
    void reserve(_Size _n_qcd)
    {
      auto count = _n_in + _n_qcd;
      if(!_M_begin) _M_create_storage_for(count);
      else if(capacity() < count) _M_rebase_reserve(_n_qcd);
    }
    
    
    /*    resizing the array    */
    void resize(_Size _n_qcd)
    {
      if(!_M_begin) {
        _M_create_storage_for(_n_in + _n_qcd);
        _M_lay_down_for(_n_qcd);
        return;
      }
      
      if(capacity() < _n_in + _n_qcd) {
        _M_rebase_resize(_n_qcd);
        return;
      }
      
      _M_lay_down_for(_n_qcd);
    }
    
    /*   resize the state by n element   */
    void resize_by(_Diff k) {
      resize(static_cast<_Size>(std::max<_Diff>(0, upper()+k)));
    }
    
    /*   shrink the array   */
    void shrink_to_fit() {
      if(_M_begin && _M_end != _M_end_of_storage)
        _M_rebase_reserve(number_of_out());
    }
    
    
    /*    functions for inserting open and closed color strings   */
    void insert_open_string(const int *, const int *);
    void insert_closed_string(const int *, const int *);
    
    /*    inserting open and closed color strings using initializer list    */
    void insert_open_string(std::initializer_list<int> ll) { insert_open_string(ll.begin(), ll.end());}
    void insert_closed_string(std::initializer_list<int> ll) { insert_closed_string(ll.begin(), ll.end());}
    
    /*    check the LC+ the color connections   */
    double chi(int k, int l) const
    {
      if(l == k) return 0.0;
      
      int ret = 0;
      auto& cL = operator[](l);
      
      if(cL.prev == k) ++ret;
      if(cL.next == k) ++ret;
      
      return ret*(!is_octet(l) ? 0.5*qcd::Cf : 0.25*qcd::Ca);
    }
    
    
    /*   test if it is triplet, antitriplet or octet   */
    bool is_octet(int l) const { return operator[](l).is_octet();}
    bool is_triplet(int l) const { return operator[](l).is_triplet();}
    bool is_antitriplet(int l) const { return operator[](l).is_antitriplet();}
    
    
  private:
    /*    allocator   */
    allocator_type _M_memory;
    
    /*    it is like an std::vector   */
    parton *_M_begin, *_M_end, *_M_end_of_storage;
    
    
  private:
    /*   create storage    */
    void _M_create_storage_for(_Size _count)
    {
      auto count = std::max(_n_reserve, _count);
      _M_begin = _M_end = _M_memory.allocate(count);
      _M_end_of_storage = _M_begin + count;
    }
    
    /*   allocate memory    */
    void _M_lay_down_for(_Size _n_qcd) {
      _M_end = _M_begin + _n_in + _n_qcd;
    }
    
    /*    rebase    */
    void _M_rebase_resize(_Size _n_qcd)
    {
      auto count = std::max(2*size(), _n_in + _n_qcd);
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_end = _M_begin + _n_in + _n_qcd;
    }
    
    void _M_rebase_reserve(_Size nqcd)
    {
      auto _n_qcd = number_of_out();
      auto count = std::max(2*size(), _n_in + nqcd);
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_end = _M_begin + _n_in + _n_qcd;
    }
    
    /*   deallocate memory    */
    void _M_deallocate() {
      if(_M_begin) _M_memory.deallocate(_M_begin, capacity());
    }
    
    /*    copy the head infomation     */
    void _M_copy_head(const quantum_color_state& x)
    {
      weight = x.weight;
    }
  };
  
  
  /*   move constructor    */
  template<ProcessTag _Proc>
  quantum_color_state<_Proc>::quantum_color_state(quantum_color_state&& x, allocator_type u)
  : _M_memory{u}
  {
    _M_begin = _M_end = _M_end_of_storage = nullptr;
    
    if(x.begin())
    {
      _M_copy_head(x);
      
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin, x._M_begin);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of_out());
        std::copy(x.begin(), x.end(), _M_begin);
      }
    }
  }
  
  
  /*  copy assignment doesn't change the memory management,
   *  it only copies the values and adjust the boundaries
   */
  template<ProcessTag _Proc>
  quantum_color_state<_Proc>& quantum_color_state<_Proc>::operator=(const quantum_color_state<_Proc>& x)
  {
    if(this != &x && x.begin())
    {
      if(auto count = x.size(); capacity() < count) {
        _M_deallocate();
        _M_create_storage_for(count);
      }
      
      _M_copy_head(x);
      _M_lay_down_for(x.number_of_out());
      std::copy(x.begin(), x.end(), _M_begin);
    }
    
    return *this;
  }
  
  
  /*   move assignment    */
  template<ProcessTag _Proc>
  quantum_color_state<_Proc>& quantum_color_state<_Proc>::operator=(quantum_color_state<_Proc>&& x)
  {
    if(this != &x && x.begin())
    {
      _M_copy_head(x);
      
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin, x._M_begin);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else
      {
        if(auto count = x.size(); capacity() < count) {
          _M_deallocate();
          _M_create_storage_for(count);
        }
        
        _M_lay_down_for(x.number_of_out());
        std::copy(x.begin(), x.end(), _M_begin);
      }
    }
    
    return *this;
  }
  
  
  template<ProcessTag _Proc>
  bool operator==(const quantum_color_state<_Proc>& x, const quantum_color_state<_Proc>& y)
  {
    if(x.size() != y.size()) return false;
    
    auto ix = x.begin();
    auto iy = y.begin();
    auto end = x.end();
    
    while(ix != end)
      if(*(ix++) != *(iy++))
        return false;
    
    return true;
  }
  
  
  template<ProcessTag _Proc>
  inline bool operator!=(const quantum_color_state<_Proc>& x, const quantum_color_state<_Proc>& y) {
    return !(x == y);
  }
  
  
  /*    inserting color strings   */
  template<ProcessTag _Proc>
  void quantum_color_state<_Proc>::insert_closed_string(const int *begin, const int *end)
  {
    quantum_color_state<_Proc>& p = *this;
    const int *i;
    
    p[*begin].prev = *(end-1);
    p[*begin].next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].prev = *(i-1);
      p[*i].next = *(i+1);
    }
    
    p[*i].prev = *(i-1);
    p[*i].next = *begin;
  }
  
  
  template<ProcessTag _Proc>
  void quantum_color_state<_Proc>::insert_open_string(const int *begin, const int *end)
  {
    quantum_color_state<_Proc>& p = *this;
    const int *i;
    
    p[*begin].prev = parton_cstate::colorlink::quarkend;
    p[*begin].next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].prev = *(i-1);
      p[*i].next = *(i+1);
    }
    
    p[*i].prev = *(i-1);
    p[*i].next = parton_cstate::colorlink::antiquarkend;
  }
  
  
  /*        Elementary color operators       */
  template<ProcessTag _Proc>
  void color_a_dagger(int sigma, int emitter, quantum_color_state<_Proc>& p)
  {
    static constexpr int QUARK = parton_cstate::colorlink::quarkend;
    static constexpr int ANTI_QUARK = parton_cstate::colorlink::antiquarkend;
    
    int spectator, m = p.upper();
    
    switch(sigma) {
      case 1:
        spectator = p[emitter].next;
        p[emitter].next  = m;
        p[m].prev = emitter;
        p[m].next = spectator;
        p[spectator].prev = m;
        break;
      case -1:
        spectator = p[emitter].prev;
        p[spectator].next = m;
        p[m].prev = spectator;
        p[m].next = emitter;
        p[emitter].prev = m;
        break;
      case 2:
        spectator = p[emitter].prev;
        p[emitter].prev  = QUARK;
        p[spectator].next  = m;
        p[m].prev = spectator;
        p[m].next = ANTI_QUARK;
        break;
      case -2:
        p[p[emitter].prev].next = p[emitter].next;
        p[p[emitter].next].prev = p[emitter].prev;
        p[emitter].prev = QUARK;
        p[emitter].next  = m;
        p[m].prev = emitter;
        p[m].next = ANTI_QUARK;
        break;
      case 3:
        spectator = p[emitter].next;
        p[emitter].next = ANTI_QUARK;
        p[spectator].prev = m;
        p[m].next = spectator;
        p[m].prev = QUARK;
        break;
      case -3:
        p[p[emitter].prev].next = p[emitter].next;
        p[p[emitter].next].prev = p[emitter].prev;
        p[emitter].next = ANTI_QUARK;
        p[emitter].prev  = m;
        p[m].next = emitter;
        p[m].prev = QUARK;
        break;
      case 4:
        p[emitter].prev  = m;
        p[m].next = emitter;
        p[m].prev = QUARK;
        break;
      case -4:
        p[emitter].next  = m;
        p[m].prev = emitter;
        p[m].next = ANTI_QUARK;
        break;
      default:
        //   do nothing
        break;
    }
  }
  
  
  
  /*    print operator for the color state   */
  template<ProcessTag _Proc>
  inline std::ostream& operator<<(std::ostream& os, const quantum_color_state<_Proc>& c)
  {
    os
    <<"Weight                   : "<<c.weight<<'\n'
    <<"-------------------------------------\n";
    for(auto k = c.lower(); k <= c.upper(); ++k)
      os<<k<<": |"<<c[k]<<">\n";
    
    return os;
  }
  
  
  /*    check for tadpole graphs   */
  template<ProcessTag _Proc>
  bool check_for_tadpole(quantum_color_state<_Proc>& c)
  {
    /*  for SU(3) set the weight to zero if there is any tadpole   */
    int lo = c.lower(), hi = c.upper();
    
    for(int l = lo; l <= hi; ++l)
      if(c[l].is_octet() && c[l].next == l)
      {
        c.weight = 0.0;
        return true;
      }
    
    return false;
  }
  
  
}   //    namespace duct

#endif
