/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 23/02/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_ordering_kT_h__
#define __mc_ordering_kT_h__ 1


/*   std headers   */
#include <cmath>
#include <utility>

/*   project headers   */
#include <cxx/math-support.h>


namespace duct {
  
  
  /*   Defining the ordering dependent functions for the INITIAL state splittings  */
  template<>
  struct ordering_traits<ini, ordering::kT>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*    calculates the z limit   */
      static double zmax(const hard_params<ini> *, double v) {
        return 4.0/sqr(std::sqrt(4.0+v)+std::sqrt(v));
      }
      
      /*  calculates the limits on the variable v    */
      static void vlimits(double& vmin, double& vmax, const hard_params<ini> *pars) {
        vmin = pars->mT;
        vmax = 1.0/pars->eta-2.0 + pars->eta;
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y
       *    returns y and the corresponding jacobian
       */
      static std::pair<double, double> mapping_to_y(const hard_params<ini> *, double v, double z) {
        return {v/(1.0-z), 1.0};
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y
       *    returns only y (no jacobian)
       */
      static double mapping_to_y_nojac(const hard_params<ini> *, double v, double z) {
        return v/(1.0-z);
      }
      
      /*  helps to define the shower time: exp(-t) = v/vnull   */
      static double vnull(double Q2, const lorentzvector<double>& qnull, const lorentzvector<double>&) {
        return qnull.mag2()/Q2;
      }
    };
  
    
    /*    these functions gives the v and z limits in the integrals of the phi color state    */
    struct color_evolution
    {
      /*  calculates the limits on the variable v    */
      static _v_range vlimits(const hard_params<ini> *pars) {
        return {pars->mT, std::min({1.0/pars->eta-2.0 + pars->eta, pars->mTmax})};
      }
      
      /*   limits on the z integral   */
      static _z_range zlimits(const hard_params<ini> *pars, double v) {
        return {pars->eta, 4.0/sqr(std::sqrt(4.0+v)+std::sqrt(v))};
      }
    };
    
    
    
    /*    these functions gives the v and z limits in the integrals of threshold corrections    */
    struct threshold
    {
      /*  calculates the limits on the variable v    */
      static _v_range vlimits(const hard_params<ini> *pars) {
        return {pars->mT, 1.0e+100};
      }
      
      /*   limits on the z integral in the main part  */
      struct main
      {
        static _z_range zlimits(const hard_params<ini> *, double v) {
          return {4.0/sqr(std::sqrt(v) + std::sqrt(4.0+v)), 1.0};
        }
      };
      
      /*   limits on the z integral in the soft part  */
      struct soft
      {
        static _z_range zlimits(const hard_params<ini> *, double v) {
          return {0, 4.0/sqr(std::sqrt(v) + std::sqrt(4.0+v))};
        }
      };
    };
  };
  
  
  
  /*   Defining the ordering dependent functions for the FINAL state splittings  */
  template<>
  struct ordering_traits<fin, ordering::kT>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*  calculates the limits on the variable v    */
      static void vlimits(double& vmin, double& vmax, const hard_params<fin> *pars) {
        vmin = pars->mT;
        vmax = 0.25*sqr(pars->ymax)/(pars->ymax + pars->bl)*0.9998;
      }
      
      /*  calculates the z limits  */
      static double zmax(const hard_params<fin> *pars, double v) {
        double ymax = 0.9999*pars->ymax;
        return 0.5*(1.0 + std::sqrt(1.0-4.0*v*(ymax + pars->bl)/sqr(ymax)));
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y
       *  It also returns the jacobian of  v --> y mapping.
       */
      static std::pair<double, double> mapping_to_y(const hard_params<fin> *pars, double v, double z) {
        double Y = v/(z*(1.0-z));
        double yb = 0.25*sqr(std::sqrt(Y) + std::sqrt(Y+4.0*pars->bl));
        return {yb-pars->bl, yb/(yb+pars->bl)};
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y   */
      static double mapping_to_y_nojac(const hard_params<fin> *pars, double v, double z) {
        double Y = v/(z*(1.0-z));
        double yb = 0.25*sqr(std::sqrt(Y) + std::sqrt(Y+4.0*pars->bl));
        return yb-pars->bl;
      }
      
      /*  helps to define the shower time: exp(-t) = v/vnull   */
      static double vnull(double pQ, const lorentzvector<double>& qnull, const lorentzvector<double>&) {
        return qnull.mag2()/pQ;
      }
    };
    
    
    /*    these functions gives the v and z limits in the integrals of the phi color state    */
    struct color_evolution
    {
      /*  calculates the limits on the variable v    */
      static _v_range vlimits(const hard_params<fin> *pars) {
        return {pars->mT*1.0001, std::min(0.25*sqr(pars->ymax)/(pars->ymax + pars->bl)*0.9998, pars->mTmax)};
      }
      
      /*   calculate the z limits   */
      static auto zlimits(const hard_params<fin> *pars, double v)
      {
        stackvector<_z_range, 2u> ret;
        
        double ymax = 0.9999*pars->ymax;
        double zmax = 0.5*(1.0 + std::sqrt(1.0-4.0*v*(ymax + pars->bl)/sqr(ymax)));
        
        ret.push_back({1.0-zmax, zmax});
        
        return ret;
      }
    };
    
  };
  
}

#endif
