/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 19/02/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_shower_stage_h__
#define __mc_shower_stage_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-event.h>
#include <bits/mc-color-state.h>
#include <bits/mc-ordering.h>
#include <bits/mc-engine-opts.h>
#include <cxx/memory-resource.h>



namespace duct {
  

  
  /*   Simple square matrix to be able to implement the chi matrix for the full color state   */
  template<ProcessTag _Proc>
  class chi_matrix
  {
    /*   private types   */
    using _Size = std::size_t;
    using _Diff = std::ptrdiff_t;
    
    /*   number of QCD incomings   */
    static constexpr _Size _n_in = number_of_qcd_in<_Proc>;
    static constexpr _Diff _n_lo = 1-static_cast<_Diff>(number_of_qcd_in<_Proc>);
    static constexpr _Size _n_reserve = 100u;
    
  public:
    /*   we want to manage the memory   */
    using allocator_type = pmr::polymorphic_allocator<double>;
    
    /*   default constructor   */
    chi_matrix() : chi_matrix{allocator_type{}} {}
    
    explicit chi_matrix(allocator_type a)
    : _M_size{0}, _M_shift{_n_lo}, _M_chi{a}
    {
      _M_chi.reserve(_n_reserve*_n_reserve);
    }
    
    
    /*   copy   */
    chi_matrix(const chi_matrix&) = default;
    chi_matrix& operator=(const chi_matrix&) = default;
    
    chi_matrix(const chi_matrix& c, allocator_type a)
    : _M_size{c._M_size}, _M_shift{c._M_shift}, _M_chi{c._M_chi, a} {}
    
    
    /*    move    */
    chi_matrix(chi_matrix&&) = default;
    chi_matrix& operator=(chi_matrix&&) = default;
    
    chi_matrix(chi_matrix&& c, allocator_type a)
    : _M_size{c._M_size}, _M_shift{c._M_shift}, _M_chi{std::move(c._M_chi), a} {}
    
    
    /*   element access    */
    const double& operator()(_Diff k, _Diff l) const { return _M_chi[_M_size*k+l-_M_shift];}
    
    /*   structural information   */
    constexpr _Diff lower() const { return _n_lo;}
    _Diff upper() const { return static_cast<_Diff>(_M_size)-static_cast<_Diff>(_n_in);}
    
    
    /*    resizes the matrix   */
    void resize(_Size _n_qcd)
    {
      _M_size  = _n_qcd+_n_in;
      _M_shift = static_cast<_Diff>(_M_size+1)*_n_lo;
      _M_chi.resize(_M_size*_M_size);
    }
    
    
    /*   reseting the matrix    */
    void reset() {
      for(double& x : _M_chi) x = 0.0;
    }
    
    
    /*    update the chi matrix   */
    void update_chi(const shower_color_state<_Proc>& c)
    {
      /*   nothing to do   */
      if(c.empty()) return;
      
      auto ic = c.begin();
      auto last = c.end();
      
      /*   reseting chi matrix   */
      auto lo = ic->lower();
      auto hi = ic->upper();
      
      resize(ic->number_of_out());
      reset();
      
    
      /*   update chi matrix   */
      double norm = 0.0, wprob;
      
      while(ic != last)
      {
        wprob = std::abs(ic->weight)/std::pow(qcd::Nc, ic->overlap_color_suppression_index);
        norm += wprob;
        
        for(auto k = lo; k <= hi; ++k)
          for(auto l = lo; l <= hi; ++l)
            elem(k,l) += wprob*ic->chi(k,l);
        
        ++ic;
      }
      
      /*   normalization   */
      for(double& x : _M_chi)
        x /= norm;
    }

    
    void update_xi(const shower_color_state<_Proc>& c)
    {
      /*   nothing to do   */
      if(c.empty()) return;
      
      auto& ci = *c.begin();
      
      /*   reseting chi matrix   */
      auto lo = ci.lower();
      auto hi = ci.upper();
      
      resize(ci.number_of_out());
      reset();
      
      for(auto k = lo; k <= hi; ++k)
        for(auto l = lo; l <= hi; ++l)
          elem(k,l) = ci.xi(k,l);
    }
    

  private:
    /*   size parameters    */
    _Size _M_size;
    _Diff _M_shift;
    
    /*   storage   */
    pmr::vector<double> _M_chi;
    
    /*    non-const element access   */
    double& elem(_Diff k, _Diff l) { return _M_chi[_M_size*k+l-_M_shift];}
  };
  
  

  
  /*
   *    Simple class for providing the kT and virtuality veto parameters.
   */
  struct shower_vetos
  {
    /*  This is maximum allowed transverse momentum square in the initial
     *  and final state splitting. These usually set by the hard state.
     */
    double max_k2T = std::numeric_limits<double>::max();
    
    /*  This is maximum allowed virtuality in the initial
     *  state splitting. These usually set by the hard state.
     */
    double max_q2 = std::numeric_limits<double>::max();
  };
  
 
  
  /*   This class defines a set of weights those are carried through the shower evolution.   */
  struct weight_set
  {
    struct weight_t
    {
      /*   Main weight
       *   In the current implementation this weight comes from the
       *   hard event and doesn't changes during the shower evolution.
       */
      double main;
      
      /*   Threshold correction factor
       *   This threshold correction factor comes from the LC+ threshold
       *   shower operator. There are also additive threshold corrections
       *   and those are counted in the color state.
       */
      double threshold;
    };
    
    /*   returns the full weight    */
    explicit operator double() const { return born.main*born.threshold + nlo.main*nlo.threshold;}

   
    /*   At hard level this is the Born term.   */
    weight_t born;
    
    /*   At hard level this gives the contributions of the NLO matrix elements.
     *   We have to do the matching to have this.
     */
    weight_t nlo;
  };
  
  
  /*    binary I/O for the weight_set class   */
  inline std::ostream& write(std::ostream& os, const weight_set::weight_t& w) {
    return duct::write(os, w.main, w.threshold);
  }
  
  inline std::istream& read(std::istream& is, weight_set::weight_t& w) {
    return duct::read(is, w.main, w.threshold);
  }
  
  inline std::ostream& write(std::ostream& os, const weight_set& w) {
    return duct::write(os, w.born, w.nlo);
  }
  
  inline std::istream& read(std::istream& is, weight_set& w) {
    return duct::read(is, w.born, w.nlo);
  }
  
  
  
  
  
  /*  Class for representing a stage of the shower evolution.
   *  This class can be seen by the users in the user function.
   *  It consists of
   *   state         => the event
   *   cstates       => list of color states
   *   scale         => scale of the event (equivalent to current time)
   *   weight        => weight of the current event
   */
  template<ProcessTag _Proc>
  struct shower_stage
  {
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::polymorphic_allocator<>;
    
    /*   default constructors    */
    shower_stage() = default;
    explicit shower_stage(allocator_type a) : state{a}, cstate{a} {}
    
    
    /*   copy   */
    shower_stage(const shower_stage&) = default;
    shower_stage& operator=(const shower_stage&) = default;
    
    shower_stage(const shower_stage& x, allocator_type a)
    : state{x.state, a}
    , cstate{x.cstate, a}
    , scale{x.scale}
    , weights{x.weights}
    {}

    /*    move  */
    shower_stage(shower_stage&&) = default;
    shower_stage& operator=(shower_stage&&) = default;
    
    shower_stage(shower_stage&& x, allocator_type a)
    : state{std::move(x.state), a}
    , cstate{std::move(x.cstate), a}
    , scale{std::move(x.scale)}
    , weights{std::move(x.weights)}
    {}
    
    
    /*  The event of the state. This contains every informations about the
     *  event, momenta, flavors, color connections, incoming objects,...
     */
    event<_Proc> state;
    
    
    /*   We can have more than one color basis state associated to a momentum and
     *   flavor configuration. Thus we have a vector (technically a list) of color states.
     */
    shower_color_state<_Proc> cstate;
    
    
    /*   The scale of the shower stage. It is typically corresponds to the
     *   shower time, but at the hard state we don't have emitter parton and
     *   the definition of the shower time is ambigous...
     */
    double scale;
    
    /*  The posibble weight of the event. This weight can come from color
     *  factors or from spin correlation (when we add spin correlations).
     */
    weight_set weights;
  };

  
  /*    binary I/O for the shower stage class    */
  template<ProcessTag _Proc>
  inline std::ostream& write(std::ostream& os, const shower_stage<_Proc>& s) {
    return duct::write(os, s.state, s.cstate, s.scale, s.weights);
  }
  
  template<ProcessTag _Proc>
  inline std::istream& read(std::istream& is, shower_stage<_Proc>& s) {
    return duct::read(is, s.state, s.cstate, s.scale, s.weights);
  }
  
  
  
  
  /*  Class for representing a stage of the shower evolution.
   *  In this class we record more information and they are basically
   *  used only during the evolution.
   *
   *  Inhereted from
   *                 => shower_stage
   *
   *  and it consists of
   *   vars          => spitting variable for the next stage
   *   qnull         => reference momentum for defining shower time
   *   current_time  => shower time when the event was created
   *   next_time     => shower time of the next splitting
   *   vetos         => veto parameters
   *   params        => hard parameters
   *   averaged_chi  => averaged chi matrix
   */
  template<ProcessTag _Proc>
  class shower_history_stage : public shower_stage<_Proc>
  {
    /*   private types   */
    using _Base = shower_stage<_Proc>;
    
  public:
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::polymorphic_allocator<>;
    
    /*   default constructors    */
    shower_history_stage() = default;

    explicit shower_history_stage(allocator_type a)
    : shower_stage<_Proc>{a}, params{a}, averaged_chi{a} {}
    
    
    /*   copy   */
    shower_history_stage(const shower_history_stage&) = default;
    shower_history_stage& operator=(const shower_history_stage&) = default;
  
    shower_history_stage(const shower_history_stage& x, allocator_type a)
    : shower_stage<_Proc>{x, a}
    , vars{x.vars}
    , qnull{x.qnull}
    , current_time{x.current_time}
    , next_time{x.next_time}
    , vetos{x.vetos}
    , params{x.params, a}
    , averaged_chi{x.averaged_chi, a}
    {}
    
    /*    move  */
    shower_history_stage(shower_history_stage&&) = default;
    shower_history_stage& operator=(shower_history_stage&&) = default;
    
    shower_history_stage(shower_history_stage&& x, allocator_type a)
    : shower_stage<_Proc>{std::move(x), a}
    , vars{std::move(x.vars)}
    , qnull{std::move(x.qnull)}
    , current_time{std::move(x.current_time)}
    , next_time{std::move(x.next_time)}
    , vetos{std::move(x.vetos)}
    , params{std::move(x.params), a}
    , averaged_chi{std::move(x.averaged_chi), a}
    {}
    
    
    /*   returns the scale (this is a derived quantity)   */
    double time_to_scale() const {
      return qnull.mag2()*std::exp(-current_time);
    }
    
    
    /*   update hard parameters    */
    template<OrderingTag _Ordering>
    void update_params(evolution_type type, double kT_cutoff, double virt_cutoff)
    {
      /*    update the chi matrix according to the evolution type   */
      switch(type) {
        case evolution_type::typeI: averaged_chi.update_chi(_Base::cstate); break;
        case evolution_type::typeII: averaged_chi.update_xi(_Base::cstate); break;
      };
      
      /*    update the hard parameters    */
      if constexpr(is_hadron_collision<_Proc>)
        _M_params_ini<_Ordering>(kT_cutoff, virt_cutoff);
      
      _M_params_fin<_Ordering>(kT_cutoff, virt_cutoff);
    }
    
    
    /*  Splitting varibles: This gives the splitting varibles of the next stage.
     *  It consists of y, z, phi, emitter, spectator and flavors.
     */
    splitting_vars vars;
    
    /*  This is the refernce momentum that help us to define shower time.
     *  We have a copy in very stage, since we want to resume shower from
     *  an intermediate state. In passive view this vector is boosted in
     *  every step of the shower.
     */
    lorentzvector<double> qnull;
    
    /*  The shower time when the event was created.
     */
    double current_time;
    
    /*  The shower time. This state is evolved without emission upto this time.
     */
    double next_time;
    
    
    /*   Since a history can have more than one hard state, we have to add
     *   the veto parameters to this class. The won't change during the
     *   evolution (I cannot imagine such a scenario... but who knows...).
     *   Maybe the upper limits on kT  or virtuality can change...
     *   Anyway, it is better to keep it here.
     */
    shower_vetos vetos;
    
    /*   In the shower evolution we always need some derived parameters.
     *   We can save some CPU time by calculating them only once and store them here.
     */
    parameters<_Proc> params;
    
    
    /*   Averaged chi matrix   */
    chi_matrix<_Proc> averaged_chi;
    
    
  private:
    /*   update final state parameters    */
    template<OrderingTag> void _M_params_ini(double, double);
    template<OrderingTag> void _M_params_fin(double, double);
  };

  
  template<ProcessTag _Proc>
  template<OrderingTag _Ordering>
  void shower_history_stage<_Proc>::_M_params_ini(double kT_cutoff, double virt_cutoff)
  {
    /*   the lower index of the array   */
    static constexpr int lo = 1-static_cast<int>(number_of_incoming_hadrons<_Proc>);
    
    /*   nice aliases    */
    auto& p = _Base::state;
    auto mdl = p.get_model();
    
    
    /*   making space  */
    params.ini.resize(number_of_incoming_hadrons<_Proc>);
    
    
    /*  Some simple information   */
    int hi = p.upper();
    auto Q = momentum_sum<incomings>(p);
    double Q2 = Q.mag2(), m2T = std::max({mdl.alpha_qcd().q2min(), sqr(kT_cutoff)}), q2cut = sqr(virt_cutoff);
    
    
    for(int emitter = lo; emitter < 1; emitter++)
    {
      /*    setting up the parameters for this emitter   */
      hard_params<ini> &pars = params.ini[emitter-lo];
      
      
      /*   model and pdf tables   */
      pars.mdl = mdl;
      pars.aux = mdl.aux(hadron{p[hadron{emitter}].key});
      
      /*   basic variables   */
      pars.emitter = emitter;
      pars.fl = p[emitter].flavor;
      pars.Q2 = Q2;
      pars.eta = p.eta(emitter);
      pars.mT = std::max(m2T, pars.aux->thrsT(pars.fl))/Q2;
      pars.mY = q2cut/Q2;
      
      /*  Variable to be able to map the shower time to v   */
      pars.vnull = ordering_traits<ini, _Ordering>::main::vnull(Q2, qnull, p[emitter].momentum);
      
      /*   veto parameters   */
      pars.mTmax = vetos.max_k2T/Q2;
      pars.mYmax = vetos.max_q2/Q2;
      
      
      /*   clear the spectator list   */
      pars.spectators.clear();
      pars.spectators.reserve(hi-lo+1);
      
      
      /*    considering all the possible spectator    */
      int ib = (emitter == -1 ? 0 : -1);

      for(int spectator = lo; spectator <= hi; ++spectator)
        if(spectator != emitter)
        {
          double xi = 0.0, xT = 0.5, zk = 0.0;
          
          if(spectator > 0)
          {
            double ska = (p[spectator].momentum*p[emitter].momentum);
            double skb = (p[spectator].momentum*p[ib].momentum);
            double m2k = p[spectator].momentum.mag2();
            
            zk = 1.0-2.0*skb/Q2;
            xi = skb/(ska+skb);
            xT = std::sqrt(std::abs(1.0-m2k*Q2/(4.0*ska*skb)));
          }
          
          pars.spectators.push_back({spectator, zk, xi, xT, averaged_chi(spectator, emitter)});
        }
    }
  }
  
  
  
  template<ProcessTag _Proc>
  template<OrderingTag _Ordering>
  void shower_history_stage<_Proc>::_M_params_fin(double kT_cutoff, double virt_cutoff)
  {
    /*   the lower index of the array   */
    static constexpr int lo = 1-static_cast<int>(number_of_incoming_hadrons<_Proc>);
    
    /*   nice aliases    */
    auto& p = _Base::state;
    auto mdl = p.get_model();
    
    
    /*   making space  */
    int hi = p.upper();
    params.fin.resize(hi);
    
    
    /*  Some simple information   */
    auto Q = momentum_sum<incomings>(p);
    double Q2 = Q.mag2(), m2T = std::max({mdl.alpha_qcd().q2min(), sqr(kT_cutoff)}), q2cut = sqr(virt_cutoff);
    
    
    /*   loop over all the possible final state emitters    */
    for(int emitter = 1; emitter <= hi; ++emitter)
    {
      hard_params<fin> &pars = params.fin[emitter-1];
      
      /*   is it decay?  &  model   */
      pars.is_decay = is_decay_process<_Proc>;
      pars.mdl = mdl;
      
      /*   basic parameters   */
      pars.emitter = emitter;
      pars.fl = p[emitter].flavor;
      
      pars.pQ = 2.0*(p[emitter].momentum*Q);
      pars.mT = m2T/pars.pQ;
      pars.mY = q2cut/pars.pQ;

      pars.al = Q2/pars.pQ;
      pars.bl = std::abs(p[emitter].momentum.mag2())/pars.pQ;
      pars.rl = std::sqrt(1.0-4.0*pars.al*pars.bl);
      pars.ymax = sqr((1.0-pars.bl)/(std::sqrt(pars.al)+std::sqrt(std::abs(pars.al+pars.bl-1.0)))) - pars.bl;
      
      
      /*  Variable to be able to map the shower time to v   */
      pars.vnull = ordering_traits<fin, _Ordering>::main::vnull(pars.pQ, qnull, p[emitter].momentum);
      
      /*   veto parameters   */
      pars.mTmax = vetos.max_k2T/pars.pQ;
      pars.mYmax = vetos.max_q2/pars.pQ;
      
      
      
      /*   check if it is a special case   */
      bool is_special = (std::abs(1.0-pars.ymax/pars.rl) < 1.0e-5);
      
      auto pl = (2.0*pars.al/(1.0+pars.rl))*p[emitter].momentum;
      auto nl = Q - pl; pl -= ((1.0-pars.rl)/(1.0+pars.rl))*Q;
      
      
      /*   clear the spectator list   */
      pars.spectators.clear();
      pars.spectators.reserve(hi-lo+1);
      
      
      /*    considering all the possible spectator    */
      for(int spectator = lo; spectator <= hi; ++spectator)
        if(spectator != emitter)
        {
          double chikl = averaged_chi(spectator, emitter);
          
          if(pars.is_decay && 0 == spectator)
            pars.spectators.push_back({spectator, 0.5, 0.25, chikl});
          else if(is_special && spectator > 0)
            pars.spectators.push_back({spectator, 1.0, 0.0, chikl});
          else
          {
            double kQ = (Q*p[spectator].momentum)*2.0*pars.rl/(1.0+pars.rl);
            double xi = (pl*p[spectator].momentum)/kQ;
            double xT = std::abs(p[spectator].momentum.mag2())*(pl*nl)/(kQ*kQ);
            
            pars.spectators.push_back({spectator, xi, xT, chikl});
          }
        }
    }
  }
  
}

#endif
