/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 21.09.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_soft_function_h__
#define __mc_soft_function_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-shower-stage.h>
#include <cxx/math/chebyshev.h>



namespace duct {
  
  namespace __details {
    
    /*    Gauss integrals    */
    inline const gauss_integral<10u> _cevol_integralZ{};
  }   //   namespace __details

  
  
  /*   We have a chebyshev interpolation for the soft functions
   *   We interpolate for every emitter, thus the function is an array.   *
   */
  template<ProcessTag _Proc>
  class soft_function
  {
    using _Diff = std::ptrdiff_t;
    using _Size = std::size_t;
    
    /*   order of the chebyshev interpolation    */
    static constexpr _Size _n_cheb = 10u;
    
    using chebyshev_t = std::array<double, _n_cheb+1u>;
    using _Chebyshev = chebyshev_interpolation_basic<_n_cheb>;
    
    
    /*   the lower index of the array   */
    static constexpr _Diff _n_low = 1-static_cast<_Diff>(number_of_incoming_hadrons<_Proc>);
    
    /*   Every emitter has different interpolation range in tau
     *   and we need some other emitter related parameters
     *   related parameter.
     */
    struct emitter_pars {
      bool is_zero;
      double tauP, tauM, Q2null;
    };
    
    
    /*   matrix to store the interpolation of the full soft function   */
    class matrix_t
    {
    public:
      /*    resizing the matrix    */
      void resize(_Diff upper)
      {
        _M_upper = upper;
        _M_nrow = upper-_n_low + 1;
        _M_shift = (1 + _M_nrow)*_n_low;
        _M_sudakov.resize(_M_nrow*_M_nrow);
        _M_delta.resize(_M_nrow*_M_nrow);
        _M_pars.resize(_M_nrow);
      }
      
      /*   reset the matrix    */
      void reset()
      {
        for(chebyshev_t& ch : _M_sudakov) ch.fill(0.0);
        for(chebyshev_t& ch : _M_delta) ch.fill(0.0);

        for(emitter_pars& p : _M_pars)
          p = {true, 0.0, 0.0, 0.0};
      }
      
      /*    element acces   */
      chebyshev_t * sudakov_exp(_Diff l) { return _M_sudakov.data() + _M_nrow*l - _M_shift;}
      const chebyshev_t * sudakov_exp(_Diff l) const { return _M_sudakov.data() + _M_nrow*l - _M_shift;}
      
      chebyshev_t * delta_vop(_Diff l) { return _M_delta.data() + _M_nrow*l - _M_shift;}
      const chebyshev_t * delta_vop(_Diff l) const { return _M_delta.data() + _M_nrow*l - _M_shift;}
      
      
      /*   acces to the emitter related parameters   */
      emitter_pars& params(_Diff l) { return _M_pars[l-_n_low];}
      const emitter_pars& params(_Diff l) const { return _M_pars[l-_n_low];}
      
      
      /*   upper and lower indices     */
      constexpr _Diff lower() const { return _n_low;}
      _Diff upper() const { return _M_upper;}
      
      
      /*   make the chebyshev coefficients   */
      void calculate_coefficients()
      {
        chebyshev_t tmp;
        
        for(chebyshev_t& ch : _M_sudakov) {
          _Chebyshev::calculate_coefficients(tmp.data(), ch.data());
          ch = tmp;
        }

        for(chebyshev_t& ch : _M_delta) {
          _Chebyshev::calculate_coefficients(tmp.data(), ch.data());
          ch = tmp;
        }
      }
      
    private:
      /*   structural parameters   */
      _Diff _M_nrow, _M_upper, _M_shift;
      
      /*   data is strored in a vector   */
      std::vector<chebyshev_t> _M_sudakov, _M_delta;
      
      /*   every emitter has its own parameters   */
      std::vector<emitter_pars> _M_pars;
    };
    
    
  public:
    /*   public static informations   */
    using chebyshev_grid_t = _Chebyshev;
    
    /*    matrix to store the result    */
    class result_t
    {
    public:
      /*    resizing the matrix    */
      void resize(_Diff upper)
      {
        _M_size = upper-_n_low + 1;
        _M_shift = _n_low*(_M_size+1);
        _M_lmd.resize(_M_size*_M_size);
      }
      
      /*   element acces   */
      double * operator[](_Diff l) { return _M_lmd.data() - _M_shift + _M_size*l;}
      const double * operator[](_Diff l) const { return _M_lmd.data() - _M_shift + _M_size*l;}
      
      /*   setting all elements to zero    */
      void reset() {
        for(double& x : _M_lmd) x = 0.0;
      }
      
      /*   structural information   */
      constexpr _Diff lower() const { return _n_low;}
      _Diff upper() const { return _M_size + _n_low - 1;}
      
      
      /*   the coefficients of the i*pi terms    */
      double omega;
      
      
    private:
      /*   structural parameters   */
      _Diff _M_size, _M_shift;
      
      /*   data is strored in a vector   */
      std::vector<double> _M_lmd;
    };
    
    
    /*   initialize the intrpolation    */
    template<class _InitFunc>
    void init(const shower_history_stage<_Proc>& stage)
    {
      /*   set the coupling   */
      _M_as = std::addressof(stage.state.get_model().alpha_qcd());
      _M_q2null = stage.qnull.mag2();
      
      /*   resize the matrix   */
      _M_cs.resize(stage.state.upper());
      _M_cs.reset();
      
      _M_tmL =  1.0e+180;
      _M_tmH = -1.0e+180;
      
      /*   initializes the interpolation   */
      _InitFunc::do_initialize(*this, stage);
      
      
      /*   setting up the interpolational polinomials   */
      _M_cs.calculate_coefficients();
    }
    
    
    /*   doing the interpolations   */
    void function(result_t& r, double tmT) const
    {
      auto upper = _M_cs.upper();
      
      /*    calculate the integral of the soft functions over the evolution range [tmT, tmH]  */
      r.resize(upper);
      
      for(_Diff L = _n_low; L <= upper; ++L)
        _M_function(r[L], L, tmT);
      
      r.omega = 2.0*M_PI*alphas_ipi(tmT);
    }
    
    void integral(result_t& r, double tmT) const
    {
      auto upper = _M_cs.upper();
      
      /*    calculate the integral of the soft functions over the evolution range [tmT, tmH]  */
      r.resize(upper);
      
      for(_Diff L = _n_low; L <= upper; ++L)
        _M_integral(r[L], L, tmT);

      r.omega = 2.0*M_PI*tau_ipi(tmT);
    }
    
    /*    integral over the whole interpolation range   */
    void integral(result_t& r) const
    {
      auto upper = _M_cs.upper();
      
      /*    calculate the integral of the soft functions over the evolution range [tmT, tmH]  */
      r.resize(upper);
      
      for(_Diff L = _n_low; L <= upper; ++L)
        _M_integral(r[L], L);
  
      r.omega = 2.0*M_PI*tau_ipi(_M_tmL);
    }
    
    
    
    /*   available range   */
    double tm_min() const { return _M_tmL;}
    double tm_max() const { return _M_tmH;}
    
    /*   coupling for the i*pi terms    */
    double alphas_ipi(double tm) const {
      return _M_as->operator()(_M_q2null*std::exp(-tm));
    }
    
    double tau_ipi(double tm) const {
      return (_M_tmH > tm ? _M_as->tau(_M_q2null*std::exp(-_M_tmH)) - _M_as->tau(_M_q2null*std::exp(-tm)) : 0.0);
    }

    double tau_ipi(double tmH, double tmL) const {
      return (tmH > tmL ? _M_as->tau(_M_q2null*std::exp(-tmH)) - _M_as->tau(_M_q2null*std::exp(-tmL)) : 0.0);
    }
    
    /*   direct access to the Chebyshev intepolational gridpoints   */
    chebyshev_t * sudakov_exp(_Diff l) { return _M_cs.sudakov_exp(l);}
    const chebyshev_t * sudakov_exp(_Diff l) const { return _M_cs.sudakov_exp(l);}

    chebyshev_t * delta_vop(_Diff l) { return _M_cs.delta_vop(l);}
    const chebyshev_t * delta_vop(_Diff l) const { return _M_cs.delta_vop(l);}
    

    /*   acces to the emitter related parameters   */
    emitter_pars& params(_Diff l) { return _M_cs.params(l);}
    const emitter_pars& params(_Diff l) const { return _M_cs.params(l);}
    
    
    /*   sets the smallest & largest times   */
    void tm_min(double tm) {
      if(tm < _M_tmL) _M_tmL = tm;
    }
    
    void tm_max(double tm) {
      if(tm > _M_tmH) _M_tmH = tm;
    }
    
    
  private:
    /*   we need the strong coupling    */
    const qcd_coupling *_M_as;
    double _M_q2null;
    
    /*    storage for interpolation    */
    matrix_t _M_cs;
    
    /*   range of shower time where there is at least one splitting function that is defined   */
    double _M_tmL, _M_tmH;
    
  private:
    /*   doing the interpolations   */
    void _M_function(double *, _Diff, double) const;
    void _M_integral(double *, _Diff, double) const;
    
    /*    integral over the whole interpolation range   */
    void _M_integral(double *ret, _Diff l) const
    {
      auto row = _M_cs.sudakov_exp(l);
      const double &jac = _M_cs.params(l).tauM;
      
      for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
        ret[k] = jac*_Chebyshev::integral(row[k].data());
    }
  };
  
  
  template<ProcessTag _Proc>
  double soft_exponent(const typename soft_function<_Proc>::result_t& r, const quantum_color_state<_Proc>& ci)
  {
    /*   calculate the \lambda({c}) exponent   */
    double lcc = 0.0;
    auto upper = r.upper();
    
    for(auto L = r.lower(); L <= upper; L++)
      if(ci.is_triplet(L)) lcc += 0.5*qcd::Cf*r[L][ci[L].next];
      else if(ci.is_antitriplet(L)) lcc += 0.5*qcd::Cf*r[L][ci[L].prev];
      else lcc += 0.25*qcd::Ca*(r[L][ci[L].prev] + r[L][ci[L].next]);
  
    return lcc;
  }
  
  
  template<ProcessTag _Proc>
  double soft_exponent(const typename soft_function<_Proc>::result_t& r, const color_state<_Proc>& ci)
  {
    /*   calculate the \lambda({c',c}) exponent   */
    double lcc = 0.0;
    auto upper = r.upper();
    
    for(auto L = r.lower(); L <= upper; L++)
      for(auto K = r.lower(); K <= upper; K++)
        if(L != K) lcc += r[L][K]*ci.chi(K,L);
    
    return lcc;
  }
  
  
  template<ProcessTag _Proc>
  double soft_exponent(const typename soft_function<_Proc>::result_t& r, const chi_matrix<_Proc>& c)
  {
    double lcc = 0.0;
    const auto upper = r.upper(), lower = r.lower();
    
    for(auto L = lower; L <= upper; L++)
      for(auto K = lower; K <= upper; K++)
        if(L != K) lcc += r[L][K]*c(K,L);
    
    return lcc;
  }
  
  
  template<ProcessTag _Proc>
  void soft_function<_Proc>::_M_function(double *ret, _Diff l, double tm) const
  {
    /*   it is a zero function   */
    auto& pars = _M_cs.params(l);
    
    if(pars.is_zero)
    {
      for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
        ret[k] = 0.0;
      
      return;
    }
    
    /*    the function is well defined    */
    double tau = _M_as->tauk(pars.Q2null*std::exp(-tm));
    double y = (tau - pars.tauP)/pars.tauM;
    
    
    /*   outside of the interpolation region it returns zero   */
    if(-1.0 > y || y > 1.0)
    {
      for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
        ret[k] = 0.0;
      
      return;
    }
    
    /*   interpolate    */
    auto row = _M_cs.delta_vop(l);
    double as = _M_as->ask(tau);
    
    for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
      ret[k] = as*_Chebyshev::function(row[k].data(), y);
  }
  
  
  template<ProcessTag _Proc>
  void soft_function<_Proc>::_M_integral(double *ret, _Diff l, double tm) const
  {
    /*   it is a zero function   */
    auto& pars = _M_cs.params(l);
    
    if(pars.is_zero)
    {
      for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
        ret[k] = 0.0;
      
      return;
    }
    
    
    /*    the function is well defined    */
    const double &jac = _M_cs.params(l).tauM;
    double y = (_M_as->tauk(pars.Q2null*std::exp(-tm)) - pars.tauP)/jac;
    
    
    /*   outside of the interpolation region the integrand is zero   */
    if(y < -1.0) y = -1.0;
    else if(y >= 1.0)
    {
      for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
        ret[k] = 0.0;
      
      return;
    }
    
    /*   interpolate    */
    auto row = _M_cs.sudakov_exp(l);
    
    for(_Diff k = _M_cs.lower(); k <= _M_cs.upper(); ++k)
      ret[k] = jac*_Chebyshev::integral_lower(row[k].data(), y);
  }
  
  
  
  /*   helper functions to initialize soft_function   */
  namespace __details {

    template<ProcessTag _Proc, OrderingTag _Ordering>
    void virt_color_op_init_fin(soft_function<_Proc>& sf, const hard_params<fin>& pars, double tmH, double tmL)
    {
      /*   aliases    */
      const qcd_coupling& as = pars.mdl.alpha_qcd();
      const double &vnull = pars.vnull, &Q2 = pars.pQ;
      const int &emitter = pars.emitter;
      
      
      /*    kinematical constraints for the splitting     */
      auto[vmin, vmax] = ordering_traits<fin, _Ordering>::color_evolution::vlimits(&pars);
      
      double vH = std::min(vmax, vnull*std::exp(-tmL));
      double vL = std::max(vmin, vnull*std::exp(-tmH));
      
      
      /*    the function is zero    */
      if(vL >= vH) return;
      
      
      sf.tm_max(std::log(vnull/vL));
      sf.tm_min(std::log(vnull/vH));
      
      
      /*   mimicing the same safty cuts that were used in the H operator    */
      double kinL = as.tauk(Q2*vmax)*1.005, kinH = as.tauk(Q2*vmin)*0.995;
      
      /*   setting the emitter related parametes of the interpolation matrix    */
      double tauL = std::max(kinL, as.tauk(Q2*vH*0.999)), tauH = std::min(kinH, as.tauk(Q2*vL));
      double tauP = 0.5*(tauH + tauL), tauM = 0.5*(tauH - tauL);
      
      if(tauH <= tauL) return;
      
      sf.params(emitter) = {false, tauP, tauM, Q2*vnull};
      
      
      /*    loop over the Chebyshev base points   */
      auto delta_vop = sf.delta_vop(emitter);
      auto sudakov_exp = sf.sudakov_exp(emitter);
      auto base = soft_function<_Proc>::chebyshev_grid_t::base();
      
      
      for(std::size_t k = 0; k < base.size; ++k)
      {
        double tau = base[k]*tauM + tauP;
        double q2 = as.q2(tau);
        double v = q2/Q2;
        
        auto z_ranges = ordering_traits<fin, _Ordering>::color_evolution::zlimits(&pars, v);
        
        /*   nothing to do   */
        if(z_ranges.empty()) continue;
        
        
        /*   otherwise calculate the z integrals   */
        double asb = 0.0, ask = as.ask(tau);
        double ras = as.as(tau)/ask;
        
        if constexpr(static_cast<unsigned>(_Ordering::alpha))
          asb = static_cast<unsigned>(_Ordering::alpha)*ask*qcd::beta1(as.nf(tau));
        
        
        for(auto [zmin, zmax] : z_ranges)
        {
          double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
          
          for(const auto& gauss : _cevol_integralZ.basepoints())
          {
            double fac = Jm, asfac = 1.0, z = 1.0-(1.0-zmin)*std::exp(-(1.0+gauss.x)*Jm);
            auto [y, yjac] = ordering_traits<fin, _Ordering>::main::mapping_to_y(&pars, v, z);
            
            fac *= yjac;
            fac *= gauss.weight;
            
            if constexpr(static_cast<unsigned>(_Ordering::alpha))
              asfac /= 1.0+asb*std::log(1.0-z);
            
            
            /*   loop over spectators   */
            for(const auto& spec : pars.spectators)
            {
              double tmp = fac*splitting_functions_base<fin>::parton_gluon(&pars, spec, y, z);
              
              delta_vop[spec.spectator][k] += ras*tmp;
              sudakov_exp[spec.spectator][k] += asfac*tmp;
            }
          }
        }
      }
    }
    
    
    template<ProcessTag _Proc, OrderingTag _Ordering>
    void virt_color_op_init_ini(soft_function<_Proc>& sf, const hard_params<ini>& pars, double tmH, double tmL)
    {
      /*   aliases    */
      const qcd_coupling& as = pars.mdl.alpha_qcd();
      const double &vnull = pars.vnull, &Q2 = pars.Q2, &eta = pars.eta;
      const int &emitter = pars.emitter, fa = -pars.fl;
      
      
      /*    kinematical constraints for the splitting     */
      auto[vmin, vmax] = ordering_traits<ini, _Ordering>::color_evolution::vlimits(&pars);
      
      double vH = std::min(vmax, vnull*std::exp(-tmL));
      double vL = std::max(vmin, vnull*std::exp(-tmH));
      
      
      /*    the function is zero    */
      if(vL >= vH) return;
      
      sf.tm_max(std::log(vnull/vL));
      sf.tm_min(std::log(vnull/vH));
      
      
      /*   setting the emitter related parametes of the interpolation matrix    */
      double tauL = as.tauk(Q2*vH*0.999), tauH = as.tauk(Q2*vL*1.01);
      double tauP = 0.5*(tauH + tauL), tauM = 0.5*(tauH - tauL);
      
      if(tauH <= tauL) return;
      
      sf.params(emitter) = {false, tauP, tauM, Q2*vnull};
      
      
      
      /*    loop over the Chebyshev base points   */
      auto delta_vop = sf.delta_vop(emitter);
      auto sudakov_exp = sf.sudakov_exp(emitter);
      auto base = soft_function<_Proc>::chebyshev_grid_t::base();
      
      
      for(std::size_t k = 0; k < base.size; ++k)
      {
        double tau = base[k]*tauM + tauP;
        double q2 = as.q2(tau);
        double v = q2/Q2;
        
        /*   the z limits are scheme dependent    */
        auto [zmin, zmax] = ordering_traits<ini, _Ordering>::color_evolution::zlimits(&pars, v);
        
        
        /*   nothing to do   */
        if(zmin >= zmax) continue;
        
        
        /*   pdf and alpha_s factors     */
        double pdf = pars.aux->parton(fa, eta, q2);
        double asb = 0.0, ask = as.ask(tau);
        double ras = as.as(tau)/ask;
        
        if constexpr(static_cast<unsigned>(_Ordering::alpha))
          asb = static_cast<unsigned>(_Ordering::alpha)*ask*qcd::beta1(as.nf(tau));
        
        
        /*     jacobians    */
        double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
        
        for(const auto& gauss : _cevol_integralZ.basepoints())
        {
          double fac = Jm, asfac = 1.0, z = 1.0-(1.0-zmin)*std::exp(-(1.0+gauss.x)*Jm);
          auto [y, yjac] = ordering_traits<ini, _Ordering>::main::mapping_to_y(&pars, v, z);
          
          fac *= yjac;
          fac *= gauss.weight;
          fac *= pars.aux->parton(fa, eta/z, q2)/pdf;
          
          if constexpr(static_cast<unsigned>(_Ordering::alpha))
            asfac /= 1.0+asb*std::log(1.0-z);
          
          
          /*   loop over spectators   */
          for(const auto& spec : pars.spectators)
          {
            double tmp = fac*splitting_functions_base<ini>::parton_gluon(spec, y, z);
            
            delta_vop[spec.spectator][k] += ras*tmp;
            sudakov_exp[spec.spectator][k] += asfac*tmp;
          }
        }
      }
    }
    
    
    
    template<ProcessTag _Proc, OrderingTag _Ordering>
    void virt_color_op_init_thr_ini(soft_function<_Proc>& sf, const hard_params<ini>& pars, double tmH, double tmL)
    {
      /*   aliases    */
      const qcd_coupling& as = pars.mdl.alpha_qcd();
      const double &vnull = pars.vnull, &Q2 = pars.Q2, &eta = pars.eta;
      const int &emitter = pars.emitter, fa = -pars.fl;
      
      
      /*    kinematical constraints for the splitting     */
      auto[vmin, vmax] = ordering_traits<ini, _Ordering>::threshold::vlimits(&pars);
      
      double vH = std::min(vmax, vnull*std::exp(-tmL));
      double vL = std::max(vmin, vnull*std::exp(-tmH));
      
      
      /*    the function is zero    */
      if(vL >= vH) return;
      
      sf.tm_max(std::log(vnull/vL));
      sf.tm_min(std::log(vnull/vH));
      
      
      /*   setting the emitter related parametes of the interpolation matrix    */
      double tauL = as.tauk(Q2*vH*0.999), tauH = as.tauk(Q2*vL*1.01);
      double tauP = 0.5*(tauH + tauL), tauM = 0.5*(tauH - tauL);
      
      if(tauH <= tauL) return;
      
      sf.params(emitter) = {false, tauP, tauM, Q2*vnull};
      
      
      /*    loop over the Chebyshev base points   */
      auto delta_vop = sf.delta_vop(emitter);
      auto sudakov_exp = sf.sudakov_exp(emitter);
      auto base = soft_function<_Proc>::chebyshev_grid_t::base();
      
      
      for(std::size_t k = 0; k < base.size; ++k)
      {
        double tau = base[k]*tauM + tauP;
        double q2 = as.q2(tau);
        double v = q2/Q2;
        
        /*   the z limits are scheme dependent    */
        auto [zmin, zmax] = ordering_traits<ini, _Ordering>::threshold::soft::zlimits(&pars, v);
        
        
        /*   nothing to do   */
        if(zmin >= zmax) continue;
        
        
        /*   pdf and alpha_s factors     */
        double pdf = pars.aux->parton(fa, eta, q2);
        double asb = 0.0, ask = as.ask(tau);
        double ras = as.as(tau)/ask;
        
        if constexpr(static_cast<unsigned>(_Ordering::alpha))
          asb = static_cast<unsigned>(_Ordering::alpha)*ask*qcd::beta1(as.nf(tau));
        
        
        /*     jacobians    */
        double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
        
        for(const auto& gauss : _cevol_integralZ.basepoints())
        {
          double fac = Jm, asfac = 1.0, pdffac = 1.0, z = 1.0-(1.0-zmin)*std::exp(-(1.0+gauss.x)*Jm);
          auto [y, yjac] = ordering_traits<ini, _Ordering>::main::mapping_to_y(&pars, v, z);
          
          fac *= yjac;
          fac *= gauss.weight;
          
          if constexpr(static_cast<unsigned>(_Ordering::alpha))
            asfac /= 1.0+asb*std::log(1.0-z);
          
          /*   PDF factor   */
          if(z > pars.eta)
            pdffac -= pars.aux->parton(fa, pars.eta/z, q2)*z/pdf;
          
          
          /*    soft splitting function in the threshold terms    */
          auto __parton_gluon_thr = [=](double xi) -> double {
            double delta = (1.0+z*y)*(1.0+y)*xi;
            double od = 1.0-delta, x2 = sqr(z*y/(1.0-z));
            
            /*   NOTE: the minus sign is important to get the correct sign
             *   when we exponentiate in class virtual_shower_color_operator
             */
            return -2.0*z*(1.0+y)/(1.0+z*y)*(1.0 - od/std::sqrt(sqr(od) + 4.0*x2*delta));
          };
          
          
          /*   loop over spectators   */
          for(const auto& spec : pars.spectators)
            if(spec.spectator > 0 && spec.xi > 1.0e-6)
            {
              double tmp = pdffac*fac*__parton_gluon_thr(spec.xi);
              
              delta_vop[spec.spectator][k] += ras*tmp;
              sudakov_exp[spec.spectator][k] += asfac*tmp;
            }
        }
      }
    }
    
  }   //   namespace __details
  
  

  
  /*    helps to initialize the chebyshev interpolations    */
  template<OrderingTag _Ordering>
  struct virtual_color_operator_initializer
  {
    struct shower
    {
      /*   function to do the initialization of the gauss interpolation   */
      template<ProcessTag _Proc>
      static void do_initialize(soft_function<_Proc>& sf, const shower_history_stage<_Proc>& stage)
      {
        /*   calculate the function in the grid points (initial state)   */
        if constexpr(is_hadron_collision<_Proc>)
          for(const auto& pars : stage.params.ini)
            __details::virt_color_op_init_ini<_Proc, _Ordering>(sf, pars, stage.next_time, stage.current_time);
        
        
        /*   calculate the function in the grid points (final state)   */
        for(const auto& pars : stage.params.fin)
          __details::virt_color_op_init_fin<_Proc, _Ordering>(sf, pars, stage.next_time, stage.current_time);
      }
      
      
      /*   this is not threshold contribution    */
      static constexpr bool is_threshold_contribution = false;
    };
    
    struct threshold
    {
      /*   function to do the initialization of the gauss interpolation   */
      template<ProcessTag _Proc>
      static void do_initialize(soft_function<_Proc>& sf, const shower_history_stage<_Proc>& stage)
      {
        if constexpr(is_hadron_collision<_Proc>)
          for(const auto& pars : stage.params.ini)
            __details::virt_color_op_init_thr_ini<_Proc, _Ordering>(sf, pars, stage.next_time, stage.current_time);
      }
      
      /*   this is threshold contribution    */
      static constexpr bool is_threshold_contribution = true;
    };
  };
  
}   //   namespace duct

#endif 
