/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitfunc_base_h__
#define __mc_splitfunc_base_h__ 1

/*   std includes   */

/*   project headers   */
#include <bits/qcd-color.h>
#include <bits/mc-hard-params.h>
#include <cxx/stackvector.h>


namespace duct {

  /*   Helper class to implement the default splitting functions for shower   */
  template<class _Type>
  struct splitting_functions;
  
  
  /*  Final state splitting functions for shower evolution.
   *  This representations is rather cumbersome, but it is optimalized.
   *  It is useful to define some adaptors to call these functions.
   */
  template<>
  struct splitting_functions<fin>
  {
    /*   small class to store the result for each channels   */
    struct result
    {
      int emitter, spectator;
      int fl, fm;
      double weight;
    };
    
    /*   calculates all the possible splitting functions    */
    auto splitfunc(const hard_params<fin>&, double, double) -> stackvector<result, 6u>;
    
    
    
    /*   Class represents the f -> f + g processes in the final state.   */
    struct emitting_gluon
    {
      /*   calculate the splitting function (azimuthal averaged) and set the state
       *   No color factor included in the return value!!!
       */
      double operator()(const lorentzvector<double>&, const lorentzvector<double>&,
                        const lorentzvector<double>&, double, double);

      /*   generate the azimuthal angle    */
      double inverse_phi(double) const;
      
    private:
      /*   The functions to generate the phi azimuthal angle.   */
      double function_phi(double) const;
      double primitive_phi(double) const;
      
      /*   parameters of the current state   */
      bool is_simple;
      double A, B, C, S, I1, R1;
    };
  };
  
  
  
}   /*   namespace duct    */

















#if 0
namespace duct {

  namespace __details {
   
    /*   Helper class to implement the default splitting functions for shower   */
    template<class _Type>
    struct splitting_functions_base;

    
    /*  Initial state splitting functions for shower evolution.
     *  This representations is rather cumbersome, but it is optimalized.
     *  It is useful to define some adaptors to call these functions.
     */
    template<>
    struct splitting_functions_base<ini>
    {
      /*  This class represents the f -> f + g splitting in the initial state (soft and soft-coll part only).  */
      class emitting_gluon
      {
      public:
        /*   calculate the splitting function (azimuthal averaged) and set the state
         *   No color factor included in the return value!!!
         */
        double operator()(double, double, double, double);
        
        /*   generate the azimuthal angle    */
        double inverse_phi(double) const;
        
      private:
        /*   The function and its integral   */
        double function_phi(double) const;
        double primitive_phi(double) const;
        
       /*    cached variables    */
        bool isSpec;
        double w2, R1, R2, R3, R4;
        double Ap, Am, Bp, Bm, Cp, Cm, rAp, rAm, rCp, rCm;
      };

  
      /*  This is the splitting function of the g -> g + g in the initial state (pure collinear part only).  */
      static double gluon_gluon(double y, double z) {
        double oza = (1.0 + y)*z;
        double oz = 1.0-z, za = 1.0-oza;
        
        return 2.0*qcd::Ca*za*(za*za/oza + (oz+za)/oz)/oza;
      }
      
      /*  This is the splitting function of the q -> q + g in the initial state (pure collinear part only).  */
      static double quark_gluon(double y, double z) {
        return qcd::Cf*(1.0/z-1.0-y);
      }
      
      /*  This is the splitting function of the q -> g + q in the initial state.  */
      static double gluon_quark(double y, double z)
      {
        double xa = 1.0/z-1.0-y;
        double w = 1.0-xa*z;
        double retval = 1.0 + 2.0*xa*z/(w*w);
        
        /*  The color factor is Tr and we have a factor
         *  from the color averaging in the initial state, Na/Nc
         */
        retval *= 0.5*qcd::Na/qcd::Nc;
        
        return retval;
      }
      
      /*  This is the splitting function of the g -> q + qbar in the initial state.  */
      static double quark_antiquark(double y, double z)
      {
        double zxa = 1.0-z*(1.0+y);
        double retval = (1.0-2.0*(1.0-zxa)*zxa)/z;
        
        /*  The color factor is Cf and the we have a factor
         *  from the color averaging in the initial state, Nc/Na
         */
        retval *= qcd::Cf*qcd::Nc/qcd::Na;
        
        return retval;
      }
      
      
      
      /*  This is the splitting function for f->f+g (soft-collinear only).     */
      static double parton_gluon(const hard_params<ini> *, double, double);

      /*  This is the splitting function for f->f+g (soft-collinear only). 
       *  We use this in the color evolution to calculate the phi_exponent.
       */
      static double parton_gluon(const hard_params<ini>::spectator_vars&, double, double);
    };
    

    
    /*  Final state splitting functions for shower evolution.
     *  This representations is rather cumbersome, but it is optimalized.
     *  It is useful to define some adaptors to call these functions.
     */
    template<>
    struct splitting_functions_base<fin>
    {
      /*   Class represents the f -> f + g processes in the final state.   */
      struct emitting_gluon
      {
        /*   calculate the splitting function (azimuthal averaged) and set the state
         *   No color factor included in the return value!!!
         */
        double operator()(const hard_params<fin> *, const hard_params<fin>::spectator_vars&, double, double);
        
        /*   generate the azimuthal angle    */
        double inverse_phi(double) const;
        
      private:
        /*   The functions to generate the phi azimuthal angle.   */
        double function_phi(double) const;
        double primitive_phi(double) const;
        
        /*   parameters of the current state   */
        bool is_simple;
        double A, B, C, S, I1, R1;
      };

      
      /*  This is the splitting function for g -> g + g process (collinear only).  */
      static double gluon_gluon(const hard_params<fin> *, double, double);
      
      /*  This is the splitting function for q -> q + g process (collinear only).  */
      static double quark_gluon(const hard_params<fin> *, double, double);
      
      /*  This is the splitting function for g -> q + qbar process.  */
      static double quark_antiquark(const hard_params<fin> *, double, double);
      static double quark_antiquark(int, const hard_params<fin> *, double, double);
 
    
      /*   This is the splitting function for f -> f + g processes (soft and soft-collinear part only).   */
      static double parton_gluon(const hard_params<fin> *, double, double);
      
      
      /*  This is the splitting function for f->f+g (soft-collinear only).
       *  We use this in the color evolution to calculate the phi_exponent.
       */
      static double parton_gluon(const hard_params<fin> *, const hard_params<fin>::spectator_vars&, double, double);
    };
    
  }   /*   namespace __details    */

}   /*   namespace duct    */

#endif


#endif

