/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitfunc_h__
#define __mc_splitfunc_h__ 1

/*   std includes   */
#include <array>

/*   project headers   */
#include <bits/mc-splitfunc-base.h>
#include <bits/mc-splitfunc-approx.h>
#include <cxx/stackvector.h>
#include <cxx/math-support.h>



namespace duct {

  
  template<class _Ordering>
  class splitting_prob<fin, _Ordering>
  {
    /*   size of the interpolation grid    */
    static constexpr unsigned int _N_grid = 8u;
    
  public:
    /*    public types    */
    using ordering_type = _Ordering;

    /*    simple class to handle z mappings    */
    struct z_mapping
    {
      /*    binding to the hard parameters and to the v variable    */
      void bind(const hard_params<fin>& pars, double v)
      {
        /*    set the z boundaries   */
        auto[_zmin, _zmax] = ordering_traits<fin, _Ordering>::main::zlimits(pars, v);
     
        zmin = _zmin;
        zmax = _zmax;
        zjac = (zmin < zmax ? std::log(zmax/zmin) : 0.0);
      }
      
      /*    do the mapping    */
      std::pair<double, double> w2z(double w) const
      {
        double z = 1.0-zmax*std::exp(-w*zjac);
        return {z, (1.0-z)*zjac};
      }
      
      /*   is possible   */
      bool is_possible() const { return (zmax > zmin);}
      
      double zmin, zmax, zjac;
    };
    

    /*   possible splitting state  */
    struct state
    {
      /*    check if the splitting possible with a given v   */
      bool is_possible(double v) const {
        return (vmin < v && v < vmax);
      }
      
      /*    hard parameters   */
      hard_params<fin> pars;
      
      /*   available range in variable v   */
      double vmin, vmax;
      
      /*   splitting function object  */
      stackvector<splitting_functions_base<fin>::result, 6u> weights;
      
      /*   z mapping   */
      z_mapping zmap;
    
      /*   hardness and momentum fraction variables   */
      double y, z;
    };
      
    
    /*   storing the details of the last call   */
    struct lastcall_t
    {
      /*    list of all the possible states    */
      std::vector<state> states;
      
      /*    total weight of the states   */
      double weight_tot = 0.0;
      
      /*   the v and u variables   */
      double vprev = -1.0, uprev = -1.0;
    };
    
  public:
    /*   default constructor   */
    splitting_prob()
    : _M_mdl{get_default_model()}
    , _M_vmin{1.0}, _M_vmax{0.0}
    , _M_z_spline{_N_grid}
    {}

    /*   with custom model resource    */
    explicit splitting_prob(model_ref mdl)
    : _M_mdl{mdl}
    , _M_vmin{1.0}, _M_vmax{0.0}
    , _M_z_spline{_N_grid}
    {}
    
    
    /*   binding the functions to the hard state   */
    template<ProcessTag _Proc>
    bool bind(const shower_stage<_Proc>&);
    
    /*    calculates the full splitting function    */
    double operator()(double, double);
    
    
    /*   binding the approx spline  to the current v variable    */
    void init_z_spline(double v) {
      _M_z_spline.assign(unitweight{}, [=, this](double w) -> double {
        return this -> operator()(v, w);
      });
    }
    
    /*   hit or miss ratio   */
    double hit_or_miss_ratio(double v, double rz)
    {
      /*    initializing the z-spline    */
      _M_z_spline.assign(unitweight{}, [=, this](double w) -> double {
        return this -> operator()(v, w);
      });
      
      /*    generating the variable for the z integral from a random number    */
      double u = _M_z_spline.integral_inverse(rz*_M_z_spline.integral());
     
      /*   returning the ration of the exact and the appoximated functions    */
      return operator()(v, u)/_M_z_spline(u);
    }
    
    
  private:
    /*   model   */
    model_ref _M_mdl;
    
    /*   v limits   */
    double _M_vmin, _M_vmax;
 
    /*   details of the last call   */
    lastcall_t _M_last_call;
 
    
  private:
    /*   spline of the approximation     */
    spline<spline_type::constant> _M_z_spline;
 
    /*   spline for the tau grid   */
    spline<spline_type::linear> _M_v_spline;
  };
    
  
  template<class _Ordering> template<ProcessTag _Proc>
  bool splitting_prob<fin, _Ordering>::bind(const shower_stage<_Proc>& hstage)
  {
    /*   nice aliases    */
    constexpr int lo = 1-static_cast<int>(number_of_incoming_hadrons<_Proc>);
    const auto& p = hstage.state;
    int hi = p.upper();


    /*   make a fresh start   */
    _M_states.clear();
    _M_last_call._weight_tot = 0.0;
    _M_last_call.vprev = -1.0;
    _M_last_call.uprev = -1.0;
    _M_vmin = 1.0e+120;
    _M_vmax = 0.0;

    
    /*   scale of the previous stage  */
    auto Q = momentum_sum<incomings>(p);
    double Q2 = Q.mag2(), Q2null = mag2(hstage.qnull);
    double vperv = hstage.scale/Q2null;
    
    
    /*    calculating the hard parameters for each legs    */
    for(int emitter = 1; emitter <= p.upper(); ++emitter)
    {
      state& current = _M_last_call.states.emplace_back();
      hard_params<fin>& pars = current.pars;
      const auto& pl = p[emitter].momentum;
            
      /*   basic parameters   */
      pars.emitter = emitter;
      pars.fl = p[emitter].flavor;
      
      pars.pQ = 2.0*(pl*Q);
      pars.Q2null = Q2null;
      pars.pQnull = 2.0*(pl*hstage.qnull);
      
      pars.mT = m2T/pars.pQ;
      pars.al = Q2/pars.pQ;
      pars.ymax = 1.0/sqr(std::sqrt(pars.al)+std::sqrt(std::abs(pars.al-1.0)));
      
      /*   veto parameters   */
      pars.mTmax = hstage.vetos.max_k2T/pars.pQ;
      pars.mYmax = hstage.vetos.max_q2/pars.pQ;

      
      /*   the allowed region in the v variable   */
      auto[vmin, vmax] = ordering_traits<fin, _Ordering>::main::vlimits(pars);
      if(vmax > vprev) vmax = vprev;
    
      if(vmin < vmax)
      {
        /*    adjusting the possible limits in v   */
        if(_M_vmin > vmin) _M_vmin = vmin;
        if(_M_vmax < vmax) _M_vmax = vmax;
        
        /*    possible spectators   */
        pars.spectators.clear();
        
        for(int spectator = lo; spectator <= hi; ++spectator)
          if(double chi = p.chi(spectator, emitter); chi > 0.1)
            pars.spectators.push_back({spectator,
              (spectator != emitter ? pars.al*(p[spectator].momentum*pl)/(p[spectator].momentum*Q) : 0.0),
              chi});
        
        /*   caching the vmin and vmax variables    */
        current.vmin = vmin;
        current.vmax = vmax;
      }
      else _M_last_call.states.pop_back();
    }
 
    /*   the splitting is not possible    */
    if(_M_last_call.states.empty()) return false;
    
    
    /*     setting up the v-spline     */
    
    
    return true;
  }


  template<class _Ordering>
  double splitting_prob<fin, _Ordering>::operator()(double v, double u)
  {
    /*   trival case  */
    if(_M_vmin > _M_vmax) return 0.0;
    if(v < _M_vmin || _M_vmax < v) return 0.0;
    
    if(_M_last_call.vprev == v && _M_last_call.uprev == u)
      return _M_last_call.weight_tot;
    
    
    /*   binding z mappings to v   */
    if(_M_last_call.vprev != v)
    {
      for(auto& sf : _M_last_call.states)
        if(sf.is_possible(v))
          sf.zmap.bind(sf.pars, v);
      
      _M_vprev = v;
    }
    
    
    /*    calculating the function   */
    _M_last_call.uprev = u;
    _M_last_call.weight_tot = 0.0;
    
    for(auto& sf : _M_last_call.states)
      if(sf.is_possible(v) && sf.zmap.is_possible())
      {
        auto[z, zjac] = sf.zmap.w2z(u);
        auto[y, yjac] = ordering_traits<fin, _Ordering>::main::mapping_to_y(pars, v, z);
        
        sf.z = z;
        sf.y = y;
        sf.weights = splitting_functions_base<fin>::splitfunc(sf.pars, y, z);
    
        /*    adjusting the weights   */
        for(auto& w : sf.weights)
          _M_last_call.weight_tot += (w.weight *= zjac*yjac);
      }
      else sf.weights.clear();
    
    return _M_last_call.weight_tot;
  }
    
    
  
  
  
  

}  //  namespace duct

#endif
  
  
  
  
  
  
  
  
  
  

#if 0

namespace duct {

  
  template<class _Ordering>
  struct splitting_functions
  {
    /*    public types    */
    using ordering_type = _Ordering;
    
    /*   The initial state splitting functions     */
    class initial_state
    {
      /*    type aliases   */
      using _Base = __details::splitting_functions_base<ini>;
      using _EmitG = __details::splitting_functions_base<ini>::emitting_gluon;

    public:
      /*    class for representing the soft state   */
      class soft_state
      {
      public:
        /*   constructor   */
        soft_state(const hard_params<ini>::spectator_vars& vars, double y, double z, double fac)
        {
          _M_weight = _M_state(vars.xi, vars.xT, y, z);
          _M_weight *= fac*vars.cfac;
          _M_spectator = vars.spectator;
        }
        
        /*    return the index of the spectator parton   */
        constexpr int spectator() const { return _M_spectator;}
        
        /*   gets full weight with the color factor   */
        constexpr double weight() const { return _M_weight;}
        
        /*   gets the azimuthal angle from a random number   */
        double azimuthal_angle(double rho) const { return _M_state.inverse_phi(rho);}
        
      private:
        /*   private data members   */
        int _M_spectator;
        double _M_weight;
        _EmitG _M_state;
      };
      
      /*    calculating the spilitting function   */
      double operator()(splitting_type, const hard_params<ini> *, double, double);
      
      /*   get the pdf ratio   */
      double pdf_ratio(int k) const { return _M_pdf_ratio[k+5];}
      
      /*   returns the soft states   */
      const std::vector<soft_state>& soft_states() const { return _M_soft;}
      
      /*    Defines the vetos. This is needed when the hard part is not uniquely defined
       *    like in process p + p -> jets.
       */
      static bool veto(const hard_params<ini> *pars, double v, double z)
      {
        double y = ordering_traits<ini,_Ordering>::main::mapping_to_y_nojac(pars, v, z);
        return ((1.0-z)*y < pars->mTmax && y < pars->mYmax);
      }

    private:
      /*   stores the soft gluon states for the possible spectators   */
      std::vector<soft_state> _M_soft;
      
      /*    array to store the pdf ration when the emitter is gluon and it turns to a quark    */
      std::array<double,11> _M_pdf_ratio;
    };
    
    
    /*   The final state splitting functions     */
    class final_state
    {
      /*    type aliases   */
      using _Base = __details::splitting_functions_base<fin>;
      using _EmitG = __details::splitting_functions_base<fin>::emitting_gluon;
      
    public:
      /*    class for representing the soft state   */
      class soft_state
      {
      public:
        /*    constructor   */
        soft_state(const hard_params<fin> *pars, const hard_params<fin>::spectator_vars& vars, double y, double z, double fac)
        {
          _M_weight = _M_state(pars, vars, y, z);
          _M_weight *= fac*vars.cfac;
          _M_spectator = vars.spectator;
        }
        
        /*    return the index of the spectator parton   */
        int spectator() const { return _M_spectator;}
        
        /*   gets full weight with the color factor   */
        double weight() const { return _M_weight;}
        
        /*   gets the azimuthal angle from a random number   */
        double azimuthal_angle(double rho) const { return _M_state.inverse_phi(rho);}
        
      private:
        /*   private data members   */
        int _M_spectator;
        double _M_weight;
        _EmitG _M_state;
      };
      
      /*   calculating the soft part   */
      double operator()(splitting_type, const hard_params<fin> *, double, double);
    
      /*   returns the soft states   */
      const std::vector<soft_state>& soft_states() const { return _M_soft;}
      
      /*   returns the collinear probabilities   */
      const stackvector<double, 7>& collinear_weights() const { return _M_coll;}
    
      /*    Defines the vetos. This is needed when the hard part is not uniquely defined
       *    like in process p + p -> jets.
       */
      static bool veto(const hard_params<fin> *pars, double v, double z)
      {
        double y = ordering_traits<fin,_Ordering>::main::mapping_to_y_nojac(pars, v, z);
        return (z*(1.0-z)*y*y/(y+pars->bl) < pars->mTmax && y < pars->mYmax);
      }
      
    private:
      /*   stores the soft gluon states for the possible spectators   */
      std::vector<soft_state> _M_soft;
 
      /*   in the collinear splitting this array gives the probabilities of each channels    */
      stackvector<double, 7> _M_coll;
   };
    
    
    /*   The approximation of the splitting function. This classes can do the mapping
     *   from uniform random numbers to the splitting variables.
     */
    struct approx {
      using initial_state = __details::splitting_functions_approx<ini, _Ordering>;
      using final_state = __details::splitting_functions_approx<fin, _Ordering>;
    };
  };

  
  template<class _Ordering>
  double splitting_functions<_Ordering>::initial_state::
  operator()(splitting_type idx, const hard_params<ini> *pars, double v, double z)
  {
    /*   transform variables from (v,z) to (y,z)    */
    auto [y, jac] = ordering_traits<ini,_Ordering>::main::mapping_to_y(pars, v, z);
  
    
    /*    calculate the splitting functions   */
    switch(idx)
    {
      case splitting_type::soft:
      {
        /*   make sopace for the soft states   */
        _M_soft.clear();
        
        /*    loop over the possible spectators   */
        for(const auto& spec : pars->spectators)
          if(spec.cfac > 1e-10) _M_soft.push_back({spec, y, z, jac});
        
        return std::accumulate(_M_soft.begin(), _M_soft.end(), 0.0,
                               [](double sum, const soft_state& s) {
                                 return sum + s.weight();
                               });
      } break;
        
      case splitting_type::collinear:
        return jac*(pars->fl ? _Base::quark_gluon(y, z) : _Base::gluon_gluon(y, z));
        break;
        
      case splitting_type::flavor_changing_collinear:
      {
        /*   quark emitter    */
        if(pars->fl) return jac*_Base::quark_antiquark(y, z);
        
        /*   gluon emitter    */
        else {
          double singlet = 0.0;
          
          for(int k = -3; k <= 5; ++k)
            if(k) singlet += (_M_pdf_ratio[k+5] = pars->aux->parton(-k, pars->eta/z, pars->Q2*v));
          
          singlet += (_M_pdf_ratio[0] = _M_pdf_ratio[10]);
          singlet += (_M_pdf_ratio[1] = _M_pdf_ratio[9]);
          _M_pdf_ratio[5] = 0.0;
          
          for(double& r : _M_pdf_ratio)
            r /= singlet;
          
          return jac*_Base::gluon_quark(y, z);
        }
      } break;
    }
    
    return 0.0;
  }

  
  template<class _Ordering>
  double splitting_functions<_Ordering>::final_state::
  operator()(splitting_type idx, const hard_params<fin> *pars, double v, double z)
  {
    /*   transform variables from (v,z) to (y,z)    */
    auto [y, jac] = ordering_traits<fin,_Ordering>::main::mapping_to_y(pars, v, z);

    
    /*    calculate the splitting functions   */
    switch(idx) {
      case splitting_type::soft:
      {
        /*   make sopace for the soft states   */
        _M_soft.clear();
        
        /*    loop over the possible spectators   */
        for(const auto& spec : pars->spectators)
          if(spec.cfac > 1e-10) _M_soft.push_back({pars, spec, y, z, jac});
        
        return std::accumulate(_M_soft.begin(), _M_soft.end(), 0.0,
                               [](double sum, const soft_state& s) {
                                 return sum + s.weight();
                               });
      } break;
        
      case splitting_type::collinear:
      {
        /*    quark emitter   */
        if(pars->fl)
        {
          _M_coll.resize(1u);
          _M_coll[0] = jac*_Base::quark_gluon(pars, y, z);
          return _M_coll[0];
        }
 
        /*   gluon emitter    */
        else {
          _M_coll.resize(7u);
 
          /*   g -> g + g  (pure collinear)    */
          _M_coll[0] = jac*_Base::gluon_gluon(pars, y, z);
          
          /*   g -> q + q~      */
          double norm = _M_coll[0], wqq = jac*_Base::quark_antiquark(81, pars, y, z);
          
          /*    we have five light flavors in this splitting...    */
          for(int k = 1; k < 6; ++k) norm += (_M_coll[k] = wqq);
          
          /*   and we only consider the top quark contribution in g -> Q + Q~ vertex   */
          norm += (_M_coll[6] = jac*_Base::quark_antiquark(6, pars, y, z));
          
          return norm;
        }
      } break;
        
      default:
        return 0.0;
    }
  }

  
  
  
}   /*   namespace duct    */


#endif

