/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 28.03.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_threshold_h__
#define __mc_threshold_h__ 1



/*   std includes   */
#include <cmath>


/*   project headers   */
#include <bits/mc-ordering.h>



namespace duct {
 
  namespace __details {
    
    /*   gauss integral for threshold calculations    */
    inline const gauss_integral<10u> _threshold_integral{};

    
    /*   the full soft function    */
    inline double __parton_gluon_sum(const hard_params<ini> *pars, double y, double z)
    {
      /*   spectator independent quantities   */
      double retval = 0.0, x2 = 4.0*sqr(z*y/(1.0-z));
      
      /*   loop over spectators   */
      for(const auto& spec : pars->spectators)
        if(spec.spectator > 0 && spec.xi > 1.0e-6)
        {
          //  NOTE:  The xi in the write up is 1.0-spec.xi
          double delta = (1.0+z*y)*(1.0+y)*spec.xi;
          double od = 1.0-delta;
          
          retval += spec.cfac*(1.0 - od/std::sqrt(sqr(od) + x2*delta));
        }
      
      return 2.0*z*z*(1.0+y)/(1+z*y)*retval;
    }
    
    
    
    /*  This is the splitting function of the g -> g + g in the initial state (pure collinear part only).  */
    inline double gluon_gluon(double y, double z)
    {
      double oza = (1.0 + y)*z;
      double oz = 1.0-z, za = 1.0-oza;
      
      return 2.0*qcd::Ca*(z*za*(za*za/oza + (oz+za)/oz)/oza - (oz/z+z*oz));
    }
    
 
    /*  This is the splitting function of the q -> g + q in the initial state.  */
    inline double gluon_quark(double y, double z)
    {
      double xa = 1.0/z-1.0-y;
      double w = 1.0-xa*z;
      double retval = 2.0*xa*z/(w*w) - 2.0*(1.0-z)/z/z;
      
      /*  The color factor is Tr and we have a factor
       *  from the color averaging in the initial state, Na/Nc
       */
      retval *= 0.5*qcd::Na/qcd::Nc;
      
      return z*retval;
    }
    
    
    
    
    
    /*   integrand of the exponent   */
    template<OrderingTag _Ordering>
    double threshold(const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      
      double q2 = as.q2(tau);
      double v = q2/pars->Q2;
      
      /*   the z limits are scheme dependent    */
      auto [zmin, zmax] = ordering_traits<ini, _Ordering>::threshold::main::zlimits(pars, v);
      auto [umin, umax] = ordering_traits<ini, _Ordering>::threshold::soft::zlimits(pars, v);
      
      
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2), asb = 0.0;
      
      if constexpr(static_cast<unsigned>(_Ordering::alpha))
        asb = static_cast<unsigned>(_Ordering::alpha)*as.ask(tau)*qcd::beta1(as.nf(tau));
      
      
      /*   for transverse momentum ordering we have zmax = 1.0
       *   and we have treat the soft term carefully
       */
      double soft1 = 0.0;
      
      if constexpr(0u == static_cast<unsigned>(_Ordering::alpha))
      {
        zmax = (1.0-pars->eta)*0.999 + pars->eta;
        
        /*   soft term    */
        soft1  = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
        soft1 *= (1.0 - pars->aux->parton(fa, pars->eta/zmax, q2)*zmax/pdf);
      }
      
      
      /*     jacobians    */
      double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
      double Jd = 0.5*std::log((1.0-umin)/(1.0-umax));
      
      
      return soft1 + _threshold_integral([=](double xi) -> double {
        /*    color diagonal integrand    */
        double soft = 0.0, reg = 0.0;
        
        
        if(zmax > zmin)
        {
          /*   integral transformation    */
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*Jm);
          double oz = 1.0-z;
          
          /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
           *   In this case the integrand is simple.
           */
          if(z < pars->eta) {
            soft  = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
            soft *= Jm;
            
            if constexpr(static_cast<unsigned>(_Ordering::alpha))
              soft /= 1.0+asb*std::log(oz);
          }
          
          /*   Otherwise we have a more complicated function   */
          else {
            double ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
            
            /*   soft term    */
            soft  = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
            soft *= Jm*(1.0 - z*ratio);
            
            if constexpr(static_cast<unsigned>(_Ordering::alpha))
              soft /= 1.0+asb*std::log(oz);
            
            /*   quark emitter   */
            if(fa) {
              reg  = qcd::Tr*(1.0-2.0*z*oz);
              reg *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
              reg += qcd::Cf*oz*ratio;
            }
            
            /*   gluon emitter   */
            else {
              reg  = qcd::Cf*(z + 2.0*oz/z);
              reg *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
              reg += 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
            }
            
            reg *= -Jm*oz;
          }
        }
        
        
        /*    color off diagonal integrand    */
        double delta = 0.0;
        
        if(umax > umin)
        {
          /*   integral transformation    */
          double z = 1.0-(1.0-umin)*std::exp(-(1.0+xi)*Jd);
          auto [y, yjac] = ordering_traits<ini, _Ordering>::main::mapping_to_y(pars, v, z);
          
          
          /*    sum over the specators  (only LC+ connections)    */
          delta = __parton_gluon_sum(pars, y, z)/z;
          delta *= Jd;
          delta *= yjac;
          
          /*   PDF factor   */
          if(z > pars->eta)
            delta *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
          
          if constexpr(static_cast<unsigned>(_Ordering::alpha))
            delta /= 1.0+asb*std::log(1.0-z);
        }
        
        
        /*    subleading correction   */
        double sub = 0.0;
        
        if(0 == fa && umax > pars->eta)
        {
          /*   integral transformation    */
          double Jr = 0.5*std::log((1.0-pars->eta)/(1.0-umax));
          double z = 1.0-(1.0-pars->eta)*std::exp(-(1.0+xi)*Jr);
          auto [y, yjac] = ordering_traits<ini, _Ordering>::main::mapping_to_y(pars, v, z);
          
          sub  = gluon_quark(y,z)*(pars->aux->parton(81, pars->eta/z, q2))/pdf;
          sub += gluon_gluon(y,z)*(pars->aux->parton(0, pars->eta/z, q2))/pdf;

          sub *= Jr*(1.0-z);
          sub *= yjac;
        }
        
        
        return soft + reg + delta + sub;
      });
    }
    
  }   //  namespace __details
  
  
  
  template<ProcessTag _Proc, OrderingTag _Ordering>
  void averaged_lcplus_threshold_sudakov(shower_history_stage<_Proc>& stage)
  {
    /*   sum over initial state emitters    */
    double result = 0.0, pdf_ratio = 1.0;
    
    if constexpr(is_hadron_collision<_Proc>)
      for(const auto& pars : stage.params.ini)
      {
        const qcd_coupling& as = pars.mdl.alpha_qcd();
        auto [vL, vH] = ordering_traits<ini, _Ordering>::threshold::vlimits(&pars);
        
        if(double v = pars.vnull*std::exp(-stage.current_time); v < vH) vH = v;
        
        if(vH > vL)
        {
          double tauL = as.tauk(pars.Q2*vH);
          double tauH = as.tauk(pars.Q2*vL);
          
          result += 0.5*(tauH-tauL)*__details::_threshold_integral([=, &pars](double xi) -> double {
            return __details::threshold<_Ordering>(&pars, (tauH-tauL)*0.5*(xi+1.0) + tauL);
          });
        }
      
        /*   PDF threshold effects    */
        pdf_ratio *= pars.mdl.hard_thrs_weight(stage.state[hadron{pars.emitter}].key, -pars.fl, pars.eta, stage.scale);
      }
    
    /*   At some point we need final state heavy flavor threshold to be considered
     *    .....
     */
    
    
    /*   applying the threshold weight   */
    double thrs_weight = pdf_ratio*std::exp(result);
    
    stage.weights.born.threshold *= thrs_weight;
    stage.weights.nlo.threshold *= thrs_weight;
  }
  
  
  
}   //   namespace duct

#endif /* __mc_threshold_h__ */
