/*  Copyright © 2021 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10.01.21 at 14:06.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_dimensionful_h__
#define __duct_dimensionful_h__ 1


/*   std includes   */
#include <iostream>
#include <complex>
#include <ratio>


/*   project headers   */
#include <cxx/concepts.h>
#include <cxx/binary-io.h>


namespace duct {

  
  
  template<class _Tp, class _Dim>
  requires complex_and_real_numbers<_Tp>
  class dimensionful
  {
  public:
    /*   underlying value type   */
    using underlying_type = _Tp;
    
    /*   default constructor   */
    dimensionful() : _M_val{0.0} {}

    /*   constructors  */
    constexpr explicit dimensionful(const _Tp& v) : _M_val(v) {}

    template<typename _Xp>
    requires std::convertible_to<_Xp, _Tp>
    constexpr explicit dimensionful(const _Xp& v) : _M_val(v) {}
    
    /*    copy   */
    dimensionful(const dimensionful& x) = default;
    
    template<complex_and_real_numbers _Xp>
    requires std::convertible_to<_Xp, _Tp>
    constexpr dimensionful(const dimensionful<_Xp, _Dim>& x)
    : _M_val(x.value()) {}
    
    
    /*    assignments     */
    dimensionful& operator=(const dimensionful& x) = default;

    template<complex_and_real_numbers _Xp>
    requires std::convertible_to<_Xp, _Tp>
    constexpr dimensionful& operator=(const dimensionful<_Xp, _Dim>& x) {
      _M_val = x.value();
      return *this;
    }
    

    /*    computed assignements    */
    constexpr dimensionful& operator+=(const dimensionful& x) {
      _M_val += x._M_val;
      return *this;
    }
    
    constexpr dimensionful& operator-=(const dimensionful& x) {
      _M_val -= x._M_val;
      return *this;
    }
    
    template<typename _Xp>
    constexpr dimensionful& operator*=(const _Xp& x) {
      _M_val *= x;
      return *this;
    }
    
    template<typename _Xp>
    constexpr dimensionful& operator/=(const _Xp& x) {
      _M_val /= x;
      return *this;
    }
    
    /*     comparison operators   */
    auto operator<=>(const dimensionful&) const = default;


    /*    returns the underlying value    */
    constexpr const _Tp& value() const { return _M_val;}
    
  private:
    _Tp _M_val;
  };
  
  
  /*   unary minus   */
  template<typename _Tp, class _Dim>
  constexpr inline auto operator-(const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<_Tp, _Dim>{-x.value()};
  }

  
  /*     multiplications     */
  template<typename _Tp, class _Dim, typename _Xp, class _DimX>
  requires (0 != std::ratio_add<_Dim, _DimX>::num)
  constexpr inline auto operator*(const dimensionful<_Tp, _Dim>& x, const dimensionful<_Xp, _DimX>& y) {
    return dimensionful<result_of_mul_t<_Tp,_Xp>, std::ratio_add<_Dim,_DimX>>{x.value()*y.value()};
  }
  
  template<typename _Tp, class _Dim, typename _Xp, class _DimX>
  requires (0 == std::ratio_add<_Dim, _DimX>::num)
  constexpr inline auto operator*(const dimensionful<_Tp, _Dim>& x, const dimensionful<_Xp, _DimX>& y) {
    return x.value()*y.value();
  }
  
  template<typename _Tp, class _Dim, complex_and_real_numbers _Xp>
  constexpr inline auto operator*(const dimensionful<_Tp, _Dim>& x, const _Xp& y) {
    return dimensionful<result_of_mul_t<_Tp,_Xp>, _Dim>{x.value()*y};
  }
  
  template<typename _Tp, class _Dim, complex_and_real_numbers _Xp>
  constexpr inline auto operator*(const _Xp& y, const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<result_of_mul_t<_Tp,_Xp>, _Dim>{x.value()*y};
  }
  
  
  /*     divisions     */
  template<typename _Tp, class _Dim, typename _Xp, class _DimX>
  requires (0 != std::ratio_subtract<_Dim, _DimX>::num)
  constexpr inline auto operator/(const dimensionful<_Tp, _Dim>& x, const dimensionful<_Xp, _DimX>& y) {
    return dimensionful<result_of_div_t<_Tp,_Xp>, std::ratio_subtract<_Dim,_DimX>>{x.value()/y.value()};
  }
    
  template<typename _Tp, typename _Xp, class _Dim>
  constexpr inline auto operator/(const dimensionful<_Tp, _Dim>& x, const dimensionful<_Xp, _Dim>& y) {
    return x.value()/y.value();
  }
  
  template<typename _Tp, class _Dim, complex_and_real_numbers _Xp>
  constexpr inline auto operator/(const dimensionful<_Tp, _Dim>& x, const _Xp& y) {
    return dimensionful<result_of_div_t<_Tp,_Xp>, _Dim>{x.value()/y};
  }
  
  template<complex_and_real_numbers _Xp, typename _Tp, class _Dim>
  constexpr inline auto operator/(const _Xp& y, const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<result_of_div_t<_Tp,_Xp>, typename std::ratio<-_Dim::num, _Dim::den>::type>{y/x.value()};
  }
  
  
  /*    addition and subtraction  */
  template<typename _Tp, typename _Xp, class _Dim>
  constexpr inline auto operator+(const dimensionful<_Tp, _Dim>& x, const dimensionful<_Xp, _Dim>& y) {
    return dimensionful<result_of_add_t<_Tp,_Xp>, _Dim>{x.value()+y.value()};
  }
 
  template<typename _Tp, typename _Xp, class _Dim>
  constexpr inline auto operator-(const dimensionful<_Tp, _Dim>& x, const dimensionful<_Xp, _Dim>& y) {
    return dimensionful<result_of_sub_t<_Tp,_Xp>, _Dim>{x.value()-y.value()};
  }
  
  
  /*    mathematical functions     */
  template<typename _Tp, class _Dim>
  constexpr inline auto abs(const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<_Tp, _Dim>{std::abs(x.value())};
  }
 
  template<typename _Tp, class _Dim>
  inline auto sqrt(const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<_Tp, typename std::ratio<_Dim::num, 2*_Dim::den>::type>{std::sqrt(x.value())};
  }
 
  template<typename _Tp, class _Dim>
  inline auto sqr(const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<_Tp, typename std::ratio<2*_Dim::num, _Dim::den>::type>{x.value()*x.value()};
  }

  template<int _Pow, typename _Tp, class _Dim>
  inline auto pow(const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<_Tp, typename std::ratio<_Pow*_Dim::num, _Dim::den>::type>{std::pow(x.value(), _Pow)};
  }

  template<int _Pow, typename _Tp, class _Dim>
  inline auto nthroot(const dimensionful<_Tp, _Dim>& x) {
    return dimensionful<_Tp, typename std::ratio<_Dim::num, _Pow*_Dim::den>::type>{std::pow(x.value(), 1.0/_Pow)};
  }

  template<typename _Tp, class _Dim>
  inline auto atan2(const dimensionful<_Tp, _Dim>& y, const dimensionful<_Tp, _Dim>& x) {
    return std::atan2(y.value(), x.value());
  }

  
  
  /*   concepts  */
  template<typename Test>
  struct is_dimensionful : std::false_type {};

  template<complex_and_real_numbers _Tp, intmax_t Num, intmax_t Den>
  struct is_dimensionful<dimensionful<_Tp, std::ratio<Num,Den>>> : std::true_type {};

  template<typename Test>
  concept dimensionful_values = is_dimensionful<Test>::value;
  
  
  template<typename _Tp>
  concept real_values = real_numbers<_Tp>
  || (dimensionful_values<_Tp> && real_numbers<typename _Tp::underlying_type>);
 
  template<typename _Tp>
  concept complex_values = complex_numbers<_Tp>
  || (dimensionful_values<_Tp> && complex_numbers<typename _Tp::underlying_type>);
  
  
  
  
  /*    Complex operators   */
  template<complex_values _Tp, class _Dim>
  inline auto real(const dimensionful<_Tp, _Dim>& a) {
    return dimensionful<typename _Tp::value_type, _Dim>{std::real(a.value())};
  }
  
  template<complex_values _Tp, class _Dim>
  inline auto imag(const dimensionful<_Tp, _Dim>& a) {
    return dimensionful<typename _Tp::value_type, _Dim>{std::imag(a.value())};
  }
  
  template<complex_values _Tp, class _Dim>
  inline auto conj(const dimensionful<_Tp, _Dim>& a) {
    return dimensionful<_Tp, _Dim>{std::conj(a.value())};
  }

  
  /*     I/O   operations   */
  template<typename _Tp, class _Dim>
  inline std::ostream& operator<<(std::ostream& os, const dimensionful<_Tp, _Dim>& x)
  {
    os<<x.value();
    
    if(_Dim::num*_Dim::den > 0)
    {
      if(_Dim::num == 1 && _Dim::den == 1) os<<" [GeV]";
      else if(_Dim::num > 1 && _Dim::den == 1) os<<" [GeV^"<<_Dim::num<<"]";
      else os<<" [GeV^{"<<_Dim::num<<"/"<<_Dim::den<<"}]";
    }
    else
    {
      if(std::abs(_Dim::num) == 1 && std::abs(_Dim::den) == 1) os<<" [1/GeV]";
      else if(std::abs(_Dim::num) > 1 && std::abs(_Dim::den) == 1) os<<" [1/GeV^"<<_Dim::num<<"]";
      else os<<" [1/GeV^{"<<std::abs(_Dim::num)<<"/"<<std::abs(_Dim::den)<<"}]";
    }
    
    return os;
  }
  
  template<typename _Tp, class _Dim>
  inline std::istream&  read(std::istream& is, dimensionful<_Tp, _Dim>& x)
  {
    _Tp v;
    duct::read(is, v);
    x = dimensionful<_Tp, _Dim>{v};
    return is;
  }
  
  template<typename _Tp, class _Dim>
  inline std::ostream&  write(std::ostream& os, const dimensionful<_Tp, _Dim>& x) {
    duct::write(os, x.value());
    return os;
  }


 
  /*    Obvious types that we usually use as dimensionful variables    */
  template<class _Tp>
  using energy = dimensionful<_Tp, std::ratio<1>>;
  
  template<class _Tp>
  using energy_sqr = dimensionful<_Tp, std::ratio<2>>;
  
  /*   literals   */
  constexpr energy<double> operator"" _MeV(long double s) { return energy<double>{s/1000.0};}
  constexpr energy<double> operator"" _GeV(long double s) { return energy<double>{s};}
  constexpr energy<double> operator"" _TeV(long double s) { return energy<double>{1000.0*s};}
  
  constexpr energy_sqr<double> operator"" _MeV2(long double s) { return energy_sqr<double>{s/1000000.0};}
  constexpr energy_sqr<double> operator"" _GeV2(long double s) { return energy_sqr<double>{s};}
  constexpr energy_sqr<double> operator"" _TeV2(long double s) { return energy_sqr<double>{1000000.0*s};}

  constexpr energy<std::complex<double>> operator"" _iMeV(long double s) {
    return energy<std::complex<double>>{std::complex<double>{0.0, static_cast<double>(s)/1000.0}};
  }
  
  constexpr energy<std::complex<double>> operator"" _iGeV(long double s) {
    return energy<std::complex<double>>{std::complex<double>{0.0, static_cast<double>(s)}};
  }
  
  constexpr energy<std::complex<double>> operator"" _iTeV(long double s) {
    return energy<std::complex<double>>{std::complex<double>{0.0, static_cast<double>(s)*1000.0}};
  }
  
  constexpr energy_sqr<std::complex<double>> operator"" _iMeV2(long double s) {
    return energy_sqr<std::complex<double>>{std::complex<double>{0.0, static_cast<double>(s)/1000000.0}};
  }
  
  constexpr energy_sqr<std::complex<double>> operator"" _iGeV2(long double s) {
    return energy_sqr<std::complex<double>>{std::complex<double>{0.0, static_cast<double>(s)}};
  }
  
  constexpr energy_sqr<std::complex<double>> operator"" _iTeV2(long double s) {
    return energy_sqr<std::complex<double>>{std::complex<double>{0.0, static_cast<double>(s)*1000000.0}};
  }
  
}  //  namespace duct

#endif

