/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 19/12/2016.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_diracalgebra_h__
#define __duct_diracalgebra_h__ 1



/*   std includes   */

/*   project headers   */
#include <bits/duct-lorentzvector.h>


namespace duct {
  
  
  
  
  /*    class for representing the elements of the Dirac algebra   */
  template<RealNumbers _Tp>
  class diracalgebra
  {
  public:
    /*   types   */
    using value_type = std::complex<_Tp>;
    
    /*   default constructor    */
    diracalgebra() = default;
    
    /*   move   */
    diracalgebra(diracalgebra&&) = default;
    diracalgebra& operator=(diracalgebra&&) = default;
    
    /*   copy   */
    diracalgebra(const diracalgebra&) = default;
    diracalgebra& operator=(const diracalgebra&) = default;
    
    explicit diracalgebra(const diracalgebra<RealNumbers>& x) {
      for(int i = 0; i < 16; i++) _M_g[i] = x[i];
    }

    diracalgebra& operator=(const diracalgebra<RealNumbers>& x) {
      for(int i = 0; i < 16; i++) _M_g[i] = x[i];
      return *this;
    }
    
    
    /*    creates the mass term   */
    explicit diracalgebra(value_type m) {
      _M_g[0] = _M_g[5] = _M_g[10] = _M_g[15] = std::move(m);
    }
    
   
    /*     making it from lorentz vector    */
    explicit diracalgebra(const lorentzvector<ComplexandRealNumbers>& p)
    {
      static constexpr std::complex<_Tp> I{0.0, 1.0};
      
      _M_g[2] = _M_g[13] = p[0]-p[3];
      _M_g[7] = _M_g[8] = p[0]+p[3];
      _M_g[3] = -(_M_g[9] = p[1] - I*p[2]);
      _M_g[6] = -(_M_g[12] = p[1] + I*p[2]);
    }
    
    /*    from lorentz vector with mass term    */
    diracalgebra(const lorentzvector<ComplexandRealNumbers>& p, value_type m)
    : diracalgebra{p}
    {
      _M_g[0] = _M_g[5] = _M_g[10] = _M_g[15] = std::move(m);
    }
    
    
    /*   computed assignments    */
    diracalgebra& operator*=(const diracalgebra<RealNumbers>& x);
    
    diracalgebra& operator+=(const diracalgebra<RealNumbers>& x) {
      for(int i = 0; i < 16; i++) _M_g[i] += x[i];
      return *this;
    }
    
    diracalgebra& operator-=(const diracalgebra<RealNumbers>& x) {
      for(int i = 0; i < 16; i++) _M_g[i] -= x[i];
      return *this;
    }
    
    diracalgebra& operator*=(const ComplexandRealNumbers& x) {
      for(int i = 0; i < 16; i++) _M_g[i] *= x;
      return *this;
    }
    
    diracalgebra& operator/=(const ComplexandRealNumbers& x) {
      for(int i = 0; i < 16; i++) _M_g[i] /= x;
      return *this;
    }
    

    /*    element access    */
    value_type& operator[](int i) {
      return _M_g[i];
    }
    
    const value_type& operator[](int i) const {
      return _M_g[i];
    }
    
  private:
    /*   data member    */
    value_type _M_g[16];
  };
  
  
  /*   special instances   */
  template<RealNumbers _Tp>
  struct standard_dirac_matrices
  {
    /*   default constructor   */
    standard_dirac_matrices();

    /*    Gamma^k and Gamma^5   */
    const diracalgebra<_Tp>& operator[](int k) const {
      return gamma[k == 5 ? 4 : k];
    }
    
  private:
    static diracalgebra<_Tp> gamma[5];
    static bool is_initialized;
  };
  
  
  template<RealNumbers _Tp>
  diracalgebra<_Tp> standard_dirac_matrices<_Tp>::gamma[5];
  
  template<RealNumbers _Tp>
  bool standard_dirac_matrices<_Tp>::is_initialized = false;
  
  
  template<RealNumbers _Tp>
  standard_dirac_matrices<_Tp>::standard_dirac_matrices()
  {
    if(!is_initialized)
    {
      /*   gamma0   */
      gamma[0][2] = gamma[0][7] = gamma[0][8] = gamma[0][13] = 1.0;
      
      /*   gamma1   */
      gamma[1][3] = gamma[1][6] = 1.0;
      gamma[1][9] = gamma[1][12] = -1.0;
      
      /*   gamma2   */
      gamma[2][3] = gamma[2][12] = {0.0,-1.0};
      gamma[2][9] = gamma[2][6] = {0.0,1.0};
      
      /*   gamma3   */
      gamma[3][2] = gamma[3][13] = 1.0;
      gamma[3][7] = gamma[3][8] = -1.0;
      
      /*   gamma5   */
      gamma[4][0] = gamma[4][5] = -1.0;
      gamma[4][10] = gamma[4][15] = 1.0;
      
      is_initialized = true;
    }
  }
  
  template<RealNumbers _Tp>
  static const standard_dirac_matrices<_Tp> dirac_matrix;
  
  
  
  /*     unary operators    */
  template<RealNumbers _Tp>
  inline constexpr const diracalgebra<_Tp>& operator+(const diracalgebra<_Tp>& x) {
    return x;
  }

  template<RealNumbers _Tp>
  inline  diracalgebra<_Tp> operator-(const diracalgebra<_Tp>& x) {
    diracalgebra<_Tp> r;
    for(int i = 0; i < 16; i++) r[i] = -x[i];
    return r;
  }
  
  
  /*    arithmetic operators     */
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto operator+(const diracalgebra<_Tp>& a, const diracalgebra<_Xp>& b)
  {
    diracalgebra<std::common_type_t<_Tp, _Xp>> r;
    for(int i = 0; i < 16; i++) r[i] = a[i]+b[i];
    return r;
  }
  
  
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto operator-(const diracalgebra<_Tp>& a, const diracalgebra<_Xp>& b)
  {
    diracalgebra<std::common_type_t<_Tp, _Xp>> r;
    for(int i = 0; i < 16; i++) r[i] = a[i]-b[i];
    return r;
  }
  

  template<RealNumbers _Tp, ComplexandRealNumbers _Xp>
  auto operator*(const diracalgebra<_Tp>& a, const _Xp& b)
  {
    using _Vp = typename traits::complex_and_real_numbers<_Xp>::value_type;
    diracalgebra<std::common_type_t<_Tp, _Vp>> r;
    for(int i = 0; i < 16; i++) r[i] = a[i]*b;
    return r;
  }
  
  
  template<RealNumbers _Tp, ComplexandRealNumbers _Xp>
  auto operator*(const _Xp& b, const diracalgebra<_Tp>& a)
  {
    using _Vp = typename traits::complex_and_real_numbers<_Xp>::value_type;
    diracalgebra<std::common_type_t<_Tp, _Vp>> r;
    for(int i = 0; i < 16; i++) r[i] = a[i]*b;
    return r;
  }
  
  
  template<RealNumbers _Tp, ComplexandRealNumbers _Xp>
  auto operator/(const diracalgebra<_Tp>& a, const _Xp& b)
  {
    using _Vp = typename traits::complex_and_real_numbers<_Xp>::value_type;
    diracalgebra<std::common_type_t<_Tp, _Vp>> r;
    for(int i = 0; i < 16; i++) r[i] = a[i]/b;
    return r;
  }
  

  /*     product of the dirac matrices     */
  template<RealNumbers _Tp>
  diracalgebra<_Tp>& diracalgebra<_Tp>::operator*=(const diracalgebra<RealNumbers>& x)
  {
    value_type r[16];
    
    for(int i = 0; i < 16; i += 4)
      for(int j = 0; j < 4; j++)
      {
        r[i+j] = _M_g[i]*x[j];
        
        for(int k = 1; k < 4; k++)
          r[i+j] += _M_g[i+k]*x[4*k+j];
      }
    
    for(int i = 0; i < 16; i++) _M_g[i] = r[i];
    
    return *this;
  }

  
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto operator*(const diracalgebra<_Tp>& a, const diracalgebra<_Xp>& b)
  {
    diracalgebra<std::common_type_t<_Tp, _Xp>> r;
    
    for(int i = 0; i < 16; i += 4)
      for(int j = 0; j < 4; j++)
      {
        r[i+j] = a[i]*b[j];
        
        for(int k = 1; k < 4; k++)
          r[i+j] += a[i+k]*b[4*k+j];
      }
    
    return r;
  }
  
  
  /*    comparisons    */
  template<RealNumbers _Tp, RealNumbers _Xp>
  bool operator==(const diracalgebra<_Tp>& a, const diracalgebra<_Xp>& b)
  {
    for(int i = 0; i < 16; i++)
      if(a[i] != b[i]) return false;
    
    return true;
  }
  
  template<RealNumbers _Tp>
  inline bool operator!=(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) {
    return !(a == b);
  }
  
  
  /*    trace of the matrix    */
  template<RealNumbers _Tp>
  inline std::complex<_Tp> Trace(const diracalgebra<_Tp>& a) {
    return a[0] + a[5] + a[10] + a[15];
  }
  
  
  /*    commutators   */
  template<RealNumbers _Tp, RealNumbers _Xp>
  inline auto commutator(const diracalgebra<_Tp>& a, const diracalgebra<_Xp>& b)
  {
    return a*b-b*a;
  }
  
  template<RealNumbers _Tp, RealNumbers _Xp>
  inline auto anticommutator(const diracalgebra<_Tp>& a, const diracalgebra<_Xp>& b)
  {
    return a*b+b*a;
  }


  /*   I/O operatons   */
  template<RealNumbers _Tp>
  std::ostream& operator<<(std::ostream& os, const diracalgebra<_Tp>& d)
  {
    for(int i = 0; i < 4; i++) {
      for(int j = 0; j < 4; j++)
        os<<d[4*i+j]<<"  ";
      os<<'\n';
    }
    
    return os;
  }

}

#endif
