/*  Copyright (C) 2002-2017 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_lorentzvector_h__
#define __duct_lorentzvector_h__ 1


/*   std includes   */


/*  project includes   */
#include <bits/duct-threevector.h>



namespace duct {
  
  
  /*  Template class for representing lorentz vectors.
   *  (object with single lorentz index)
   */
  template<typename _Tp>
  requires real_values<_Tp> || complex_values<_Tp>
  class lorentzvector : public threevector<_Tp>
  {
    /*   private types   */
    using _Base = threevector<_Tp>;
    
  public:
    using value_type = _Tp;
    using threevector_type = _Base;
    
    
    /*   default constructor  */
    lorentzvector() = default;
    
    /*   defaulted copy and move    */
    lorentzvector(lorentzvector&&) = default;
    lorentzvector(const lorentzvector&) = default;
    lorentzvector& operator=(lorentzvector&&) = default;
    lorentzvector& operator=(const lorentzvector&) = default;
    

    /*   constructors   */
    lorentzvector(value_type E, value_type px, value_type py, value_type pz)
    {
      _Base::_M_c[0] = E;
      _Base::_M_c[1] = px;
      _Base::_M_c[2] = py;
      _Base::_M_c[3] = pz;
    }
    
    template<typename _Ep, typename _Xp>
    requires std::convertible_to<_Xp, value_type> && std::convertible_to<_Ep, value_type>
    lorentzvector(_Ep E, const threevector<_Xp>& v)
    : _Base{v} { _Base::_M_c[0] = E;}
    
     
    
    /*   copy with conversion   */
    template<typename _Xp>
    requires std::convertible_to<_Xp, value_type>
    explicit lorentzvector(const lorentzvector<_Xp>& p) {
      _Base::_M_copy(p[0], p);
    }
        
    template<typename _Xp>
    requires std::convertible_to<_Xp, value_type>
    lorentzvector& operator=(const lorentzvector<_Xp>& p) {
      _Base::_M_copy(p[0], p);
      return *this;
    }
    
    
    /*   computed assinments   */
    template<typename _Xp>
    lorentzvector& operator+=(const lorentzvector<_Xp>& p) {
      _Base::_M_c[0] += p[0];
      _Base::operator+=(p);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator-=(const lorentzvector<_Xp>& p) {
      _Base::_M_c[0] -= p[0];
      _Base::operator-=(p);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator*=(const _Xp& x) {
      _Base::_M_c[0] *= x;
      _Base::operator*=(x);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator/=(const _Xp& x) {
      _Base::_M_c[0] /= x;
      _Base::operator/=(x);
      return *this;
    }
    
    
    /*   light cone coordinates   */
    value_type plus () const { return _Base::_M_c[0] + _Base::_M_c[3];}
    value_type minus() const { return _Base::_M_c[0] - _Base::_M_c[3];}
    
    
    /*    rapidities    */
    auto rapidity() const requires real_values<value_type> {
      return 0.5*std::log(plus()/minus());
    }
    
    auto pseudo_rapidity() const requires real_values<value_type> {
      return -std::log(std::tan(0.5*_Base::theta()));
    }
    
    /*   invariant mass square   */
    auto mag2() const requires real_values<value_type> {
      return _Base::_M_c[0]*_Base::_M_c[0] - _Base::mag2();
    }
    
    /*   invariant mass    */
    auto mag() const requires real_values<value_type>
    {
      using std::sqrt;
      auto p =  mag2();
      return p >= decltype(p){0.0} ? sqrt(p) : -sqrt(-p);
    }
    
    /*   tranverse momentum   */
    template<real_values _Xp, real_values _Yp>
    lorentzvector transverse(const lorentzvector<_Xp>&, const lorentzvector<_Yp>&) const;
    
    /*   boost vector    */
    auto boostVector() const requires real_values<value_type> {
      return threevector{_Base::_M_c[1]/_Base::_M_c[0], _Base::_M_c[2]/_Base::_M_c[0], _Base::_M_c[3]/_Base::_M_c[0]};
    }
      
    /*   Lorentz boost   */
    template<real_numbers _Xp, real_numbers _Yp, real_numbers _Zp>
    void boost(_Xp, _Yp, _Zp);
    
    template<real_numbers _Xp>
    void boost(const threevector<_Xp>& a) {
      boost(a[1], a[2], a[3]);
    }
  };
  
  
  /*   Lorentz boost   */
  template<typename _Tp>
  template<real_numbers _Xp, real_numbers _Yp, real_numbers _Zp>
  void lorentzvector<_Tp>::boost(_Xp bx, _Yp by, _Zp bz)
  {
    auto b2 = bx*bx + by*by + bz*bz;
    auto gamma = 1.0/std::sqrt(1.0 - b2);
    auto gamma2 = (b2 > 0.0 ? (gamma - 1.0)/b2 : 0.0);
    
    auto bp = bx*_Base::_M_c[1] + by*_Base::_M_c[2] + bz*_Base::_M_c[3];
    
    _Base::_M_c[1] += (gamma2*bx)*bp + (gamma*bx)*_Base::_M_c[0];
    _Base::_M_c[2] += (gamma2*by)*bp + (gamma*by)*_Base::_M_c[0];
    _Base::_M_c[3] += (gamma2*bz)*bp + (gamma*bz)*_Base::_M_c[0];
    _Base::_M_c[0]  = gamma*(_Base::_M_c[0] + bp);
  }
  
  
  /*   Transverse projection    */
  template<typename _Tp>
  template<real_values _Xp, real_values _Yp>
  lorentzvector<_Tp> lorentzvector<_Tp>::
  transverse(const lorentzvector<_Xp>& na, const lorentzvector<_Yp>& nb) const
  {
    auto sab = na*nb;
    auto na2 = na.mag2();
    auto nb2 = nb.mag2();
    
    const auto& p = *this;
    auto ra = na - nb*(na2/sab);
    auto rb = nb - na*(nb2/sab);
 
    return (na*dot(rb, p) + nb*dot(ra, p))*(sab/(sab*sab-na2*nb2)) - p;
  }
  
  
  
  /*   lightcone variables   */
  inline auto plus(const lorentzvector<auto>& p) {
    return p.plus();
  }
  
  inline auto minus(const lorentzvector<auto>& p) {
    return p.minus();
  }
  
  
  /*    rapidities    */
  template<real_values _Xp>
  inline auto rapidity(const lorentzvector<_Xp>& p) {
    return p.rapidity();
  }
  
  template<real_values _Xp>
  inline auto pseudo_rapidity(const lorentzvector<_Xp>& p) {
    return p.pseudo_rapidity();
  }
  
  /*   invariant mass square   */
  template<real_values _Xp>
  inline auto mag2(const lorentzvector<_Xp>& p) {
    return p.mag2();
  }
  
  /*   invariant mass   */
  template<real_values _Xp>
  inline auto mag(const lorentzvector<_Xp>& p) {
    return p.mag();
  }
  
  /*   boost vector   */
  template<real_values _Xp>
  inline auto boostVector(const lorentzvector<_Xp>& p) {
    return p.boostVector();
  }
  
  
  /*   unary operators    */
  template<typename _Tp> inline
  constexpr const lorentzvector<_Tp>& operator+(const lorentzvector<_Tp>& p) {
    return p;
  }
  
  template<typename _Tp>
  inline lorentzvector<_Tp> operator-(const lorentzvector<_Tp>& p) {
    return lorentzvector{-p[0], -p[1], -p[2], -p[3]};
  }
  
  
  /*   arithmetic binary operators   */
  template<typename _Tp, typename _Xp>
  inline auto operator+(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return lorentzvector{a[0]+b[0], a[1]+b[1], a[2]+b[2], a[3]+b[3]};
  }
  
  template<typename _Tp, typename _Xp>
  inline auto operator-(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return lorentzvector{a[0]-b[0], a[1]-b[1], a[2]-b[2], a[3]-b[3]};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const lorentzvector<_Tp>& a, const _Xp& b) {
    return lorentzvector{a[0]*b, a[1]*b, a[2]*b, a[3]*b};
  }
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const _Xp& b, const lorentzvector<_Tp>& a) {
    return lorentzvector{a[0]*b, a[1]*b, a[2]*b, a[3]*b};
  }
  
  template<typename _Tp, typename _Xp>
  inline auto operator/(const lorentzvector<_Tp>& a, const _Xp& b) {
    return lorentzvector{a[0]/b, a[1]/b, a[2]/b, a[3]/b};
  }
  
  
  /*    comparisons    */
  template<typename _Tp, typename _Xp>
  inline bool operator==(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return a[0] == b[0] && a[1] == b[1] && a[2] == b[2] && a[3] == b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline bool operator!=(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return a[0] != b[0] || a[1] != b[1] || a[2] != b[2] || a[3] != b[3];
  }
  
  
  /*    dot products   */
  template<typename _Tp, typename _Xp>
  inline auto operator*(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return a[0]*b[0] - a[1]*b[1] - a[2]*b[2] - a[3]*b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto dot(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return a[0]*b[0] - a[1]*b[1] - a[2]*b[2] - a[3]*b[3];
  }
  
  
  template<real_values _Tp, real_values _Xp>
  auto scalar_mp(lorentzvector<_Tp> a, lorentzvector<_Xp> b)
  {
    using _ComplexD = std::complex<std::common_type_t<_Tp, _Xp>>;
    
    constexpr _ComplexD I{0.0,1.0};
    
    _ComplexD ea{1.0}, eb{1.0}, ph{1.0};
    auto pa = a.perp();
    auto pb = b.perp();
    
    if(a[0] < 0.0) a *= -1.0, ph *= I;
    if(b[0] < 0.0) b *= -1.0, ph *= I;
    
    if(pa != 0.0) ea = _ComplexD{a[1], a[2]}/pa;
    if(pb != 0.0) eb = _ComplexD{b[1], b[2]}/pb;
    
    auto mp = a.minus()*b.plus();
    auto pm = a.plus()*b.minus();
    
    mp = (mp <= 0.0 ? 0.0 : mp);
    pm = (pm <= 0.0 ? 0.0 : pm);
    
    return ph*(std::sqrt(mp)*ea - std::sqrt(pm)*eb);
  }
  
  
  /*    Complex operators   */
  template<complex_values _Tp>
  inline auto real(const lorentzvector<_Tp>& a) {
    using std::real;
    return lorentzvector{real(a[0]), real(a[1]), real(a[2]), real(a[3])};
  }
  
  template<complex_values _Tp>
  inline auto imag(const lorentzvector<_Tp>& a) {
    using std::imag;
    return lorentzvector{imag(a[0]), imag(a[1]), imag(a[2]), imag(a[3])};
  }
  
  template<complex_values _Tp>
  inline auto conj(const lorentzvector<_Tp>& a) {
    using std::conj;
    return lorentzvector{conj(a[0]), conj(a[1]), conj(a[2]), conj(a[3])};
  }
  
 
  
  /*    I/O operations    */
  template<typename _Tp>
  inline std::ostream& operator<<(std::ostream& os, const lorentzvector<_Tp>& q) {
    return os<<"["<<q[0]<<", ("<<q[1]<<", "<<q[2]<<", "<<q[3]<<")]";
  }
  
  template<typename _Tp, class _Dim>
  inline std::ostream& operator<<(std::ostream& os, const lorentzvector<dimensionful<_Tp, _Dim>>& q)
  {
    os<<"["<<q[0].value()<<", ("<<q[1].value()<<", "<<q[2].value()<<", "<<q[3].value()<<")] ";
    
    if(_Dim::num*_Dim::den > 0)
    {
      if(_Dim::num == 1 && _Dim::den == 1) os<<"[GeV]";
      else if(_Dim::num > 1 && _Dim::den == 1) os<<"[GeV^"<<_Dim::num<<"]";
      else os<<"[GeV^{"<<_Dim::num<<"/"<<_Dim::den<<"}]";
    }
    else
    {
      if(std::abs(_Dim::num) == 1 && std::abs(_Dim::den) == 1) os<<"[1/GeV]";
      else if(std::abs(_Dim::num) > 1 && std::abs(_Dim::den) == 1) os<<"[1/GeV^"<<_Dim::num<<"]";
      else os<<"[1/GeV^{"<<std::abs(_Dim::num)<<"/"<<std::abs(_Dim::den)<<"}]";
    }
      
    return os;
  }
 
  /*   binary I/O    */
  std::ostream& write(std::ostream& os, const lorentzvector<auto>& q) {
    return os.write(static_cast<const char *>(&q), sizeof(q));
  }
 
  std::istream& read(std::istream& is, lorentzvector<auto>& q) {
    return is.read(static_cast<char *>(&q), sizeof(q));
  }
  
  
}  //   namespace duct
#endif
