/*  Copyright (C) 2002-2017 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_threevector_h__
#define __duct_threevector_h__ 1


/*   std includes   */
#include <utility>
#include <cmath>
#include <iostream>


/*  project includes   */
#include <bits/duct-dimensionful.h>



namespace duct {
 
  
  /*
   *  Template class definition for representing threevector like objects.
   *  In our application most of the time we need actually a lorentzvector,
   *  thus we store four components instead of three.
   */
  template<typename _Tp>
  requires real_values<_Tp> || complex_values<_Tp>
  class threevector
  {
  public:
    using value_type = _Tp;
    
    /*   default constructor   */
    threevector() {
      _M_c[0] = _M_c[1] = _M_c[2] = _M_c[3] = value_type{0.0};
    }
    
    /*   constructors   */
    threevector(value_type x, value_type y, value_type z) {
      _M_c[0] = value_type{0.0};
      _M_c[1] = x;
      _M_c[2] = y;
      _M_c[3] = z;
    }
    
    /*   defaulted copy and move    */
    threevector(threevector&&) = default;
    threevector(const threevector&) = default;
    threevector& operator=(threevector&&) = default;
    threevector& operator=(const threevector&) = default;
    
    
    /*   copy with conversion   */
    template<typename _Xp>
    explicit threevector(const threevector<_Xp>& v) {
      _M_copy(v);
    }
    
    template<typename _Xp>
    threevector& operator=(const threevector<_Xp>& v) {
      _M_copy(v);
      return *this;
    }
    
    
    /*   computed assignments   */
    template<typename _Xp>
    threevector& operator+=(const threevector<_Xp>& v)
    {
      _M_c[1] += v[1];
      _M_c[2] += v[2];
      _M_c[3] += v[3];
      
      return *this;
    }
    
    template<typename _Xp>
    threevector& operator-=(const threevector<_Xp>& v)
    {
      _M_c[1] -= v[1];
      _M_c[2] -= v[2];
      _M_c[3] -= v[3];
      
      return *this;
    }
    
    template<typename _Xp>
    threevector& operator*=(const _Xp& v)
    {
      _M_c[1] *= v;
      _M_c[2] *= v;
      _M_c[3] *= v;
      
      return *this;
    }
    
    template<typename _Xp>
    threevector& operator/=(const _Xp& v)
    {
      _M_c[1] /= v;
      _M_c[2] /= v;
      _M_c[3] /= v;
      
      return *this;
    }
    
    
    
    /*    element access via subscript operator   */
    value_type& operator[](int k) { return _M_c[k];}
    const value_type& operator[](int k) const { return _M_c[k];}
    
    
    /*   magnitude and the transvers component squared   */
    auto mag2() const {
      return _M_c[1]*_M_c[1] + _M_c[2]*_M_c[2] + _M_c[3]*_M_c[3];
    }
    
    auto perp2() const {
      return _M_c[1]*_M_c[1] + _M_c[2]*_M_c[2];
    }
    
    
    /*   magnitude and the transvers component squared   */
    auto mag() const requires real_values<value_type> {
      using std::sqrt;
      return sqrt(mag2());
    }
    
    auto perp() const requires real_values<value_type> {
      using std::sqrt;
     return sqrt(perp2());
    }
    
    
    /*   unit vector   */
    auto unit() const requires real_values<value_type>
    {
      auto m = mag();
      return threevector<result_of_div_t<value_type, decltype(m)>>{_M_c[1]/m, _M_c[2]/m, _M_c[3]/m};
    }
    
    
    /*        azimuth and polar angles     */
    auto phi() const requires real_values<value_type> {
      return _S_atan2(_M_c[2], _M_c[1]);
    }
    
    auto theta() const requires real_values<value_type> {
      return _S_atan2(perp(), _M_c[3]);
    }
    
    
    /*   ratate along one of the axis   */
    template<real_numbers _Xp>
    void rotate(int, const _Xp&);
    
    /*   rotate about an axis   */
    template<real_numbers _Xp, real_numbers _Yp>
    void rotate(const threevector<_Xp>&, const _Yp&);
    
    
  protected:
    /*  We actually store a lorentzvector and we use only the
     *  last three elements of this array as threevector.
     */
    value_type _M_c[4];
    
    
    /*   helper functions to optimize the operators   */
    template<typename _Xp>
    void _M_copy(const threevector<_Xp>& v) {
      _M_c[1] = v[1];
      _M_c[2] = v[2];
      _M_c[3] = v[3];
    }
    
    template<typename _Vp, typename _Xp>
    void _M_copy(const _Vp& E, const threevector<_Xp>& v) {
      _M_c[0] = E;
      _M_c[1] = v[1];
      _M_c[2] = v[2];
      _M_c[3] = v[3];
    }
    
    static auto _S_atan2(value_type y, value_type x)
    {
      using std::atan2;
      return value_type{0.0} == x && y == value_type{0.0} ? 0.0 : atan2(y, x);
    }
  };
  
  
  
  /*   magnitude and the transvers component squared   */
  template<real_values _Tp>
  inline auto mag2(const threevector<_Tp>& v) {
    return v.mag2();
  }
  
  template<real_values _Tp>
  inline auto perp2(const threevector<_Tp>& v) {
    return v.perp2();
  }
  

  /*   magnitude and the transvers component squared   */
  template<real_values _Tp>
  inline auto mag(const threevector<_Tp>& v) {
    return v.mag();
  }
  
  template<real_values _Tp>
  inline auto perp(const threevector<_Tp>& v) {
    return v.perp();
  }
  
  
  /*   unit vector   */
  template<real_values _Tp>
  inline auto unit(const threevector<_Tp>& v) {
    return v.unit();
  }
  
  
  /*   azimuth and polar angles     */
  template<real_values _Tp>
  inline auto phi(const threevector<_Tp>& v) {
    return v.phi();
  }
  
  template<real_values _Tp>
  inline auto theta(const threevector<_Tp>& v) {
    return v.theta();
  }
    
  
  /*   rotate about a coordinate axis   */
  template<typename _Tp> template<real_numbers _Xp>
  void threevector<_Tp>::rotate(int k, const _Xp& psi)
  {
    auto s = std::sin(psi), c = std::cos(psi);
    
    switch(k) {
      case 1: {
        const value_type py = _M_c[2], pz = _M_c[3];
        
        _M_c[2] = c*py - s*pz;
        _M_c[3] = s*py + c*pz;
      } break;
        
      case 2: {
        const value_type px = _M_c[1], pz = _M_c[3];
        
        _M_c[1] =  c*px + s*pz;
        _M_c[3] = -s*px + c*pz;
      } break;
        
      case 3: {
        const value_type px = _M_c[1], py = _M_c[2];
        
        _M_c[1] = c*px - s*py;
        _M_c[2] = s*px + c*py;
      } break;
    }
  }
  
  
  /*   rotate about an axis   */
  template<typename _Tp> template<real_numbers _Xp, real_numbers _Yp>
  void threevector<_Tp>::rotate(const threevector<_Xp>& axis, const _Yp& psi)
  {
    auto ph = axis.phi();
    auto th = axis.theta();
    
    rotate(3, -ph);
    rotate(2, -th);
    rotate(3, psi);
    rotate(2, th);
    rotate(3, ph);
  }
  
  
  
  /*   unary operators    */
  template<typename _Tp>
  constexpr const threevector<_Tp>& operator+(const threevector<_Tp>& x) {
    return x;
  }
  
  template<typename _Tp>
  inline threevector<_Tp> operator-(const threevector<_Tp>& x) {
    return {-x[1], -x[2], -x[3]};
  }
  
  
  
  /*    arithmetic operators    */
  template<typename _Tp, typename _Xp>
  inline auto operator+(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return threevector{a[1]+b[1], a[2]+b[2], a[3]+b[3]};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator-(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return threevector{a[1]-b[1], a[2]-b[2], a[3]-b[3]};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const threevector<_Tp>& a, const _Xp& b) {
    return threevector{a[1]*b, a[2]*b, a[3]*b};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const _Xp& b, const threevector<_Tp>& a) {
    return threevector{a[1]*b, a[2]*b, a[3]*b};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator/(const threevector<_Tp>& a, const _Xp& b) {
    return threevector{a[1]/b, a[2]/b, a[3]/b};
  }
  
  
  /*    comparisons    */
  template<typename _Tp, typename _Xp>
  inline bool operator==(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return a[1] == b[1] && a[2] == b[2] && a[3] == b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline bool operator!=(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return a[1] != b[1] || a[2] != b[2] || a[3] != b[3];
  }
  
  
  /*  products of threevectors   */
  template<typename _Tp, typename _Xp>
  inline auto operator*(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto dot(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto cross(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return threevector{a[2]*b[3]-a[3]*b[2], a[3]*b[1]-a[1]*b[3], a[1]*b[2]-a[2]*b[1]};
  }
  
  
  
  /*    I/O operations   */
  template<typename _Tp>
  inline std::ostream& operator<<(std::ostream& os, const threevector<_Tp>& q) {
    return os<<"("<<q[1]<<", "<<q[2]<<", "<<q[3]<<")";
  }
  
  template<typename _Tp, class _Dim>
  inline std::ostream& operator<<(std::ostream& os, const threevector<dimensionful<_Tp, _Dim>>& q)
  {
    os<<"("<<q[1].value()<<", "<<q[2].value()<<", "<<q[3].value()<<") ";
    
    if(_Dim::num*_Dim::den > 0)
    {
      if(_Dim::num == 1 && _Dim::den == 1) os<<"[GeV]";
      else if(_Dim::num > 1 && _Dim::den == 1) os<<"[GeV^"<<_Dim::num<<"]";
      else os<<"[GeV^{"<<_Dim::num<<"/"<<_Dim::den<<"}]";
    }
    else
    {
      if(std::abs(_Dim::num) == 1 && std::abs(_Dim::den) == 1) os<<"[1/GeV]";
      else if(std::abs(_Dim::num) > 1 && std::abs(_Dim::den) == 1) os<<"[1/GeV^"<<_Dim::num<<"]";
      else os<<"[1/GeV^{"<<std::abs(_Dim::num)<<"/"<<std::abs(_Dim::den)<<"}]";
    }
      
    return os;
  }
  

  
  /*    Complex operators   */
  template<complex_values _Tp>
  inline auto real(const threevector<_Tp>& a) {
    using std::real;
    return threevector{real(a[1]), real(a[2]), real(a[3])};
  }
  
  template<complex_values _Tp>
  inline auto imag(const threevector<_Tp>& a) {
    using std::imag;
    return threevector{imag(a[1]), imag(a[2]), imag(a[3])};
  }
  
  template<complex_values _Tp>
  inline auto conj(const threevector<_Tp>& a) {
    using std::conj;
    return threevector{conj(a[1]), conj(a[2]), conj(a[3])};
  }
  
}  //  namespace duct

#endif
