/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __core_mc_coupling_h__
#define __core_mc_coupling_h__ 1


/*   std headers   */
#include <cmath>
#include <memory>
#include <array>
#include <utility>
#include <type_traits>


/*   project headers   */
#include <bits/duct-dimensionful.h>
#include <bits/qcd-color.h>
#include <bits/qcd-coupling.h>
#include <cxx/binary-io.h>



namespace duct {
  
  namespace __details {
    
    /*   simple MSbar NLO alpha_s implementation    */
    std::pair<double, unsigned> as_msbar(double);

    
    /*    Interpolation grid for QCD strong coupling   */
    class coupling_grid
    {
      /*    size of the grid    */
      static constexpr std::size_t _N_grid = 200;
      
      /*   structure for storing the gridpoint   */
      struct gridpoint {
        unsigned nf;
        double t, as, tau, ask, tauk;
       };
      
      friend std::istream& read(std::istream& is, gridpoint& p);
      friend std::ostream& write(std::ostream& os, const gridpoint& p);
      
    public:
      /*   default constructor    */
      coupling_grid() = default;
      
      /*   creating from a simple coupling functions    */
      template<class _Func>
      coupling_grid(energy_sqr<double>, energy_sqr<double>, _Func&&);
      
      
      /*   returns the number of active flavors as function of tauk */
      unsigned nf(double) const;

      /*   returns the coupling as function of tauk */
      double as(double) const;
      
      /*   returns the coupling with the usual Kg correction as function of tauk */
      double ask(double) const;
      
      /*   tauk => q2 mapping   */
      energy_sqr<double> q2(double) const;
      
      /*   q2 => tau mapping  */
      double tau(energy_sqr<double>) const;

      /*   q2 => tauk mapping  */
      double tauk(energy_sqr<double>) const;
      
      /*    observers    */
      energy_sqr<double> q2max() const { return _M_q2max;}
      energy_sqr<double> q2min() const { return _M_q2min;}
      
      /*   I/O operations   */
      void write(std::ostream& os) const {
        duct::write(os, _M_dt, _M_dtau, _M_q2min, _M_q2max, _M_grid);
      }
      
      void read(std::istream& is) {
        duct::read(is, _M_dt, _M_dtau, _M_q2min, _M_q2max, _M_grid);
      }
      
    private:
      /*      grid distances and limits     */
      double _M_dt, _M_dtau;
      energy_sqr<double> _M_q2min, _M_q2max;
      
      /*   array of grid points   */
      std::array<gridpoint, _N_grid+1> _M_grid;

      /*   private helper functions   */
      double _M_tauk_inv(double) const;
    };
  
    
    template<class _Func>
    coupling_grid::coupling_grid(energy_sqr<double> _q2min, energy_sqr<double> _q2max, _Func&& __as)
    : _M_q2min{_q2min}, _M_q2max{_q2max}
    {
      /*   fixed parameters   */
      _M_grid[0].t = 0.0;
      _M_grid[_N_grid].t = std::log(_q2max/_q2min);
      _M_dt = _M_grid[_N_grid].t/_N_grid;
      
      /*    calculating the tauk values in the gridpoints   */
      auto aa = __as(_q2max);
      
      _M_grid[0].tau = 0.0;
      _M_grid[0].as = aa.first;

      _M_grid[0].tauk = 0.0;
      _M_grid[0].ask = aa.first/(1.0 - aa.first*qcd::Kg(aa.second));
      
      for(std::size_t k = 1; k <= _N_grid; ++k)
      {
        aa = __as(_q2max*std::exp(-_M_dt*k));

        _M_grid[k].as = aa.first;
        _M_grid[k].tau = _M_grid[k-1].tau + 0.5*_M_dt*(_M_grid[k-1].as+_M_grid[k].as);

        _M_grid[k].ask = aa.first/(1.0 - aa.first*qcd::Kg(aa.second));
        _M_grid[k].tauk = _M_grid[k-1].tauk + 0.5*_M_dt*(_M_grid[k-1].ask+_M_grid[k].ask);
      }
      
      /*    q2 on an equidistant tauk grid     */
      _M_dtau = _M_grid[_N_grid].tauk/_N_grid;
      
      for(std::size_t k = 1; k <= _N_grid; ++k)
        _M_grid[k].t = _M_tauk_inv(k*_M_dtau);

      /*   as and ask are also on the tauk grid   */
      for(std::size_t k = 0; k <= _N_grid; ++k)
      {
        aa = __as(_q2max*std::exp(-_M_grid[k].t));
        
        _M_grid[k].nf = aa.second;
        _M_grid[k].as = aa.first;
        _M_grid[k].ask = aa.first/(1.0 - aa.first*qcd::Kg(aa.second));
      }
    }
    
    
    /*   I/O operations   */
    inline std::istream& read(std::istream& is, coupling_grid::gridpoint& p) {
      return duct::read(is, p.nf, p.t, p.as, p.tau, p.ask, p.tauk);
    }
    
    inline std::ostream& write(std::ostream& os, const coupling_grid::gridpoint& p) {
      return duct::write(os, p.nf, p.t, p.as, p.tau, p.ask, p.tauk);
    }
  }   //   namespace __details
  
  
  /*    Grid based QCD coupling  
   *    We use this in the shower program becaoce we need some 
   *    other quantities and it is hard to provide them analyticaly.
   */
  class qcd_coupling
  {
    template<typename T> static constexpr bool
    _InitConcept = (!std::is_same<qcd::coupling, typename std::decay<T>::type>::value
                    && !std::is_base_of<qcd::coupling, typename std::decay<T>::type>::value
                    && std::is_same<std::pair<double, unsigned>, std::result_of_t<T(energy_sqr<double>)>>::value);
    
    /*    the underlying grid alpha_s   */
    using _Base = __details::coupling_grid;
    
  public:
    /*    Default constructor   */
    qcd_coupling() : _M_as{nullptr} {}

    /*    Creating from analytic expression     */
    qcd_coupling(energy_sqr<double> _q2min, energy_sqr<double> _q2max, const qcd::coupling& _as)
    : _M_as{std::make_shared<_Base>(_q2min, _q2max, [&_as](energy_sqr<double> q2) { return _as.asnf(q2);})} {}
    
    
    /*    Creating from a callable    */
    template<class _Func>
    requires _InitConcept<_Func>
    qcd_coupling(double _q2min, double _q2max, _Func&& _as)
    : _M_as{std::make_shared<_Base>(_q2min, _q2max, std::forward<_Func>(_as))} {}
    
    
    /*   returns the coupling as function of q2   */
    double operator()(energy_sqr<double> q2in) const { return _M_as->as(_M_as->tauk(q2in));}
    
    /*   returns the coupling as function of tauk */
    unsigned nf(double tk) const { return _M_as->nf(tk);}
    
    /*   returns the coupling as function of tauk */
    double as(double tk) const { return _M_as->as(tk);}
    
    /*   returns the coupling with the usual Kg correction as function of tauk */
    double ask(double tk) const { return _M_as->ask(tk);}
    
    /*   tauk => q2 mapping   */
    energy_sqr<double> q2(double tk) const { return _M_as->q2(tk);}
    
    /*   q2 => tauk mapping  */
    double tau(energy_sqr<double> q2in) const { return _M_as->tau(q2in);}
    
    /*   q2 => tauk mapping  */
    double tauk(energy_sqr<double> q2in) const { return _M_as->tauk(q2in);}
    
    /*    observers    */
    energy_sqr<double> q2max() const { return _M_as->q2max();}
    energy_sqr<double> q2min() const { return _M_as->q2min();}

    
    /*   I/O operations   */
    void write(std::ostream&) const;
    void read(std::istream&);
    
    
    /*   check if it an empty object    */
    bool is_empty() const { return nullptr == _M_as;}
    
  private:
    /*   Shared grid. The coupling doesn't change after creation
     *   thus it make sense to share the underlying grid.
     */
    std::shared_ptr<_Base> _M_as;
  };

  
  /*   I/O operatoions     */
  inline std::ostream& write(std::ostream& os, const qcd_coupling& as)  {
    as.write(os);
    return os;
  }
  
  inline std::istream& read(std::istream& is, qcd_coupling& as) {
    as.read(is);
    return is;
  }
  
}   /*   namespace duct   */

#endif
