/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 23/02/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_ordering_kT_h__
#define __mc_ordering_kT_h__ 1



/*   std headers   */
#include <cmath>
#include <utility>

/*   project headers   */
#include <bits/mc-hard-params.h>



namespace duct {
  
  template<>
  struct ordering_traits<fin, ordering::kT>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*  calculates the limits on the variable v    */
      static auto v2limits(const hard_params<fin>& pars) {
        return std::pair{pars.mT*pars.pQ, std::min(0.25*pars.ymax, pars.mTmax)*pars.pQ};
      }
   
      /*    calculates the z limit   */
      static auto zlimits(const hard_params<fin>& pars, energy_sqr<double> v2)
      {
        double vT = v2/(pars.pQ*pars.ymax);
        double zp = (vT >= 0.25 ? 0.4999999 : 0.5+std::sqrt(0.25-vT));
        return std::pair{1.0-zp, zp};
      }
      
      /*    mapping the independent variables to y   */
      static auto mapping_to_y(const hard_params<fin>& pars, energy_sqr<double> v2, double z) {
        return std::pair{v2/(pars.pQ*z*(1.0-z)), 1.0};
      }
    };
    
    /*    we might need other components    */
  };
  
  
  
  template<>
  struct ordering_traits<ini, ordering::kT>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*  calculates the limits on the variable v    */
      static auto v2limits(const hard_params<ini>& pars)
      {
        const double &eta = pars.eta;
        return std::pair{pars.mT*pars.Q2, std::min(sqr(1.0-eta)/eta, pars.mTmax)*pars.Q2};
      }
      
      /*    calculates the z limit   */
      static auto zlimits(const hard_params<ini>& pars, energy_sqr<double> v2)
      {
        const double &eta = pars.eta, v = v2/pars.Q2;
        return std::pair{eta, std::min(1.0 - v*eta/(1.0-eta), 4.0/sqr(std::sqrt(pars.mT)+std::sqrt(pars.mT+4.0)))};
      }
      
      /*    mapping the independent variables to y   */
      static auto mapping_to_y(const hard_params<ini>& pars, energy_sqr<double> v2, double z) {
        return std::pair{v2/pars.Q2/(1.0-z), 1.0};
      }
    };
    
    /*    we might need other components    */
  };




}   //   namespace duct

#endif
