/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 25/09/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_shower_aux_h__
#define __mc_shower_aux_h__ 1


/*   std headers   */
#include <cmath>
#include <iostream>
#include <memory>
#include <algorithm>
#include <array>

/*   project headers   */
#include <bits/duct-tags.h>
#include <bits/mc-dglap.h>
#include <cxx/math-support.h>
#include <cxx/fixed-array-view.h>



namespace duct {
  
  namespace __details {
    
    
    class shower_pdf_tbls
    {
      /*   size of the grids (these can be modified)   */
      static constexpr std::size_t _N_tau = 100;
      static constexpr std::size_t _N_eta = 100;
      static constexpr std::size_t _N_chi = 100;
      
      
      /*** DON'T CHANGE BEYOND THIS POINT!!! ***/
      using _Table = std::array<double, (_N_tau+1)*(_N_eta+1)>;
      using _SoftTable = std::array<double, (_N_tau+1)*(_N_eta+1)*(_N_chi+1)>;
      
      using _Array = span<double, _N_eta+1, _N_tau+1>;
      using _SoftArray = span<double, _N_eta+1, _N_tau+1, _N_chi+1>;
      
      using _ConstArray = span<const double, _N_eta+1, _N_tau+1>;
      using _ConstSoftArray = span<const double, _N_eta+1, _N_tau+1, _N_chi+1>;
      
    public:
      /*   hard wired eta max   */
      static constexpr double _eta_max = 0.9999;
      
      /*    constructors   */
      shower_pdf_tbls() = default;
      explicit shower_pdf_tbls(const dglap&);
      
      /*   pdf functions    */
      double parton(int, double, double) const;
      
      /*   soft functions    */
      double integral_soft(double, int, double, double) const;
      double inverse_soft(double, int, double, double) const;
      
      
      /*   collinear functions   */
      double integral_coll(int, double, double) const;
      double inverse_coll(double, int, double) const;
      
      
      /*    binary I/O operations   */
      void read(std::istream&);
      void write(std::ostream&) const;
      
    private:
      /*   xi = log(1/x-1) limits   */
      double _M_ximin, _M_ximax;
      
      /*   grid distances    */
      double _M_dxi, _M_dtl, _M_dtc, _M_dtb, _M_dchi;
      
      /*  tables (pdfs, collinear integrals, soft integrals)   */
      _Table _M_pdf[10], _M_coll[11];
      _SoftTable _M_soft[9];
    };
    

  
    class shower_aux_impl
    {
    public:
      /*    simple default constructor   */
      shower_aux_impl() = default;
      
      /*    constructors   */
      explicit shower_aux_impl(const dglap& d)
      : _M_m2T{d.q2min()}, _M_m2c{d.thrs(4)}, _M_m2b{d.thrs(5)}, _M_scheme{d.scheme()}, _M_ipdf{d} {}
      
      /*   return the pdf    */
      double parton(const qcd_coupling&, int, double, energy_sqr<double>) const;
      
      /*    inverse function of the integrals   */
      std::pair<double,double>
      inverse(const qcd_coupling&, double, splitting_type, double, int, double, energy_sqr<double>) const;
      
      
       /*   cutoff scale   */
      energy_sqr<double> q2min() const { return _M_m2T;}
      
      /*   heavy flavor threshold   */
      energy_sqr<double> thrs(int f, double eta) const {
        switch(_M_scheme) {
          case ordering::scheme::kT:      return thrs(f); break;
          case ordering::scheme::lambda:  return thrs(f)/(1.0-eta); break;
          case ordering::scheme::angular: return thrs(f)/sqr(1.0-eta); break;
        }
        
        return 0.0_GeV2;
      }
      
      energy_sqr<double> thrsT(int f, double eta) const {
        switch(_M_scheme) {
          case ordering::scheme::kT:      return thrsT(f); break;
          case ordering::scheme::lambda:  return thrsT(f)/(1.0-eta); break;
          case ordering::scheme::angular: return thrsT(f)/sqr(1.0-eta); break;
        }
        
        return 0.0_GeV2;
      }
      
      /*   it is the absolute threshold; it is equivalent to thrs(f, eta=0.0)    */
      energy_sqr<double> thrs(int f) const {
        switch(std::abs(f)) {
          case 4: return _M_m2c; break;
          case 5: return _M_m2b; break;
        }
        
        return 0.0_GeV2;
      }
      
      energy_sqr<double> thrsT(int f) const {
        switch(std::abs(f)) {
          case 4: return  std::max(_M_m2T, _M_m2c); break;
          case 5: return  std::max(_M_m2T, _M_m2b); break;
        }
        
        return _M_m2T;
      }
      
      /*   returns the scheme    */
      ordering::scheme scheme() const { return _M_scheme;}
      

      /*    binary I/O operations   */
      void read(std::istream& is) {
        unsigned sch;
        
        duct::read(is, _M_m2T, _M_m2c, _M_m2b, sch);
        _M_ipdf.read(is);
        _M_scheme = static_cast<ordering::scheme>(sch);
      }
      
      void write(std::ostream& os) const {
        duct::write(os, _M_m2T, _M_m2c, _M_m2b, static_cast<unsigned>(_M_scheme));
        _M_ipdf.write(os);
      }
      
    private:
      /*   cutoff & threshold scales    */
      energy_sqr<double> _M_m2T, _M_m2c, _M_m2b;
      
      /*   factorization scheme     */
      ordering::scheme _M_scheme;
      
      /*    integrals of the pdfs for shower evolution   */
      shower_pdf_tbls _M_ipdf;
      
      
      /*    inverse functions    */
      std::pair<double,double> inverse_soft(const qcd_coupling&, double, double, int, double, energy_sqr<double>) const;
      std::pair<double,double> inverse_coll(const qcd_coupling&, double, double, int, double, energy_sqr<double>) const;
    };
    
  }   /*   namespace __details   */

  
  
  class shower_aux
  {
  public:
    /*    empty object   */
    shower_aux() = default;
    
    /*   creating from dglap evolution    */
    explicit shower_aux(const dglap& d)
    : _M_as{d.coupling()}
    , _M_aux{std::make_shared<__details::shower_aux_impl>(d)}
    , _M_anti{false} {}
    
    
    /*   return the pdf    */
    double parton(int f, double x, energy_sqr<double> q2) const {
      return _M_aux->parton(_M_as, (_M_anti ? -f : f), x, q2);
    }
    
    /*    inverse function of the integrals   */
    std::pair<double,double>
    inverse(double xi, splitting_type idx, double x, int f, double eta, energy_sqr<double> q2) const {
      return _M_aux->inverse(_M_as, xi, idx, x, (_M_anti ? -f : f), eta, q2);
    }
    
    
    /*   observers   */
    energy_sqr<double> q2min() const { return _M_aux->q2min();}
    energy_sqr<double> thrs(int f) const { return _M_aux->thrs(f);}
    energy_sqr<double> thrsT(int f) const { return _M_aux->thrsT(f);}
    energy_sqr<double> thrs(int f, double eta) const { return _M_aux->thrs(f, eta);}
    energy_sqr<double> thrsT(int f, double eta) const { return _M_aux->thrsT(f, eta);}
    ordering::scheme scheme() const { return _M_aux->scheme();}
    
  
    /*   turn it to anti hadron   */
    void anti() { _M_anti = !_M_anti;}
  
    
    /*   I/O  operations   */
    void read(std::istream&, const qcd_coupling&);
    void write(std::ostream&) const;
    
    
    /*   check if it is empty   */
    bool is_empty() const { return _M_aux == nullptr;}
    
  private:
    /*   "reference to" strong coupling   */
    qcd_coupling _M_as;
    
    /*   shared pointer to the pdf tables    */
    std::shared_ptr<__details::shower_aux_impl> _M_aux;
  
    /*   true if it an anti-hardorn of the underlying one    */
    bool _M_anti;
  };
  
  
}    /*   namespace duct    */

#endif

