/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 19/02/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_shower_stage_h__
#define __mc_shower_stage_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-event.h>
#include <bits/mc-ordering.h>
#include <cxx/memory-resource.h>



namespace duct {
  
  
  /*
   *    Simple class for providing the kT and virtuality veto parameters.
   */
  struct shower_vetos
  {
    /*  This is maximum allowed transverse momentum square in the initial
     *  and final state splitting. These usually set by the hard state.
     */
    energy_sqr<double> max_k2T = energy_sqr<double>{std::numeric_limits<double>::max()};
    
    /*  This is maximum allowed virtuality in the initial
     *  state splitting. These usually set by the hard state.
     */
    energy_sqr<double> max_q2 = energy_sqr<double>{std::numeric_limits<double>::max()};
  };
  
 
  
  /*   This class defines a set of weights those are carried through the shower evolution.   */
  struct weight_set
  {
    struct weight_t
    {
      /*   Main weight
       *   In the current implementation this weight comes from the
       *   hard event and doesn't changes during the shower evolution.
       */
      double main;
      
      /*   Threshold correction factor
       *   This threshold correction factor comes from the LC+ threshold
       *   shower operator. There are also additive threshold corrections
       *   and those are counted in the color state.
       */
      double threshold;
    };
    
    /*   returns the full weight    */
    explicit operator double() const { return born.main*born.threshold + nlo.main*nlo.threshold;}

   
    /*   At hard level this is the Born term.   */
    weight_t born;
    
    /*   At hard level this gives the contributions of the NLO matrix elements.
     *   We have to do the matching to have this.
     */
    weight_t nlo;
  };
  
  
  /*    binary I/O for the weight_set class   */
  inline std::ostream& write(std::ostream& os, const weight_set::weight_t& w) {
    return duct::write(os, w.main, w.threshold);
  }
  
  inline std::istream& read(std::istream& is, weight_set::weight_t& w) {
    return duct::read(is, w.main, w.threshold);
  }
  
  inline std::ostream& write(std::ostream& os, const weight_set& w) {
    return duct::write(os, w.born, w.nlo);
  }
  
  inline std::istream& read(std::istream& is, weight_set& w) {
    return duct::read(is, w.born, w.nlo);
  }
  
  
  
  
  
  /*  Class for representing a stage of the shower evolution.
   *  This class can be seen by the users in the user function.
   *  It consists of
   *   state         => the event
   *   cstates       => list of color states
   *   scale         => scale of the event (equivalent to current time)
   *   weight        => weight of the current event
   */
  template<ProcessTag _Proc>
  struct shower_stage
  {
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::polymorphic_allocator<>;
    
    /*   default constructors    */
    shower_stage() = default;
    explicit shower_stage(allocator_type a) : state{a} {}
    
    
    /*   copy   */
    shower_stage(const shower_stage&) = default;
    shower_stage& operator=(const shower_stage&) = default;
    
    shower_stage(const shower_stage& x, allocator_type a)
    : state{x.state, a}
    , scale{x.scale}
    , weights{x.weights}
    {}

    /*    move  */
    shower_stage(shower_stage&&) = default;
    shower_stage& operator=(shower_stage&&) = default;
    
    shower_stage(shower_stage&& x, allocator_type a)
    : state{std::move(x.state), a}
    , scale{std::move(x.scale)}
    , weights{std::move(x.weights)}
    {}
    
    
    /*  The event of the state. This contains every informations about the
     *  event, momenta, flavors, color connections, incoming objects,...
     */
    event<_Proc> state;
    
    
    /*   The scale of the shower stage. It is typically corresponds to the
     *   shower time, but at the hard state we don't have emitter parton and
     *   the definition of the shower time is ambigous...
     */
    energy_sqr<double> scale;
    
    /*  The posibble weight of the event. This weight can come from color
     *  factors or from spin correlation (when we add spin correlations).
     */
    weight_set weights;
  };

  
  /*    binary I/O for the shower stage class    */
  template<ProcessTag _Proc>
  inline std::ostream& write(std::ostream& os, const shower_stage<_Proc>& s) {
    return duct::write(os, s.state, s.scale, s.weights);
  }
  
  template<ProcessTag _Proc>
  inline std::istream& read(std::istream& is, shower_stage<_Proc>& s) {
    return duct::read(is, s.state, s.scale, s.weights);
  }
  
  

  /*  Class for representing a stage of the shower evolution.
   *  In this class we record more information and they are basically
   *  used only during the evolution.
   *
   *  Inhereted from
   *                 => shower_stage
   *
   *  and it consists of
   *   vars          => spitting variable for the next stage
   *   cvars         => color spitting variable for the next stage
   *   qnull         => reference momentum for defining shower time
   *   current_time  => shower time when the event was created
   *   next_time     => shower time of the next splitting
   *   vetos         => veto parameters
   */
  template<ProcessTag _Proc>
  class shower_history_stage : public shower_stage<_Proc>
  {
    /*   private types   */
    using _Base = shower_stage<_Proc>;
    
  public:
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::polymorphic_allocator<>;
    
    /*   default constructors    */
    shower_history_stage() = default;

    explicit shower_history_stage(allocator_type a)
    : shower_stage<_Proc>{a} {}
    
    
    /*   copy   */
    shower_history_stage(const shower_history_stage&) = default;
    shower_history_stage& operator=(const shower_history_stage&) = default;
  
    shower_history_stage(const shower_history_stage& x, allocator_type a)
    : shower_stage<_Proc>{x, a}
    , vars{x.vars}
    , cvars{x.cvars}
    , current_scale{x.current_scale}
    , next_scale{x.next_scale}
    , vetos{x.vetos}
    {}
    
    /*    move  */
    shower_history_stage(shower_history_stage&&) = default;
    shower_history_stage& operator=(shower_history_stage&&) = default;
    
    shower_history_stage(shower_history_stage&& x, allocator_type a)
    : shower_stage<_Proc>{std::move(x), a}
    , vars{std::move(x.vars)}
    , cvars{std::move(x.cvars)}
    , current_scale{std::move(x.current_scale)}
    , next_scale{std::move(x.next_scale)}
    , vetos{std::move(x.vetos)}
    {}
    
    
    /*  Splitting varibles: This gives the splitting varibles of the next stage.
     *  It consists of y, z, phi, emitter, spectator and flavors.
     */
    splitting_vars vars;
    
    /*  Color splitting varibles: This gives the splitting varibles of the next stage
     *  for the color mapping.
     */
    splitting_cvars cvars;
    
    
    /*  The shower time when the event was created.
     */
    energy_sqr<double> current_scale;
    
    /*  The shower time. This state is evolved without emission upto this time.
     */
    energy_sqr<double> next_scale;
    
    
    /*   Since a history can have more than one hard state, we have to add
     *   the veto parameters to this class. The won't change during the
     *   evolution (I cannot imagine such a scenario... but who knows...).
     *   Maybe the upper limits on kT  or virtuality can change...
     *   Anyway, it is better to keep it here.
     */
    shower_vetos vetos;
  };
}

#endif
