/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 15/01/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <set>

/*   project headers   */
#include "bits/mc-dglap.h"



namespace duct {
  
  namespace __details {
     
    /*-*-*-*-*-*- Primitive functions of the scheme dependent  kernels -*-*-*-*-*-*/
    
    static std::pair<double, double> I0qq(double xb, double loz)
    {
      std::pair<double, double> res;
      double oz = std::exp(loz);
      double z = 1.0-oz;
      
      /*   soft part   */
      if(xb > 1e-15)
      {
        res.second = -2.0*qcd::Cf*std::log(1.0 + xb*loz)/xb;
        res.first = res.second + 2.0*qcd::Cf*oz*expint_Ei_scaled(1.0/xb + loz)/xb;
      }
      else {
        res.first = 2.0*qcd::Cf*(oz - loz);
        res.second = -2.0*qcd::Cf*loz;
      }
      
      /*    pure collinear part   */
      res.first += qcd::Cf*z*(1.0-0.5*z);
      res.second += qcd::Cf*(-z+std::log(z));
      
      return res;
    }
    
    static std::pair<double, double> I0gg(double xb, double loz)
    {
      std::pair<double, double> res;
      double oz = std::exp(loz);
      double z = 1.0-oz;
      double lz = std::log(z);
      
      /*   soft part   */
      if(xb > 1e-15)
      {
        res.second = -2.0*qcd::Ca*std::log(1.0 + xb*loz)/xb;
        res.first = res.second + 2.0*qcd::Ca*oz*expint_Ei_scaled(1.0/xb + loz)/xb;
      }
      else {
        res.first = 2.0*qcd::Ca*(oz - loz);
        res.second = -2.0*qcd::Ca*loz;
      }
      
      /*    pure collinear part   */
      res.first += 2.0*qcd::Ca*(z*(0.5*z - 1.0 - z*z/3.0) + lz);
      res.second += 2.0*qcd::Ca*(z - 0.5*z*z - 1.0/z - lz);
      
      return res;
    }
    
    static std::pair<double, double> I0gq(double loz)
    {
      std::pair<double, double> res;
      double oz = std::exp(loz);
      double z = 1.0-oz;
      double lz = 2.0*std::log(z);
      
      res.first = qcd::Cf*((0.5*z - 2.0)*z + lz);
      res.second = qcd::Cf*(z - 2.0/z - lz);
      
      return res;
    }
    
    static std::pair<double, double> I0qg(double loz)
    {
      std::pair<double, double> res;
      double oz = std::exp(loz);
      double z = 1.0-oz;
      
      res.first = 0.5*z*(1.0 - z + 2.0*z*z/3.0);
      res.second = 0.5*(z*(z-2.0) + std::log(z));
      
      return res;
    }
    
    
    /*   This is the regularized integral at x = 1 and xi = 0. Thus
     *
     *     lim{x->1} [Iqq(xb=0, x) + 2*qcd::Cf*log(1-x)] = Iqq()
     *     lim{x->1} [Igg(xb=0, x) + 2*qcd::Ca*log(1-x)] = Igg()
     *
     *     lim{x->1} Igq(x) = Igq()
     *     lim{x->1} Iqg(x) = Iqg()
     */
    constexpr std::pair<double, double> I0qq() { return {0.5*qcd::Cf, -qcd::Cf};}
    constexpr std::pair<double, double> I0gg() { return {-5.0*qcd::Ca/3.0, -qcd::Ca};}
    constexpr std::pair<double, double> I0gq() { return {-3.0*qcd::Cf/2.0, -qcd::Cf};}
    constexpr std::pair<double, double> I0qg() { return {1.0/3.0, -0.5};}
    
    
    /*   This is the regularized integral. Thus we have
     *
     *    Iqq(xb, x) = rIqq(xb, x) - 2.0*qcd::Cf*std::log(1.0 + xb*log(1-x))/xb
     *    Igg(xb, x) = rIgg(xb, x) - 2.0*qcd::Ca*std::log(1.0 + xb*log(1-x))/xb
     */
    static std::pair<double, double> rI0qq(double xb, double loz)
    {
      std::pair<double, double> res;
      double oz = std::exp(loz);
      double z = 1.0-oz;
      
      /*   soft part   */
      res.first = 2.0*qcd::Cf*oz*(xb > 1e-15 ? expint_Ei_scaled(1.0/xb + loz)/xb : 1.0);
      
      /*    pure collinear part   */
      res.first += qcd::Cf*z*(1.0-0.5*z);
      res.second = qcd::Cf*(-z+std::log(z));
      
      return res;
    }
    
    static std::pair<double, double> rI0gg(double xb, double loz)
    {
      std::pair<double, double> res;
      double oz = std::exp(loz);
      double z = 1.0-oz;
      double lz = std::log(z);
      
      /*   soft part   */
      res.first = 2.0*qcd::Ca*oz*(xb > 1e-15 ? expint_Ei_scaled(1.0/xb + loz)/xb : 1.0);
      
      /*    pure collinear part   */
      res.first += 2.0*qcd::Ca*(z*(0.5*z - 1.0 - z*z/3.0) + lz);
      res.second = 2.0*qcd::Ca*(z - 0.5*z*z - 1.0/z - lz);
      
      return res;
    }
    
    
    inline std::pair<double, double> I0qq(double lr, double xb, double v)
    {
      double ov = 1.0-v;
      if(xb < 1e-13) return ov < 1e-14 ? I0qq() : I0qq(xb, std::log(ov));
      if(ov < 1e-14) return rI0qq(xb, lr/xb);
      return I0qq(xb, std::max(std::log(ov), lr/xb));
    }
    
    inline std::pair<double, double> I0gg(double lr, double xb, double v)
    {
      double ov = 1.0-v;
      if(xb < 1e-13) return ov < 1e-14 ? I0gg() : I0gg(xb, std::log(ov));
      if(ov < 1e-14) return rI0gg(xb, lr/xb);
      return I0gg(xb, std::max(std::log(ov), lr/xb));
    }
    
    inline std::pair<double, double> I0gq(double lr, double xb, double v)
    {
      double ov = 1.0-v;
      if(xb < 1e-13) return ov < 1e-14 ? I0gq() : I0gq(std::log(ov));
      if(ov < 1e-14) return I0gq(lr/xb);
      return I0gq(std::max(std::log(ov), lr/xb));
    }
    
    inline std::pair<double, double> I0qg(double lr, double xb, double v)
    {
      double ov = 1.0-v;
      if(xb < 1e-13) return ov < 1e-14 ? I0qg() : I0qg(std::log(ov));
      if(ov < 1e-14) return I0qg(lr/xb);
      return I0qg(std::max(std::log(ov), lr/xb));
    }
    
    
    /*-*-*-*-*-*- Integrals of the scheme dependent kernels -*-*-*-*-*-*/
    
    static double lo1qq(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0qq(xb, lu);
      auto Iv = I0qq(lr, xb, v);
      
      double res = Iv.first-u*Iv.second;
      res -= Iu.first-u*Iu.second;
      res *= v/(v-u);
      
      return res;
    }
    
    static double lo1gq(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0gq(lu);
      auto Iv = I0gq(lr, xb, v);
      
      double res = Iv.first-u*Iv.second;
      res -= Iu.first-u*Iu.second;
      res *= v/(v-u);
      
      return res;
    }
    
    static double lo1qg(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0qg(lu);
      auto Iv = I0qg(lr, xb, v);
      
      double res = Iv.first-u*Iv.second;
      res -= Iu.first-u*Iu.second;
      res *= v/(v-u);
      
      return res;
    }
    
    static double lo1gg(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0gg(xb, lu);
      auto Iv = I0gg(lr, xb, v);
      
      double res = Iv.first-u*Iv.second;
      res -= Iu.first-u*Iu.second;
      res *= v/(v-u);
      
      return res;
    }
    
    
    static double lo2qq(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0qq(xb, lu);
      auto Iv = I0qq(lr, xb, v);
      
      double res = v*Iv.second - Iv.first;
      res -= v*Iu.second - Iu.first;
      res *= u/(v-u);
      
      return res;
    }
    
    static double lo2gg(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0gg(xb, lu);
      auto Iv = I0gg(lr, xb, v);
      
      double res = v*Iv.second - Iv.first;
      res -= v*Iu.second - Iu.first;
      res *= u/(v-u);
      
      return res;
    }
    
    static double lo2gq(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0gq(lu);
      auto Iv = I0gq(lr, xb, v);
      
      double res = v*Iv.second - Iv.first;
      res -= v*Iu.second - Iu.first;
      res *= u/(v-u);
      
      return res;
    }
    
    static double lo2qg(double lr, double xb, double u, double v)
    {
      double lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 0.0;
      
      auto Iu = I0qg(lu);
      auto Iv = I0qg(lr, xb, v);
      
      double res = v*Iv.second - Iv.first;
      res -= v*Iu.second - Iu.first;
      res *= u/(v-u);
      
      return res;
    }
    
    
    static double lo3qq(double lr, double xb, double u)
    {
      double res, lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 2.0*qcd::Cf*std::log(1.0 + lr)/xb;
      
      auto Iu = I0qq(xb, lu);
      auto Iv = xb <= 1e-15 ? I0qq() : rI0qq(xb, lr/xb);
      
      res = Iv.first-u*Iv.second;
      res -= Iu.first-u*Iu.second;
      res /= 1.0-u;
      
      return res;
    }
    
    static double lo3gg(double lr, double xb, double u)
    {
      double res, lu = std::log(1.0-u);
      if(xb > 1e-15 && xb*lu <= lr) return 2.0*qcd::Ca*std::log(1.0 + lr)/xb;
      
      auto Iu = I0gg(xb, lu);
      auto Iv = xb <= 1e-15 ? I0gg() : rI0gg(xb, lr/xb);
      
      res = Iv.first-u*Iv.second;
      res -= Iu.first-u*Iu.second;
      res /= 1.0-u;
      
      return res;
    }
    
  }   //   namespace __details

  
  /*-*-*-*-*-*- Making the grids -*-*-*-*-*-*/
  
  void dglap::_M_xt_grid()
  {
    /*  These parameters are hard wired. I don't think
     *  that the users should change them. We can set up
     *  a large grid over a big interval and that should
     *  be OK for every processes.
     */
    constexpr unsigned nt = 256u;
    constexpr double dx = 0.005;
    constexpr double xmin = 1e-8, xmax = 0.9999;
    
    /****  Don't change beyond this point!   ****/
    {
      double lx = std::log(xmin), lox = std::log(1.0-xmax);
      double dlx = -dx*lx, dlox = dx*lox;
      
      unsigned int k = 0U;
      double xk, x1 = xmin;
      std::set<double> xx({1.0});
      
      while((xk = xmin*std::exp((k++)*dlx))-x1 < dx)
        xx.insert(x1 = xk);
      
      k = 0U;
      double x2 = xmax;
      while(x2-(xk = 1.0-std::exp(lox-(k++)*dlox)) < dx)
        xx.insert(x2 = xk);
      
      unsigned int nx = 1 + (x2-x1)/dx;
      double dmx = (x2-x1)/nx;
      
      for(k = 1; k < nx; k++)
        xx.insert(dmx*k+x1);
      
      
      /*   copy the basepoint to the _M_x array   */
      _M_x.assign(xx.cbegin(), xx.cend());
    }
    
    /*  t grid   */
    {
      double tmax = _M_as.tauk(_M_m2T);
      double dt = tmax/nt, tc = tmax-_M_as.tauk(_M_m2c), tb = tmax-_M_as.tauk(_M_m2b);
      std::set<double> tt{0.0, tmax};
      
      for(unsigned k = 1; k < nt; k++)
        tt.insert(dt*k);
      
      if(0 < tc && tc < tmax) tt.insert(tc);
      if(0 < tb && tb < tmax) tt.insert(tb);
      
      /*    copy the basepoint to the _M_t array   */
      _M_t.assign(tt.cbegin(), tt.cend());
    }
  }
  
  
  /*-*-*-*-*-*-*-*-*-*-*-*- weights of the DGLAP kernels  -*-*-*-*-*-*-*-*-*-*-*-*/
  
  __details::pdf_kernel dglap::_M_weights(ordering::scheme sch, unsigned i, unsigned k, unsigned it) const
  {
    double sp = _M_x[i]/_M_x[k+1];

    /*    set all elements of the kernel to zero   */
    __details::pdf_kernel res{0.0};
    
    /*   calculate the scale   */
    double tauk = _M_t.back()-_M_t[it];
    energy_sqr<double> q2 = _M_as.q2(tauk);
    
    /*   number of active flavours   */
    unsigned nf = 3u;
    
    if(q2 >= _M_m2c) ++nf;
    if(q2 >= _M_m2b) ++nf;
    
    /*   coupling and the cutoff parameter   */
    double asb = qcd::beta1(nf)*_M_as.ask(tauk);
    double lr = asb*std::log(_M_m2T/q2), xb = asb*static_cast<unsigned>(sch);
    double lrc = asb*std::log(std::max(_M_m2T, _M_m2c)/q2);
    double lrb = asb*std::log(std::max(_M_m2T, _M_m2b)/q2);
    
    
    if(i < k) {
      double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
      
      /*    gluon-gluon vertex      */
      res[0][0] = __details::lo1gg(lr, xb, sp, s) + __details::lo2gg(lr, xb, s, sm);
      
      /*    gluon-quark vertices    */
      res[0][1] = res[0][2] = res[0][3] = res[0][4] = __details::lo1gq(lr, xb, sp, s) + __details::lo2gq(lr, xb, s, sm);
      
      /*    quark-gluon vertices    */
      double sqg = __details::lo1qg(lr, xb, sp, s) + __details::lo2qg(lr, xb, s, sm);
      
      res[1][0] = 4.0*sqg;
      res[2][0] = 2.0*sqg;
      
      if(lrc <= 0.0) res[3][0] = 2.0*(__details::lo1qg(lrc, xb, sp, s) + __details::lo2qg(lrc, xb, s, sm));
      if(lrb <= 0.0) res[4][0] = 2.0*(__details::lo1qg(lrb, xb, sp, s) + __details::lo2qg(lrb, xb, s, sm));
      
      /*    quark-quark vertices    */
      res[1][1] = res[2][2] = __details::lo1qq(lr, xb, sp, s) + __details::lo2qq(lr, xb, s, sm);
      if(lrc <= 0.0) res[3][3] = __details::lo1qq(lrc, xb, sp, s) + __details::lo2qq(lrc, xb, s, sm);
      if(lrb <= 0.0) res[4][4] = __details::lo1qq(lrb, xb, sp, s) + __details::lo2qq(lrb, xb, s, sm);
      
      /*    non-singlet    */
      res[5][0] = res[1][1];
      res[5][1] = res[1][1];
    }
    
    /*    diagonal elements    */
    else if(i == k) {
      /*    gluon-gluon vertex      */
      res[0][0] = __details::lo3gg(lr, xb, sp) + qcd::G1g(nf);
      
      /*    gluon-quark vertices    */
      res[0][1] = res[0][2] = res[0][3] = res[0][4] = __details::lo1gq(lr, xb, sp, 1.0);
      
      /*    quark-gluon vertices    */
      double sqg = __details::lo1qg(lr, xb, sp, 1.0);
      
      res[1][0] = 4.0*sqg;
      res[2][0] = 2.0*sqg;

      if(lrc <= 0.0) res[3][0] = 2.0*__details::lo1qg(lrc, xb, sp, 1.0);
      if(lrb <= 0.0) res[4][0] = 2.0*__details::lo1qg(lrb, xb, sp, 1.0);
      
      /*    quark-quark vertices    */
      res[1][1] = res[2][2] = __details::lo3qq(lr, xb, sp) + qcd::G1q;
      if(lrc <= 0.0) res[3][3] = __details::lo3qq(lrc, xb, sp) + qcd::G1q;
      if(lrb <= 0.0) res[4][4] = __details::lo3qq(lrb, xb, sp) + qcd::G1q;
      
      
      /*   non-singlet   */
      res[5][0] = res[1][1];
      res[5][1] = res[1][1];
    }
    
    /*    we shouldn't call this function with k < i    */
    else throw "dglap::_M_weights(scheme, ...) : We have some trouble...";
    
    return res;
  }
  
  
  /*-*-*-*-*-*-*-*-*-*-*-*- do the evolution  -*-*-*-*-*-*-*-*-*-*-*-*/

  void dglap::evolve(ordering::scheme sch, std::function<void(double, double *)> func)
  {
    /*   setting the factorization scheme   */
    _M_scheme = sch;
    
    /*   setting up the grid    */
    _M_xt_grid();
    
    /*    upper indices on the x,t grids   */
    unsigned nx = _M_x.size()-1;
    unsigned nt = _M_t.size();
    
    /*    at x = 1 the pdf and its derivative is zero  */
    __details::pdf_matrix dh{nx+1, nt};
    
    _M_h.resize(nx+1, nt);
    
    /*    the function values at the fitting scale   */
    double __f[11], dp, up;
    double *f = __f+5;
    
    /*   flavors in the pdf_array
     *      0 => gluon
     *      1 => singlet = u+ubar + d+dbar
     *      2 => s+sbar
     *      3 => c+cbar
     *      4 => b+bbar
     *      5 => d-dbar
     *      6 => u-ubar
     *      7 => s-sbar
     *      8 => T3 = u+ubar - (d+dbar)
     */
    for(unsigned k = 0; k <= nx; k++)
    {
      __details::pdf_array& h0 = _M_h[k][0];
      
      func(_M_x[k], f);
      
      dp = f[-1]+f[1];
      up = f[-2]+f[2];
      
      //---- singlet ----
      h0[0] = f[0];
      h0[1] = dp+up;
      h0[2] = f[-3]+f[3];
      h0[3] = f[-4]+f[4];
      h0[4] = f[-5]+f[5];
      
      //---- valence ----
      h0[5] = f[1]-f[-1];
      h0[6] = f[2]-f[-2];
      h0[7] = f[3]-f[-3];
      
      //---- T3 ----
      h0[8] = up-dp;
    }
    
    /*     the t derivatives at fitting scale    */
    for(unsigned k = 0; k <= nx; k++) {
      dh[k][0] = __details::pdf_array{0.0};
      
      for(unsigned l = k; l < nx; l++)
        dh[k][0] += _M_weights(sch, k, l, 0)*_M_h[l][0];
    }
    
    /*    now we solve the equation iteratively   */
    static const __details::pdf_kernel one{1.0};
    
    for(unsigned int it = 1; it < nt; it++)
    {
      double dt = 0.5*(_M_t[it] - _M_t[it-1]);
      
      /*   the PDFs at x = 1 are always zero   */
      dh[nx][it] = __details::pdf_array{0.0};
      
      /*   we have to solve the equation for x < 1   */
      for(int k = nx-1; k >= 0; --k)
      {
        __details::pdf_array S{0.0};
        __details::pdf_kernel w{_M_weights(sch, k, k, it)};

        for(unsigned l = k+1; l < nx; l++)
          S += _M_weights(sch, k, l, it)*_M_h[l][it];
        
        S += w*(_M_h[k][it-1] + dt*dh[k][it-1]);
        w *= -dt;
        w += one;
        w.inverse();
        
        dh[k][it] = w*S;
        _M_h[k][it] = _M_h[k][it-1] + dt*(dh[k][it] + dh[k][it-1]);
      }
    }
  }
  

  /*-*-*-*-*-*-*-*-*- do some interpolation on the table after the evolution -*-*-*-*-*-*-*-*-*/
  
  namespace __details {
    
    template<typename RandomIterator>
    static inline unsigned locate(RandomIterator xb, RandomIterator xe, double x)
    {
      auto n = xe-xb;
      int ju = n, jm, jl = 0;
      
      while(ju-jl > 1) {
        jm=(ju+jl)/2;
        if(x >= xb[jm]) jl = jm;
        else ju = jm;
      }
      
      return static_cast<unsigned>(jl == n-1 ? jl-1 : jl);
    }
  }
  
  
  double dglap::_M_parton(int fl, unsigned ix, unsigned it) const
  {
    /*    trivial case    */
    if(std::abs(fl) == 6 || ix >= _M_x.size() || it >= _M_t.size()) return 0.0;
    
    const __details::pdf_array& h = _M_h[ix][it];
    
    /*   flavor combinations   */
    if(10 <= fl && fl < 19) return h[fl-10];
    
    /*   generic flavors   */
    switch(std::abs(fl)) {
      case 81: return h[1]+h[2]+h[3]+h[4]; break;
      case 82: return h[3]; break;
      case 83: return h[4]; break;
    }
    
    /*   usual flavors   */
    switch(fl) {
      //----- b and c quarks -----
      case -5: case 5: return 0.5*h[4]; break;
      case -4: case 4: return 0.5*h[3]; break;
        
      //----- strange quark -----
      case -3: return (h[2] - h[7])/2.0; break;
      case  3: return (h[2] + h[7])/2.0; break;
        
      //----- up quark -----
      case -2: return (h[1] + h[8] - 2.0*h[6])/4.0; break;
      case  2: return (h[1] + h[8] + 2.0*h[6])/4.0; break;
 
      //----- down quark -----
      case -1: return (h[1] - h[8] - 2.0*h[5])/4.0; break;
      case  1: return (h[1] - h[8] + 2.0*h[5])/4.0; break;
        
      //----- gluon -----
      case  0: return h[0]; break;
    }
    
    return 0.0;
  }
  
  
  double dglap::parton(int fl, double x, double tauk) const
  {
    /*    trivial case    */
    if(std::abs(fl) == 6 || x >= 0.999999) return 0.0;
    
    /*    finding the x, t interval   */
    double t = _M_t.back()-tauk;
    unsigned l = __details::locate(_M_x.begin(), _M_x.end(), x);
    unsigned k = __details::locate(_M_t.begin(), _M_t.end(), t);
    
    /*    the of the function in the corners of the interpolation square   */
    double f[] = {
      _M_parton(fl, l,   k), _M_parton(fl, l,   k+1),
      _M_parton(fl, l+1, k), _M_parton(fl, l+1, k+1)
    };
    
    /*    normalized variables    */
    double tt = (_M_t[k+1]-t)/(_M_t[k+1]-_M_t[k]);
    
    /*   When every point is well defined
     *   we intepolate the log of the function.
     */
    if(f[0] > 0.0 && f[1] > 0.0 && f[2] > 0.0 && f[3] > 0.0)
    {
      double xi = std::log(1.0/x-1.0);
      double xi1 = std::log(1.0/_M_x[l+1]-1.0);
      double xi2 = std::log(1.0/_M_x[l]-1.0);
      double xx = (xi2-xi)/(xi2-xi1);
      
      /*    interpolation in variable t    */
      double h1 = std::log(f[0])*tt + std::log(f[1])*(1.0-tt);
      double h2 = std::log(f[2])*tt + std::log(f[3])*(1.0-tt);
      
      /*    interpolation in variable x    */
      return std::exp(h1*(1.0-xx) + h2*xx);
    }
    
    /*     interpolation in variable t     */
    double h1 = f[0]*tt + f[1]*(1.0-tt);
    double h2 = f[2]*tt + f[3]*(1.0-tt);
    
    /*     interpolation in variable x     */
    double xx = (_M_x[l+1]-x)/(_M_x[l+1]-_M_x[l]);
    return h1*xx + h2*(1.0-xx);
  }
  
 
  double dglap::parton_for_mc(int fl, double xi, double tk) const
  {
    /*    transverse momentum odering and massless partons are simple    */
    int f = std::abs(fl);
    bool is_heavy = (4 == f || 5 == f || 12 == f || 13 == f || 82 == f || 83 == f);
    
    if(_M_scheme == ordering::scheme::kT || !is_heavy)
      return parton(fl, xi, tk);
        
    /*    massive parton in lambda and angular ordered scheme    */
    double r, m;
    double oxi = 1.0-xi;
    
    // XXX This is not pretty here! r,m should be dimensionful!!!
    if(_M_scheme == ordering::scheme::angular) {
      r = sqrt(_M_as.q2(tk)).value();
      m = sqrt(thrsT(fl)).value();
    } else {
      r = _M_as.q2(tk).value();
      m = thrsT(fl).value();
    }
    
    /*    absolute threshold scale   */
    if(r <= m) return 0.0;
    
    /*      eta and q2 variables      */
    double q = 0.5*(oxi*r + std::sqrt(sqr(oxi*r) + 4.0*m*(r-m)*oxi*xi))/oxi;
    double eta = 1.0 - (q-m)*oxi/(r-m);
    double kT = q*(1.0-eta);
    
    if(_M_scheme == ordering::scheme::angular)
      q *= q;

    energy_sqr<double> q2{q};
    
    return kT < 1.3*m ? heavy_quark_approx(fl, eta, _M_as.tauk(q2)) : operator()(fl, eta, q2);
  }
  

  
  double dglap::heavy_quark_approx(int fl, double eta, double tk) const
  {
    static const gauss_integral<50> integral;
    energy_sqr<double> m2 = thrsT(fl);
    double tauT = _M_as.tauk(thrsT(fl, eta));
    double fac = 1.0;
    
    if(tauT <= tk) return 0.0;
    
    /*   factor of two when it is needed     */
    switch(std::abs(fl)) {
      case 4: case 5: fac = 1.0; break;
      case 12: case 13: case 82: case 83: fac = 2.0; break;
      default: return 0.0;
    }
    
    /*    calculate the integrals    */
    return fac*(tauT-tk)*0.5*integral([=,this](double y) -> double {
      double tau = (tauT-tk)*0.5*(1+y) + tk, zmax = 1.0;
      
      switch(_M_scheme) {
        case ordering::scheme::kT: zmax = 1.0; break;
        case ordering::scheme::lambda: zmax = 1.0-m2/std::max(_M_as.q2(tau), m2); break;
        case ordering::scheme::angular: zmax = 1.0-std::sqrt(m2/std::max(_M_as.q2(tau), m2)); break;
       }
      
      if(zmax <= eta) return 0.0;
      
      return 0.5*(zmax-eta)*integral([=,this](double x) -> double {
        double z = 0.5*(1.0+x)*(zmax-eta) + eta;
        return 0.5*(1.0-2.0*z*(1.0-z))*parton(0, eta/z, tau);
      });
    });
  }
  

  

}  //   namespace duct



