/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 16/02/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <cmath>

/*   project headers   */
#include "bits/qcd-coupling.h"



namespace duct {
  namespace qcd {
  
    coupling::coupling(qcd::order ord, double asf, energy_sqr<double> q2f,
                       energy_sqr<double> m2c, energy_sqr<double> m2b, energy_sqr<double> m2t)
    : _M_ord{ord}
    {
      /*     thresholds    */
      _M_thrs[0] = _M_thrs[1] = _M_thrs[2] = _M_thrs[3] = 0.0_GeV2;
      _M_thrs[4] = m2c; _M_thrs[5] = m2b; _M_thrs[6] = m2t;
      
      /*      number of active flavors at the fitting scale    */
      unsigned nf = number_of_flavors(q2f);
      
      /*      setting the lambda_QCD for nf flavors    */
      _M_lmd[nf] = _M_astolmd(nf, asf, q2f);
      
      /*      setting the lambda_QCD for k > nf flavors    */
      for(unsigned k = nf+1; k <= 6; k++) {
        double as = _M_base_as(k-1, _M_thrs[k], _M_lmd[k-1]);
        _M_lmd[k] = _M_astolmd(k, as, _M_thrs[k]);
      }
      
      /*      setting the lambda_QCD for k < nf flavors    */
      for(unsigned k = nf; k > 3; --k)
        if(_M_thrs[k] > _M_lmd[k]) {
          double as = _M_base_as(k, _M_thrs[k], _M_lmd[k]);
          _M_lmd[k-1] = _M_astolmd(k-1, as, _M_thrs[k]);
        } else {
          _M_lmd[k-1] = _M_lmd[k];
        }
      
      /*     Just in case ....    */
      _M_lmd[0] = _M_lmd[1] = _M_lmd[2] = _M_lmd[3];
    }
    
    
    energy_sqr<double> coupling::_M_astolmd(unsigned int nf, double as, energy_sqr<double> q2) const
    {
      double b1 = beta1(nf);
      double t = 1.0/(b1*as);
      
      /*  it is a leading order evolution  */
      if(_M_ord == qcd::order::lo) return q2*std::exp(-t);
      
      /*  at NLO or higer order level returns with the NLO Lambda_QCD  */
      double b2 = beta2(nf);
      
      t += 1.0*b2/(b1*b1)*std::log(b2*as/(b1+b2*as));
      
      return q2*std::exp(-t);
    }

    
    double coupling::_M_base_as(unsigned int nf, energy_sqr<double> q2, energy_sqr<double> lmd2) const
    {
      double b1 = beta1(nf);
      double t = b1*log(q2/lmd2);
      double as = 1.0/t;
      
      /*  it is a leading order evolution  */
      if(_M_ord == qcd::order::lo) return as;
      
      /*  at NLO or higer order level returns with the NLO alpha_s  */
      double ox, br = beta2(nf)/b1;
      double x = (1.0+as*br)/as, lbr = br*std::log(br)-br;
      
      do {
        ox = x;
        x -= x*(x-br*std::log(x)+lbr-t)/(x-br);
      } while(std::abs(ox-x)/ox > 1e-6);
      
      return 1.0/(x-br);
    }

    
  }
}
