/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 27/09/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */

/*   project headers   */
#include <bits/mc-shower-aux.h>


namespace duct {
  
  namespace __details {
    
    
    shower_pdf_tbls::shower_pdf_tbls(const dglap& pdf)
    {
      /*   gauss integration to calculate the soft and collinear integrals   */
      static const gauss_integral<10> _S_integral;
      
      const dglap *d = &pdf;
      const qcd_coupling *as = std::addressof(pdf.coupling());
      
      /*   limits in xi  */
      _M_ximin = std::log(1.0/_eta_max - 1.0);
      _M_ximax = std::log(1.0/pdf.xmin() - 1.0);
      
      /*   grid distances   */
      _M_dtl = as->tauk(pdf.q2min())/_N_tau;
      _M_dtc = as->tauk(pdf.thrsT(4))/_N_tau;
      _M_dtb = as->tauk(pdf.thrsT(5))/_N_tau;
      
      _M_dxi = (_M_ximax-_M_ximin)/_N_eta;
      _M_dchi = 1.0/_N_chi;
      
      
      /*   the eta grid   */
      std::array<double, _N_eta+1> eta;
      
      for(std::size_t k = 0; k <= _N_eta; ++k)
        eta[k] = 1.0/(1.0+std::exp(k*_M_dxi + _M_ximin));
      
      
      /*   PDF tables
       *
       *         loop over light flavours     */
      for(int f = 0; f < 10; ++f)
      {
        double dt = _M_dtl;
        int fl = (9 == f ? 81 : f-3);
        _Array grid{_M_pdf[f]};
        
        if(7 == f) dt = _M_dtc;
        else if(8 == f) dt = _M_dtb;
        
        /*   loop over the grid points   */
        for(std::size_t k = 0; k <= _N_eta; k++)
          for(std::size_t l = 0; l <= _N_tau; l++)
            grid[k][l] = pdf.parton_for_mc(fl, eta[k], l*dt);
      }
      
      
      /*   COLLINEAR integrals
       *
       */
      auto icoll = [=](double eta1, double eta2, int fl, double tau) -> double
      {
        return _S_integral([=](double x) -> double {
          double jac = (0.5/eta1-0.5/eta2);
          double z = eta1*eta2/(eta2-0.5*(x+1.0)*(eta2-eta1));
          
          if(6 == fl) return jac*z*d->parton(0, z, tau);
          else if(7 == fl) return jac*z*d->parton(81, z, tau);
          else return jac*d->parton(fl, z, tau);
        });
      };
      
      
      /*         loop over light flavors     */
      for(int f = 0; f < 11; ++f)
      {
        double dt = _M_dtl;
        _Array grid{_M_coll[f]};
        
        if(7 == f) dt = _M_dtc;
        else if(8 == f) dt = _M_dtb;
        
        /*   loop over tau points    */
        for(std::size_t l = 0; l <= _N_tau; ++l)
        {
          grid[0][l] = icoll(_eta_max, 1.0, f-3, l*dt);
          for(std::size_t k = 0; k < _N_eta; ++k)
            grid[k+1][l] = grid[k][l] + icoll(eta[k+1], eta[k], f-3, l*dt);
        }
      }
      
      
      /*   SOFT integrals
       *
       */
      static const double cmx = -std::log(1.0-_eta_max);
      
      auto isoft = [=](double chi1, double chi2, int fl, double eta, double tau) -> double
      {
        double ret;
        unsigned alfa = static_cast<unsigned>(d->scheme());
        
        /*    transverse momentum ordering   */
        if(alfa == 0)
          ret = _S_integral([=](double x) -> double {
            double z = 1.0-(1.0-eta)*std::exp(-0.5*((chi2-chi1)*x+chi1+chi2)*cmx);
            return 0.5*(chi2-chi1)*z*d->parton(fl, eta/z, tau);
          });
        
        /*   lambda or angular ordering    */
        else {
          double asb = alfa*qcd::beta1(as->nf(tau))*as->ask(tau);
          double zmax = 1.0 - std::pow(as->q2min()/as->q2(tau), 1.0/alfa);
          
          ret = _S_integral([=](double x) -> double {
            double z = 1.0-(1.0-eta)*std::exp(-0.5*((chi2-chi1)*x+chi1+chi2)*cmx);
            return 0.5*(chi2-chi1)*z*d->parton(fl, eta/z, tau)/(1.0+asb*std::log(1.0-std::min(z,zmax)));
          });
        }
        
        return ret/std::max(1e-20, d->parton(fl, eta, tau));
      };
      
      
      /*    loop over the light flavors     */
      for(int f = 0; f < 9; ++f)
      {
        double dt = _M_dtl;
        _SoftArray grid{_M_soft[f]};
        
        if(7 == f) dt = _M_dtc;
        else if(8 == f) dt = _M_dtb;
        
        /*   loop over eta and tau points   */
        for(unsigned k = 0; k <= _N_eta; ++k)
          for(unsigned l = 0; l <= _N_tau; ++l)
          {
            /*   loop over chi points   */
            double tau = l*dt;
            const auto& g = grid[k][l];
            
            g[0] = 0.0;
            for(unsigned n = 0; n < _N_chi; ++n)
              g[n+1] = g[n] + isoft(std::pow(n*_M_dchi,10), std::pow((n+1)*_M_dchi,10), f-3, eta[k], tau);
          }
      }
    }
    
    
    double shower_pdf_tbls::parton(int fl, double eta, double tau) const
    {
      /*    trivial case    */
      if(fl == -4) fl = 4;
      if(fl == -5) fl = 5;
      if(std::abs(fl) == 6 || eta >= 0.9999999) return 0.0;
      if(std::abs(fl) == 81) fl = 6;
      
      /*    grid distance in variable tau   */
      double dt;
      
      switch(fl) {
        case 4:  dt = _M_dtc; break;
        case 5:  dt = _M_dtb; break;
        default: dt = _M_dtl;
      }
      
      /*    find the xi and tau interval   */
      double xi = (std::log(1.0/eta-1.0)-_M_ximin)/_M_dxi, t = tau/dt;
      int l = std::min((int)_N_eta-1, std::max(0, static_cast<int>(xi)));
      int k = std::min((int)_N_tau-1, std::max(0, static_cast<int>(t)));
      
      /*    the function in the corners    */
      _ConstArray grid{_M_pdf[fl+3]};
      double f[] = {grid[l][k], grid[l][k+1], grid[l+1][k], grid[l+1][k+1]};
      
      /*   When every point is well defined
       *   we intepolate the log of the function.
       */
      double xx = l+1-xi, tt = k+1-t;
      
      
      /*   When every point is well defined
       *   we intepolate the log of the function.
       */
      if(f[0] > 0.0 && f[1] > 0.0 && f[2] > 0.0 && f[3] > 0.0)
      {
        /*    interpolation in variable t    */
        double h1 = std::log(f[0])*tt + std::log(f[1])*(1.0-tt);
        double h2 = std::log(f[2])*tt + std::log(f[3])*(1.0-tt);
        
        /*    interpolation in variable x    */
        return std::exp(h1*xx + h2*(1.0-xx));
      }
      
      /*     interpolation in variable t     */
      double h1 = f[0]*tt + f[1]*(1.0-tt);
      double h2 = f[2]*tt + f[3]*(1.0-tt);
      
      /*     interpolation in variable x     */
      return h1*xx + h2*(1.0-xx);
    }
    
    
    double shower_pdf_tbls::integral_coll(int fl, double eta, double tau) const
    {
      /*   taking care of the symmetric c and b quarks   */
      if(fl == -4) fl = 4;
      if(fl == -5) fl = 5;

      /*    grid distance in variable tau   */
      double dt;
      
      switch(fl) {
        case 4:  dt = _M_dtc; break;
        case 5:  dt = _M_dtb; break;
        default: dt = _M_dtl;
      }
      
      /*    find the xi and tau interval   */
      double xi = (std::log(1.0/eta-1.0)-_M_ximin)/_M_dxi, t = tau/dt;
      int l = std::min((int)_N_eta-1, std::max(0, static_cast<int>(xi)));
      int k = std::min((int)_N_tau-1, std::max(0, static_cast<int>(t)));
      
      /*    the function in the corners    */
      _ConstArray grid{_M_coll[fl+3]};
      double f[] = {grid[l][k], grid[l][k+1], grid[l+1][k], grid[l+1][k+1]};
      
      /*   When every point is well defined
       *   we intepolate the log of the function.
       */
      double xx = l+1-xi, tt = k+1-t;
      
      if(f[0] > 0.0 && f[1] > 0.0 && f[2] > 0.0 && f[3] > 0.0) {
        double h1 = std::log(f[0])*tt + std::log(f[1])*(1.0-tt);
        double h2 = std::log(f[2])*tt + std::log(f[3])*(1.0-tt);
        
        return std::exp(h1*xx + h2*(1.0-xx));
      }
      
      /*    otherwise  simple linear interpolation   */
      double h1 = f[0]*tt + f[1]*(1.0-tt);
      double h2 = f[2]*tt + f[3]*(1.0-tt);
      
      return h1*xx + h2*(1.0-xx);
    }
    
    
    double shower_pdf_tbls::inverse_coll(double F, int fl, double tau) const
    {
      /*   taking care of the symmetric c and b quarks   */
      if(fl == -4) fl = 4;
      if(fl == -5) fl = 5;
      
      /*    grid distance in variable tau   */
      double dt;
      
      switch(fl) {
        case 4:  dt = _M_dtc; break;
        case 5:  dt = _M_dtb; break;
        default: dt = _M_dtl;
      }
      
      /*   finding the t interval and the normalized tau   */
      int k = static_cast<int>(tau /= dt);
      double tt = k+1-tau;
      
      /*   getting the right grid   */
      _ConstArray grid{_M_coll[fl+3]};
      
      /*   Finding the relevant x interval   */
      int ju = (int)_N_eta+1, jm, jl = 0;
      double f1, f2, F1 = 0.0, F2 = 0.0, integ, xx;
      
      while(ju-jl > 1) {
        jm = (ju+jl)/2;
        
        /*    Calculate the integral in the x gridpoint   */
        f1 = grid[jm][k];
        f2 = grid[jm][k+1];
        
        /*   When every point is well defined we intepolate the log of the function. */
        if(f1 > 0.0 && f2 > 0.0) integ = std::exp(std::log(f1)*tt + std::log(f2)*(1.0-tt));
        else integ = f1*tt + f2*(1.0-tt);
        
        if(F < integ) {ju = jm; F2 = integ;}
        else {jl = jm; F1 = integ;}
      }
      
      if(jl == (int)_N_eta) --jl;
      
      /*   get root  */
      if(F1 > 0.0 && F2 > 0.0) xx = std::log(F/F1)/std::log(F2/F1);
      else xx = (F-F1)/(F2-F1);
      
      return 1.0/(1.0+std::exp(_M_dxi*(jl+xx) + _M_ximin));
    }
    
    
    double shower_pdf_tbls::integral_soft(double chi, int fl, double eta, double tau) const
    {
      /*   taking care of the symmetric c and b quarks   */
      if(fl == -4) fl = 4;
      if(fl == -5) fl = 5;
      
      /*    grid distance in variable tau   */
      double dt;
      
      switch(fl) {
        case 4:  dt = _M_dtc; break;
        case 5:  dt = _M_dtb; break;
        default: dt = _M_dtl;
      }
      
      /*   adjusting aruments   */
      double c = std::pow(chi,0.1)/_M_dchi;
      
      /*    find the xi and tau interval   */
      double xi = (std::log(1.0/eta-1.0)-_M_ximin)/_M_dxi, t = tau/dt;
      int l = std::min((int)_N_eta-1, std::max(0, static_cast<int>(xi)));
      int k = std::min((int)_N_tau-1, std::max(0, static_cast<int>(t)));
      
      /*   simple linear interpolation in variable chi    */
      _ConstSoftArray grid{_M_soft[fl+3]};
      
      int i = static_cast<int>(c);
      double h1, h2, F1, F2, cc = i+1-c, xx = l+1-xi, tt = k+1-t;
      
      const double f1[] = {grid[l][k][i],   grid[l][k+1][i],   grid[l+1][k][i],   grid[l+1][k+1][i]};
      const double f2[] = {grid[l][k][i+1], grid[l][k+1][i+1], grid[l+1][k][i+1], grid[l+1][k+1][i+1]};
      
      /*   When every point is well defined
       *   we intepolate the log of the function.
       */
      h1 = f1[0]*tt + f1[1]*(1.0-tt);
      h2 = f1[2]*tt + f1[3]*(1.0-tt);
      F1 = h1*xx + h2*(1.0-xx);
      
      h1 = f2[0]*tt + f2[1]*(1.0-tt);
      h2 = f2[2]*tt + f2[3]*(1.0-tt);
      F2 = h1*xx + h2*(1.0-xx);
      
      if(F1 > 0.0 && F2 > 0.0)
        return std::exp(std::log(F1)*cc + std::log(F2)*(1.0-cc));
      
      return F1*cc + F2*(1.0-cc);
    }
    
    
    double shower_pdf_tbls::inverse_soft(double F, int fl, double eta, double tau) const
    {
      /*   taking care of the symmetric c and b quarks   */
      if(fl == -4) fl = 4;
      if(fl == -5) fl = 5;
      
      /*    grid distance in variable tau   */
      double dt;
      
      switch(std::abs(fl)) {
        case 4:  dt = _M_dtc; break;
        case 5:  dt = _M_dtb; break;
        default: dt = _M_dtl;
      }
      
      /*   finding the t interval and the normalized tau   */
      double xi = (std::log(1.0/eta-1.0)-_M_ximin)/_M_dxi, t = tau/dt;
      int l = std::min((int)_N_eta-1, std::max(0, static_cast<int>(xi)));
      int k = std::min((int)_N_tau-1, std::max(0, static_cast<int>(t)));
      double cc, xx = l+1-xi, tt = k+1-t;
      
      /*   getting the right grid   */
      _ConstSoftArray grid{_M_soft[fl+3]};
      
      /*   Finding the relevant x interval   */
      int ju = (int)_N_chi+1, jm, jl = 0;
      double f[4], h1, h2, F1 = 0.0, F2 = 0.0, integ;
      
      while(ju-jl > 1) {
        jm = (ju+jl)/2;
        
        f[0] = grid[l][k][jm];
        f[1] = grid[l][k+1][jm];
        f[2] = grid[l+1][k][jm];
        f[3] = grid[l+1][k+1][jm];
        
        /*   simple linear interpolation   */
        h1 = f[0]*tt + f[1]*(1.0-tt);
        h2 = f[2]*tt + f[3]*(1.0-tt);
        integ = h1*xx + h2*(1.0-xx);
        
        if(F < integ) {ju = jm; F2 = integ;}
        else {jl = jm; F1 = integ;}
      }
      
      if(jl == (int)_N_chi) --jl;
      
      /*   get the root  */
      if(F1 > 0.0 && F2 > 0.0) cc = std::log(F/F1)/std::log(F2/F1);
      else cc = (F-F1)/(F2-F1);
      
      return std::pow(_M_dchi*(jl+cc), 10);
    }
    
    
    void shower_pdf_tbls::write(std::ostream& os) const {
      duct::write(os, _M_ximin, _M_ximax, _M_dxi, _M_dtl, _M_dtc, _M_dtb, _M_dchi, _M_pdf, _M_coll, _M_soft);
    }
    
    void shower_pdf_tbls::read(std::istream& is) {
      duct::read(is, _M_ximin, _M_ximax, _M_dxi, _M_dtl, _M_dtc, _M_dtb, _M_dchi, _M_pdf, _M_coll, _M_soft);
    }
    
 
  
  
    double shower_aux_impl::parton(const qcd_coupling& as, int fl, double eta, energy_sqr<double> q2) const
    {
      int f = std::abs(fl);
      bool is_heavy = (4 == f || 5 == f);
      
      /*     light quark or kT scheme   */
      if(_M_scheme == ordering::scheme::kT || !is_heavy)
        return _M_ipdf.parton(fl, eta, as.tauk(q2));
      
      /*   mapping to (xi,r2)   */
      double xi;
      energy_sqr<double> r2, m2 = thrsT(fl);
      
      switch(_M_scheme) {
        case ordering::scheme::lambda:
          xi = eta/(1.0-m2/q2);
          r2 = q2-eta*m2/(1.0-eta);
          break;
          
        case ordering::scheme::angular:
          xi = eta/(1.0-std::sqrt(m2/q2));
          r2 = sqr(sqrt(q2)-eta*sqrt(m2)/(1.0-eta));
          break;
          
        default:
          return 0.0;
      }
      
      return (r2 <= m2 ? 0.0 : _M_ipdf.parton(fl, xi, as.tauk(r2)));
    }
    
    
    
    std::pair<double,double>
    shower_aux_impl::
    inverse_coll(const qcd_coupling& as, double chi, double x, int fl, double eta, energy_sqr<double> q2) const
    {
      double xmin = eta;
      
      switch(_M_scheme) {
        case ordering::scheme::kT: xmin /= x; break;
        case ordering::scheme::lambda: xmin /= std::min(x, 1.0-_M_m2T/q2); break;
        case ordering::scheme::angular: xmin /= std::min(x, 1.0-std::sqrt(_M_m2T/q2)); break;
      }
      
      /*   check if it is possible to make the splitting    */
      if(xmin >= 1.0 || q2 < _M_m2T) return {eta, 0.0};
      
      /*   calculate the inverse   */
      double tau = as.tauk(q2);
      double jac = _M_ipdf.integral_coll(fl, xmin, tau);
      double z = eta/_M_ipdf.inverse_coll(chi*jac, fl, tau);
      
      /*   check if it is in the allowed region XXX Understand this!!!   */
      if(eta/z <= xmin) return {(eta/xmin-eta)*0.99+eta, 0.0};
      
      return {z, z*(fl == 6 || fl == 7 ? z*jac/eta : jac)};
    }
    
    
    std::pair<double,double>
    shower_aux_impl::
    inverse_soft(const qcd_coupling& as, double xi, double x, int fl, double eta, energy_sqr<double> q2) const
    {
      static const double cmx = -std::log(1.0-shower_pdf_tbls::_eta_max);
      
      /*    adjusting the upper limit   */
      double xmax = x;
      
      switch(_M_scheme) {
        case ordering::scheme::kT: xmax = x; break;
        case ordering::scheme::lambda: xmax = std::min(x, 1.0-_M_m2T/q2); break;
        case ordering::scheme::angular: xmax = std::min(x, 1.0-std::sqrt(_M_m2T/q2)); break;
      }
      
      /*   check if it is possible to make the splitting    */
      if(xmax*0.99999 <= eta) return {eta, 0.0};
      
      /*   calculate the inverse   */
      double tau = as.tauk(q2);
      double chi_max = std::log((1.0-eta)/(1.0-xmax))/cmx;
      
      double jac = _M_ipdf.integral_soft(chi_max, fl, eta, tau);
      double chi = _M_ipdf.inverse_soft(xi*jac, fl, eta, tau);
      double z = 1.0-(1.0-eta)*std::exp(-chi*cmx);
      
      /*   check if it is in the allowed region  XXX Understand this!!!  */
      if(z >= xmax) return {(xmax-eta)*0.99+eta, 0.0};
      
      return {z, jac*cmx*(1.0-z)};
    }
    
    
    std::pair<double,double>
    shower_aux_impl::
    inverse(const qcd_coupling& as, double xi, splitting_type idx, double x, int fl, double eta, energy_sqr<double> q2) const
    {
      /*   freezing the scale close to the heavy flavor threshold   */
      double tau = as.tauk(q2), tauT = as.tauk(thrsT(fl,eta));
      double tauF = tauT*0.95, xmax = x;
      
      /*   freeze the scale only for the flavour changing collinear vertex    */
      if(splitting_type::flavor_changing_collinear == idx && tauF < tau) q2 = as.q2(tauF);
      
  
      switch(_M_scheme) {
        case ordering::scheme::kT: xmax = x; break;
        case ordering::scheme::lambda: xmax = std::min(x, 1.0-thrsT(fl)/q2); break;
        case ordering::scheme::angular: xmax = std::min(x, 1.0-std::sqrt(thrsT(fl)/q2)); break;
      }
      
      /*    calculating the inverse   */
      std::pair<double,double> res{eta, 0.0};
      
      switch(idx) {
        case splitting_type::soft:
          res = inverse_soft(as, xi, xmax, fl, eta, q2);
          res.second *= tauT - tau;
          break;
          
        case splitting_type::collinear: {
          res = inverse_coll(as, xi, xmax, (fl ? fl : 6), eta, q2);
          if(res.second > 0.0) {
            res.second *= eta/parton(as, fl, eta, q2);
            res.second *= tauT - tau;
          }
        } break;
          
        case splitting_type::flavor_changing_collinear: {
          res = inverse_coll(as, xi, xmax, (fl ? 0 : 7), eta, q2);
          
          if(res.second > 0.0) {
            res.second *= eta/parton(as, fl, eta, q2);
            if(std::abs(fl) >= 4) res.second *= tauT - (tau > tauF ? tauF : tau);
            else res.second *= tauT - tau;
          }
        } break;
      }
      
      if(res.first != res.first || res.second != res.second) {
        std::cout<<"NaN : shower_aux_impl::inverse(...) : "<<res.first<<", "<<res.second<<", "<<xi<<", "<<xmax<<", "<<eta<<", "<<q2<<std::endl;
        return {(xmax-eta)*xi+eta, 0.0};
      }

      
      return res;
    }
    
    
    
  }   //   namespace __details
  

  
  
  void shower_aux::read(std::istream& is, const qcd_coupling& as)
  {
    char c;
    duct::read(is, _M_anti, c);
    
    /*   read if it is in the file    */
    if('A' == c) {
      _M_aux = std::make_shared<__details::shower_aux_impl>();
      _M_aux -> read(is);
      _M_as = as;
    }
    
    /*   otherwise reset the object   */
    else {
      _M_as = qcd_coupling{};
      _M_aux.reset();
      _M_anti = false;
    }
  }
  
  
  void shower_aux::write(std::ostream& os) const
  {
    if(_M_aux) {
      duct::write(os, _M_anti, 'A');
      _M_aux->write(os);
    } else duct::write(os, _M_anti, 'D');
  }
  
}   //   namespace duct



