/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */



/*   std includes   */
#include <cmath>

/*   project headers   */
#include "bits/qcd-color.h"
#include "bits/mc-splitting-func.h"
#include "cxx/math-support.h"


constexpr double __one_pi = std::numbers::pi;
constexpr double __two_pi = 2.0*std::numbers::pi;


namespace duct {

  
  auto splitting_functions<fin>::
  splitfunc(const hard_params<fin>& pars, double y, double u) -> stackvector<result, 6u>
  {
    stackvector<result, 6u> res;
    
    /*    aliases   */
    const double &ymax = pars.ymax, &al = pars.al;
    
    /*   safety   */
    if(y >= ymax) return res;
    
    /*    standard variables    */
    bool is_special = (std::abs(1.0-ymax) < 1.0e-5);
    
    double oy = 1.0+y;
    double lmd = is_special ? 1.0 - y : std::sqrt(oy*oy-4.0*al*y);
    double x0 = (oy-lmd)/(oy+lmd);
    
    
    /*    gluon emission   */
    for(const auto& s : pars.spectators)
    {
      double weight = 0.0;
      
      /*    soft part of the gluon emission   */
      if(pars.emitter != s.spectator)
      {
        if(is_special && s.spectator > 0)
        {
          weight = 2.0*u/(x0/(x0 + u*(1.0-x0)) + 1.0-u);
        }
        else
        {
          /*    do lorentz transformation if the spectator is in the final state    */
          double xi = s.xi;
          
          if(s.spectator > 0)
            xi /= xi + (1.0-xi)*sqr((2.0*al-oy-lmd)/(2.0*al-2.0));
          
          /*    basic variables   */
          double A = 2.0*(xi+(1.0-xi)*x0);
          double B = x0/(x0 + u*(1.0-x0));
          double C = xi*(1.0-u) + (1.0-xi)*u*x0;
          double S2 = 4.0*xi*(1.0-xi)*u*(1.0-u)*x0;
          
          weight = (A + 2.0*B)/std::sqrt(std::abs(sqr(B+C)-S2))-2.0;
        }
      }
      
      /*   collinear part of the gluon emission   */
      else
      {
        /*    g -> g + g splitting   */
        if(0 == pars.fl)
        {
          double dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
          weight = u*(1.0-u)*(1.0 - 0.5*dlt*(2.0-dlt));
        }
  
        /*    q -> q + g splitting   */
        else
        {
          double rg = lmd+y-1.0;
          double h = (rg*rg + 4.0*y)/(2.0*lmd), gam = 0.5*(oy+lmd);
          double u1 = 1.0-u+x0*u;
          
          weight = h*(-1.0 + (1.0 + x0 - x0/u1)/u1) + gam*(1.0-u);
        }
      }
      
      /*   setting the return value   */
      res.push_back({pars.emitter, s.spectator, pars.fl, 0, lmd*s.cfac*weight});
    }
    
    /*    g -> q + q~ splitting  */
    if(0 == pars.fl)
    {
      /*    light flavors    */
      constexpr unsigned int nf = 5u;

      /*  The net color factors is Tr = 1/2.   */
      res.push_back({pars.emitter, pars.emitter, 81, -81, 0.5*lmd*nf*(1.0 - 2.0*u*(1.0-u))});
    }
    
    return res;
  }
  

  double splitting_functions<fin>::
  splitfunc_sum(const hard_params<fin>& pars, double y, double u)
  {
    double res = 0.0;
    
    /*    aliases   */
    const double &ymax = pars.ymax, &al = pars.al;
    
    /*   safety   */
    if(y >= ymax) return res;
    
    /*    standard variables    */
    bool is_special = (std::abs(1.0-ymax) < 1.0e-5);
    
    double oy = 1.0+y;
    double lmd = is_special ? 1.0 - y : std::sqrt(oy*oy-4.0*al*y);
    double x0 = (oy-lmd)/(oy+lmd);
    
    
    /*    gluon emission   */
    for(const auto& s : pars.spectators)
    {
      double weight = 0.0;
      
      /*    soft part of the gluon emission   */
      if(pars.emitter != s.spectator)
      {
        if(is_special && s.spectator > 0)
        {
          weight = 2.0*u/(x0/(x0 + u*(1.0-x0)) + 1.0-u);
        }
        else
        {
          /*    do lorentz transformation if the spectator is in the final state    */
          double xi = s.xi;
          
          if(s.spectator > 0)
            xi /= xi + (1.0-xi)*sqr((2.0*al-oy-lmd)/(2.0*al-2.0));
          
          /*    basic variables   */
          double A = 2.0*(xi+(1.0-xi)*x0);
          double B = x0/(x0 + u*(1.0-x0));
          double C = xi*(1.0-u) + (1.0-xi)*u*x0;
          double S2 = 4.0*xi*(1.0-xi)*u*(1.0-u)*x0;
          
          weight = (A + 2.0*B)/std::sqrt(std::abs(sqr(B+C)-S2))-2.0;
        }
      }
      
      /*   collinear part of the gluon emission   */
      else
      {
        /*    g -> g + g splitting   */
        if(0 == pars.fl)
        {
          double dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
          weight = u*(1.0-u)*(1.0 - 0.5*dlt*(2.0-dlt));
        }
  
        /*    q -> q + g splitting   */
        else
        {
          double rg = lmd+y-1.0;
          double h = (rg*rg + 4.0*y)/(2.0*lmd), gam = 0.5*(oy+lmd);
          double u1 = 1.0-u+x0*u;
          
          weight = h*(-1.0 + (1.0 + x0 - x0/u1)/u1) + gam*(1.0-u);
        }
      }
      
      /*   setting the return value   */
      res +=  s.cfac*weight;
    }
    
    /*    g -> q + q~ splitting  */
    if(0 == pars.fl)
    {
      /*    light flavors    */
      constexpr unsigned int nf = 5u;

      /*  The net color factors is Tr = 1/2.   */
      res += 0.5*nf*(1.0 - 2.0*u*(1.0-u));
    }
    
    return lmd*res;
  }
  


  /*   Function to generate the azimuthal angle   */
  double splitting_functions<fin>::
  azimuthal_angle(int spectator, double al, double xi, double y, double u, double rho)
  {
    /*    standard variables    */
    double oy = 1.0+y;
    double lmd = (std::abs(1.0-al) < 1.0e-6 ? 1.0 - y : std::sqrt(oy*oy-4.0*al*y));
    double x0 = (oy-lmd)/(oy+lmd);

    /*    do lorentz transformation if the spectator is in the final state    */
    if(spectator > 0)
      xi /= xi + (1.0-xi)*sqr((2.0*al-oy-lmd)/(2.0*al-2.0));
    
    
    /*    basic variables   */
    double A = 2.0*(xi+(1.0-xi)*x0);
    double B = x0/(x0 + u*(1.0-x0));
    double C = xi*(1.0-u) + (1.0-xi)*u*x0;
    double S = 2.0*std::sqrt(std::abs(xi*(1.0-xi)*u*(1.0-u)*x0));
    double I1 = (A + 2.0*B)/std::sqrt(std::abs(sqr(B+C)-S*S));
    double R1 = std::sqrt((B+C+S)/(B+C-S));

    
    /*   we need two lamdas   */
    auto function_phi = [=](double phi) -> double {
      return (A + 2.0*B)/(B+C - S*std::cos(__two_pi*phi)) - 2.0;
    };
    
    auto primitive_phi = [=](double phi) -> double {
      phi *= __two_pi;
      
      /*   if phi is 2pi then we have simpler expression   */
      if(std::abs(1.0-phi/__two_pi) < 1e-13 || phi >= __two_pi)
        return I1 - 2.0;
      
      return (I1*((phi <= __one_pi ? 0.0 : __one_pi) + std::atan(R1*std::tan(0.5*phi))) - phi)/__one_pi;
    };
    
    
    /*  we use Newton method to find azimuthal angle     */
    double mi = rho;
    
    //==============
    {
      /*   otherwise we have to use Newton method to find the root   */
      double lo = 0.0, hi = 1.0;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      //----- normalization ----
      rho *= primitive_phi(1.0);
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = primitive_phi(mi)))/function_phi(mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
    }
    
    
    return mi*__two_pi;
  }
  
 }   //   namespace duct
  



namespace duct {

  
  auto splitting_functions<ini>::
  splitfunc(const hard_params<ini>& pars, double y, double z) -> stackvector<result, 6u>
  {
    stackvector<result, 6u> res;

    /*   y & z dependent variables   */
    const double oz = 1.0-z, oy = 1.0+y;
    const double x = z*y/oz, ozy = 1.0+z*y, za = z*oy;
    const double oza = 1.0-za;


    /*   loop over spectators   */
    for(const auto& s : pars.spectators)
    {
      double weight;
      
      /*   soft collinear part of the gluon emission   */
      if(pars.emitter != s.spectator)
      {
        if(std::abs(1.0-s.xi) < 1.0e-6)
          weight = 2.0*z/oz;
        else
        {
          double od = 1.0-ozy*oy*(1.0-s.xi);
          double w = od/std::sqrt(od*od + 4.0*x*x*(1.0-od)) - 1.0;
          weight = 2.0*z*(z*oy/ozy*w + 1.0)/oz;
        }
      }
      
      /*    pure collinear part of the gluon emission   */
      else
      {
         weight = (0 == pars.fl ? 2.0*oza/oy*(1.0/za + z*oza/oz) : oza);
      }
      
      /*   setting the return value   */
      res.push_back({pars.emitter, s.spectator, pars.fl, 0, s.cfac*weight});
    }
    
    /*     g(a) -> q(ahat) + q(m+1)  flavor changing splitting   */
    if(0 == pars.fl)
      res.push_back({pars.emitter, pars.emitter, 85, -85, qcd::Cf*(z + 2.0*oza/za/oy)});
  
    /*     q(a) -> g(ahat) + q~(m+1)  flavor changing splitting  (the color factor is Tr = 0.5) */
    else res.push_back({pars.emitter, pars.emitter, 0, pars.fl, 0.5*(1.0 - 2.0*za*oza)});
  
    
    return res;
  }
  
 
  std::pair<double, double>
  splitting_functions<ini>::splitfunc_sum(const hard_params<ini>& pars, double y, double z)
  {
    std::pair<double, double> res{0.0, 0.0};

    /*   y & z dependent variables   */
    const double oz = 1.0-z, oy = 1.0+y;
    const double x = z*y/oz, ozy = 1.0+z*y, za = z*oy;
    const double oza = 1.0-za;


    /*   loop over spectators   */
    for(const auto& s : pars.spectators)
    {
      double weight;
      
      /*   soft collinear part of the gluon emission   */
      if(pars.emitter != s.spectator)
      {
        if(std::abs(1.0-s.xi) < 1.0e-6)
          weight = 2.0*z/oz;
        else
        {
          double od = 1.0-ozy*oy*(1.0-s.xi);
          double w = od/std::sqrt(od*od + 4.0*x*x*(1.0-od)) - 1.0;
          weight = 2.0*z*(z*oy/ozy*w + 1.0)/oz;
        }
      }
      
      /*    pure collinear part of the gluon emission   */
      else
      {
         weight = (0 == pars.fl ? 2.0*oza/oy*(1.0/za + z*oza/oz) : oza);
      }
      
      /*   setting the return value   */
      res.first += s.cfac*weight;
    }
    
    /*     g(a) -> q(ahat) + q(m+1)  flavor changing splitting   */
    if(0 == pars.fl)
      res.second = qcd::Cf*(z + 2.0*oza/za/oy);
  
    /*     q(a) -> g(ahat) + q~(m+1)  flavor changing splitting  (the color factor is Tr = 0.5) */
    else res.second = 0.5*(1.0 - 2.0*za*oza);
    
    
    return res;
  }
  

  double splitting_functions<ini>::azimuthal_angle(double xi_in, double y, double z, double rho)
  {
    /*    xa and xb coordinates of the emitted patons   */
    double saQ = 1.0/z, xi = 1.0-xi_in;
    double xa = (1.0/z-1.0-y), xb = z*y;
    
    /*    boost parameters    */
    double kT = std::sqrt(xa*xb), omega = 1.0+y;
    double alpha = omega*xi, ysaQ = saQ+y;
    
    /*    coordinates of the spectator after the boost     */
    double sxi = std::sqrt(xi*(1.0-xi));
    double betap = (1.0-xi + (xi*kT+2.0*sxi)*kT)/omega;
    double betam = (1.0-xi + (xi*kT-2.0*sxi)*kT)/omega;
    
    /*    cached variables     */
    double Cp = alpha*xb + betap*xa + (sxi + xi*kT)*2.0*kT;
    double Cm = alpha*xb + betam*xa - (sxi - xi*kT)*2.0*kT;
    
    double Bp = betap/z;
    double Bm = betam/z;
    
    double Ap = Cp*saQ + y*(alpha + betap/z);
    double Am = Cm*saQ + y*(alpha + betam/z);
    
    double rAp = std::sqrt(Ap);
    double rAm = std::sqrt(Am);
    
    double R2 = 1.0/ysaQ;
    double R1 = -xi*y + R2*(saQ*(1.0-xi) - xi*y);


    /*    we need the some lambdas   */
    auto function_phi = [=](double phi) -> double {
      double c = std::cos(__two_pi*phi);
      double gp = 0.5*(1.0+c), gm = 0.5*(1.0-c);
      return (Bp*gp+Bm*gm)/(Ap*gp+Am*gm);
    };
    
    
    auto primitive_phi = [=](double phi) -> double {
      /*    the full integral   */
      if(phi == 1.0) return R1/(rAp*rAm) + R2;
      
      /*    the primitive function in the general case     */
      phi *= __two_pi;
      
      double t = std::tan(0.5*phi), one = (phi <= __one_pi ? 0.0 : __one_pi);
      double iA = (one + std::atan(rAm*t/rAp))/(rAp*rAm);
      
      return (R1*iA + R2*0.5*phi)/__one_pi;
    };

    /*  we use Newton method to find azimuthal angle     */
    double mi = rho;
    
    //==============
    {
      /*   otherwise we have to use Newton method to find the root   */
      double lo = 0.0, hi = 1.0;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      //----- normalization ----
      rho *= primitive_phi(1.0);
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = primitive_phi(mi)))/function_phi(mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
    }
    
    return mi*__two_pi;
  }
  

}   //namespace duct

