/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 06.10.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_buffer_h__
#define __cxx_buffer_h__ 1



/*   std includes   */
#include <iterator>
#include <stdexcept>
#include <algorithm>
#include <type_traits>
#include <initializer_list>


/*   project headers   */



namespace duct {
  
  template<class _Tp, std::size_t _BufferSize>
  class buffer
  {
  public:
    /*    ususal typedefs   */
    using value_type = _Tp;
    using size_type = std::size_t;
    
    using pointer = value_type *;
    using iterator = value_type *;
    using reference = value_type&;
    
    using const_pointer = const value_type *;
    using const_iterator = const value_type *;
    using const_reference = const value_type&;
    
    using reverse_iterator = std::reverse_iterator<iterator>;
    using const_reverse_iterator = std::reverse_iterator<const_iterator>;
    
    
    /*   constructors   */
    explicit buffer(std::size_t s = 0u)
    : _M_size{std::min(s, _BufferSize)} {}
    
    buffer(std::size_t s, const _Tp& c)
    : _M_size{std::min(s, _BufferSize)}
    {
      for(std::size_t k = 0; k < _M_size; ++k)
        _M_data[k] = c;
    }
    
    buffer(std::initializer_list<value_type> il)
    : _M_size{std::min(il.size(), _BufferSize)}
    {
      std::size_t k = 0;
      for(auto ik = il.begin(); ik < il.begin()+_M_size; ++ik)
        _M_data[k++] = *ik;
    }
    
    
    /*   iterators   */
    iterator begin() { return reinterpret_cast<pointer>(_M_data);}
    iterator end() { return begin() + _M_size;}
    
    const_iterator begin() const { return reinterpret_cast<const_pointer>(_M_data);}
    const_iterator end() const { return begin() + _M_size;}
    
    const_iterator cbegin() const { return reinterpret_cast<const_pointer>(_M_data);}
    const_iterator cend() const { return cbegin()+_M_size;}
    
    reverse_iterator rbegin() { return reverse_iterator(end());}
    reverse_iterator rend() { return reverse_iterator(begin());}
    
    const_reverse_iterator rbegin() const { return reverse_iterator(end());}
    const_reverse_iterator rend() const { return reverse_iterator(begin());}
    
    const_reverse_iterator crbegin() const { return reverse_iterator(end());}
    const_reverse_iterator crend() const { return reverse_iterator(begin());}
    
    
    /*   resizeing the vector   */
    void resize(std::size_t new_size) {
      _M_size = std::min(_BufferSize, new_size);
    }
    
    void push_back(const value_type& v) {
      if(_M_size < _BufferSize) _M_data[_M_size++] = v;
    }
    
    void push_back(value_type&& v) {
      if(_M_size < _BufferSize) _M_data[_M_size++] = std::move(v);
    }
    
    
    template<typename... _Args>
    void emplace_back(_Args&&... args) {
      if(_M_size < _BufferSize) _M_data[_M_size++] = value_type{std::forward<_Args>(args)...};
    }
    
    void pop_back() { --_M_size;}
    
    /*   remove all the elements   */
    void clear() { _M_size = 0;}
    
    
    /*   element access    */
    reference operator[](std::size_t k) { return _M_data[k];}
    constexpr const_reference operator[](std::size_t k) const { return _M_data[k];}
    
    reference at(std::size_t k) {
      if(k < _M_size) return _M_data[k];
      else throw std::out_of_range{"duct::buffer<...> : out of range"};
    }
    
    const_reference at(std::size_t k) const {
      if(k < _M_size) return _M_data[k];
      else throw std::out_of_range{"duct::buffer<...> : out of range"};
    }
    
    pointer data() { return reinterpret_cast<pointer>(_M_data);}
    const_pointer data() const { return reinterpret_cast<const_pointer>(_M_data);}
    
    reference back() { return operator[](_M_size-1);}
    const_reference back() const { return operator[](_M_size-1);}
    
    /*   structural informations   */
    constexpr std::size_t size() const { return _M_size;}
    constexpr std::size_t capacity() const { return _BufferSize;}
    constexpr bool empty() const { return !_M_size;}
    
  private:
    /*   actual size of the vector   */
    std::size_t _M_size;
    
    /*   the underlying buffer   */
    value_type _M_data[_BufferSize];
  };
  

  
  
}   //   namespace duct

#endif /* __cxx_buffer_h__ */
