/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 27.12.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __math_linear_algebra_h__
#define __math_linear_algebra_h__ 1



/*   std includes   */
#include <algorithm>
#include <iostream>
#include <optional>


/*   project headers   */
#include <cxx/concepts.h>


namespace duct {
  
  /*   Defining vector as arithmetic type    */
  namespace linalg {
    
    
    template<ComplexandRealNumbers _Tp, std::size_t _Dim, BraKetTags _BraKet>
    class linvector
    {
    public:
      /*    usual public types    */
      using value_type = _Tp;
      using size_type = std::size_t;
      using reference = value_type&;
      using const_reference = const value_type&;
      using pointer = value_type*;
      using const_pointer = const value_type*;
      using iterator = value_type*;
      using const_iterator = const value_type*;
      
      
      /*   default constructor    */
      linvector() {
        std::fill(begin(), end(), 0.0);
      }
      
      linvector(std::initializer_list<_Tp> c)
      {
        const auto s = std::min(c.size(), _Dim);
        auto ic = c.begin();
        
        for(std::size_t k = 0; k < s; ++k, ++ic)
          _M_elem[k] = *ic;
        
        for(std::size_t k = s; k < _Dim; ++k)
          _M_elem[k] = 0.0;
      }
      
      
      /*   copy   */
      linvector(const linvector&) = default;
      linvector& operator=(const linvector&) = default;
      
      template<ComplexandRealNumbers _Xp>
      explicit linvector(const linvector<_Xp, _Dim, _BraKet>& v) {
        std::copy(v.begin(), v.end(), begin());
      }
      
      template<ComplexandRealNumbers _Xp>
      linvector& operator=(const linvector<_Xp, _Dim, _BraKet>& v) {
        std::copy(v.begin(), v.end(), begin());
        return *this;
      }
      
      /*   move   */
      linvector(linvector&&) = default;
      linvector& operator=(linvector&&) = default;
      
      
      /*   iterators   */
      iterator begin() { return _M_elem;}
      iterator end() { return _M_elem+_Dim;}
      
      const_iterator begin() const { return _M_elem;}
      const_iterator end() const { return _M_elem+_Dim;}
      
      const_iterator cbegin() const { return _M_elem;}
      const_iterator cend() const { return _M_elem+_Dim;}
      
      /*   element access   */
      reference operator[](size_type k) { return _M_elem[k];}
      const_reference operator[](size_type k) const { return _M_elem[k];}
      
      /*   structural informations   */
      constexpr size_type size() const { return _Dim;}
      
      
      /*   mathematical operations   */
      template<ComplexandRealNumbers _Xp>
      linvector& operator+=(const linvector<_Xp, _Dim, _BraKet>& v)
      {
        for(size_type k = 0; k < _Dim; ++k)
          _M_elem[k] += v[k];
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      linvector& operator-=(const linvector<_Xp, _Dim, _BraKet>& v)
      {
        for(size_type k = 0; k < _Dim; ++k)
          _M_elem[k] -= v[k];
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      linvector& operator*=(const _Xp& x)
      {
        for(size_type k = 0; k < _Dim; ++k)
          _M_elem[k] *= x;
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      linvector& operator/=(const _Xp& x)
      {
        for(size_type k = 0; k < _Dim; ++k)
          _M_elem[k] /= x;
        
        return *this;
      }
      
    private:
      /*   data member is a simple C array    */
      value_type _M_elem[_Dim];
    };
    
    template<ComplexandRealNumbers _Tp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto operator-(const linvector<_Tp, _Dim, _BraKet>& a)
    {
      linvector<_Tp, _Dim, _BraKet> res;
      
      for(std::size_t k = 0; k < _Dim; ++k)
        res[k] = -a[k];
      
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto operator+(const linvector<_Tp, _Dim, _BraKet>& a, const linvector<_Xp, _Dim, _BraKet>& b)
    {
      linvector<result_of_add_t<_Tp, _Xp>, _Dim, _BraKet> res;
      
      for(std::size_t k = 0; k < _Dim; ++k)
        res[k] = a[k] + b[k];
      
      return res;
    }
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto operator-(const linvector<_Tp, _Dim, _BraKet>& a, const linvector<_Xp, _Dim, _BraKet>& b)
    {
      linvector<result_of_sub_t<_Tp, _Xp>, _Dim, _BraKet> res;
      
      for(std::size_t k = 0; k < _Dim; ++k)
        res[k] = a[k] - b[k];
      
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto operator*(const linvector<_Tp, _Dim, _BraKet>& a, const _Xp& b)
    {
      linvector<result_of_mul_t<_Tp, _Xp>, _Dim, _BraKet> res;
      
      for(std::size_t k = 0; k < _Dim; ++k)
        res[k] = a[k]*b;
      
      return res;
    }
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto operator/(const linvector<_Tp, _Dim, _BraKet>& a, const _Xp& b)
    {
      linvector<result_of_div_t<_Tp, _Xp>, _Dim, _BraKet> res;
      
      for(std::size_t k = 0; k < _Dim; ++k)
        res[k] = a[k]/b;
      
      return res;
    }
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto operator*(const _Xp& b, const linvector<_Tp, _Dim, _BraKet>& a)
    {
      linvector<result_of_mul_t<_Tp, _Xp>, _Dim, _BraKet> res;
      
      for(std::size_t k = 0; k < _Dim; ++k)
        res[k] = a[k]*b;
      
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto transpose(const linvector<_Tp, _Dim, _BraKet>& a)
    {
      linvector<_Tp, _Dim, adjoint_tag_of<_BraKet>> res;
      
      for(std::size_t r = 0; r < _Dim; ++r)
        res[r] = a[r];
      
      return res;
    }
    
    template<ComplexNumbers _Tp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto adjoint(const linvector<_Tp, _Dim, _BraKet>& a)
    {
      linvector<_Tp, _Dim, adjoint_tag_of<_BraKet>> res;
      
      for(std::size_t r = 0; r < _Dim; ++r)
        res[r] = std::conj(a[r]);
      
      return res;
    }
    
    template<RealNumbers _Tp, std::size_t _Dim, BraKetTags _BraKet>
    inline auto adjoint(const linvector<_Tp, _Dim, _BraKet>& a) {
      return tranpose(a);
    }
    
    
    /*    specializations    */
    template<ComplexandRealNumbers _Tp, std::size_t _Dim>
    using bravector = linvector<_Tp, _Dim, bra>;
    
    template<ComplexandRealNumbers _Tp, std::size_t _Dim>
    using ketvector = linvector<_Tp, _Dim, ket>;
    
    
    template<ComplexNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim>
    inline auto operator*(const bravector<_Tp, _Dim>& a, const ketvector<_Xp, _Dim>& b)
    {
      result_of_mul_t<_Tp, _Xp> res = 0.0;
      
      for(std::size_t r = 0; r < _Dim; ++r)
        res += a[r]*b[r];
      
      return res;
    }
    
    template<ComplexandRealNumbers _Tp,  std::size_t _Dim>
    inline std::ostream& operator<<(std::ostream& os, const bravector<_Tp, _Dim>& a)
    {
      os<<"<";
      
      for(std::size_t r = 0; r < _Dim-1; ++r)
        os<<a[r]<<", ";
      
      os<<a[_Dim-1]<<"|";
      
      return os;
    }
    
    template<ComplexandRealNumbers _Tp,  std::size_t _Dim>
    inline std::ostream& operator<<(std::ostream& os, const ketvector<_Tp, _Dim>& a)
    {
      os<<"|";
      
      for(std::size_t r = 0; r < _Dim-1; ++r)
        os<<a[r]<<", ";
      
      os<<a[_Dim-1]<<">";
      
      return os;
    }
    
  }   //  namespace linalg
  
  
  
  /*   Defining a matrix class as an arithmetic type    */
  namespace linalg {
    
    
    template<ComplexandRealNumbers _Tp>
    struct unit_matrix_arg
    {
      unit_matrix_arg() : coeff{1.0} {}
      explicit unit_matrix_arg(const _Tp& c) : coeff{c} {}
      _Tp coeff;
    };
    
    
    template<ComplexandRealNumbers _Tp, std::size_t _Rows, std::size_t _Cols = _Rows>
    class matrix
    {
    public:
      /*    usual public types    */
      using value_type = _Tp;
      using size_type = std::size_t;
      using reference = value_type&;
      using const_reference = const value_type&;
      using pointer = value_type*;
      using const_pointer = const value_type*;
      using iterator = value_type*;
      using const_iterator = const value_type*;
      
      
      /*   default constructor   */
      matrix() {
        std::fill(begin(), end(), 0.0);
      }
      
      matrix(std::initializer_list<_Tp> il)
      {
        std::copy_n(il.begin(), std::min(il.size(), _Rows*_Cols), begin());
        if(il.size() < _Rows*_Cols) std::fill(begin()+il.size(), end(), 0.0);
      }


      template<ComplexandRealNumbers _Xp>
      matrix(const unit_matrix_arg<_Xp>& d)
      {
        std::fill(begin(), end(), 0.0);
        
        for(std::size_t r = 0; r < std::min(_Rows, _Cols); r++)
          _M_elem[r][r] = d.coeff;
      }
      
      
      /*   copy   */
      matrix(const matrix&) = default;
      matrix& operator=(const matrix&) = default;
      
      template<ComplexandRealNumbers _Xp>
      explicit matrix(const matrix<_Xp, _Rows, _Cols>& m) {
        std::copy(m.begin(), m.end(), begin());
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator=(const matrix<_Xp, _Rows, _Cols>& m) {
        std::copy(m.begin(), m.end(), begin());
        return *this;
      }
      
      matrix& operator=(std::nullptr_t) {
        std::fill(begin(), end(), 0.0);
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator=(const _Xp& d)
      {
        std::fill(begin(), end(), 0.0);
        
        for(std::size_t r = 0; r < std::min(_Rows, _Cols); r++)
          _M_elem[r][r] = d;
        
        return *this;
      }
      
      matrix& operator=(std::initializer_list<_Tp> il)
      {
        std::copy_n(il.begin(), std::min(il.size(), _Rows*_Cols), begin());
        if(il.size() < _Rows*_Cols) std::fill(begin()+il.size(), end(), 0.0);
        return *this;
      }

      
      /*   move   */
      matrix(matrix&&) = default;
      matrix& operator=(matrix&&) = default;
      
      
      /*   iterators   */
      iterator begin() { return _M_elem[0];}
      iterator end() { return _M_elem[0] + _Rows*_Cols;}
      
      const_iterator begin() const { return _M_elem[0];}
      const_iterator end() const { return _M_elem[0] + _Rows*_Cols;}
      
      const_iterator cbegin() const { return _M_elem[0];}
      const_iterator cend() const { return _M_elem[0] + _Rows*_Cols;}
      
      
      /*    element access   */
      pointer operator[](size_type l) { return _M_elem[l];}
      const_pointer operator[](size_type l) const { return _M_elem[l];}
      
      
      /*   structural informations    */
      constexpr size_type nrows() const { return _Rows;}
      constexpr size_type ncols() const { return _Cols;}
      
      /*    arithmetic operations    */
      template<ComplexandRealNumbers _Xp>
      matrix& operator+=(const matrix<_Xp, _Rows, _Cols>& m)
      {
        pointer curr = begin(), last = end();
        auto x = m.begin();
        
        for(; curr < last; ++curr, ++x)
          *curr += *x;
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator-=(const matrix<_Xp, _Rows, _Cols>& m)
      {
        pointer curr = begin(), last = end();
        auto x = m.begin();
        
        for(; curr < last; ++curr, ++x)
          *curr -= *x;
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator*=(const _Xp& a)
      {
        pointer curr = begin(), last = end();
        
        for(; curr < last; ++curr)
          *curr *= a;
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator/=(const _Xp& a)
      {
        pointer curr = begin(), last = end();
        
        for(; curr < last; ++curr)
          *curr /= a;
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator+=(const _Xp& d)
      {
        for(std::size_t r = 0; r < std::min(_Rows, _Cols); r++)
          _M_elem[r][r] += d;
        
        return *this;
      }
      
      template<ComplexandRealNumbers _Xp>
      matrix& operator-=(const _Xp& d)
      {
        for(std::size_t r = 0; r < std::min(_Rows, _Cols); r++)
          _M_elem[r][r] -= d;
        
        return *this;
      }
      
    private:
      /*   data member is just a raw C matrix   */
      value_type _M_elem[_Rows][_Cols];
    };
    
    
    template<ComplexandRealNumbers _Tp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator-(const matrix<_Tp, _Rows, _Cols>& a)
    {
      matrix<_Tp, _Rows, _Cols> res;
      
      auto curr = res.begin(), last = res.end();
      auto ia = a.begin();
      
      for(; curr < last; ++curr, ++ia)
        *curr = -(*ia);
      
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator+(const matrix<_Tp, _Rows, _Cols>& a, const matrix<_Xp, _Rows, _Cols>& b)
    {
      matrix<result_of_add_t<_Tp, _Xp>, _Rows, _Cols> res;
      std::transform(a.begin(), a.end(), b.begin(), res.begin(), [](const _Tp& x, const _Xp& y) { return x+y;});
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator-(const matrix<_Tp, _Rows, _Cols>& a, const matrix<_Xp, _Rows, _Cols>& b)
    {
      matrix<result_of_sub_t<_Tp, _Xp>, _Rows, _Cols> res;
      std::transform(a.begin(), a.end(), b.begin(), res.begin(), [](const _Tp& x, const _Xp& y) { return x-y;});
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Dim, std::size_t _Cols>
    inline auto operator*(const matrix<_Tp, _Rows, _Dim>& a, const matrix<_Xp, _Dim, _Cols>& b)
    {
      matrix<result_of_add_t<_Tp, _Xp>, _Rows, _Cols> res;
      
      for(std::size_t r = 0; r < _Rows; ++r)
        for(std::size_t c = 0; c < _Cols; ++c)
        {
          res[r][c] = 0.0;
          
          for(std::size_t x = 0; x < _Dim; ++x)
            res[r][c] += a[r][x]*b[x][c];
        }
      
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator*(const matrix<_Tp, _Rows, _Cols>& a, const _Xp& b)
    {
      matrix<result_of_mul_t<_Tp, _Xp>, _Rows, _Cols> res;
      std::transform(a.begin(), a.end(), res.begin(), [&b](const _Tp& x) { return x*b;});
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator/(const matrix<_Tp, _Rows, _Cols>& a, const _Xp& b)
    {
      matrix<result_of_div_t<_Tp, _Xp>, _Rows, _Cols> res;
      std::transform(a.begin(), a.end(), res.begin(), [&b](const _Tp& x) { return x/b;});
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator*(const _Xp& b, const matrix<_Tp, _Rows, _Cols>& a)
    {
      matrix<result_of_mul_t<_Tp, _Xp>, _Rows, _Cols> res;
      std::transform(a.begin(), a.end(), res.begin(), [&b](const _Tp& x) { return x*b;});
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, std::size_t _Rows, std::size_t _Cols>
    inline auto transpose(const matrix<_Tp, _Rows, _Cols>& a)
    {
      matrix<_Tp, _Cols, _Rows> res;
      
      for(std::size_t r = 0; r < _Rows; ++r)
        for(std::size_t c = 0; c < _Cols; ++c)
          res[c][r] = a[r][c];
      
      return res;
    }
    
    
    template<ComplexNumbers _Tp, std::size_t _Rows, std::size_t _Cols>
    inline auto adjoint(const matrix<_Tp, _Rows, _Cols>& a)
    {
      matrix<_Tp, _Cols, _Rows> res;
      
      for(std::size_t r = 0; r < _Rows; ++r)
        for(std::size_t c = 0; c < _Cols; ++c)
          res[c][r] = std::conj(a[r][c]);
      
      return res;
    }
    
    
    template<RealNumbers _Tp, std::size_t _Rows, std::size_t _Cols>
    inline auto adjoint(const matrix<_Tp, _Rows, _Cols>& a) {
      return transpose(a);
    }
    
    
    template<ComplexandRealNumbers _Tp, std::size_t _Rows, std::size_t _Cols>
    inline std::ostream& operator<<(std::ostream& os, const matrix<_Tp, _Rows, _Cols>& a)
    {
      for(std::size_t r = 0; r < _Rows; ++r)
      {
        for(std::size_t c = 0; c < _Cols; ++c)
          os<<a[r][c]<<" ";
        
        os<<'\n';
      }
      
      return os;
    }
    
  }   //   namespace linalg
  
 
  
  /*   operations with matrices and vectors    */
  namespace linalg {
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    inline auto operator*(const ketvector<_Tp, _Rows>& a, const bravector<_Xp, _Cols>& b)
    {
      matrix<result_of_mul_t<_Tp, _Xp>, _Rows, _Cols> res;
      
      for(std::size_t r = 0; r < _Rows; ++r)
        for(std::size_t c = 0; c < _Cols; ++c)
          res[r][c] = a[r]*b[c];
      
      return res;
    }
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    auto operator*(const matrix<_Tp, _Rows, _Cols>& m, const ketvector<_Xp, _Cols>& v)
    {
      ketvector<result_of_mul_t<_Tp, _Xp>, _Rows> res;
      
      for(std::size_t r = 0; r < _Rows; ++r)
      {
        res[r] = 0.0;
        
        for(std::size_t c = 0; c < _Cols; ++c)
          res[r] += m[r][c]*v[c];
      }
      
      return res;
    }
    
    
    template<ComplexandRealNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Rows, std::size_t _Cols>
    auto operator*(const bravector<_Tp, _Rows>& v, const matrix<_Xp, _Rows, _Cols>& m)
    {
      bravector<result_of_mul_t<_Tp, _Xp>, _Cols> res;
      
      for(std::size_t c = 0; c < _Cols; ++c)
      {
        res[c] = 0.0;
        
        for(std::size_t r = 0; r < _Rows; ++r)
          res[c] += v[r]*m[r][c];
      }
      
      return res;
    }
    
  }   //   namespace linalg
  
 
  
  
  /*   diagonalization, inverse, Putzer's algorithm,....    */
  namespace linalg {
    
    /*   This function diagonalizes a general complex n-by-n matrix.
     *     Input :  A = n-by-n matrix,
     *     Output:  d = vector of eigenvalues,
     *              U = transformation matrix,
     *   these fulfill
     *     d = U A U^-1,  A = U^-1 d U,  U A = d U
     */
    template<ComplexNumbers _Xp, std::size_t _Dim>
    int eigensystem(matrix<_Xp, _Dim, _Dim> A, matrix<_Xp, _Dim, _Dim>& d, matrix<_Xp, _Dim, _Dim>& U)
    {
      using real_type = typename traits::complex_and_real_numbers<_Xp>::value_type;
      using complex_type = _Xp;
      
      constexpr real_type red = 0.01/(_Dim*_Dim*_Dim*_Dim);
      constexpr real_type EPS = 1.97215e-31;
      
      /*   initializing   */
      complex_type ev[_Dim][2];
      
      d = 0.0;
      U = 1.0;
      
      for(std::size_t p = 0; p < _Dim; ++p) {
        ev[p][0] = 0.0;
        ev[p][1] = d[p][p] = A[p][p];
      }
      
      
      /*   do the iterations    */
      int nsweeps = 1;
      
      for(; nsweeps <= 50; ++nsweeps)
      {
        /*   chechking the size of the off-diagonal elements   */
        real_type thresh = 0.0;
        
        for(std::size_t q = 1; q < _Dim; ++q)
          for(std::size_t p = 0; p < q; ++p)
            thresh += std::norm(A[p][q]) + std::norm(A[q][p]);
        
        /*   matrix A is already diagonal stop the iteration   */
        if(thresh < EPS) goto done;
        
        
        /*   we still have to work   */
        thresh = (nsweeps < 4 ? thresh*red : 0.0);
        
        for(std::size_t q = 1; q < _Dim; ++q)
          for(std::size_t p = 0; p < q; ++p)
          {
            const auto& Apq = A[p][q], Aqp = A[q][p];
            const auto off = std::norm(Apq) + std::norm(Aqp);
            
            if(nsweeps > 4 && off < EPS*(std::norm(ev[p][1]) + std::norm(ev[q][1])))
              A[p][q] = A[q][p] = 0.0;
            
            else if(off > thresh)
            {
              auto delta = Apq*Aqp;
              auto x = 0.5*(ev[p][1] - ev[q][1]);
              auto y = std::sqrt(x*x + delta);
              auto t = x - y;
              auto s = x + y;
              
              if(std::abs(t) < std::abs(s))
                t = s;
              
              t = 1.0/t;
              delta *= t;
              ev[p][0] += delta;
              ev[q][0] -= delta;
              
              ev[p][1] = d[p][p] + ev[p][0];
              ev[q][1] = d[q][q] + ev[q][0];
              
              auto invc = std::sqrt(delta*t + 1.0);
              
              s = t/invc;
              t /= invc + 1.0;
              
              auto sx = s*Apq;
              auto ty = t*Apq;
              auto sy = s*Aqp;
              auto tx = t*Aqp;
              
              
              for(std::size_t j = 0; j < _Dim; ++j)
              {
                auto xx = A[j][p];
                auto yy = A[j][q];
                
                A[j][p] += sy*(yy - ty*xx);
                A[j][q] -= sx*(xx + tx*yy);
                
                xx = A[p][j];
                yy = A[q][j];
                
                A[p][j] += sx*(yy - tx*xx);
                A[q][j] -= sy*(xx + ty*yy);
              }
              
              A[p][q] = A[q][p] = 0.0;
              
              for(std::size_t j = 0; j < _Dim; ++j)
              {
                auto xx = U[p][j];
                auto yy = U[q][j];
                
                U[p][j] += sx*(yy - tx*xx);
                U[q][j] -= sy*(xx + ty*yy);
              }
            }
          }
        
        for(std::size_t p = 0; p < _Dim; ++p)
        {
          ev[p][0] = 0.0;
          d[p][p] = ev[p][1];
        }
      }
      
      /*   it return with -51 if it doesn't converge   */
      return -nsweeps;
      
    done:
      
      /* normalize the eigenvectors */
      for(std::size_t  p = 0; p < _Dim; ++p)
      {
        real_type norm = 0.0;
        
        for(std::size_t q = 0; q < _Dim; ++q)
          norm += std::norm(U[p][q]);
        
        norm = std::sqrt(norm);
        
        for(std::size_t q = 0; q < _Dim; ++q)
          U[p][q] *= norm;
      }
      
      return nsweeps;
    }
    
    
    
    /*    inverse of a matrix    */
    template<ComplexandRealNumbers _Tp, std::size_t _Dim>
    auto inverse(matrix<_Tp, _Dim, _Dim> a)
    {
      using real_type = typename traits::complex_and_real_numbers<_Tp>::value_type;
      
      std::size_t ipiv[_Dim], indxc[_Dim], indxr[_Dim];
      
      std::fill(std::begin(ipiv), std::end(ipiv), 0);
      
      
      for(std::size_t i = 0; i < _Dim; ++i)
      {
        real_type big = 0.0;
        std::size_t irow = 0, icol = 0;
        
        for(std::size_t j = 0; j < _Dim; ++j)
          if(ipiv[j] != 1)
            for(std::size_t k = 0; k < _Dim; ++k)
            {
              if(ipiv[k] == 0)
              {
                if(std::abs(a[j][k]) >= big)
                {
                  big = std::abs(a[j][k]);
                  irow = j;
                  icol = k;
                }
              }
              
              /*   matrix is singular    */
              else if(ipiv[k] > 1)
                return std::optional<matrix<_Tp, _Dim, _Dim>>{};
            }
        
        ipiv[icol]++;
        
        if(irow != icol)
          for(std::size_t l = 0; l < _Dim; ++l)
            std::swap(a[irow][l], a[icol][l]);
        
        indxr[i] = irow;
        indxc[i] = icol;
        
        /*   matrix is singular   */
        if(a[icol][icol] == 0.0)
          return std::optional<matrix<_Tp, _Dim, _Dim>>{};
        
        auto pivinv = 1.0/a[icol][icol];
        
        a[icol][icol] = 1.0;
        
        for(std::size_t l = 0; l < _Dim; ++l)
          a[icol][l] *= pivinv;
        
        for(std::size_t ll = 0; ll < _Dim; ++ll)
          if(ll != icol)
          {
            auto dum = a[ll][icol];
            a[ll][icol] = 0.0;
            
            for(std::size_t l = 0; l < _Dim; ++l)
              a[ll][l] -= a[icol][l]*dum;
          }
      }
      
      for(std::size_t l = _Dim; l >= 1;)
      {
        --l;
        
        if(indxr[l] != indxc[l])
          for(std::size_t k = 0; k < _Dim; ++k)
            std::swap(a[k][indxr[l]], a[k][indxc[l]]);
      }
      
      return std::optional<matrix<_Tp, _Dim, _Dim>>{std::move(a)};
    }
 
    
    
    
    /*    Putzer's algorithm : exponentiating a matrix   */
    template<RealNumbers _Fp, ComplexNumbers _Tp, std::size_t _Dim>
    matrix<_Tp, _Dim, _Dim> putzer(_Fp t, matrix<_Tp, _Dim, _Dim> A, const _Tp *lmd)
    {
      using complex_t = _Tp;
      using real_t = typename traits::complex_and_real_numbers<_Tp>::value_type;
      
      /*   we need some workspace and temporaries   */
      std::size_t kmax[_Dim];
      complex_t r[_Dim][_Dim], elmd[_Dim], rr[_Dim];
      
      for(std::size_t l = 0; l < _Dim; ++l)
      {
        kmax[l] = 0;
        elmd[l] = std::exp(t*lmd[l]);
      }
      
      
      /*    the first eigenvalue   */
      r[0][0] = 1.0;
      rr[0] = elmd[0];
      
      
      /*   loop over the eigenvalues    */
      for(std::size_t n = 1; n < _Dim; ++n)
      {
        r[n][0] = 0.0;
        
        for(std::size_t i = 0; i < n; ++i)
        {
          auto delta = lmd[i]-lmd[n];
          
          /*   different eigenvalue   */
          if(std::norm(delta) > 1e-12)
          {
            for(std::size_t x = 0; x <= kmax[i]; ++x)
            {
              complex_t rs = 0.0, dfac = delta;
              
              for(std::size_t k = x; k <= kmax[i]; ++k)
              {
                rs += r[i][k]/dfac;
                dfac *= -delta;
              }
              
              r[i][x] = rs;
            }
            
            r[n][0] -= r[i][0];
          }
          
          /*   same eigenvalue    */
          else
          {
            kmax[i]++;
            
            for(std::size_t k = kmax[i]; k > 0; --k)
              r[i][k] = r[i][k-1];
            
            r[i][0] = 0.0;
          }
          
          /*   putting together the full ri(t) function    */
          complex_t ri = r[i][0], tfac = 1.0;
          
          for(std::size_t x = 1; x <= kmax[i]; ++x)
          {
            tfac *= t/static_cast<real_t>(x);
            ri += r[i][x]*tfac;
          }
          
          rr[n] += ri*elmd[i];
        }
        
        rr[n] += r[n][0]*elmd[n];
      }
      
      
      /*    calculation the exponential of the matrix   */
      matrix<_Tp, _Dim, _Dim> res(rr[0]), Pi{A -= lmd[0]};
      
      res += rr[1]*Pi;
      
      for(std::size_t n = 2; n < _Dim; ++n)
      {
        A += lmd[n-2]-lmd[n-1];
        Pi = A*Pi;
        
        res += rr[n]*Pi;
      }
      
      return res;
    }
    
    
    
    /*    Putzer's algorithm with initial condition resulting a vector   */
    template<RealNumbers _Fp, ComplexNumbers _Tp, ComplexandRealNumbers _Xp, std::size_t _Dim>
    ketvector<_Tp, _Dim> putzer(_Fp t, matrix<_Tp, _Dim, _Dim> A, const _Tp *lmd, const ketvector<_Xp, _Dim>& vin)
    {
      using complex_t = _Tp;
      using real_t = typename traits::complex_and_real_numbers<_Tp>::value_type;
      
      
      /*   we need some workspace and temporaries   */
      std::size_t kmax[_Dim];
      complex_t r[_Dim][_Dim], elmd[_Dim], rr[_Dim];
      
      for(std::size_t l = 0; l < _Dim; ++l)
      {
        kmax[l] = 0;
        elmd[l] = std::exp(t*lmd[l]);
      }
      
      
      /*    the first eigenvalue   */
      r[0][0] = 1.0;
      rr[0] = elmd[0];
      
      
      /*   loop over the eigenvalues    */
      for(std::size_t n = 1; n < _Dim; ++n)
      {
        r[n][0] = 0.0;
        
        for(std::size_t i = 0; i < n; ++i)
        {
          auto delta = lmd[i]-lmd[n];
          
          /*   different eigenvalue   */
          if(std::norm(delta) > 1e-12)
          {
            for(std::size_t x = 0; x <= kmax[i]; ++x)
            {
              complex_t rs = 0.0, dfac = delta;
              
              for(std::size_t k = x; k <= kmax[i]; ++k)
              {
                rs += r[i][k]/dfac;
                dfac *= -delta;
              }
              
              r[i][x] = rs;
            }
            
            r[n][0] -= r[i][0];
          }
          
          /*   same eigenvalue    */
          else
          {
            kmax[i]++;
            
            for(std::size_t k = kmax[i]; k > 0; --k)
              r[i][k] = r[i][k-1];
            
            r[i][0] = 0.0;
          }
          
          /*   putting together the full ri(t) function    */
          complex_t ri = r[i][0], tfac = 1.0;
          
          for(std::size_t x = 1; x <= kmax[i]; ++x)
          {
            tfac *= t/static_cast<real_t>(x);
            ri += r[i][x]*tfac;
          }
          
          rr[n] += ri*elmd[i];
        }
        
        rr[n] += r[n][0]*elmd[n];
      }
      
      
      /*    calculating the result   */
      ketvector<_Tp, _Dim> v{vin};
      ketvector<_Tp, _Dim> res{v};
      
      res *= rr[0];
      res += rr[1]*(v = (A -= lmd[0])*v);
      
      for(std::size_t n = 2; n < _Dim; ++n)
        res += rr[n]*(v = (A += lmd[n-2]-lmd[n-1])*v);
      
      return res;
    }

  }
  
}   //   namespace duct

#endif /* __math_linear_algebra_h__ */
