/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 26.12.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   std includes   */


/*   project headers   */
#include "poly-solve.h"

namespace duct {
  
  void solve_quadratic(double b, double c, std::complex<double> *roots)
  {
    const auto Q = std::sqrt(std::complex<double>{b*b - 4.0*c});
    
    roots[0] = (-b + Q)/2.0;
    roots[1] = (-b - Q)/2.0;
  }

  
  void solve_quadratic(const std::complex<double>& b, const std::complex<double>& c, std::complex<double> *roots)
  {
    const auto Q = std::sqrt(std::complex<double>{b*b - 4.0*c});
    
    roots[0] = (-b + Q)/2.0;
    roots[1] = (-b - Q)/2.0;
  }
  
  
  
  void solve_qubic(double a, double b, double c, std::complex<double> *roots)
  {
    constexpr std::complex<double> C1{0.5, 0.5*std::sqrt(3.0)}, C2{0.5, -0.5*std::sqrt(3.0)};
    
    const std::complex<double> D = 6.75*(4.0*b*b*b+(4.0*a*c-b*b)*a*a - 18.0*a*b*c + 27.0*c*c);
    const std::complex<double> Q1 = std::pow(((4.5*b-a*a)*a - 13.5*c) + std::sqrt(D), 1.0/3.0);
    const double Q2 = a*a-3.0*b;
    
    roots[0] = (-a + Q2/Q1 + Q1)/3.0;
    roots[1] = (-a - C1*Q2/Q1 - C2*Q1)/3.0;
    roots[2] = (-a - C2*Q2/Q1 - C1*Q1)/3.0;
  }
  
 
  void solve_qubic(const std::complex<double>& a, const std::complex<double>& b, const std::complex<double>& c, std::complex<double> *roots)
  {
    constexpr std::complex<double> C1{0.5, 0.5*std::sqrt(3.0)}, C2{0.5, -0.5*std::sqrt(3.0)};
    
    const std::complex<double> D = 6.75*(4.0*b*b*b+(4.0*a*c-b*b)*a*a - 18.0*a*b*c + 27.0*c*c);
    const std::complex<double> Q1 = std::pow(((4.5*b-a*a)*a - 13.5*c) + std::sqrt(D), 1.0/3.0);
    const std::complex<double> Q2 = a*a-3.0*b;
    
    roots[0] = (-a + Q2/Q1 + Q1)/3.0;
    roots[1] = (-a - C1*Q2/Q1 - C2*Q1)/3.0;
    roots[2] = (-a - C2*Q2/Q1 - C1*Q1)/3.0;
  }
  
  

  void solve_quartic(double b, double c, double d, double e, std::complex<double> *roots)
  {
    const double p = c - 3.0*b*b/8.0;
    const double q = (b*b*b - 4.0*b*c + 8.0*d)/8.0;
    const double r = (16.0*b*b*c + 256.0*e - 64.0*b*d - 3.0*b*b*b*b)/256.0;
    
    /*    check for the trivial case    */
    if(double tiny = 1e-15*std::abs(b); std::abs(p) < tiny && std::abs(q) < tiny && std::abs(r) < tiny) {
      roots[0] = roots[1] = roots[2] = roots[3] = -0.25*b;
      return;
    }
    
    /*  solve cubic equation : x^3 + p*x^2 + (p*p/4.0-r)*x - q*q/8.0 == 0   */
    std::complex<double> x[3];
    solve_qubic(p, p*p/4.0-r, -q*q/8.0, x);
    
    /*    find the "bigest" root   */
    std::complex<double> m = x[0];
    
    if(std::norm(m) < std::norm(x[1])) m = x[1];
    if(std::norm(m) < std::norm(x[2])) m = x[2];
    
    /*   calculating the roots   */
    const std::complex<double> Q1 = std::sqrt(2.0*m);
    const std::complex<double> Q2 = std::sqrt(-2.0*(p+m+q/Q1));
    const std::complex<double> Q3 = std::sqrt(-2.0*(p+m-q/Q1));
    
    roots[0] = -0.25*b + 0.5*(Q1 + Q2);
    roots[1] = -0.25*b + 0.5*(Q1 - Q2);
    roots[2] = -0.25*b - 0.5*(Q1 + Q3);
    roots[3] = -0.25*b - 0.5*(Q1 - Q3);
  }

  
  
  void solve_quartic(const std::complex<double>& b, const std::complex<double>& c,
                     const std::complex<double>& d, const std::complex<double>& e,
                     std::complex<double> *roots)
  {
    const std::complex<double> p = c - 3.0*b*b/8.0;
    const std::complex<double> q = (b*b*b - 4.0*b*c + 8.0*d)/8.0;
    const std::complex<double> r = (16.0*b*b*c + 256.0*e - 64.0*b*d - 3.0*b*b*b*b)/256.0;
    
    /*    check for the trivial case    */
    if(double tiny = 1e-15*std::abs(b); std::abs(p) < tiny && std::abs(q) < tiny && std::abs(r) < tiny) {
      roots[0] = roots[1] = roots[2] = roots[3] = -0.25*b;
      return;
    }
    
    /*  solve cubic equation : x^3 + p*x^2 + (p*p/4.0-r)*x - q*q/8.0 == 0   */
    std::complex<double> x[3];
    solve_qubic(p, p*p/4.0-r, -q*q/8.0, x);
    
    /*    find the "bigest" root   */
    std::complex<double> m = x[0];
    
    if(std::norm(m) < std::norm(x[1])) m = x[1];
    if(std::norm(m) < std::norm(x[2])) m = x[2];
    
    /*   calculating the roots   */
    const std::complex<double> Q1 = std::sqrt(2.0*m);
    const std::complex<double> Q2 = std::sqrt(-2.0*(p+m+q/Q1));
    const std::complex<double> Q3 = std::sqrt(-2.0*(p+m-q/Q1));
    
    roots[0] = -0.25*b + 0.5*(Q1 + Q2);
    roots[1] = -0.25*b + 0.5*(Q1 - Q2);
    roots[2] = -0.25*b - 0.5*(Q1 + Q3);
    roots[3] = -0.25*b - 0.5*(Q1 - Q3);
  }
  
}   //   namespace duct


