/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10/12/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __math_polylog_h__
#define __math_polylog_h__ 1


/*   std includes   */
#include <cmath>


/*   project headers   */
#include <cxx/math/pow.h>
#include <cxx/math/zeta.h>
#include <cxx/math/harmonic-numbers.h>



namespace duct {
  
  namespace __details {
    
    /*   Helper class to implement the various polylog series   */
    template<std::size_t N>
    struct polylog_series
    {
      static constexpr double pi2 = M_PI*M_PI;
      static constexpr double pi4 = pi2*pi2;
      
      /*   standatd sum, use it on [-0.5, 0.5]  */
      static double series1(double x);
      
      /*   standatd sum, use it on [0.5, 1] */
      static double series2(double x);
      
      /*    This is the Bernoulli polynomial part to be able to calculate x < -1
       *    I don't ave a recursive algorithm to generate this series. I used Mathematica.
       *    If higher polylogs are needed one can use this:
       *
       *      b[n_] := Simplify[-((2 Pi I)^n/n!) BernoulliB[n, (Log[-x] + I Pi)/(2 Pi I)]]
       */
      static double series3(double x) { return 0.0;}
      
      /*   this is the polylog on the [-1,1] interval   */
      static double series(double x)
      {
        if(std::abs(x) <= 0.5) return series1(x);
        else if(x > 0.5) return series2(x);
        
        double ret;
        constexpr double fac = meta::pow<N-1>(2.0);
        
        if(x > -0.707106781186548)
          ret = series1(x*x)/fac;
        else ret = series2(x*x)/fac;
        
        return ret - series2(-x);
      }
    };
    
    
    template<std::size_t N>
    double polylog_series<N>::series1(double x)
    {
      constexpr unsigned kmax = 1000;
      
      double sum  = x, term = x;
      
      for(unsigned k = 2; k < kmax; ++k)
      {
        term *= x;
        term *= meta::pow<N>(1.0-1.0/k);
        sum += term;
        if(std::abs(term/sum) < 1e-16) break;
      }
      
      return sum;
    }
    
    extern const double polylog_series2_coeff[35];
    
    template<std::size_t N>
    double polylog_series<N>::series2(double x)
    {
      
      /*   trivial case    */
      if(x == 1.0) return zeta(N);
      
      /*   otherwise calculate   */
      double lx = std::log(x);
      double term = 1.0, lx2 = lx*lx;
      double sum = zeta(N), tmp;
      
      for(unsigned k = 1; k < N-1; ++k)
      {
        term *= lx;
        term /= k;
        sum += zeta(N-k)*term;
      }
      
      term *= lx;
      term /= N-1;
      
      sum += term*(meta::harmonic_number<N-1> - std::log(-lx));
      
      term *= lx;
      term /= N;
      
      sum -= 0.5*term;
      
      term *= lx/(N+1);
      sum += polylog_series2_coeff[0]*term;
      
      for(unsigned k = 1; k < 34; ++k)
      {
        term *= lx2;
        term *= 2.0*k/(N+2.0*k);
        term *= (2.0*k+1.0)/(N+2.0*k+1.0);
        tmp = polylog_series2_coeff[k]*term;
        sum += tmp;
        
        if(std::abs(tmp/sum) < 1e-16) break;
      }
      
      return sum;
    }
    
    
    /*    This is the Bernoulli polynomial part to be able to calculate x < -1
     *    I don't have a recursive algorithm to generate this series. I used Mathematica.
     *    If higher polylogs are needed one can use this:
     *
     *      b[n_] := Simplify[-((2 Pi I)^n/n!) BernoulliB[n, (Log[-x] + I Pi)/(2 Pi I)]]
     */
    template<>
    inline double polylog_series<2>::series3(double x) {
      return -pi2/6.0 - 0.5*sqr(std::log(-x));
    }
    
    template<>
    inline double polylog_series<3>::series3(double x)
    {
      double lx = std::log(-x);
      return -lx*(pi2 + sqr(lx))/6.0;
    }
    
    template<>
    inline double polylog_series<4>::series3(double x)
    {
      double lx = sqr(std::log(-x));
      return -7.0*pi4/360.0 - lx*(pi2 + 0.5*lx)/12.0;
    }
    
    template<>
    inline double polylog_series<5>::series3(double x)
    {
      double lx = std::log(-x);
      double lx2 = lx*lx;
      
      return -lx*(7.0*pi4/360.0 + (pi2/36.0 + lx2/120.0)*lx2);
    }
    
    
    template<>
    inline double polylog_series<6>::series3(double x)
    {
      double lx2 = sqr(std::log(-x));
      double lx4 = lx2*lx2;
      
      return -31.0*pi4*pi2/15120.0 - 7.0/720.0*pi4*lx2 - (pi2/144.0 + lx2/720.0)*lx4;
    }
    
  }  /*   namespace __details   */
  
  
  namespace meta {
  
    /*  We put the static version of math function on namespace meta.
     *  This way we can implement their runtime version too.
     */
    template<std::size_t N>
    inline double polylog(double x)
    {
      /*    returns meaningful value on [-1,1] interval for every N   */
      if(std::abs(x) <= 1.0) return __details::polylog_series<N>::series(x);
      
      /*   for x < -1 it is meaningful only for 2 <= N <= 6   */
      return __details::polylog_series<N>::series3(x) - (N%2 ? -1.0 : 1.0)*__details::polylog_series<N>::series(1.0/x);
    }
  }   /*   namespace meta   */

  
  /*   Li2(x) and Li3(x)  functions   */
  inline double dilog(double x) { return meta::polylog<2>(x);}
  inline double trilog(double x) { return meta::polylog<3>(x);}
  
}  /*   namespace duct   */

#endif