/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán qcd::Nagy on 12/27/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>
#include <algorithm>

//   project headers
#include "proc-hhcjets.h"


namespace duct {
  namespace __details {
    
    static inline double f1(double r, double x1, double x2)
    {
      double xn, x = 0.5*(x1+x2);
      
      do {
        xn = x;
        x += (2.0*r - x - std::sinh(x))/(1.0+std::cosh(x));
      } while(std::abs(x-xn) > 1e-8*x);
      
      return x;
    }
  }   //   namespace __details
  
  
  
  /*   Table of all the possible subprocesses
   */
  const hhcjets::subprocess hhcjets::_M_proc_tbl[] = {
    //----- process 0 -----
    {{0,0,0,0}, &hhcjets::_M_process_gggg, &hhcjets::_S_select_color_gggg},
    
    /*    Process 1-6: g + g --> q + qbar, where q = d,u,s,c,b,t
     */
    {{0,0,1,-1}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,2,-2}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,3,-3}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,4,-4}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,5,-5}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
//    {{0,0,6,-6}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    
    /*    Process  7-11: g + q --> g + q, where q = d,u,s,c,b
     */
    {{0,-1,0,1}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-2,0,2}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-3,0,3}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-4,0,4}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-5,0,5}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    
    /*    Process 12-16: g + qbar --> g + qbar where q = d,u,s,c,b
     */
    {{0,1,0,-1}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,2,0,-2}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,3,0,-3}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,4,0,-4}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,5,0,-5}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    
    /*    Process 17-21: q + g --> g + q, where q = d,u,s,c,b
     */
    {{-1,0,0,1}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-2,0,0,2}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-3,0,0,3}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-4,0,0,4}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-5,0,0,5}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    
    /*    Process 22-26: qbar + g --> g + qbar, where q = d,u,s,c,b
     */
    {{1,0,0,-1}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{2,0,0,-2}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{3,0,0,-3}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{4,0,0,-4}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{5,0,0,-5}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    
    /*    Process 27-31: q + qbar --> g + g, where q = d,u,s,c,b
     */
    {{-1,1,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-2,2,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-3,3,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-4,4,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-5,5,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    
    /*    Process 32-36: qbar + q --> g + g, where q = d,u,s,c,b
     */
    {{1,-1,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{2,-2,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{3,-3,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{4,-4,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{5,-5,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    
    /*    process 37-41: d + dbar --> r + rbar, where r =   u,s,c,b,t
     *    process 42-46: u + ubar --> r + rbar, where r = d,  s,c,b,t
     *    process 47-51: s + sbar --> r + rbar, where r = d,u,  c,b,t
     *    process 52-56: c + cbar --> r + rbar, where r = d,u,s,  b,t
     *    process 57-61: b + bbar --> r + rbar, where r = d,u,s,c,  t
     */
    {{-1,1,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
//    {{-1,1,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-2,2,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
//    {{-2,2,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-3,3,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
//    {{-3,3,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-4,4,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
//    {{-4,4,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-5,5,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
//    {{-5,5,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    /*    process 62-66: dbar + d --> rbar + r, where r =   u,s,c,b,t
     *    process 67-71: ubar + u --> rbar + r, where r = d,  s,c,b,t
     *    process 72-76: sbar + s --> rbar + r, where r = d,u,  c,b,t
     *    process 77-81: cbar + c --> rbar + r, where r = d,u,s,  b,t
     *    process 81-86: bbar + b --> rbar + r, where r = d,u,s,c,  t
     */
    {{1,-1,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
//    {{1,-1,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{2,-2,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
//    {{2,-2,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{3,-3,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
//    {{3,-3,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{4,-4,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
//    {{4,-4,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{5,-5,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
//    {{5,-5,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    
    /*    Process 87-90:   d + r --> d + r, where r =   u,s,c,b
     *    Process 91-94:   u + r --> u + r, where r = d,  s,c,b
     *    Process 95-98:   s + r --> s + r, where r = d,u,  c,b
     *    Process 99-102:  c + r --> c + r, where r = d,u,s,  b
     *    Process 103-106: b + r --> b + r, where r = d,u,s,c
     */
    {{-1,-2,1,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-1,-3,1,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-1,-4,1,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-1,-5,1,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-2,-1,2,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-2,-3,2,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-2,-4,2,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-2,-5,2,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-3,-1,3,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-3,-2,3,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-3,-4,3,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-3,-5,3,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-4,-1,4,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-4,-2,4,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-4,-3,4,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-4,-5,4,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-5,-1,5,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-5,-2,5,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-5,-3,5,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-5,-4,5,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    /*    Process 107-110: dbar + rbar --> dbar + rbar, where r =   u,s,c,b
     *    Process 111-114: ubar + rbar --> ubar + rbar, where r = d,  s,c,b
     *    Process 115-118: sbar + rbar --> sbar + rbar, where r = d,u,  c,b
     *    Process 119-122: cbar + rbar --> cbar + rbar, where r = d,u,s,  b
     *    Process 123-126: bbar + rbar --> bbar + rbar, where r = d,u,s,c
     */
    {{1,2,-1,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{1,3,-1,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{1,4,-1,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{1,5,-1,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{2,1,-2,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{2,3,-2,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{2,4,-2,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{2,5,-2,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{3,1,-3,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{3,2,-3,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{3,4,-3,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{3,5,-3,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{4,1,-4,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{4,2,-4,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{4,3,-4,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{4,5,-4,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{5,1,-5,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{5,2,-5,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{5,3,-5,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{5,4,-5,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    /*    Process 127-130: d + rbar --> d + rbar, where r =   u,s,c,b
     *    Process 131-134: u + rbar --> u + rbar, where r = d,  s,c,b
     *    Process 135-138: s + rbar --> s + rbar, where r = d,u,  c,b
     *    Process 139-142: c + rbar --> c + rbar, where r = d,u,s,  b
     *    Process 143-146: b + rbar --> b + rbar, where r = d,u,s,c
     */
    {{-1,2,1,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-1,3,1,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-1,4,1,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-1,5,1,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-2,1,2,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-2,3,2,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-2,4,2,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-2,5,2,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-3,1,3,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-3,2,3,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-3,4,3,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-3,5,3,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-4,1,4,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-4,2,4,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-4,3,4,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-4,5,4,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-5,1,5,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-5,2,5,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-5,3,5,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-5,4,5,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    /*    Process 147-150: dbar + r --> dbar + r, where r =   u,s,c,b
     *    Process 151-154: ubar + r --> ubar + r, where r = d,  s,c,b
     *    Process 155-158: sbar + r --> sbar + r, where r = d,u,  c,b
     *    Process 159-162: cbar + r --> cbar + r, where r = d,u,s,  b
     *    Process 163-166: bbar + r --> bbar + r, where r = d,u,s,c
     */
    {{1,-2,-1,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{1,-3,-1,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{1,-4,-1,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{1,-5,-1,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{2,-1,-2,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{2,-3,-2,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{2,-4,-2,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{2,-5,-2,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{3,-1,-3,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{3,-2,-3,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{3,-4,-3,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{3,-5,-3,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{4,-1,-4,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{4,-2,-4,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{4,-3,-4,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{4,-5,-4,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{5,-1,-5,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{5,-2,-5,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{5,-3,-5,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{5,-4,-5,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    
    /*    Process 167-169:   d + dbar --> d + dbar
     *    Process 170-172:   u + ubar --> u + ubar
     *    Process 173-175:   s + sbar --> s + sbar
     *    Process 176-168:   c + cbar --> c + cbar
     *    Process 179-181:   b + bbar --> b + bbar
     */
    {{-1,1,1,-1}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-1,1,1,-1}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-1,1,1,-1}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-2,2,2,-2}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-2,2,2,-2}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-2,2,2,-2}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-3,3,3,-3}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-3,3,3,-3}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-3,3,3,-3}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-4,4,4,-4}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-4,4,4,-4}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-4,4,4,-4}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-5,5,5,-5}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-5,5,5,-5}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-5,5,5,-5}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    /*    Process 182-184:   dbar + d --> dbar + d
     *    Process 185-187:   ubar + u --> ubar + u
     *    Process 188-190:   sbar + s --> sbar + s
     *    Process 191-193:   cbar + c --> cbar + c
     *    Process 194-196:   bbar + b --> bbar + b
     */
    {{1,-1,-1,1}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{1,-1,-1,1}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{1,-1,-1,1}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{2,-2,-2,2}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{2,-2,-2,2}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{2,-2,-2,2}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{3,-3,-3,3}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{3,-3,-3,3}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{3,-3,-3,3}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{4,-4,-4,4}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{4,-4,-4,4}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{4,-4,-4,4}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{5,-5,-5,5}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{5,-5,-5,5}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{5,-5,-5,5}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    /*    Process 197-199:   d + d --> d + d
     *    Process 200-202:   u + u --> u + u
     *    Process 203-205:   s + s --> s + s
     *    Process 206-208:   c + c --> c + c
     *    Process 209-211:   b + b --> b + b
     */
    {{-1,-1,1,1}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-1,-1,1,1}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-1,-1,1,1}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-2,-2,2,2}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-2,-2,2,2}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-2,-2,2,2}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-3,-3,3,3}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-3,-3,3,3}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-3,-3,3,3}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-4,-4,4,4}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-4,-4,4,4}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-4,-4,4,4}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-5,-5,5,5}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-5,-5,5,5}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-5,-5,5,5}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    /*    Process 212-214:   dbar + dbar --> dbar + dbar
     *    Process 215-217:   ubar + ubar --> ubar + ubar
     *    Process 218-220:   sbar + sbar --> sbar + sbar
     *    Process 221-223:   cbar + cbar --> cbar + cbar
     *    Process 224-226:   bbar + bbar --> bbar + bbar
     */
    {{1,1,-1,-1}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{1,1,-1,-1}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{1,1,-1,-1}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{2,2,-2,-2}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{2,2,-2,-2}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{2,2,-2,-2}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{3,3,-3,-3}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{3,3,-3,-3}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{3,3,-3,-3}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{4,4,-4,-4}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{4,4,-4,-4}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{4,4,-4,-4}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{5,5,-5,-5}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{5,5,-5,-5}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{5,5,-5,-5}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    /*  End of list
     */
    {{0,0,0,0}, nullptr, nullptr}
  };
  
  
  /*  Tables for the color supression indices
   */
  const unsigned int hhcjets::_M_color_suppr_idx_4g[6][6] = {
    {0U, 2U, 2U, 2U, 2U, 2U},
    {2U, 0U, 2U, 2U, 2U, 2U},
    {2U, 2U, 0U, 2U, 2U, 2U},
    {2U, 2U, 2U, 0U, 2U, 2U},
    {2U, 2U, 2U, 2U, 0U, 2U},
    {2U, 2U, 2U, 2U, 2U, 0U}
  };
  
  const unsigned int hhcjets::_M_color_suppr_idx_2q2g[2][2] = {
    {0U, 2U}, {2U, 0U}
  };
  
  const unsigned int hhcjets::_M_color_suppr_idx_4q[2][2] = {
    {0U, 1U}, {1U, 0U}
  };
  
  const unsigned int hhcjets::_M_color_suppr_idx_4qC[2][2] = {
    {0U, 1U}, {1U, 0U}
  };
  
  
  
  hhcjets::hhcjets(const model_ref& mdl, energy<double> Ecm,
                   std::initializer_list<energy<double>> pTlist, std::initializer_list<double> y12list,
                   double xpert, double xshw, int hdr1, int hdr2)
  : _M_mdl{mdl}, _M_s{Ecm*Ecm}, _M_scalefac_pert{xpert*xpert}, _M_scalefac_shw{xshw*xshw}
  {
    /*   incoming hadrons   */
    _M_hadronA = hdr1;
    _M_hadronB = hdr2;
    
    
    /*    the pdf functions   */
    _M_pdf1 = _M_mdl.pdf(qcd::order::nlo, hdr1);
    _M_pdf2 = _M_mdl.pdf(qcd::order::nlo, hdr2);
    
    
    /*   p2T ranges  */
    std::vector<double> xT{0.25};
    
    for(auto pT : pTlist)
      if(pT < 0.49*Ecm && pT > 0.0_GeV)
        xT.push_back(pT*pT/_M_s);
    
    std::sort(xT.begin(), xT.end());
    
    
    /*   y12 ranges  */
    double ymax = 2.0*std::acosh(0.5/std::sqrt(xT[0]));
    std::vector<double> y12{0.0, ymax};
    
    
    for(auto y : y12list)
      if(0.0 < y && y < ymax)
        y12.push_back(y);
    
    std::sort(y12.begin(), y12.end());
    
    
    /*   finding the active (xT,y) ranges    */
    for(std::size_t i = 0; i < xT.size()-1; ++i)
    {
      double yi = 2.0*std::acosh(0.5/std::sqrt(xT[i]));
      
      for(std::size_t k = 0; k < y12.size()-1; ++k)
      if(double yh = std::min(y12[k+1], yi); yh > y12[k])
      _M_xTy.push_back({xT[i], xT[i+1], y12[k], yh});
    }
  }
  
  
  void hhcjets::generate(random_engine_type& random, shower_stage<hhc>& hard, shower_vetos& vetos)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    /*   resizing the state, setting the model  */
    hard.state.resize(0,2);
    
    
    /*    incoming hadrons   */
    auto Ecm = sqrt(_M_s);
    
    hard.state.hadron[-1].momentum = {0.5*Ecm, 0.0_GeV,0.0_GeV, 0.5*Ecm};
    hard.state.hadron[ 0].momentum = {0.5*Ecm, 0.0_GeV,0.0_GeV,-0.5*Ecm};
    hard.state.hadron[-1].key = _M_hadronA;
    hard.state.hadron[ 0].key = _M_hadronB;
    
    
    /*   generating the (most of the) event   */
    hard.weights.born.main = _M_generate_event(random, hard, vetos);
    hard.weights.nlo.main = 0.0;
  }
  
  
  double hhcjets::_M_generate_event(random_engine_type& random, shower_stage<hhc>& hard, shower_vetos& vetos)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    /*    generating the independent kinematical variables   */
    double tau, x, eta, weight = 4.0;
    
    //-----
    {
      /*   pick a pT and a y range   */
      auto nr = _M_xTy.size();
      const auto& r = _M_xTy[static_cast<unsigned>(nr*rng(random)*0.9999)];
      
      /*    generating the rapidity variable   */
      const double sh1 = 0.5*(r.ymin + std::sinh(r.ymin)), sh2 = 0.5*(r.ymax+std::sinh(r.ymax));
      const double y = __details::f1(rng(random)*(sh2-sh1) + sh1, r.ymin, r.ymax);
      
      /*   gnerating tau variable   */
      const double chy = 4.0*sqr(std::cosh(0.5*y));
      const double lmin = sqr(std::log(r.xTmin*chy)), lmax = sqr(std::log(std::min(1.0, r.xTmax*chy)));
      
      tau = std::exp(-std::sqrt(rng(random)*(lmax-lmin) + lmin));
      x = 1.0/(1.0+std::exp(rng(random) < 0.5 ? y : -y));
      eta = std::pow(tau, rng(random));
      
      
      weight *= nr;
      weight *= 2.0*(sh2-sh1);
      weight *= 0.5*(lmin-lmax);
      weight *= sqr(x*(1.0-x));
    }
    
    
    /*    generating variable eta and etb    */
    double etb = tau/eta;
    energy_sqr<double> shat = _M_s*tau;
    energy_sqr<double> scale = shat*x*(1.0-x);
    
    
    /*      the scale of the hard event    */
    hard.scale = _M_scalefac_shw*scale;
    
    
    /*    pre-calculating the pdf factors    */
    for(int k = -5; k <= 5; ++k) {
      _M_fa[k+5] = _M_pdf1 -> parton(k, eta, _M_scalefac_pert*scale);
      _M_fb[k+5] = _M_pdf2 -> parton(k, etb, _M_scalefac_pert*scale);
    }
    
    
    /*    QCD coupling    */
    double as = 2.0*M_PI*_M_mdl.alpha_qcd(_M_scalefac_pert*scale);
    
    weight *= 4.0*M_PI*as*as*1.0_GeV2/shat;
    
    
    
    /* Here we select the flavor configuration.
     * For this we use rather complicated weights those
     * consist of pdfs and some x dependent function.
     */
    weight *= this->_M_select_subprocess(x, rng(random));
    
    hard.state[-1].flavor = _M_active->fl[0];
    hard.state[ 0].flavor = _M_active->fl[1];
    hard.state[ 1].flavor = _M_active->fl[2];
    hard.state[ 2].flavor = _M_active->fl[3];
    
    
    /*     +/- and transverse momenta      */
    double phi = 2.0*M_PI*rng(random);
    auto mT = sqrt(x*(1.0-x)*shat);
    
    lorentzvector<energy<double>> na = hard.state.hadron[-1].momentum*eta;
    lorentzvector<energy<double>> nb = hard.state.hadron[ 0].momentum*etb;
    lorentzvector<energy<double>> kT{0.0_GeV, mT*std::cos(phi), mT*std::sin(phi), 0.0_GeV};
    
    
    /*    reconstructing the parton momenta    */
    hard.state[-1].momentum = na;
    hard.state[ 0].momentum = nb;
    hard.state[ 1].momentum = (1.0-x)*na +      x *nb + kT;
    hard.state[ 2].momentum =      x *na + (1.0-x)*nb - kT;
    
    
    /*    selecting the color states    */
    (this ->* (_M_active->select_color))(hard, rng(random));
    
    
    /*   vetos    */
    vetos = shower_vetos{};
    vetos.max_k2T = shat*x*(1.0-x);
    
    
    return weight;
  }
  
  
  double hhcjets::_M_select_subprocess(double x, double rho)
  {
    /*     calculate the weight of the subprocesses    */
    double weight[250], norm = 0.0;
    double *iwgt = weight;
    
    for(auto iproc = _M_proc_tbl; iproc->weight_function; ++iproc, ++iwgt)
    norm += (*iwgt = (this ->* (iproc->weight_function))(x, iproc->fl));
    
    
    /*     selecting subprocess     */
    unsigned ip = 0U;
    double cproc = weight[0];
    
    while(cproc < rho*norm)
      cproc += weight[++ip];
    
    /*    active subprocess     */
    _M_active = _M_proc_tbl + ip;
    
    
    return norm;
  }
  
  
  
  /*  Process 0: g + g --> g + g
   */
  double hhcjets::_M_process_gggg(double x, const int *) const
  {
    constexpr double cfac = qcd::Na2/(qcd::Nc2*qcd::Nc2);
    
    //----- calculate the weight ----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double x2 = x*x, ox2 = 1.0-x;
    ox2 *= ox2;
    
    return fa[0]*fb[0]*cfac*(1.0+x2*x2+ox2*ox2)/(x2*ox2);
  }
  
  
  void hhcjets::_S_select_color_gggg(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = x*x/16.0, p12 = x*ox/16.0, p13 = x/16.0, p22 = ox*ox/16.0, p23 = ox/16.0, p33 = 1.0/16.0;
    
    const double color_prob[6][6] = {
      {p11, p12, p13, p11, p12, p13},
      {p12, p22, p23, p12, p22, p23},
      {p13, p23, p33, p13, p23, p33},
      {p11, p12, p13, p11, p12, p13},
      {p12, p22, p23, p12, p22, p23},
      {p13, p23, p33, p13, p23, p33}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w13 = -1.0, w22 = 1.0, w23 = -1.0, w33 = 1.0;
    
    static const double color_weight[6][6] = {
      {w11, w12, w13, w11, w12, w13},
      {w12, w22, w23, w12, w22, w23},
      {w13, w23, w33, w13, w23, w33},
      {w11, w12, w13, w11, w12, w13},
      {w12, w22, w23, w12, w22, w23},
      {w13, w23, w33, w13, w23, w33}
    };
    
    //----- all the possible color state -----
    static const int perm[6][4] = {
      {-1,0,1,2}, {-1,1,2,0}, {-1,2,0,1}, {2,1,0,-1}, {0,2,1,-1}, {1,0,2,-1}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_closed_stringL(perm[i], perm[i]+4, state);
    insert_closed_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  /*    Process 1-6: g + g --> q + qbar, where q = d,u,s,c,b,t
   */
  double hhcjets::_M_process_ggqQ(double x, const int *) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double x2 = x*x, ox = 1.0-x;
    double ox2 = ox*ox;
    
    //----- leading contribution -----
    double retval = x2 + ox2;
    
    return fa[0]*fb[0]*retval/(8.0*qcd::Nc*x*ox);
  }
  
  
  void hhcjets::_S_select_color_ggqQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = ox*ox, p12 = x*ox, p22 = x*x;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,-1,0,2}, {1,0,-1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  /*    Process  7-11: g + q --> g + q, where q = d,u,s,c,b
   *    Process 12-16: g + qbar --> g + qbar where q = d,u,s,c,b
   */
  double hhcjets::_M_process_gqgq(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double x2 = x*x, ox = 1.0-x;
    double ox2 = ox*ox;
    
    //----- leading contribution -----
    constexpr double cfac = qcd::Na/(8.0*qcd::Nc2);
    double retval = (1.0+ox2)/(x2*ox);
    const int ib = -fl[1];
    
    retval *= fa[0]*fb[ib]*cfac;
    retval *= (2.0-x)*(2.0-x);         //  this is part of the color weights in the notes
    
    return retval;
  }
  
  
  void hhcjets::_S_select_color_gqgq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {2,-1,1,0}, {2,1,-1,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_color_gQgQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,1,-1,2}, {0,-1,1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  /*    Process 17-21: q + g --> g + q, where q = d,u,s,c,b
   *    Process 22-26: qbar + g --> g + qbar, where q = d,u,s,c,b
   */
  double hhcjets::_M_process_qggq(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = 1.0 - x;
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = qcd::Na/(8.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(y*y*oy);
    const int ia = -fl[0];
    
    retval *= fa[ia]*fb[0]*cfac;
    retval *= (2.0-y)*(2.0-y);      //  this is part of the color weights in the notes
    
    return retval;
  }
  
  
  void hhcjets::_S_select_color_qggq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[0].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {2,0,1,-1}, {2,1,0,-1}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  void hhcjets::_S_select_color_QggQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[0].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,0,2}, {-1,0,1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  /*    Process 27-31: q + qbar --> g + g, where q = d,u,s,c,b
   *    Process 32-36: qbar + q --> g + g, where q = d,u,s,c,b
   */
  double hhcjets::_M_process_qQgg(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = qcd::Na2/(16.0*qcd::Nc2*qcd::Nc);
    double retval = x*x+oy*oy;
    const int ia = -fl[0];
    
    retval *= cfac/(x*oy);
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  
  void hhcjets::_S_select_color_qQgg(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[1].momentum*state[2].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = x*x, p12 = x*ox, p22 = ox*ox;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,1,2,-1}, {0,2,1,-1}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  void hhcjets::_S_select_color_Qqgg(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[1].momentum*state[2].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = x*x, p12 = x*ox, p22 = ox*ox;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,1,0}, {-1,1,2,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+4, state);
    insert_open_stringR(perm[j], perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  /*    Process 37-41: d + dbar --> r + rbar, where r =   u,s,c,b,t
   *    Process 42-46: u + ubar --> r + rbar, where r = d,  s,c,b,t
   *    Process 47-51: s + sbar --> r + rbar, where r = d,u,  c,b,t
   *    Process 52-56: c + cbar --> r + rbar, where r = d,u,s,  b,t
   *    Process 57-61: b + bbar --> r + rbar, where r = d,u,s,c,  t
   
   *    Process 62-66: dbar + d --> rbar + r, where r =   u,s,c,b,t
   *    Process 67-71: ubar + u --> rbar + r, where r = d,  s,c,b,t
   *    Process 72-76: sbar + s --> rbar + r, where r = d,u,  c,b,t
   *    Process 77-81: cbar + c --> rbar + r, where r = d,u,s,  b,t
   *    Process 81-86: bbar + b --> rbar + r, where r = d,u,s,c,  t
   */
  double hhcjets::_M_process_qQrR(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(8.0*qcd::Nc2);
    double retval = x*x+oy*oy;
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_color_qQrR(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_color_QqRr(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {2,1,-1,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  
  /*    Process 87-90:   d + r --> d + r, where r =   u,s,c,b
   *    Process 91-94:   u + r --> u + r, where r = d,  s,c,b
   *    Process 95-98:   s + r --> s + r, where r = d,u,  c,b
   *    Process 99-102:  c + r --> c + r, where r = d,u,s,  b
   *    Process 103-106: b + r --> b + r, where r = d,u,s,c
   
   *    Process 107-110: dbar + rbar --> dbar + rbar, where r =   u,s,c,b
   *    Process 111-114: ubar + rbar --> ubar + rbar, where r = d,  s,c,b
   *    Process 115-118: sbar + rbar --> sbar + rbar, where r = d,u,  c,b
   *    Process 119-122: cbar + rbar --> cbar + rbar, where r = d,u,s,  b
   *    Process 123-126: bbar + rbar --> bbar + rbar, where r = d,u,s,c
   
   *    Process 127-130: d + rbar --> d + rbar, where r =   u,s,c,b
   *    Process 131-134: u + rbar --> u + rbar, where r = d,  s,c,b
   *    Process 135-138: s + rbar --> s + rbar, where r = d,u,  c,b
   *    Process 139-142: c + rbar --> c + rbar, where r = d,u,s,  b
   *    Process 143-146: b + rbar --> b + rbar, where r = d,u,s,c
   
   *    Process 147-150: dbar + r --> dbar + r, where r =   u,s,c,b
   *    Process 151-154: ubar + r --> ubar + r, where r = d,  s,c,b
   *    Process 155-158: sbar + r --> sbar + r, where r = d,u,  c,b
   *    Process 159-162: cbar + r --> cbar + r, where r = d,u,s,  b
   *    Process 163-166: bbar + r --> bbar + r, where r = d,u,s,c
   */
  double hhcjets::_M_process_qrqr(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(8.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(x*x);
    const int ia = -fl[0], ib = -fl[1];
    
    retval *= cfac;
    retval *= fa[ia]*fb[ib];
    
    return retval;
  }
  
  void hhcjets::_S_select_color_qrqr(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_color_QRQR(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_color_qRqR(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,2,0,-1}, {1,-1,0,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  void hhcjets::_S_select_color_QrQr(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,0,2,1}, {-1,1,2,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  /*    Process 167-169:   d + dbar --> d + dbar
   *    Process 170-172:   u + ubar --> u + ubar
   *    Process 173-175:   s + sbar --> s + sbar
   *    Process 176-168:   c + cbar --> c + cbar
   *    Process 179-181:   b + bbar --> b + bbar
   
   *    Process 182-184:   dbar + d --> dbar + d
   *    Process 185-187:   ubar + u --> ubar + u
   *    Process 188-190:   sbar + s --> sbar + s
   *    Process 191-193:   cbar + c --> cbar + c
   *    Process 194-196:   bbar + b --> bbar + b
   */
  double hhcjets::_M_processA_qQqQ(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = x*x+oy*oy;
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_colorA_qQqQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_colorA_QqQq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {-1,0,2,1}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  
  double hhcjets::_M_processB_qQqQ(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(x*x);
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_colorB_qQqQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  void hhcjets::_S_select_colorB_QqQq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {-1,0,2,1}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  double hhcjets::_M_processC_qQqQ(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = 2.0*oy*oy/x;
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_colorC_qQqQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  void hhcjets::_S_select_colorC_QqQq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {-1,0,2,1}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  /*    Process 197-199:   d + d --> d + d
   *    Process 200-202:   u + u --> u + u
   *    Process 203-205:   s + s --> s + s
   *    Process 206-208:   c + c --> c + c
   *    Process 209-211:   b + b --> b + b
   
   *    Process 212-214:   dbar + dbar --> dbar + dbar
   *    Process 215-217:   ubar + ubar --> ubar + ubar
   *    Process 218-210:   sbar + sbar --> sbar + sbar
   *    Process 221-223:   cbar + cbar --> cbar + cbar
   *    Process 224-226:   bbar + bbar --> bbar + bbar
   */
  double hhcjets::_M_processA_qqqq(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(x*x);
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_colorA_qqqq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_colorA_QQQQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  double hhcjets::_M_processB_qqqq(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = (1.0+x*x)/(oy*oy);
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_colorB_qqqq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  void hhcjets::_S_select_colorB_QQQQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4q[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
  
  double hhcjets::_M_processC_qqqq(double x, const int *fl) const
  {
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double oy = 1.0-x;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = 2.0/(x*oy);
    const int ia = -fl[0];
    
    retval *= cfac;
    retval *= fa[ia]*fb[ia];
    
    return retval;
  }
  
  void hhcjets::_S_select_colorC_qqqq(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {-w11, -w12},
      {-w12, -w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  void hhcjets::_S_select_colorC_QQQQ(shower_stage<hhc>& hard, double rho)
  {
    event<hhc>& state = hard.state;
    
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {-w11, -w12},
      {-w12, -w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    
    /*  MC color sum   */
    auto[i,j] = _S_pick_a_color_state(color_prob, rho);
    
    insert_open_stringL(perm[i], perm[i]+2, state);
    insert_open_stringL(perm[i]+2, perm[i]+4, state);
    
    insert_open_stringR(perm[j], perm[j]+2, state);
    insert_open_stringR(perm[j]+2, perm[j]+4, state);
    
    state.overlap_color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    state.color_weight = color_weight[i][j];
  }
  
  
}  //   namespace duct
