/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/23/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __proc_hccjets_h__
#define __proc_hccjets_h__ 1

//   std includes

//   project headers
#include <bits/duct-model.h>
#include <bits/mc-shower-history.h>
#include <cxx/random.h>


namespace duct {
  
  class hhcjets
  {
  public:
    /*     constructor    */
    hhcjets(const model_ref&, energy<double>, std::initializer_list<energy<double>>, std::initializer_list<double>,
            double = 1.0, double = 1.0, int = 2212, int = 2212);
    
    
    /*  this operator generates a new hard event   */
    void generate(random_engine_type&, shower_stage<hhc>&, shower_vetos&);
    
    
  private:
    /*   pointer to the model  */
    model_ref _M_mdl;
    
    /*   ponter to the pdfs   */
    int _M_hadronA, _M_hadronB;
    const qcd::dglap *_M_pdf1, *_M_pdf2;
    
    /*   genreate the event (most of the work is done here)    */
    double _M_generate_event(random_engine_type&, shower_stage<hhc>&, shower_vetos&);
    
    /*    select the active subprocess    */
    double _M_select_subprocess(double, double);
    
    
    /*   xT and y12 ranges  */
    struct xTyrange {
      double xTmin, xTmax;
      double ymin, ymax;
    };
    
    std::vector<xTyrange> _M_xTy;
    
    /*     structure for represnting the subprocesses    */
    struct subprocess {
      int fl[4];
      double (hhcjets::*weight_function)(double, const int *) const;
      void (hhcjets::*select_color)(shower_stage<hhc>&, double);
    };
    
    /*     table of the possible subprocesses     */
    static const subprocess _M_proc_tbl[250];
    
    /*     cached variables to communicate between the private member functions     */
    energy_sqr<double> _M_s;
 
    double _M_scalefac_pert, _M_scalefac_shw;
    mutable double _M_fa[11], _M_fb[11];
    mutable const subprocess *_M_active;
    
     
    /*      tables of color supression indices    */
    static const unsigned int _M_color_suppr_idx_4g[6][6];
    static const unsigned int _M_color_suppr_idx_2q2g[2][2];
    static const unsigned int _M_color_suppr_idx_4q[2][2];
    static const unsigned int _M_color_suppr_idx_4qC[2][2];
    
    
    /*      weight functions of the possible subprocesses     */
    double _M_process_gggg(double, const int *) const;
    double _M_process_ggqQ(double, const int *) const;
    double _M_process_gqgq(double, const int *) const;
    double _M_process_qggq(double, const int *) const;
    double _M_process_qQgg(double, const int *) const;
    double _M_process_qQrR(double, const int *) const;
    double _M_process_qrqr(double, const int *) const;
    double _M_processA_qQqQ(double, const int *) const;
    double _M_processB_qQqQ(double, const int *) const;
    double _M_processC_qQqQ(double, const int *) const;
    double _M_processA_qqqq(double, const int *) const;
    double _M_processB_qqqq(double, const int *) const;
    double _M_processC_qqqq(double, const int *) const;
    
    /*     functions for selecting color configurations    */
    void _S_select_color_gggg(shower_stage<hhc>&, double);
    void _S_select_color_ggqQ(shower_stage<hhc>&, double);
    void _S_select_color_gqgq(shower_stage<hhc>&, double);
    void _S_select_color_gQgQ(shower_stage<hhc>&, double);
    void _S_select_color_qggq(shower_stage<hhc>&, double);
    void _S_select_color_QggQ(shower_stage<hhc>&, double);
    void _S_select_color_qQgg(shower_stage<hhc>&, double);
    void _S_select_color_Qqgg(shower_stage<hhc>&, double);
    void _S_select_color_qQrR(shower_stage<hhc>&, double);
    void _S_select_color_QqRr(shower_stage<hhc>&, double);
    void _S_select_color_qrqr(shower_stage<hhc>&, double);
    void _S_select_color_QRQR(shower_stage<hhc>&, double);
    void _S_select_color_qRqR(shower_stage<hhc>&, double);
    void _S_select_color_QrQr(shower_stage<hhc>&, double);
    
    void _S_select_colorA_qQqQ(shower_stage<hhc>&, double);
    void _S_select_colorB_qQqQ(shower_stage<hhc>&, double);
    void _S_select_colorC_qQqQ(shower_stage<hhc>&, double);
    void _S_select_colorA_QqQq(shower_stage<hhc>&, double);
    void _S_select_colorB_QqQq(shower_stage<hhc>&, double);
    void _S_select_colorC_QqQq(shower_stage<hhc>&, double);
    void _S_select_colorA_qqqq(shower_stage<hhc>&, double);
    void _S_select_colorB_qqqq(shower_stage<hhc>&, double);
    void _S_select_colorC_qqqq(shower_stage<hhc>&, double);
    void _S_select_colorA_QQQQ(shower_stage<hhc>&, double);
    void _S_select_colorB_QQQQ(shower_stage<hhc>&, double);
    void _S_select_colorC_QQQQ(shower_stage<hhc>&, double);
    
    
    /*   helper function to select a color state    */
    template<unsigned N>
    static std::pair<unsigned, unsigned> _S_pick_a_color_state(const double (&cpm)[N][N], double rho)
    {
      double *cp = (double *)cpm;
      double cprob = cp[0];
      unsigned int ic = 0u;
      
      while(cprob < rho)
        cprob += cp[++ic];
      
      return {ic/N, ic%N};
    }
  };
  
}

#endif
