/*  rs-resum-event.h
 *  Created by Zoltan Nagy on 18.08.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __rs_resum_event_h__
#define __rs_resum_event_h__ 1


/*   std includes    */
#include <algorithm>
#include <iostream>

/*   project headers    */
#include <bits/duct-tags.h>
#include <bits/duct-model.h>
#include <bits/duct-lorentzvector.h>
#include <cxx/misc/cxx-iterator.h>
#include <cxx/memory-resource.h>



namespace duct {
  namespace resum {
    
    struct resum_event_vars
    {
      /*   splitting variables of the mapping    */
      int type;
      long double y, z, phi;
      int emitter, spectator;
      
      /*   scale of the strong coupling   */
      long double asscale;
      
      /*   Lambda^2/Q2 and kT^2/Q2   */
      long double lambda, kappaT;
    };

    
    struct resum_event_base
    {
      /*  Overall weight that is independent of the color states and operators.
       *  We might want to do MC summation and that has some weight.
       */
      long double weight = 0L;
      
      /*   scale of the stage   */
      long double scale = 1.0L;
      
      /*  splitting variables   */
      resum_event_vars vars;
    };
    
    
    
    /*    base class for the event classes, we used to have statistical_basis    */
    template<ProcessTag _Proc>
    class resummation_event : public resum_event_base
    {
      /*   some private types   */
      using _Diff = std::ptrdiff_t;
      using _Size = std::size_t;
      using _Base = resum_event_base;
      
      
      /*   number of incoming particles   */
      static constexpr _Size _n_in = number_of_incomings<_Proc>;
      static constexpr _Size _n_reserve = 50u;
      
    public:
      /*   public types   */
      /*  This class is for storing one particle informations.
       *    => flavor        : flavor of the parton
       *    => momentum      : four mementum of the particle
       */
      struct particle
      {
        /*   flavor information   */
        int flavor = 0;
        
        /*    momentum of the particle    */
        lorentzvector<long double> momentum;
      };
      
      
      /*   usual types   */
      using reference = particle&;
      using pointer = particle *;
      using iterator = pointer;
      
      using const_reference = const particle&;
      using const_pointer = const particle *;
      using const_iterator = const_pointer;
      
      using size_type = _Size;
      using difference_type = _Diff;
      using allocator_type = pmr::polymorphic_allocator<particle>;
      
      
      /*    default constructors reserves space for 50 particles   */
      resummation_event() : resummation_event{allocator_type{}} {}
      
      explicit resummation_event(allocator_type a)
      : _M_memory{a}, _M_model{get_default_model()}
      {
        _M_create_storage_for(_n_reserve);
      }
      
      /*   destructor   */
      ~resummation_event() { _M_deallocate();}
      
      
      /*   non-trivial constructor   */
      explicit resummation_event(_Size _n_qcd, allocator_type a = {})
      : _M_memory{a}, _M_model{get_default_model()}
      {
        _M_create_storage_for(_n_in+_n_qcd);
        _M_lay_down_for(0, _n_qcd);
      }
      
      resummation_event(_Size _n_other, _Size _n_qcd, allocator_type a = {})
      : _M_memory{a}, _M_model{get_default_model()}
      {
        _M_create_storage_for(_n_other+_n_in+_n_qcd);
        _M_lay_down_for(_n_other, _n_qcd);
      }
      
      
      /*   copy    */
      resummation_event& operator=(const resummation_event&);
      
      resummation_event(const resummation_event& x, allocator_type a = {})
      : _Base{x}, _M_memory{a}, _M_model{x._M_model}
      {
        if(x.begin()) {
          _M_create_storage_for(x.size());
          _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
          std::copy(x.begin(), x.end(), _M_begin);
        }
        else _M_begin = _M_in = _M_out = _M_end = _M_end_of_storage = nullptr;
      }
      
      
      /*   move    */
      resummation_event& operator=(resummation_event&&);
      resummation_event(resummation_event&&, allocator_type);
      
      resummation_event(resummation_event&& x)
      : _Base{std::move(x)}
      ,_M_memory{x._M_memory}
      , _M_begin{x._M_begin}
      , _M_in{x._M_in}
      , _M_out{x._M_out}
      , _M_end{x._M_end}
      , _M_end_of_storage{x._M_end_of_storage}
      , _M_model{x._M_model}
      {
        x._M_begin = x._M_in = x._M_out = x._M_end = x._M_end_of_storage = nullptr;
      }
      
      
      
      /*   element access  */
      reference operator[](_Diff k) {
        return _M_out[k-1];
      }
      
      const_reference operator[](_Diff k) const {
        return _M_out[k-1];
      }
   
      
      /*   iterators   */
      iterator begin() { return _M_begin;}
      const_iterator begin() const { return _M_begin;}
      const_iterator cbegin() const { return _M_begin;}
      
      iterator end() { return _M_end;}
      const_iterator end() const { return _M_end;}
      const_iterator cend() const { return _M_end;}
      
      
      /*   direct access to the data   */
      pointer data() { return _M_begin;}
      const_pointer data() const { return _M_begin;}
      const_pointer cdata() const { return _M_begin;}
      
      
      /*    observers   */
      bool empty() const { return (_M_begin == _M_end);}
      
      _Diff lower() const { return _M_begin-_M_out+1;}
      _Diff upper() const { return _M_end-_M_out;}
      
      _Size number_of_in() const { return static_cast<_Size>(_M_out-_M_in);}
      _Size number_of_qcd() const { return static_cast<_Size>(_M_end-_M_out);}
      _Size number_of_other() const { return static_cast<_Size>(_M_in-_M_begin);}
      
      _Size size() const { return static_cast<_Size>(_M_end-_M_begin);}
      _Size capacity() const { return static_cast<_Size>(_M_end_of_storage-_M_begin);}
      _Size capacity_qcd() const { return static_cast<_Size>(_M_end_of_storage-_M_out);}
      
      
      /*    get/set the particle table    */
      model_ref get_model() const { return _M_model;}
      void set_model(const model_ref& mdl) { _M_model = mdl;}
      
      
      /*   resreving storage    */
      void reserve(_Size _n_qcd)
      {
        if(!_M_begin) _M_create_storage_for(_n_in + _n_qcd);
        else if(capacity_qcd() < _n_qcd) _M_rebase_reserve(_n_qcd);
      }
      
      
      /*    resizing the array    */
      void resize(_Size _n_qcd)
      {
        /*   invalid resummation_events we just recreate the array   */
        if(!_M_begin) {
          _M_create_storage_for(_n_in + _n_qcd);
          _M_lay_down_for(0, _n_qcd);
          return;
        }
        
        /*   otherwise   */
        if(capacity_qcd() < _n_qcd) {
          _M_rebase_resize(number_of_other(), _n_qcd);
          return;
        }
        
        _M_end = _M_out + _n_qcd;
      }
      
      void resize(_Size _n_other, _Size _n_qcd)
      {
        /*   invalid resummation_events we just recreate the array   */
        if(!_M_begin) {
          _M_create_storage_for(_n_other + _n_in + _n_qcd);
          _M_lay_down_for(_n_other, _n_qcd);
          return;
        }
        
        /*   otherwise   */
        if(capacity() < _n_other + _n_in + _n_qcd) {
          _M_rebase_resize(_n_other, _n_qcd);
          return;
        }
        
        _M_lay_down_for(_n_other, _n_qcd);
      }
      
      void resize_by(_Diff k) {
        resize(static_cast<_Size>(std::max<_Diff>(0, upper()+k)));
      }
      
      /*   shrink the array   */
      void shrink_to_fit() {
        if(_M_begin && _M_end != _M_end_of_storage)
          _M_rebase_reserve(number_of_qcd());
      }
      
      /*   remove the last parton   */
      void remove_last() {
        if(_M_end) --_M_end;
      }
      
    private:
      /*   memory resource of this object    */
      allocator_type _M_memory;
      
      /*   pointers to mark the resummation_event layout in the memory   */
      pointer _M_begin, _M_in, _M_out, _M_end, _M_end_of_storage;
      
      /*   particle table    */
      model_ref _M_model;
      
        
    private:
      /*   create storage    */
      void _M_create_storage_for(_Size _count)
      {
        auto count = std::max(_n_reserve, _count);
        _M_in = _M_out = _M_end = _M_begin = _M_memory.allocate(count);
        _M_end_of_storage = _M_begin + count;
      }
      
      /*   allocate memory    */
      void _M_lay_down_for(_Size _n_other, _Size _n_qcd)
      {
        _M_in = _M_begin + _n_other;
        _M_out = _M_in + _n_in;
        _M_end = _M_out + _n_qcd;
      }
      
      /*   deallocate memory    */
      void _M_deallocate()
      {
        if(_M_begin)
          _M_memory.deallocate(_M_begin, capacity());
      }
      
      /*    rebase    */
      void _M_rebase_resize(_Size _n_other, _Size _n_qcd)
      {
        auto count = std::max(2*size(), _n_other + _n_in + _n_qcd);
        auto ptr = _M_memory.allocate(count);
        
        std::copy(_M_begin, _M_end, ptr);
        _M_deallocate();
        
        _M_begin = ptr;
        _M_end_of_storage = ptr + count;
        _M_lay_down_for(_n_other, _n_qcd);
      }
      
      void _M_rebase_reserve(_Size nqcd)
      {
        auto _n_qcd = number_of_qcd();
        auto _n_other = number_of_other();
        auto count = _n_other + _n_in + nqcd;
        auto ptr = _M_memory.allocate(count);
        
        std::copy(_M_begin, _M_end, ptr);
        _M_deallocate();
        
        _M_begin = ptr;
        _M_end_of_storage = ptr + count;
        _M_lay_down_for(_n_other, _n_qcd);
      }
    };
    
    
    /*     copy assignment   */
    template<ProcessTag _Proc>
    resummation_event<_Proc>& resummation_event<_Proc>::operator=(const resummation_event<_Proc>& x)
    {
      if(&x != this && x.begin())
      {
        if(auto count = x.size(); capacity() < count) {
          _M_deallocate();
          _M_create_storage_for(count);
        }
        
        _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
        std::copy(x.begin(), x.end(), _M_begin);
        _M_model = x._M_model;
        _Base::operator=(x);
      }
      
      return *this;
    }
    
    
    /*    move constructor    */
    template<ProcessTag _Proc>
    resummation_event<_Proc>::resummation_event(resummation_event<_Proc>&& x, allocator_type a)
    : _Base{std::move(x)}, _M_memory{a}, _M_model{std::move(x._M_model)}
    {
      _M_begin = _M_in = _M_out = _M_end = _M_end_of_storage = nullptr;
      
      if(x.begin())
      {
        if(_M_memory == x._M_memory)
        {
          std::swap(_M_begin ,x._M_begin);
          std::swap(_M_in, x._M_in);
          std::swap(_M_out, x._M_out);
          std::swap(_M_end, x._M_end);
          std::swap(_M_end_of_storage, x._M_end_of_storage);
        }
        else {
          _M_create_storage_for(x.size());
          _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
          std::copy(x.begin(), x.end(), _M_begin);
        }
      }
    }
    
    
    
    /*     move assignment   */
    template<ProcessTag _Proc>
    resummation_event<_Proc>& resummation_event<_Proc>::operator=(resummation_event<_Proc>&& x)
    {
      if(&x != this && x.begin())
      {
        _Base::operator=(std::move(x));
        
        if(_M_memory == x._M_memory)
        {
          std::swap(_M_begin ,x._M_begin);
          std::swap(_M_in, x._M_in);
          std::swap(_M_out, x._M_out);
          std::swap(_M_end, x._M_end);
          std::swap(_M_end_of_storage, x._M_end_of_storage);
          std::swap(_M_model, x._M_model);
        }
        else {
          if(auto count = x.size(); capacity() < count) {
            _M_deallocate();
            _M_create_storage_for(count);
          }
          
          _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
          std::copy(x.begin(), x.end(), _M_begin);
          _M_model = std::move(x._M_model);
        }
      }
      
      return *this;
    }
    
    
    
    /*    iterators    */
    template<class _Range>
    requires (std::is_same_v<_Range, incomings>
              || std::is_same_v<_Range, non_qcd>
              || std::is_same_v<_Range, qcd_out>)
    inline auto begin(resummation_event<ProcessTag auto>& p)
    {
      auto ret = p.begin();
      
      if constexpr(std::is_same_v<_Range, incomings>)
        ret += p.number_of_other();
      else if constexpr(std::is_same_v<_Range, qcd_out>)
        ret += p.number_of_other() + p.number_of_in();
      
      return ret;
    }
    
    
    template<class _Range>
    requires (std::is_same_v<_Range, incomings>
              || std::is_same_v<_Range, non_qcd>
              || std::is_same_v<_Range, qcd_out>)
    inline auto end(resummation_event<ProcessTag auto>& p)
    {
      auto ret = p.begin() + p.number_of_other();
      
      if constexpr(std::is_same_v<_Range, incomings>)
        ret += p.number_of_in();
      else if constexpr(std::is_same_v<_Range, qcd_out>)
        ret += p.number_of_in() + p.number_of_qcd();
      
      return ret;
    }
    
    
    template<class _Range>
    requires (std::is_same_v<_Range, incomings>
              || std::is_same_v<_Range, non_qcd>
              || std::is_same_v<_Range, qcd_out>)
    inline auto cbegin(const resummation_event<ProcessTag auto>& p)
    {
      auto ret = p.begin();
      
      if constexpr(std::is_same_v<_Range, incomings>)
        ret += p.number_of_other();
      else if constexpr(std::is_same_v<_Range, qcd_out>)
        ret += p.number_of_other() + p.number_of_in();
      
      return ret;
    }
    
    
    template<class _Range>
    requires (std::is_same_v<_Range, incomings>
              || std::is_same_v<_Range, non_qcd>
              || std::is_same_v<_Range, qcd_out>)
    inline auto cend(const resummation_event<ProcessTag auto>& p)
    {
      auto ret = p.begin() + p.number_of_other();
      
      if constexpr(std::is_same_v<_Range, incomings>)
        ret += p.number_of_in();
      else if constexpr(std::is_same_v<_Range, qcd_out>)
        ret += p.number_of_in() + p.number_of_qcd();
      
      return ret;
    }
    
    
    template<class _Range>
    inline auto begin(const resummation_event<ProcessTag auto>& p) {
      return cbegin<_Range>(p);
    }
    
    
    template<class _Range>
    inline auto end(const resummation_event<ProcessTag auto>& p){
      return cend<_Range>(p);
    }
    
    
    
    /*   ranges   */
    template<class _Range, ProcessTag _Proc>
    inline auto range(resummation_event<_Proc>& p)
    {
      return iterator_range<typename resummation_event<_Proc>::iterator> {
        begin<_Range>(p), end<_Range>(p)
      };
    }
    
    template<class _Range, ProcessTag _Proc>
    inline auto crange(const resummation_event<_Proc>& p)
    {
      return iterator_range<typename resummation_event<_Proc>::const_iterator> {
        cbegin<_Range>(p), cend<_Range>(p)
      };
    }
    
    template<class _Range, ProcessTag _Proc>
    inline auto range(const resummation_event<_Proc>& p) {
      return crange<_Range>(p);
    }
    
    
    /*   number of different kind of particles    */
    template<class _Range>
    requires (std::is_same_v<_Range, incomings>
              || std::is_same_v<_Range, non_qcd>
              || std::is_same_v<_Range, qcd_out>)
    inline auto number_of(const resummation_event<ProcessTag auto>& p)
    {
      if constexpr(std::is_same_v<_Range, incomings>) return p.number_of_in();
      else if constexpr(std::is_same_v<_Range, non_qcd>) return p.number_of_other();
      else return p.number_of_qcd();
    }
    
    
    /*   momentum sum over a range   */
    template<class _Range>
    requires (std::is_same_v<_Range, incomings>
              || std::is_same_v<_Range, non_qcd>
              || std::is_same_v<_Range, qcd_out>)
    inline lorentzvector<long double> momentum_sum(const resummation_event<ProcessTag auto>& p)
    {
      lorentzvector<long double> retval;
      
      for(const auto& ri : range<_Range>(p))
        retval += ri.momentum;
      
      return retval;
    }
    
    
    
    /*   binary I/O     */
    template<ProcessTag _Proc>
    inline std::ostream& write(std::ostream& os, const resummation_event<_Proc>& a) {
      return os.write(static_cast<const char *>(a.begin()), a.size()*sizeof(resummation_event<_Proc>::particle));
    }
    
    template<ProcessTag _Proc>
    inline std::istream& write(std::istream& is, resummation_event<_Proc>& a)
    {
      std::size_t nother, nqcd;
      a.resize(nother, nqcd);
      return is.read(static_cast<char *>(a.begin()), a.size()*sizeof(resummation_event<_Proc>::particle));
    }
    
    /*    print the resummation_event   */
    template<ProcessTag _Proc>
    std::ostream& operator<<(std::ostream& os, const resummation_event<_Proc>& a)
    {
      /*   code for printing the partons    */
      model_ref mdl = a.get_model();
      
      //----- the ranges -----
      auto qcd = range<qcd_out>(a);
      auto non = range<non_qcd>(a);
      auto inc = range<incomings>(a);
      
      //------ print the momenta ----
      int idx = 1-number_of<non_qcd>(a) - number_of<incomings>(a);
      
      if(non.begin() < non.end()) {
        os<<"+---- non-qcd outgoings -----\n";
        for(auto& pi : non)
          os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "
          <<pi.momentum<<"  "<<pi.momentum.mag2()<<'\n';
      }
      
      os<<"+---- incomings -----\n";
      for(auto& pi : inc)
        os<<"|  "<<mdl.name(-(pi.flavor))<<"("<<idx++<<") : "
        <<pi.momentum<<"  "<<pi.momentum.mag2()
        <<'\n';
      
      os<<"+---- qcd outgoings -----\n";
      for(auto& pi : qcd)
        os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "
        <<pi.momentum<<"  "<<pi.momentum.mag2()
        <<'\n';
      
      os<<"+----\n\n\n";
      
      auto itot = momentum_sum<incomings>(a);
      auto otot = momentum_sum<non_qcd>(a) + momentum_sum<qcd_out>(a);
      
      os<<"Total incoming momentum     : "<<itot<<'\n';
      os<<"Total outgoing momentum     : "<<otot<<'\n';
      os<<"Momentum conservation       : "<<itot-otot<<"\n\n";
      
      return os;
    }
    
    
  }  //   namespace resum
}  //   namespace duct


#endif
