/*  rs-resum-test.h
 *  Created by Zoltan Nagy on 21.01.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __rs_resum_test_h__
#define __rs_resum_test_h__ 1


/*   std includes    */
#include <type_traits>
#include <utility>


/*   project headers    */
#include <cxx/random.h>
#include <cxx/memory-resource.h>
#include <bits/rs-resum-stage.h>
#include <bits/rs-shower-scheme.h>
#include <bits/qcd-color-overlap.h>
#include <bits/hep-dist.h>




namespace duct {
  namespace resum {
   
    
    /*   Requirements for a hard state   */
    template<class _Proc, class _Hard>
    concept ResumHardState = ProcessTag<_Proc> &&
    requires(_Hard& h, random_engine_type& rne, resummation_stage<_Proc,1u>& stage)
    {
      /*   generates a hard event   */
      {h.generate(rne, stage)};
    };



    /*   Class declaration for dound resumation test for
     *   various shower schemes and processes
     *
     *   There is no definition here because we don't know the
     *   structure of the test for hadron collisions and we define
     *   this class as specialaization for every process type.
     */
    template<ProcessTag>
    class resummation_exponent;
    


    /*   Class for doing resummation test for e+e- annihilation process.
     *   The shower scheme is specified via the template parameter.
     */
    template<>
    class resummation_exponent<epa>
    {
      /*   private types   */
      using _Event = resummation_event<epa>;
      
      /*    Storage class for the variaous stages and it helps to have simpler memory management   */
      struct _Stages
      {
        /*    it has to be an allocator aware container     */
        using allocator_type = pmr::polymorphic_allocator<>;
        
        /*   default constructors    */
        explicit _Stages(allocator_type a) : hard{a}, first{a}, aterm{a}, bterm{a}
        {}
        
        /*   copy   */
        _Stages(const _Stages&) = default;
        _Stages& operator=(const _Stages&) = default;
        
        _Stages(const _Stages& x, allocator_type a)
        : hard{x.hard, a}, first{x.first, a}, aterm{x.aterm, a}, bterm{x.bterm, a}
        {}
        
        /*    move  */
        _Stages(_Stages&&) = default;
        _Stages& operator=(_Stages&&) = default;
        
        _Stages(_Stages&& x, allocator_type a)
        : hard{std::move(x.hard), a}, first{std::move(x.first), a}
        , aterm{std::move(x.aterm), a}, bterm{std::move(x.bterm), a}
        {}
        
        
        /*    hard event   */
        resummation_stage<epa,1u> hard;
        
        /*    workspace for the first and second stages    */
        resummation_stage<epa,2u> first;
        resummation_stage<epa,6u> aterm;
        resummation_stage<epa,2u> bterm;
      };
      
      
  public:
      /*   default constructors  (intital memory block is 16MB)  */
      resummation_exponent() : _M_memory{1u<<25}, _M_stages{nullptr}
      {}
      
      /*   no copy   */
      resummation_exponent(const resummation_exponent&) = delete;
      resummation_exponent& operator=(const resummation_exponent&) = delete;

      
      /*   Generates the events and calculate the corresponing weights
       *   for the S_\cV^{(2)} test. No observable includes.
       */
      template<class _Scheme, class _Hard>
      requires ShowerScheme<epa, _Scheme> && ResumHardState<epa, _Hard>
      void generate(random_engine_type&, _Hard&);
      
      
      /*   convolute to the observable    */
      template<typename _Ret, class _ORatio, typename ..._Args>
      requires std::is_invocable_r_v<_Ret, _ORatio, const resummation_event<epa>&, const resummation_event<epa>&, _Args&&...>
      auto operator()(_ORatio&&, _Args&&...) const;

      /*   this is for event shapes when we use Laplace transformation    */
      template<class _Func>
      requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&>
      auto operator()(_Func&&, std::initializer_list<int> = {0,1,2,3,4,5}, bool = true) const;

      template<class _Func>
      requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&>
      auto laplace1(_Func&&, std::initializer_list<int> = {0,1,2,3,4,5}, bool = true) const;
      
      template<class _Func>
      requires std::is_invocable_r_v<long double, _Func, const resum_event_vars&>
      auto operator()(_Func&&, std::initializer_list<int> = {0,1,2,3,4,5}, bool = true) const;

      template<class _Func>
      requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&>
      auto integrand(_Func&&, std::initializer_list<int> = {0,1,2,3,4,5}, bool = true) const;
      
      template<class _Func>
      requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&, long double>
      auto integrand2(_Func&&, std::initializer_list<int> = {0,1,2,3,4,5}, bool = true) const;
      

      /*   access to the hard state    */
      const resummation_stage<epa,1u>& hard() const { return _M_stages->hard;}

    private:
      /*   memory resouce   */
      duct::monotonic_buffer_resource _M_memory;

      /*   stages   */
      _Stages * _M_stages;
 
      
      /*    list of possible color operators      */
      std::vector<splitting_cvars_virt> _TMP_color_vertices;
  
      /*   every analyser has its own resources to calculate color overlaps   */
      mutable color_overlap _TMP_overlap;

      
    private:
      /*   do the color mapping    */
      void _M_color_mapping(int, int, shower_color_state<epa>&);
      
      
      /*   This is the overlap calculator function.    */
      long double _M_color_weight(const shower_color_state<epa>& c) const
      {
        if(c.empty()) return 0.0;
        
        /*    calculate all the overlaps    */
        long double ret = 0.0;
        
        for(auto& ci : c)
          ret += std::real(ci.weight)*_TMP_overlap(ci);
        
        return ret;
      }
 
    
      /*   generate a splitting   */
      template<class _Scheme>
      void _M_generate_splitting(int type, _Event& p, const _Event& plk, const long double *r, int l, int k, long double fac)
      {
        /*    generate the variables   */
        auto zeta = _Scheme::generate_vars(r, l, k, plk);
        
        /*   do the mapping   */
        zeta.type = type;
        _Scheme::mapping(p, plk, zeta);
        
        /*   some useful variables   */
        p.vars = _Scheme::event_vars(plk, zeta);
    
        
        /*   calculate the weight  */
        p.scale = _Scheme::ordering(plk, zeta);
        p.weight = 0.0;

        if(p.scale < plk.scale && _Scheme::safety_cuts(plk, zeta)) //&& std::abs(zeta.jac) > 1.0e-24
        {
          p.weight = _Scheme::splitting_function(p, plk, zeta);
          p.weight *= plk.weight;
          p.weight *= fac*zeta.jac;
        }
      }
    };
  

    template<class _Scheme, class _Hard>
    requires ShowerScheme<epa, _Scheme> && ResumHardState<epa, _Hard>
    void resummation_exponent<epa>::generate(random_engine_type& rne, _Hard& hardgen)
    {
      /*    local variables   */
      std::array<long double, 20u> r;
      std::uniform_real_distribution<long double> rng(0.0, 1.0);
      
      
      /*   purge the old history and construct the new one  */
      _M_memory.reset();
  
      /*   constructing the hard stage    */
      pmr::polymorphic_allocator<_Stages> alloc{&_M_memory};
      
      _M_stages = alloc.allocate(1u);
      alloc.construct(_M_stages);
      
   
      /*    aliases   */
      auto& hard = _M_stages->hard;
      auto& first = _M_stages->first;
      auto& aterm = _M_stages->aterm;
      auto& bterm = _M_stages->bterm;

      
      /*   generate the hard event    */

      //======
      {
        hard.cstate.clear();
        hardgen.generate(rne, hard);
        
        
        /*   Setting up the color states
         *   Basicaaly these variables are not used here in the resummation test.
         */
        for(auto& c : hard.cstate)
        {
          /*   make the health report   */
          c.make_health_report();
          
          /*   We always start with SU(Nc) states.   */
          c.is_color_u3 = false;
          
          /*   color supression index of the evolution (always zero in the hard state)     */
          c.color_suppression_index = 0u;
          
          /*   color supression index of the overlap     */
          c.overlap_color_suppression_index = 0u;
          
          /*   control variables    */
          c.order_of_soft_corrections = 0u;
          c.order_of_softH_corrections = 0u;
          c.order_of_softV_corrections = 0u;
          c.order_of_softT_corrections = 0u;
        }
        
        /*   hard color weight   */
        hard.cweight = _M_color_weight(hard.cstate);
      }
      
      
      /*   generating the first stage    */
      const int np = hard.state.upper();
      std::uniform_int_distribution<int> lrng(1, np-1), krng(1, np);
  
      int l = lrng(rne), k = krng(rne);
      if(l == k) l = np;
      
      //======
      {
        /*   needs a set of random numbers    */
        for(long double& ri : r) ri = rng(rne);
        
        
        /*    generating  the splitting variables    */
        _M_generate_splitting<_Scheme>(1, first.state[0], hard.state, r.data(), l, k, np*(np-1));
        
        /*    "virtual" state   */
        first.state[1] = hard.state;
        first.state[1].weight = first.state[0].weight;
        first.state[1].scale = first.state[0].scale;

        /*   mapping the colors   */
        first.cstate = hard.cstate;
        _M_color_mapping(l, k, first.cstate);
        first.cweight = _M_color_weight(first.cstate);
      }
      
      
      /*    generating the A term of the second stage    */
      {
        /*   needs a new set of random numbers    */
        for(long double& ri : r) ri = rng(rne);
        
        
        /*    calcuating the six terms    */
        constexpr long double fac = 0.25*qcd::Ca;
        
        _M_generate_splitting<_Scheme>(20, aterm.state[0], first.state[0], r.data(), l,    k,     fac);
        _M_generate_splitting<_Scheme>(21, aterm.state[1], first.state[0], r.data(), l,    np+1, -fac);
        _M_generate_splitting<_Scheme>(22, aterm.state[2], first.state[0], r.data(), k,    l,     fac);
        _M_generate_splitting<_Scheme>(23, aterm.state[3], first.state[0], r.data(), k,    np+1, -fac);
        _M_generate_splitting<_Scheme>(24, aterm.state[4], first.state[0], r.data(), np+1, l,    -fac);
        _M_generate_splitting<_Scheme>(25, aterm.state[5], first.state[0], r.data(), np+1, k,    -fac);
        
        /*   the color state is the same as in the first step   */
        aterm.cstate = first.cstate;
        aterm.cweight = first.cweight;
      }
      
      
      /*    generating the B term of the second stage    */
      {
        /*   selecting emitter and spectator randomly    */
        int i = lrng(rne), j = krng(rne);
        if(i == j) i = np;
        
        /*    generation the two terms   */
        const long double fac = 0.25*np*(np-1);
        
        _M_generate_splitting<_Scheme>(3, bterm.state[0], first.state[0], r.data(), i, j,  fac);
        _M_generate_splitting<_Scheme>(4, bterm.state[1], first.state[1], r.data(), i, j, -fac);

        /*    the color state in the B term    */
        bterm.cstate = first.cstate;
        _M_color_mapping(i, j, bterm.cstate);
        bterm.cweight = _M_color_weight(bterm.cstate);
      }
      
    }
    
    
   
    template<class _Func>
    requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&>
    auto resummation_exponent<epa>::operator()(_Func&& observable, std::initializer_list<int> ida, bool idb) const
    {
      /*    aliases   */
      auto& hard = _M_stages->hard;
      auto& first = _M_stages->first;
      auto& aterm = _M_stages->aterm;
      auto& bterm = _M_stages->bterm;
      

      /*    calculate the relevant parameters      */
      long double obshrd = observable(hard.state), obs1st = 0.0;
      long double coeff1st = 0.0, coeffs1v = 0.0;
      std::array<long double, 8u> obs, coeff;

      obs.fill(0.0);
      coeff.fill(0.0);
      
      
      /*   parameters for the S1v term     */
      obs1st = observable(first.state[0])-obshrd;
      coeffs1v = 0.0;
      
      
      if(obs1st > 0.0)
      {
        /*   weight of the first stage   */
        coeffs1v = first.state[0].weight*first.cweight;
        
        /*   parameters for the A terms  */
        for(std::size_t k : ida)
        {
          coeff[k] = aterm.cweight*aterm.state[k].weight;
          obs[k] = observable(aterm.state[k])-obshrd;
          if(obs[k] <= 0.0) coeff[k] = 0.0;
          
          coeff1st -= coeff[k];
        }
        
        /*   parameters for the  B terms   */
        if(idb)
        {
          coeff[6] = bterm.cweight*bterm.state[0].weight;
          obs[6] = observable(bterm.state[0])-obshrd;
          if(obs[6] <= 0.0) coeff[6] = 0.0;
          
          
          coeff[7] = bterm.cweight*bterm.state[1].weight;
          obs[7] = obs1st + observable(bterm.state[1])-obshrd;
          if(obs[7] <= 0.0) coeff[7] = 0.0;
          
          coeff1st -= coeff[6];
          coeff1st -= coeff[7];
        }
      }
      
      
      /*    returns a pair of lamba expression   */
      return std::pair{
        /*   S_V^{(1)}  operator   */
        [=](double L) {
          long double x = std::exp(L)*obs1st;
          long double r0 = std::exp(-x)*coeffs1v;
          long double r1 = -x*r0;

          return std::array<long double, 3u>{{r0-coeffs1v, r1, r1*(1.0-x)}};
        },

        /*   S_V^{(2)} operator   */
        [=](double L) {
          /*    contribution of the virtual part in V operator   */
          long double s = std::exp(L);
          long double x = s*obs1st, r0, r1, r2;
          std::array<long double,4u> res;

          res[0] = std::exp(-x)*coeff1st;
          res[1] = -x*res[0];
          res[2] = res[1]*(1.0-x);
          res[3] = res[2]*(1.0-x)-x*res[1];

          /*    contribution of the real emission   */
          for(std::size_t k = 0; k < 8; ++k)
          {
            x = s*obs[k];

            res[0] += (r0 = std::exp(-x)*coeff[k]);
            res[1] += (r1 = -x*r0);
            res[2] += (r2 = r1*(1.0-x));
            res[3] += r2*(1.0-x)-x*r1;
          }

          return res;
        }
      };
    }

    template<class _Func>
    requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&>
    auto resummation_exponent<epa>::laplace1(_Func&& observable, std::initializer_list<int> ida, bool idb) const
    {
      /*    aliases   */
      auto& hard = _M_stages->hard;
      auto& first = _M_stages->first;
      auto& aterm = _M_stages->aterm;
      auto& bterm = _M_stages->bterm;
      
      
      /*    calculate the relevant parameters      */
      long double obshrd = observable(hard.state);
      std::array<long double, 9u> obs, coeff, coeffL;
      
      obs.fill(0.0);
      coeff.fill(0.0);
      coeffL.fill(0.0);

      
      /*   parameters for the I2 integral     */
      obs[8] = observable(first.state[0])-obshrd;
             
      if(obs[8] > 0.0)
      {
        long double ls1 = std::log(first.state[0].vars.asscale);
        
        /*   parameters for the A terms  */
        for(std::size_t k : ida)
        {
          obs[k] = observable(aterm.state[k])-obshrd;
          coeff[8]  -= (coeff[k] = (obs[k] > 0.0 ? aterm.cweight*aterm.state[k].weight : 0.0));
          coeffL[8] -= (coeffL[k] = -coeff[k]*(ls1 + std::log(aterm.state[k].vars.asscale)));
        }
        
        /*   parameters for the  B terms   */
        if(idb)
        {
          obs[6] = observable(bterm.state[0])-obshrd;
          coeff[8]  -= (coeff[6] = (obs[6] > 0.0 ? bterm.cweight*bterm.state[0].weight : 0.0));
          coeffL[8] -= (coeffL[6] = -coeff[6]*(ls1 + std::log(bterm.state[0].vars.asscale)));
        
          obs[7] = obs[8] + observable(bterm.state[1])-obshrd;
          coeff[8]  -= (coeff[7] = (obs[7] > 0.0 ? bterm.cweight*bterm.state[1].weight : 0.0));
          coeffL[8] -= (coeffL[7] = -coeff[7]*(ls1 + std::log(bterm.state[1].vars.asscale)));
        }
      }
      
      
      /*    returns a pair of lamba expression   */
      return std::pair{
        /*   I_2^{[2]} operator   */
        [=](double L) {
          long double s = std::exp(L);
          std::array<long double,4u> res{{0.0L,0.0L,0.0L,0.0L}};
          
          /*    contribution of the real emission   */
          for(std::size_t k = 0; k < 9; ++k)
          {
            long double x = s*obs[k], r0, r1, r2;
            
            res[0] += (r0 = std::exp(-x)*coeff[k]);
            res[1] += (r1 = -x*r0);
            res[2] += (r2 = r1*(1.0-x));
            res[3] += r2*(1.0-x)-x*r1;
          }
          
          return res;
        },
        
        /*   I_3^{[2]} operator   */
        [=](double L) {
          long double s = std::exp(L);
          std::array<long double,4u> c{{0.0L,0.0L,0.0L,0.0L}};
          std::array<long double,4u> d{{0.0L,0.0L,0.0L,0.0L}};

          /*    contribution of the real emission   */
          for(std::size_t k = 0; k < 9; ++k)
          {
            long double x = s*obs[k], r0, r1, r2;

            c[0] += (r0 = std::exp(-x)*coeff[k]);
            c[1] += (r1 = -x*r0);
            c[2] += (r2 = r1*(1.0-x));
            c[3] += r2*(1.0-x)-x*r1;
    
            d[0] += (r0 = std::exp(-x)*coeffL[k]);
            d[1] += (r1 = -x*r0);
            d[2] += (r2 = r1*(1.0-x));
            d[3] += r2*(1.0-x)-x*r1;
          }
          
          d[0] -= 2.0L*(L*c[0]            );
          d[1] -= 2.0L*(L*c[1] +      c[0]);
          d[2] -= 2.0L*(L*c[2] + 2.0L*c[1]);
          d[3] -= 2.0L*(L*c[3] + 3.0L*c[2]);

          return d;
        }
      };
    }
    
    
    template<class _Func>
    requires std::is_invocable_r_v<long double, _Func, const resum_event_vars&>
    auto resummation_exponent<epa>::operator()(_Func&& observable, std::initializer_list<int> ida, bool idb) const
    {
      /*    aliases   */
      auto& first = _M_stages->first;
      auto& aterm = _M_stages->aterm;
      auto& bterm = _M_stages->bterm;


      /*    calculate the relevant parameters      */
      long double obs1st = 0.0, coeff1st = 0.0, coeffs1v = 0.0;
      std::array<long double, 8u> obs, coeff;

      obs.fill(0.0);
      coeff.fill(0.0);


      if(std::abs(first.state[0].weight) > 1e-10)
      {
        /*   parameters for the S1v term     */
        obs1st = observable(first.state[0].vars);
        coeffs1v = first.state[0].weight*first.cweight;


        /*   parameters for the A terms  */
        for(std::size_t k : ida)
        {
          coeff1st -= (coeff[k] = aterm.cweight*aterm.state[k].weight);
          if(std::abs(coeff[k]) > 1e-10) obs[k] = obs1st + observable(aterm.state[k].vars);
        }

        /*   parameters for the  B terms   */
        if(idb)
        {
          coeff1st -= (coeff[6] = bterm.cweight*bterm.state[0].weight);
          if(std::abs(coeff[6]) > 1e-10) obs[6] = obs1st + observable(bterm.state[0].vars);
          
          coeff1st -= (coeff[7] = bterm.cweight*bterm.state[1].weight);
          if(std::abs(coeff[7]) > 1e-10) obs[7] = obs1st + observable(bterm.state[1].vars);
        }
      }


      /*    returns a pair of lamba expression   */
      return std::pair{
        /*   S_V^{(1)}  operator   */
        [=](double L) {
          long double x = std::exp(L)*obs1st;
          long double r0 = std::exp(-x)*coeffs1v;
          long double r1 = -x*r0;

          return std::array<long double, 3u>{{r0-coeffs1v, r1, r1*(1.0-x)}};
        },

        /*   S_V^{(2)} operator   */
        [=](double L) {
          /*    contribution of the virtual part in V operator   */
          long double s = std::exp(L);
          long double x = s*obs1st, r0, r1, r2;
          std::array<long double,4u> res;

          res[0] = std::exp(-x)*coeff1st;
          res[1] = -x*res[0];
          res[2] = res[1]*(1.0-x);
          res[3] = res[2]*(1.0-x)-x*res[1];

          /*    contribution of the real emission   */
          for(std::size_t k = 0; k < 8; ++k)
          {
            x = s*obs[k];

            res[0] += (r0 = std::exp(-x)*coeff[k]);
            res[1] += (r1 = -x*r0);
            res[2] += (r2 = r1*(1.0-x));
            res[3] += r2*(1.0-x)-x*r1;
          }

          return res;
        }
      };
    }

    
        
    template<class _Func>
    requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&>
    auto resummation_exponent<epa>::integrand(_Func&& funcy3, std::initializer_list<int> ida, bool idb) const
    {
      /*    aliases   */
      auto& first = _M_stages->first;
      auto& aterm = _M_stages->aterm;
      auto& bterm = _M_stages->bterm;
      
      
      /*    calculate the relevant parameters      */
      long double coeff1 = 0.0, obs1 = 0.0;
      std::array<long double, 8u> obs2, coeff2;
      
      obs2.fill(0.0L);
      coeff2.fill(0.0L);
      
      
      if(std::abs(first.state[0].weight) > 1e-10)
      {
        /*   parameters for the S1v term     */
        obs1 = funcy3(first.state[0]);
        coeff1 = first.state[0].weight*first.cweight;
        
        
        /*   parameters for the A terms  */
        for(std::size_t k : ida)
        {
          obs2[k] = funcy3(aterm.state[k]);
          coeff2[k] = aterm.cweight*aterm.state[k].weight;
        }
        
        
        /*   parameters for the  B terms   */
        if(idb)
        {
          obs2[6] = funcy3(bterm.state[0]);
          coeff2[6] = bterm.cweight*bterm.state[0].weight;
          
          obs2[7] = funcy3(bterm.state[1]);
          coeff2[7] = bterm.cweight*bterm.state[1].weight;
        }
      }
  
       
      /*    returns a pair of lambda expression   */
      return std::pair{
        /*   S_V^{(1)}  operator   */
        [=](const distpoint::hist1d& h) {
          std::array<long double, 2u> ret;
          long double yl = std::exp(-h.xmax), yh = std::exp(-h.xmin);
          
          ret[0] = (obs1 > yl ? coeff1 : 0.0L);
          ret[1] = (yl <= obs1 && obs1 < yh ? coeff1/(h.xmax-h.xmin) : 0.0L);
          
          return ret;
        },
        
        /*   S_V^{(2)} operator   */
        [=](const distpoint::hist1d& h) {
          std::array<long double, 2u> ret{{0.0L,0.0L}};
          long double yl = std::exp(-h.xmax), yc = std::exp(-0.5*(h.xmin+h.xmax)), yh = std::exp(-h.xmin);
          
//          ret[0] += ((obs2[6] < yl ? 1.0L : 0.0L) - (obs1 < yl ? 1.0L : 0.0L))*coeff2[6];
//          ret[0] -= (obs2[7] > yl ? 1.0L : 0.0L)*(obs1 < yl ? 1.0L : 0.0L)*coeff2[7];
//          ret[0] -= (obs2[7]+obs1 > yl ? 1.0L : 0.0L)*(obs1 < yl ? 1.0L : 0.0L)*coeff2[7];
//
//          return ret;
        
          ret[0] = -(obs2[7] > yl ? 1.0L : 0.0L)*(obs1 < yl ? 1.0L : 0.0L)*coeff2[7];
          ret[1]  = (obs1 < yc ? 1.0L : 0.0L)*(yl <= obs2[7] && obs2[7] < yh ? -coeff2[7]/(h.xmax-h.xmin) : 0.0L);
          ret[1] -= (obs2[7] > yc ? 1.0L : 0.0L)*(yl <= obs1 && obs1 < yh ? -coeff2[7]/(h.xmax-h.xmin) : 0.0L);
          
          for(std::size_t k = 0; k < 7; ++k)
          {
            ret[0] += ((obs2[k] < yl ? 1.0L : 0.0L) - (obs1 < yl ? 1.0L : 0.0L))*coeff2[k];
            ret[1] -= (yl <= obs2[k] && obs2[k] < yh ? coeff2[k]/(h.xmax-h.xmin) : 0.0L);
            ret[1] += (yl <= obs1 && obs1 < yh ? coeff2[k]/(h.xmax-h.xmin) : 0.0L);
          }
          
          return ret;
        }
      };
    }
    

    template<class _Func>
    requires std::is_invocable_r_v<long double, _Func, const resummation_event<epa>&, long double>
    auto resummation_exponent<epa>::integrand2(_Func&& funcy3, std::initializer_list<int> ida, bool idb) const
    {
      /*    aliases   */
      auto& first = _M_stages->first;
      auto& aterm = _M_stages->aterm;
      auto& bterm = _M_stages->bterm;
      
      
      /*    calculate the relevant parameters      */
      long double coeff1 = 0.0;
      std::array<long double, 8u> coeff2;
      
      coeff2.fill(0.0L);
      
      
      if(std::abs(first.state[0].weight) > 1e-10)
      {
        /*   parameters for the S1v term     */
        coeff1 = first.state[0].weight*first.cweight;
        
        /*   parameters for the A terms  */
        for(std::size_t k : ida)
        {
          coeff2[k] = aterm.cweight*aterm.state[k].weight;
        }
        
        
        /*   parameters for the  B terms   */
        if(idb)
        {
          coeff2[6] = bterm.cweight*bterm.state[0].weight;
          coeff2[7] = bterm.cweight*bterm.state[1].weight;
        }
      }
      
      
      /*    returns a pair of lambda expression   */
      return std::pair{
        /*   S_V^{(1)}  operator   */
        [=, this](const distpoint::hist1d& h) {
          std::array<long double, 2u> ret;
          long double yl = std::exp(-h.xmax);
          
          ret[0] = (funcy3(_M_stages->first.state[0], yl)-1.0)*coeff1;
          ret[1] = 0.0L;
          
          return ret;
        },
        
        /*   S_V^{(2)} operator   */
        [=,this](const distpoint::hist1d& h) {
          std::array<long double, 2u> ret{{0.0L,0.0L}};
          long double yl = std::exp(-h.xmax);
          
          
          long double R2st = (funcy3(_M_stages->first.state[0], yl) ? 1.0L : 0.0);
          
          for(std::size_t k = 0; k < 6; ++k)
            ret[0] += (funcy3(_M_stages->aterm.state[k], yl) - R2st)*coeff2[k];
  
          ret[0] += (funcy3(_M_stages->bterm.state[0], yl) - R2st)*coeff2[6];
          ret[0] += (funcy3(_M_stages->bterm.state[1], yl) -1.0L)*R2st*coeff2[7];

          return ret;
        }
      };
    }
    

  }   //  namespace resum
}   //  namespace duct


#endif
