/*  rs-resum-stage.h
 *  Created by Zoltan Nagy on 18.12.19.
 *  Copyright © 2019 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __rs_resum_stage_h__
#define __rs_resum_stage_h__ 1


/*   std includes    */


/*   project headers    */
#include <bits/rs-resum-event.h>
#include <bits/mc-color-state.h>






namespace duct {
  namespace resum {
    
 
    
    /*   Some helper class and function    */
    namespace __details {
      
      template<ProcessTag _Proc, std::size_t _N_states>
      class event_array
      {
        using _Event = resummation_event<_Proc>;
        using _Storage = pmr::vector<_Event>;
        
      public:
        /*    ususal typedefs   */
        using value_type = _Event;
        using size_type = std::size_t;
        
        using pointer = typename _Storage::pointer;
        using iterator = typename _Storage::iterator;
        using reference = typename _Storage::reference;
        
        using const_pointer = typename _Storage::const_pointer;
        using const_iterator = typename _Storage::const_iterator;
        using const_reference = typename _Storage::const_reference;
        
        using reverse_iterator = typename _Storage::reverse_iterator;
        using const_reverse_iterator = typename _Storage::const_reverse_iterator;

        /*    it has to be an allocator aware container     */
        using allocator_type = typename _Storage::allocator_type;
        
        /*   default constructors    */
        event_array() : _M_states{_N_states} {}
        explicit event_array(allocator_type a) : _M_states{_N_states, a} {}
        
        /*   copy   */
        event_array(const event_array&) = default;
        event_array& operator=(const event_array&) = default;
        
        event_array(const event_array& x, allocator_type a)
        : _M_states{x._M_states, a} {}
        
        /*    move  */
        event_array(event_array&&) = default;
        event_array& operator=(event_array&&) = default;
        
        event_array(event_array&& x, allocator_type a)
        : _M_states{std::move(x._M_states), a} {}
        
        
        /*   acces to the states   */
        reference operator[](std::size_t k) { return _M_states[k];}
        const_reference operator[](std::size_t k) const { return _M_states[k];}
        
        
        /*   iterators   */
        iterator begin() { return _M_states.begin();}
        iterator end() { return _M_states.end();}

        const_iterator begin() const { return _M_states.begin();}
        const_iterator end() const { return _M_states.end();}
 
        const_iterator cbegin() const { return _M_states.cbegin();}
        const_iterator cend() const { return _M_states.cend();}
        
      private:
        /*   the states are stored in a vector   */
        _Storage _M_states;
      };
      
      
    }
    
    
    

    /*   It is a storage class to store several events associated with the same color structure.
     */
    template<ProcessTag _Proc, std::size_t _N_states>
    class resummation_stage
    {
    public:
      /*    it has to be an allocator aware container     */
      using allocator_type = pmr::polymorphic_allocator<>;

      /*   default constructors    */
      resummation_stage() : state{}, cstate{}, cweight{1.0} {}
      explicit resummation_stage(allocator_type a) : state{a}, cstate{a}, cweight{1.0}
      {}

      /*   copy   */
      resummation_stage(const resummation_stage&) = default;
      resummation_stage& operator=(const resummation_stage&) = default;

      resummation_stage(const resummation_stage& x, allocator_type a)
      : state{x.state, a}, cstate{x.cstate, a}, cweight{x.cweight}
      {}

      /*    move  */
      resummation_stage(resummation_stage&&) = default;
      resummation_stage& operator=(resummation_stage&&) = default;

      resummation_stage(resummation_stage&& x, allocator_type a)
      : state{std::move(x.state), a}, cstate{std::move(x.cstate), a}, cweight{x.cweight}
      {}
      
      /*   returns the total weight   */
      double weight_tot(std::size_t k) const {
        return state[k].weight*cweight;
      }
      
      /*    resets the weights    */
      void reset_weights()
      {
        for(auto& s : state)
          s.weight = 0.0;
        
        cweight = 1.0;
      }
      
    public:
      /*  The event of the state. This contains every informations about the
       *  event, momenta, flavors, incoming objects,...
       */
      __details::event_array<_Proc, _N_states> state;
      
      /*   We can have more than one color basis state associated to a momentum and
       *   flavor configuration. Thus we have a vector (technically a list) of color states.
       */
      shower_color_state<_Proc> cstate;
      
      /*   color weight   */
      long double cweight;
    };


    template<ProcessTag _Proc>
    class resummation_stage<_Proc, 1u>
    {
    public:
      /*    it has to be an allocator aware container     */
      using allocator_type = pmr::polymorphic_allocator<>;
      
      /*   default constructors    */
      resummation_stage() : state{}, cstate{}, cweight{1.0} {}
      explicit resummation_stage(allocator_type a) : state{a}, cstate{a}, cweight{1.0}
      {}
      
      /*   copy   */
      resummation_stage(const resummation_stage&) = default;
      resummation_stage& operator=(const resummation_stage&) = default;
      
      resummation_stage(const resummation_stage& x, allocator_type a)
      : state{x.state, a}, cstate{x.cstate, a}, cweight{x.cweight}
      {}
      
      /*    move  */
      resummation_stage(resummation_stage&&) = default;
      resummation_stage& operator=(resummation_stage&&) = default;
      
      resummation_stage(resummation_stage&& x, allocator_type a)
      : state{std::move(x.state), a}, cstate{std::move(x.cstate), a}, cweight{x.cweight}
      {}
      
      /*   returns the total weight   */
      double weight_tot() const {
        return state.weight*cweight;
      }

      /*    resets the weights    */
      void reset_weights() {
        state.weight = 0.0;
        cweight = 1.0;
      }
      
    public:
      /*  The event of the state. This contains every informations about the
       *  event, momenta, flavors, incoming objects,...
       */
      resummation_event<_Proc> state;
      
      /*   We can have more than one color basis state associated to a momentum and
       *   flavor configuration. Thus we have a vector (technically a list) of color states.
       */
      shower_color_state<_Proc> cstate;
      
      /*   color weight   */
      long double cweight;
    };
    

  }
}

#endif
