/*  rs-resummation.h
 *  Created by Zoltan Nagy on 31.01.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __rs_resummation_h__
#define __rs_resummation_h__


/*   std includes    */


/*   project headers    */
#include <bits/mc-deductor.h>
#include <bits/rs-resum-exp.h>
#include <bits/rs-resum-user.h>



namespace duct {
  namespace resum {
    namespace __details {

      /*    This class is to store the buffer (a set of resummation_exponent objects) and
       *    provides functions to fill with real event and analyse them.
       */
      template<ProcessTag _Proc>
      class resummation_buffer
      {
      public:
        /*   Create a buffer with n elements   */
        explicit resummation_buffer(std::size_t n = 32u)
        : _M_pool(n), _M_kill{false}
        { for(std::size_t k = 0; k < n; k++) _M_consumed.push(k);}

        /*   Copy and move are deleted because the mutex and conditional_variable
         *   cannot be copied or moved.
         */
        resummation_buffer(const resummation_buffer&) = delete;
        resummation_buffer& operator=(const resummation_buffer&) = delete;

        resummation_buffer(resummation_buffer&&) = delete;
        resummation_buffer& operator=(resummation_buffer&&) = delete;

        ~resummation_buffer() = default;


        /*  This is the shower generator. It write new events to the buffer.
         *  It stops when it receives kill signal form the analyzer routine.
         */
        template<class, class _MakeHard>
        void generator(_MakeHard&&);


        /*  This is the analyser routine. It reads elements from a buffer and call the user
         *  routines. It stops and send the kill signal when all the analysises are done.
         */
        template<class _MakeUser>
        void analyser(_MakeUser&&);

        /*   Checked if the calculation is finished   */
        bool is_killed() const { return _M_kill;}

        /*   Send the kill signal too all of the threads   */
        void kill() { _M_kill = true;}

       private:
        /*   This is the pool where we store the elements of the buffer   */
        std::vector<resummation_exponent<_Proc>> _M_pool;

        /*    This is the queue of the produced and consumed objects.  */
        std::queue<std::size_t> _M_produced, _M_consumed;

        /*   Signal to kill the generators   */
        std::atomic<bool> _M_kill;

        /*    Mutexes for protecting the queues.   */
        std::mutex _M_pmtx, _M_cmtx;

        /*   Conditional variables to syncronize threads    */
        std::condition_variable _M_prod_is_not_empty, _M_cons_is_not_empty;
      };


      /*   Function to generate the event. We can have several generator task. They can run
       *   simultaneously and resummation_buffer class takes care about the syncronisation.
       */
      template<ProcessTag _Proc>
      template<class _Scheme, class _MakeHard>
      void resummation_buffer<_Proc>::generator(_MakeHard&& mkhard)
      {
        /*     initialize the random number generator   */
        std::size_t seed = std::hash<std::thread::id>()(std::this_thread::get_id());
        seed ^= std::hash<decltype(this)>()(this) << 1;
        seed ^= std::hash<unsigned long>()(std::chrono::system_clock::to_time_t(std::chrono::system_clock::now())) << 1;
        seed ^= std::hash<_MakeHard*>()(&mkhard) << 1;
        
        random_engine_type rne(seed);
        
        /*   hard generator   */
        auto hard = mkhard();


        /*   The generator runs until it gets the kill signal.   */
        while(_M_kill == false)
        {
          /*   pop from the consumed queue   */
          unsigned int current;
          //------
          {
            std::unique_lock<std::mutex> lock(_M_cmtx);
            while(_M_consumed.empty()) {
              if(_M_kill) return;
              _M_cons_is_not_empty.wait(lock);
            }

            current = _M_consumed.front(); _M_consumed.pop();
          }

          /*   Generate random points and calculate the exponent.   */
          _M_pool[current].template generate<_Scheme>(rne, hard);


          /*   puts the product into the produced queue   */
          _M_pmtx.lock(); _M_produced.push(current); _M_pmtx.unlock();
          _M_prod_is_not_empty.notify_one();
        }
      }



      /*   This function runs the analyser routines. It takes user functions and run them
       *   sequentially on every shower histories.
       */
      template<ProcessTag _Proc>
      template<class _MakeUser>
      void resummation_buffer<_Proc>::analyser(_MakeUser&& mkuser)
      {
        std::list<std::unique_ptr<resum_user<_Proc>>> users{mkuser()};


        /*   This is a fresh calculation, let the generators run...  */
        _M_kill = false;

        /*   The analyzer runs until all the user routines are done.   */
        while(_M_kill == false)
        {
          /*   If all the analyseses are done, send the kill signal.   */
          if(users.empty()) {
            _M_kill = true;
            break;
          }

          /*   pop from the produced queue   */
          unsigned int current;
          //------
          {
            std::unique_lock<std::mutex> lock(_M_pmtx);
            while(_M_produced.empty()) _M_prod_is_not_empty.wait(lock);
            current = _M_produced.front(); _M_produced.pop();
          }

          /*   Analysing the event  */
          const auto& rsexp = _M_pool[current];

          for(auto iu = users.begin(); iu != users.end();)
          {
            (*iu)->userfunc(rsexp);
            (*iu)->end_of_event();

            if((*iu)->is_done()) {
              (*iu)->finish();
              iu = users.erase(iu);
            } else ++iu;
          }

          /*   puts the consumed item back to the consumed queue   */
          _M_cmtx.lock(); _M_consumed.push(current); _M_cmtx.unlock();
          _M_cons_is_not_empty.notify_one();
        }

        /*   There was a kill signal, notify all blocked generators to quit.   */
        _M_cons_is_not_empty.notify_all();
      }



      /*    This is a class for representing a resummation process.
       *    It starts many generator and one analyser thread and lets them to run.
       *    The instances of this class own the threads and they join in the
       *    destructor of this class.
       */
      template<ProcessTag _Proc>
      class resummation_task
      {
      public:
        /*   default constructor  */
        resummation_task() :  _M_buffer{nullptr} {}

        /*   copy is deleted   */
        resummation_task(const resummation_task&) = delete;
        resummation_task& operator=(const resummation_task&) = delete;

        /*   move is allowed    */
        resummation_task(resummation_task&& x)
        : _M_buffer(nullptr), _M_threads(std::move(x._M_threads))
        {
          std::swap(_M_buffer, x._M_buffer);
        }

        resummation_task& operator=(resummation_task&& x)
        {
          std::swap(_M_buffer, x._M_buffer);
          _M_threads = std::move(x._M_threads);
          return *this;
        }

        /*   destructor  */
        ~resummation_task()
        {
          /*   Every thread has to be joined to the main thread before
           *   the data members get destroyed.
           */
          for(auto& ti : _M_threads) ti.join();

          /*   deleting the buffer   */
          if(_M_buffer) delete _M_buffer;
        }


        /*   Creates and starts the calculation for a given hard process and with the user functions.   */
        template<class _Scheme, class _MakeHard, class _MakeUser>
        resummation_task(_Scheme&&, _MakeHard&& mkhard, _MakeUser&& mkuser, int load)
        {
          /*   if it not specified then the number of thread is the number of available CPUs   */
          unsigned nthr = std::thread::hardware_concurrency();
          unsigned ngen = nthr;

          if(load == 0u) ngen = nthr;
          else if(load < 0) ngen = 1 + nthr/std::abs(load);
          else ngen = load;

          if(ngen > nthr) ngen = nthr;
          if(ngen == 0u) ngen = 1u;

          /*   Buffer to queue the generated events.   */
          _M_buffer = new resummation_buffer<_Proc>{ngen+2u};

          /*   Vector of threads   */
          _M_threads.resize(ngen+1);

          /*   Starting the analyser process   */
          _M_threads[0] = std::thread(&resummation_buffer<_Proc>::template analyser<_MakeUser>, _M_buffer, mkuser);

          /*   Starting the generator process(es)   */
          for(unsigned i = 1; i <= ngen; ++i)
            _M_threads[i] = std::thread(&resummation_buffer<_Proc>::template generator<_Scheme, _MakeHard>, _M_buffer, mkhard);
        }


        /*   Checked if the calculation is finished   */
        bool is_killed() const { return _M_buffer ? _M_buffer->is_killed() : true;}

        /*   Send the kill signal too all of the threads   */
        void kill() { if(_M_buffer) _M_buffer->kill();}


      private:
        /*   Buffer to queue the generated events.   */
        resummation_buffer<_Proc> *_M_buffer;

        /*   Vector of threads   */
        std::vector<std::thread> _M_threads;
      };

    }   //   namespace __details




    struct resummation_opts
    {
      /*   the name must be provided  */
      resummation_opts(const char *n, int ll = 0)
      : name{n}, load{ll} {}


      /*   data members    */
      const char *name = "My wonderful calculation";
      int load = 0;
    };



    /*  This is a high level class to start and stop shower calculations.
     *  It can handle many processes simultaneously.
     */
    template<ProcessTag _Proc>
    class resummation : private duct::__details::deductor_kernel_rtd
    {
      /*    some type aliases   */
      using _Rtd = duct::__details::deductor_kernel_rtd;
      using _Task = __details::resummation_task<_Proc>;

    public:
      /*   default constructor   */
      resummation() = default;

      /*   create from model   */
      explicit resummation(model&& mdl) : _Rtd{std::forward<model>(mdl)} {};
      explicit resummation(const model& mdl) : _Rtd{mdl} {}
      explicit resummation(const qcd::input& qin) : resummation{model{qin}} {}
      explicit resummation(const qcd::predefined& qin) : resummation{model{qin}} {}

      /*   the first argument is the model dir    */
      resummation(const fs::path& dir, const qcd::input& qin)
      : resummation{model{dir, qin}} {}


      /*   copy is deleted because the process cannot be copied   */
      resummation(const resummation&) = delete;
      resummation& operator=(const resummation&) = delete;

      /*   move   */
      resummation(resummation&&) = default;
      resummation& operator=(resummation&&) = default;


      /*    destructor   */
      ~resummation() = default;


      /*   Checked if the process is killed   */
      bool is_killed(unsigned int __pid) const {
        auto p = _M_proc.find(__pid);
        if(p != _M_proc.end())
          return p->second.second.is_killed();
        return true;
      }

      /*    Send a kill signal to a process    */
      void kill(unsigned int __pid) {
        auto p = _M_proc.find(__pid);
        if(p != _M_proc.end()) {
          p->second.second.kill();
          _M_proc.erase(p);
        }
      }


      /*   returns the name of the process with the given ID    */
      std::string name(unsigned __pid) const {
        auto p = _M_proc.find(__pid);
        if(p != _M_proc.end()) return p->second.first;

        return std::string{"PID "} + std::to_string(__pid) + std::string{" not found!"};
      }


      /*    Starts a shower calculation    */
      template<class _Scheme, class _Hard, class ..._Users>
      requires ShowerScheme<_Proc, _Scheme>
      unsigned operator()(_Scheme&&, const resummation_opts& opts, _Hard&& mkh, _Users * ...usr)
      {
        using _UsrPtr = std::unique_ptr<resum_user<_Proc>>;
        using _UsrList = std::list<_UsrPtr>;
        using _Ordering = typename _Scheme::ordering_type;
        
        /*   The PID of this process   */
        unsigned int __pid = _Rtd::pid++;

        /*    start process   */
        _M_proc[__pid] = {opts.name,
          _Task{
            /*   Scheme selector   */
            _Scheme{},
            
            /*   Hard process    */
            [=, this]() {
              return mkh(model_ref{_Ordering{}, _Rtd::mdl});
            },

            /*    Analysers     */
            [=]() -> _UsrList {
              _UsrList users;

              (users.push_back(_UsrPtr{usr}), ...);

              for(const auto& u : users)
              {
                u->process_info(__pid, opts.name);
                u->initialize();
              }

              return users;
            },

            /*   number of the generators   */
            opts.load
          }
        };

        return __pid;
      }


    private:
      /*    Process table   */
      std::map<unsigned, std::pair<std::string, _Task>> _M_proc;
    };


  }  //  namespace resum
}  //  namespace duct

#endif
