/*  shower-scheme-cs.cc
 *  Created by Zoltan Nagy on 20.02.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes    */
#include <cmath>


/*   project headers    */
#include "cxx/math-support.h"
#include "bits/rs-shower-scheme.h"



namespace duct {
  namespace resum {
    
    namespace __details {
      
  
      resum_splitting_vars generate_vars_cs(const long double *rng, const long double& ymax)
      {
        /*  parameters to control the importance sampling    */
        constexpr long double epsilon = 1.0e-8, beta = 2.0;
        
        const long double eps = epsilon*ymax, lmax = -std::log(ymax);
        const long double leps = -std::log(eps), beta2 = beta*beta;
        const long double n1 = 0.5*(leps*leps-lmax*lmax), n2 = beta*(beta + leps);
        
        
        /*   generate the x0 variable  */
        long double y, z, phi, jac = n1+n2;
        
        if(*(rng++) < n1/jac) {
          y = std::exp(-std::sqrt(leps*leps - 2.0*(*(rng++))*n1));
        }
        else
        {
          long double r = -(*(rng++))*n2/beta2;
          long double fe = M_E*std::pow(eps, -1.0/beta);
          
          long double x = std::pow(r/lambert_Wm1(r/fe), beta);
          
          y = eps*x;
          jac /= std::pow(x, 1.0/beta);
        }
        
        /*   generate the z variable   */
        z = (1.0-std::pow(y, *(rng++)))/(1.0-y);
        
        /*   generating the phi variable   */
        phi = 2.0*M_PI*(*(rng++));
        
        
        /*   returning the splitting variables and the jacobian  */
        resum_splitting_vars zeta;
        
        zeta.y = y;
        zeta.z = z;
        zeta.phi = phi;
        zeta.jac = jac;
        
        return zeta;
      }
      

      resum_splitting_vars generate_vars_cs(const long double *r, int l, int k, const resummation_event<epa>& p, const long double& yL)
      {
        /*   xl and xi variable   */
        auto Q = p[-1].momentum+p[0].momentum;
        const long double xlk = (p[l].momentum*p[k].momentum)/(Q*p[l].momentum);
        
        /*    generate variables   */
        auto zeta = generate_vars_cs(r, std::min(1.0L, yL/xlk));
        
        zeta.emitter = l;
        zeta.spectator = k;
        
        zeta.fl = p[l].flavor;
        zeta.fm = 0;  //  we have only gluon emission (at least at the moment)
        
        
        return zeta;
      }
      

      
      /* ======================================================================================================== */
      
   
      static resum_splitting_vars generate_vars_cs_kT(const long double *rng, const long double& yT)
      {
        /*  parameters to control the importance sampling    */
        long double epsilon = 1.0e-9L, beta = 4.0;
        
        while(1.0e-15L + std::abs(yT*epsilon) < 1.0e-9L)
          if((epsilon *= 10.0L) > 1.0e-3L) break;

        const long double eps = epsilon*yT;
        const long double leps = -std::log(eps), beta2 = beta*beta;
        const long double n1 = 0.5*leps*leps, n2 = beta*(beta + leps);
        
        
        /*   generate the x0 variable  */
        long double y, z, phi, jac = n1+n2;
        
        if(*(rng++) < n1/jac) {
          y = std::exp(-std::sqrt(leps*leps - 2.0*(*(rng++))*n1));
        }
        else
        {
          const long double r = -(*(rng++))*n2/beta2;
          const long double fe = M_E*std::pow(eps, -1.0/beta);
          const long double x = std::pow(r/lambert_Wm1(r/fe), beta);
          
          y = eps*x;
          jac /= std::pow(x, 1.0/beta);
        }
        
        /*   generate the z variable   */
        const long double& rz = *(rng++);

        if(y < yT)
        {
          z = (1.0-std::pow(y, rz))/(1.0L-y);
        }
        else
        {
          const long double zp = 0.5L*(1.0L+std::sqrt(1.0L-yT/y));
          const long double zm = 1.0L-zp, oy = 1.0L-y;
          const long double Nm = -std::log(1.0L-zm*oy), Np = std::log((1.0L-zp*oy)/y);
          
          z = (*(rng++) < Nm/(Nm+Np) ? 1.0L-std::exp(-rz*Nm) : 1.0L-y*std::exp(rz*Np))/oy;
          jac /= -std::log(y);
          jac *= Nm+Np;
        }
        
        /*   generating the phi variable   */
        phi = 2.0*M_PI*(*(rng++));
        
        
        /*   returning the splitting variables and the jacobian  */
        resum_splitting_vars zeta;
        
        zeta.y = y;
        zeta.z = z;
        zeta.phi = phi;
        zeta.jac = jac;
        
        return zeta;
      }
      
      resum_splitting_vars generate_vars_cs_kT(const long double *r, int l, int k, const resummation_event<epa>& p, const long double& kT)
      {
        /*   xl and xi variable   */
        auto Q = p[-1].momentum+p[0].momentum;
        const long double xlk = 2.0L*(p[l].momentum*p[k].momentum)/mag2(Q);
        
        /*    generate variables   */
        auto zeta = generate_vars_cs_kT(r, std::min(1.0L, 4.0L*kT/xlk));
        
        zeta.emitter = l;
        zeta.spectator = k;
        
        zeta.fl = p[l].flavor;
        zeta.fm = 0;  //  we have only gluon emission (at least at the moment)
        
        
        return zeta;
      }
      
  
    }   //  namespace __details
  
  
  }   //  namespace resum
}   //  namespace duct


